////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Application
//
// ghmGcApp.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 11/08/2002
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCAPP_CPP__

#include <dolphin.h>
#include "ghmGcApp.h"

volatile bool	ghmGcApp::mRetraceFlag       = false;
volatile u32	ghmGcApp::mRetraceCount      = 0;
volatile u32	ghmGcApp::mRetraceFrame      = 1;
volatile u32	ghmGcApp::mRetraceFrameCount = 0;
volatile u32	ghmGcApp::mRetraceCountDiff  = 0;
volatile u32	ghmGcApp::mRefreshCount      = 0;
int				ghmGcApp::mResetProc         = 0;
bool			ghmGcApp::mResetEnable       = true;
int				ghmGcApp::mResetParam        = OS_RESET_RESTART;
u32				ghmGcApp::mResetCode         = 1;
BOOL			ghmGcApp::mResetForceMenu    = FALSE;

u32	retraceFrameCount = 0;

ghmGcApp::ghmGcApp()
{
}

ghmGcApp::~ghmGcApp()
{
}

bool ghmGcApp::Initialize(void)
{
	mPerfMeter.Initialize();
	
	mTerminate         = false;
	mIDCounter         = 0;
	mRetraceCount      = 0;
	mRetraceCountDiff  = 0;
	mRetraceFrame      = 1;
	mRetraceFrameCount = 0;
	mRefreshCount      = 0;
	
	SetRetraceCallBack();
	
	return true;
}

void ghmGcApp::PreRetraceCallBack(u32 retraceCount)
{
	mRetraceCountDiff  = retraceCount - mRetraceCount;
	mRetraceCount      = retraceCount;
	mRetraceFrameCount++;
	if( mRetraceFrameCount >= mRetraceFrame ){
		mRetraceFlag = true;
	}
}

void ghmGcApp::PostRetraceCallBack(u32 retraceCount)
{
#pragma unused(retraceCount)
}

bool ghmGcApp::SetRetraceCallBack(void)
{
	VISetPreRetraceCallback (PreRetraceCallBack );
	VISetPostRetraceCallback(PostRetraceCallBack);
	return true;
}

bool ghmGcApp::Run(void)
{
	bool	result = true;
	
	while( mTerminate==false ){
		
		// VBlank
		
		if( (result=SynchronizeVBlank())==false ){
			break;
		}
		
		// Zbg
		ResetProcess();
		
		// t@C
		ghmGcFile::FrameProcessAll();
		
		// t[
		
		if( (result=PreFrameProcess())==false ){
			break;
		}
		if( (result=FrameProcess())==false ){
			break;
		}
		if( (result=PostFrameProcess())==false ){
			break;
		}
		
		// _O
		
		if( (result=PreRenderProcess())==false ){
			break;
		}
		if( (result=RenderProcess())==false ){
			break;
		}
		if( (result=PostRenderProcess())==false ){
			break;
		}
		
		// t@C
		ghmGcFile::RenderProcessAll();
		
		// ̎gp󋵕\
		ghmGcMem::RenderAll();
		
	}
	
	return true;
}

bool ghmGcApp::ResetProcess(void)
{
	switch( mResetProc ){
	case	0:
		if( mResetEnable==false ){
			break;
		}
		if( OSGetResetButtonState()==FALSE ){
			break;
		}
		mResetProc = 1;
		break;
	case	1:
		if( OSGetResetButtonState()==TRUE ){
			break;
		}
		VISetBlack(TRUE);
		VIFlush();
		VIWaitForRetrace();
		mResetProc = 0;
		OSResetSystem(mResetParam,mResetCode,mResetForceMenu);
		return true;
	default:
		break;
	}
	return false;
}

bool ghmGcApp::PreFrameProcess(void)
{
	ghmGcAppTask		*pTask;
	ghmGcAppTask		*pNext;
	
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pNext){
		pNext = pTask->GetNext();
		if( pTask->GetFlag() & GHMGCAPPTASK_PREFRAMEPROCESS ){
			continue;
		}
		pTask->ResetFlag(GHMGCAPPTASK_PREFRAMEPROCESS);
		if( pTask->PreFrameProcess()==false ){
			pTask->Terminate();
		}
	}
	
	// ptH[}X[^[botԌvJn
	mPerfMeter.StartProc();
	
	return true;
}

bool ghmGcApp::FrameProcess(void)
{
	ghmGcAppTask		*pTask;
	ghmGcAppTask		*pNext;
	
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pNext){
		pNext = pTask->GetNext();
		if( pTask->GetFlag() & GHMGCAPPTASK_FRAMEPROCESS ){
			continue;
		}
		pTask->ResetFlag(GHMGCAPPTASK_FRAMEPROCESS);
		if( pTask->FrameProcess()==false ){
			pTask->Terminate();
		}
	}
	
	return true;
}

bool ghmGcApp::PostFrameProcess(void)
{
	ghmGcAppTask		*pTask;
	ghmGcAppTask		*pNext;
	
	// ptH[}X[^[botԌvI
	mPerfMeter.EndProc();
	
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pNext){
		pNext = pTask->GetNext();
		if( pTask->GetFlag() & GHMGCAPPTASK_POSTFRAMEPROCESS ){
			continue;
		}
		pTask->ResetFlag(GHMGCAPPTASK_POSTFRAMEPROCESS);
		if( pTask->PostFrameProcess()==false ){
			pTask->Terminate();
		}
	}
	
	return true;
}

bool ghmGcApp::PreRenderProcess(void)
{
	ghmGcAppTask		*pTask;
	ghmGcAppTask		*pNext;
	
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pNext){
		pNext = pTask->GetNext();
		if( pTask->GetFlag() & GHMGCAPPTASK_PRERENDERPROCESS ){
			continue;
		}
		pTask->ResetFlag(GHMGCAPPTASK_PRERENDERPROCESS);
		if( pTask->PreRenderProcess()==false ){
			pTask->Terminate();
		}
	}
	
	// ptH[}X[^[fotԌvJn
	mPerfMeter.StartRender();
	
	return true;
}

bool ghmGcApp::RenderProcess(void)
{
	ghmGcAppTask		*pTask;
	ghmGcAppTask		*pNext;
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pNext){
		pNext = pTask->GetNext();
		if( pTask->GetFlag() & GHMGCAPPTASK_RENDERPROCESS ){
			continue;
		}
		pTask->ResetFlag(GHMGCAPPTASK_RENDERPROCESS);
		if( pTask->RenderProcess()==false ){
			pTask->Terminate();
		}
	}
	return true;
}

bool ghmGcApp::PostRenderProcess(void)
{
	ghmGcAppTask		*pTask;
	ghmGcAppTask		*pNext;
	
	// ptH[}X[^[fotԌvI
	mPerfMeter.EndRender();
	
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pNext){
		pNext = pTask->GetNext();
		if( pTask->GetFlag() & GHMGCAPPTASK_POSTRENDERPROCESS ){
			continue;
		}
		pTask->ResetFlag(GHMGCAPPTASK_POSTRENDERPROCESS);
		if( pTask->PostRenderProcess()==false ){
			pTask->Terminate();
		}
	}
	
	return true;
}

bool ghmGcApp::SynchronizeVBlank(void)
{
	ghmGcAppTask		*pTask;
	ghmGcAppTask		*pNext;
	
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pNext){
		pNext = pTask->GetNext();
		switch( pTask->GetState() ){
		case	ghmGcAppTask::TERM:
			if( pTask->Terminate()==false ){
				break;
			}
		case	ghmGcAppTask::DELETE:
			delete pTask;
			break;
		case	ghmGcAppTask::INIT:
			pTask->Initialize();
			break;
		default:
			pTask->ResetFlag(GHMGCAPPTASK_PROCESSFLAGALL);
			break;
		}
	}
	
	// ptH[}X[^[I
	mPerfMeter.End();
	
//	for(retraceFrameCount=0 ; retraceFrameCount<mRetraceFrame ; retraceFrameCount++){
//		VIWaitForRetrace();
//	}
//	while( mRetraceFrameCount < mRetraceFrame ){
//		VIWaitForRetrace();
//	}
//	VIWaitForRetrace();
	VIWaitForRetrace();
	while( mRetraceFlag==false ){}
	mRetraceFlag = false;
	mRetraceFrameCount = 0;
	mRefreshCount++;
	
	// ptH[}X[^[vJn
	mPerfMeter.Start();
	
	return true;
}

int ghmGcApp::GenerateUniqueID(void)
{
	int				id = mIDCounter;
	ghmGcAppTask	*pTask;
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pTask->GetNext()){
		if( pTask->GetID()==id ){
			// Č
			id++;
			if( id < 0 ){
				id = 0;
			}
			pTask = &mTaskAnchor;
			break;
		}
	}
	mIDCounter = id+1;
	if( mIDCounter < 0 ){
		mIDCounter = 0;
	}
	return id;
}

ghmGcAppTask* ghmGcApp::GetTask(int id)
{
	ghmGcAppTask		*pTask;
	for(pTask=mTaskAnchor.GetNext() ; pTask!=NULL ; pTask=pTask->GetNext()){
		if( pTask->GetID()==id ){
			return pTask;
		}
	}
	return NULL;
}

