////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Collision
//
// ghmGcColl.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 06/23/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCCOLL_CPP__

#include "ghmGcColl.h"
#include "ghmGcMath.h"

int ghmGcPlane::GetCrossState(ghmGcCollObj &obj)
{
	return obj.GetCrossState(*this);
}

int ghmGcViewVolume::GetCrossState(ghmGcCollObj &obj)
{
	u32			outside = 0;
	u32			inside  = 0;
	int			i;
	u32			mask;
	ghmGcPlane	*pClipPlane;
	for(i=0,pClipPlane=mClipPlane,mask=1 ; i<6 ; i++,pClipPlane++,mask<<=1){
		switch( obj.GetCrossState(*pClipPlane) ){
		case	ghmGcCollObj::INSIDE:
			inside |= mask;
			break;
		case	ghmGcCollObj::OUTSIDE:
			outside |= mask;
			break;
		default:
			break;
		}
	}
	if( inside==(1<<6)-1 ){
		return ghmGcCollObj::INSIDE;
	}
	if( !outside ){
		return ghmGcCollObj::CROSSOVER;
	}
	return ghmGcCollObj::OUTSIDE;
}

int ghmGcViewVolume::GetFarClip(ghmGcCollObj &obj,f32 dist)
{
	return obj.GetFarClip(mCoordinate,dist);
}

bool ghmGcViewVolume::SetCurrent(HSD_CObj *pCObj)
{
	if( pCObj==NULL ){
		return false;
	}
	if( HSD_CObjGetProjectionType(pCObj)!=COBJ_PROJECTION_PERSPECTIVE ){
		return false;
	}
	
	Vec			campos;
	f32			n,f,aspect,fov;
	Mtx			dirmtx;
	
	n      = HSD_CObjGetNear  (pCObj);
	f      = HSD_CObjGetFar   (pCObj);
	fov    = HSD_CObjGetFov   (pCObj);
	aspect = HSD_CObjGetAspect(pCObj);
	
	MTXTranspose(pCObj->mtx,dirmtx);
	
	campos.x = -pCObj->mtx[0][3];
	campos.y = -pCObj->mtx[1][3];
	campos.z = -pCObj->mtx[2][3];
	MTXMultVec(dirmtx,&campos,&campos);
	
	return SetCurrent(campos,dirmtx,n,f,aspect,fov);
}

bool ghmGcViewVolume::SetCurrent(const Vec &campos,const Vec &camdir,f32 n,f32 f,f32 aspect,f32 fov)
{
	Vec		dir,vdir,hdir;
	Mtx		dirmtx;
	VECNormalize(&camdir,&dir);
	ghmMath::Roll2UpVec(dir,0.0f,vdir);
	VECCrossProduct(&dir,&vdir,&hdir);
	
	dirmtx[0][0] = hdir.x;
	dirmtx[1][0] = hdir.y;
	dirmtx[2][0] = hdir.z;
	dirmtx[0][1] = vdir.x;
	dirmtx[1][1] = vdir.y;
	dirmtx[2][1] = vdir.z;
	dirmtx[0][2] = -dir.x;
	dirmtx[1][2] = -dir.y;
	dirmtx[2][2] = -dir.z;
	dirmtx[0][3] =    0.0f;
	dirmtx[1][3] =    0.0f;
	dirmtx[2][3] =    0.0f;
	
	SetCurrent(campos,dirmtx,n,f,aspect,fov);
	
	return true;
}

bool ghmGcViewVolume::SetCurrent(const Vec &campos,Mtx dirmtx,f32 n,f32 f,f32 aspect,f32 fov)
{
	Vec			vdir,hdir,fdir;
	int			i;
	
	MTXCopy(dirmtx,mCoordinate);
	mCoordinate[0][3] = campos.x;
	mCoordinate[1][3] = campos.y;
	mCoordinate[2][3] = campos.z;
	MTXInverse(mCoordinate,mCoordinate);
	
	fov *= (0.5f * PI / 180.0f);
	
	mClipPlane[1].mNormal.x =  0.0f;
	mClipPlane[1].mNormal.y =  0.0f;
	mClipPlane[1].mNormal.z = -1.0f;
	MTXMultVecSR(dirmtx,&mClipPlane[1].mNormal,&mClipPlane[1].mNormal);
	
	fdir                 .x =  0.0f;
	fdir                 .y =  0.0f;
	fdir                 .z = -1.0f;
	mClipPlane[2].mNormal.x =  0.0f;
	mClipPlane[2].mNormal.y =  tanf(fov);
	mClipPlane[2].mNormal.z =  fdir.z;
	mClipPlane[3].mNormal.x =  mClipPlane[2].mNormal.x;
	mClipPlane[3].mNormal.y = -mClipPlane[2].mNormal.y;
	mClipPlane[3].mNormal.z =  mClipPlane[2].mNormal.z;
	mClipPlane[4].mNormal.x =  mClipPlane[2].mNormal.y * aspect;
	mClipPlane[4].mNormal.y =  0.0f;
	mClipPlane[4].mNormal.z =  mClipPlane[2].mNormal.z;
	mClipPlane[5].mNormal.x = -mClipPlane[4].mNormal.x;
	mClipPlane[5].mNormal.y =  mClipPlane[4].mNormal.y;
	mClipPlane[5].mNormal.z =  mClipPlane[4].mNormal.z;
	
	vdir.x = mClipPlane[4].mNormal.x;
	vdir.y = mClipPlane[2].mNormal.y;
	vdir.z = fdir.z;
	VECScale    (&vdir,mClipPosition+0,n);
	MTXMultVecSR(dirmtx,mClipPosition+0,mClipPosition+0);
	VECScale    (&vdir,mClipPosition+1,f);
	MTXMultVecSR(dirmtx,mClipPosition+1,mClipPosition+1);
	vdir.x = -mClipPlane[4].mNormal.x;
	VECScale    (&vdir,mClipPosition+2,n);
	MTXMultVecSR(dirmtx,mClipPosition+2,mClipPosition+2);
	VECScale    (&vdir,mClipPosition+3,f);
	MTXMultVecSR(dirmtx,mClipPosition+3,mClipPosition+3);
	vdir.y = -mClipPlane[2].mNormal.y;
	VECScale    (&vdir,mClipPosition+4,n);
	MTXMultVecSR(dirmtx,mClipPosition+4,mClipPosition+4);
	VECScale    (&vdir,mClipPosition+5,f);
	MTXMultVecSR(dirmtx,mClipPosition+5,mClipPosition+5);
	vdir.x = mClipPlane[4].mNormal.x;
	VECScale    (&vdir,mClipPosition+6,n);
	MTXMultVecSR(dirmtx,mClipPosition+6,mClipPosition+6);
	VECScale    (&vdir,mClipPosition+7,f);
	MTXMultVecSR(dirmtx,mClipPosition+7,mClipPosition+7);
	
	VECCrossProduct(&fdir,&mClipPlane[5].mNormal,&vdir);
	VECCrossProduct(&fdir,&mClipPlane[2].mNormal,&hdir);
	
	VECCrossProduct(&hdir,&mClipPlane[2].mNormal,&mClipPlane[2].mNormal);
	VECCrossProduct(&mClipPlane[3].mNormal,&hdir,&mClipPlane[3].mNormal);
	VECCrossProduct(&mClipPlane[4].mNormal,&vdir,&mClipPlane[4].mNormal);
	VECCrossProduct(&vdir,&mClipPlane[5].mNormal,&mClipPlane[5].mNormal);
	
	// up,down,right,left
	for(i=2 ; i<6 ; i++){
		MTXMultVecSR(dirmtx,&mClipPlane[i].mNormal,&mClipPlane[i].mNormal);
		VECNormalize(&mClipPlane[i].mNormal,&mClipPlane[i].mNormal);
		mClipPlane[i].mPosition = campos;
	}
	
	// near
	VECScale(&mClipPlane[1].mNormal,&mClipPlane[0].mNormal,-1.0f);
	VECScale(&mClipPlane[1].mNormal,&mClipPlane[0].mPosition,n);
	VECAdd  (&campos,&mClipPlane[0].mPosition,&mClipPlane[0].mPosition);
	
	// far
	VECScale(&mClipPlane[1].mNormal,&mClipPlane[1].mPosition,f);
	VECAdd  (&campos,&mClipPlane[1].mPosition,&mClipPlane[1].mPosition);
	
	return true;
}

int ghmGcCollPoint::GetCrossState(const ghmGcPlane &plane)
{
	Vec		dir;
	f32		ip;
	
	VECSubtract(&mPosition,&plane.mPosition,&dir);
	ip = VECDotProduct(&plane.mNormal,&dir);
	
	if( ip > 0.0f ){
		return OUTSIDE;
	}
	
	return INSIDE;
}

int ghmGcCollPoint::GetFarClip(Mtx coord,f32 dist)
{
	Vec		pos;
	MTXMultVec(coord,&mPosition,&pos);
	if( pos.z < -dist ){
		return OUTSIDE;
	}
	return INSIDE;
}

int ghmGcCollSphere::GetCrossState(const ghmGcPlane &plane)
{
	Vec		dir;
	f32		ip;
	
	VECSubtract(&mPosition,&plane.mPosition,&dir);
	ip = VECDotProduct(&plane.mNormal,&dir);
	
	if( ip > mRadius ){
		return OUTSIDE;
	}
	
	if( ip > -mRadius ){
		return CROSSOVER;
	}
	
	return INSIDE;
}

int ghmGcCollSphere::GetFarClip(Mtx coord,f32 dist)
{
	Vec		pos;
	MTXMultVec(coord,&mPosition,&pos);
	if( pos.z >= - dist + mRadius ){
		return INSIDE;
	}
	if( pos.z >= - dist - mRadius ){
		return CROSSOVER;
	}
	return OUTSIDE;
}

int ghmGcCollBox::GetCrossState(const ghmGcPlane &plane)
{
	Vec		dir;
	f32		ip,extent;
	
	dir.x = mCoordinate[0][0];
	dir.y = mCoordinate[1][0];
	dir.z = mCoordinate[2][0];
	ip = mExtent.x * VECDotProduct(&plane.mNormal,&dir);
	if( ip < 0.0f ){
		ip = -ip;
	}
	extent = ip;
	
	dir.x = mCoordinate[0][1];
	dir.y = mCoordinate[1][1];
	dir.z = mCoordinate[2][1];
	ip = mExtent.y *  VECDotProduct(&plane.mNormal,&dir);
	if( ip < 0.0f ){
		ip = -ip;
	}
	extent += ip;
	
	dir.x = mCoordinate[0][2];
	dir.y = mCoordinate[1][2];
	dir.z = mCoordinate[2][2];
	ip = mExtent.z *  VECDotProduct(&plane.mNormal,&dir);
	if( ip < 0.0f ){
		ip = -ip;
	}
	extent += ip;
	
	dir.x = mCoordinate[0][3];
	dir.y = mCoordinate[1][3];
	dir.z = mCoordinate[2][3];
	VECSubtract(&dir,&plane.mPosition,&dir);
	ip = VECDotProduct(&plane.mNormal,&dir);
	
	if( ip > extent ){
		return OUTSIDE;
	}
	if( ip > -extent ){
		return CROSSOVER;
	}
	
	return INSIDE;
}

int ghmGcCollBox::GetFarClip(Mtx coord,f32 dist)
{
	Vec		dir;
	f32		extent;
	Mtx		mtx;
	
	MTXConcat(coord,mCoordinate,mtx);
	dir.x = mExtent.x;
	dir.y = 0.0f;
	dir.z = 0.0f;
	MTXMultVecSR(mtx,&dir,&dir);
	extent = GHM_FABS(dir.z);
	
	dir.x = 0.0f;
	dir.y = mExtent.y;
	dir.z = 0.0f;
	MTXMultVecSR(mtx,&dir,&dir);
	extent += GHM_FABS(dir.z);
	
	dir.x = 0.0f;
	dir.y = 0.0f;
	dir.z = mExtent.z;
	MTXMultVecSR(mtx,&dir,&dir);
	extent += GHM_FABS(dir.z);
	
	if( mtx[2][3] >= -dist + extent ){
		return INSIDE;
	}
	if( mtx[2][3] >= -dist - extent ){
		return CROSSOVER;
	}
	return OUTSIDE;
}

