////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube List
//
// ghmGcList.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 11/09/2002
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCLIST_CPP__

#include <dolphin.h>
#include "ghmGcList.h"

ghmGcListObj::ghmGcListObj()
{
	mpPrev    = NULL;
	mpNext    = NULL;
	mPriority = GHMGCLIST_ANCHOR_PRIO;
}

ghmGcListObj::ghmGcListObj(ghmGcListObj *pAnchor,int prio)
{
	mPriority = GHMGCLIST_ANCHOR_PRIO;
	if( pAnchor==NULL ){
		// AJ[ɂȂ
		mpPrev = NULL;
		mpNext = NULL;
	}else{
		if( prio < 0 ){
			prio = 0;
		}
		ghmGcListObj::Join(pAnchor,prio);
	}
}

ghmGcListObj::~ghmGcListObj()
{
	if( mPriority < 0 ){
		ghmGcListObj	*pObj;
		ghmGcListObj	*pNext;
		for(pObj=mpNext ; pObj!=NULL ; pObj=pNext){
			pNext = pObj->mpNext;
			delete pObj;
		}
	}else{
		Remove();
	}
}

bool ghmGcListObj::LinkPrevToNext(void)
{
	if( mpNext!=NULL ){
		mpNext->mpPrev = mpPrev;
	}
	if( mpPrev!=NULL ){
		mpPrev->mpNext = mpNext;
	}
	return true;
}

bool ghmGcListObj::ChangePriority(int prio)
{
	ghmGcListObj	*pObj;
	if( prio < 0 ){
		return false;
	}
	if( prio==mPriority ){
		// ς炸
		return true;
	}
	if( prio < mPriority ){
		// ʂ̃vCIeBɕύX
		pObj = mpPrev;
		while( pObj->mPriority >= prio ){
			pObj = pObj->mpPrev;
			ASSERTMSG1(pObj!=NULL,"ghmGcListObj change priority error : no anchor 0x%08x",(int)this);
		}
		LinkPrevToNext();
		mpNext = pObj->mpNext;
		mpPrev = pObj;
	}else{
		// ʂ̃vCIeBɕύX
		if( (pObj = mpNext)==NULL ){
			// ԉʂ̂߁Aς炸
			return true;
		}
		while( pObj->mPriority < prio ){
			if( pObj->mpNext==NULL ){
				LinkPrevToNext();
				mpNext         = NULL;
				mpPrev         = pObj;
				mpPrev->mpNext = this;
				return true;
			}
			pObj = pObj->mpNext;
		}
		LinkPrevToNext();
		mpNext = pObj;
		mpPrev = pObj->mpPrev;
	}
	mPriority = prio;
	if( mpNext != NULL ){
		mpNext->mpPrev = this;
	}
	mpPrev->mpNext = this;
	return true;
}

int ghmGcListObj::GetIndex(void)
{
	ghmGcListObj	*pObj;
	if( (pObj = mpPrev)==NULL ){
		// AJ[
		return -1;
	}
	int	index = 0;
	while( pObj->mpPrev!=NULL ){
		index++;
		pObj = pObj->mpPrev;
	}
	return index;
}

int ghmGcListObj::GetIndexFromLast(void)
{
	ghmGcListObj	*pObj;
	int				index = 0;
	pObj = mpNext;
	while( pObj!=NULL ){
		index++;
		pObj = pObj->mpNext;
	}
	return index;
}

bool ghmGcListObj::Join(ghmGcListObj *pAnchor,int prio)
{
	if( pAnchor==NULL ){
		return false;
	}
	if( prio < 0 || mPriority >= 0 ){
		return false;
	}
	ghmGcListObj	*pObj = pAnchor;
	mPriority = prio;
	while( pObj->mPriority < prio ){
		if( pObj->mpNext==NULL ){
			mpNext         = NULL;
			mpPrev         = pObj;
			mpPrev->mpNext = this;
			return true;
		}
		pObj = pObj->mpNext;
	}
	mpNext         = pObj;
	mpPrev         = pObj->mpPrev;
	mpNext->mpPrev = this;
	mpPrev->mpNext = this;
	return true;
}

bool ghmGcListObj::Remove(void)
{
	if( mPriority < 0 ){
		return false;
	}
	if( mpPrev != NULL ){
		mpPrev->mpNext = mpNext;
	}else{
		ASSERTMSG2(mpNext==NULL,"ghmGcListObj destructor error : anchor release 0x%08x next 0x%08x",(int)this,(int)mpNext);
	}
	if( mpNext != NULL ){
		mpNext->mpPrev = mpPrev;
	}
	mPriority = GHMGCLIST_ANCHOR_PRIO;
	mpPrev    = NULL;
	mpNext    = NULL;
	return true;
}

