////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Mathematical Manager
//
// ghmGcMath.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 11/16/2002
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCMATH_CPP__

#include "ghmGcMath.h"

void ghmQuaternion::QuatSlerp(ghmQuaternion *pq1,ghmQuaternion *pq2,float t)
{
	
	float	fCosTheta;
	float	_t;
	
	// ]xNg̓(cos = Im(q1) * Im(q2))
	fCosTheta = pq1->x * pq2->x + pq1->y * pq2->y + pq1->z * pq2->z + pq1->w * pq2->w;
	
	// ϓpx`FbNA90xȏȂ甽]
	if( fCosTheta < 0.0f ){
		fCosTheta = -fCosTheta;
		pq2->x = -pq2->x;
		pq2->y = -pq2->y;
		pq2->z = -pq2->z;
		pq2->w = -pq2->w;
	}
	
	// q1q2܂ł̎Ԃ1.0ɐKĂ̂ŁA͈̔͂Ŏԕϓ
	if( 1.0f - fCosTheta > 0.001f ){
		
		float fTheta = acosf(fCosTheta);
		float fSin   = sinf (fTheta);
		
		_t = sinf( fTheta * (1.0f - t) ) / fSin;
		t  = sinf( fTheta *         t  ) / fSin;
		
	}else{
		
		_t = 1.0f - t;
	}
	
	// ԂꂽNEH[^jIZo
	x = _t * pq1->x + t * pq2->x;
	y = _t * pq1->y + t * pq2->y;
	z = _t * pq1->z + t * pq2->z;
	w = _t * pq1->w + t * pq2->w;
	
}

f32		ghmMath::projection[GX_PROJECTION_SZ];
f32		ghmMath::viewport  [GX_VIEWPORT_SZ  ];

bool ghmMath::TransPosition(Mtx mtx,Vec &wpos,Vec &spos)
{
	
	GXGetProjectionv(projection);
	GXGetViewportv  (viewport);
	GXProject(wpos.x,
	          wpos.y,
	          wpos.z,
	          mtx,
	          projection,
	          viewport,
	          &spos.x,
	          &spos.y,
	          &spos.z);
	
	return true;
};

int ghmMath::Roll2UpVec(const Vec &pos,Vec &interest,float roll,Vec &up)
{
	Vec		eye;
	VECSubtract(&interest,&pos,&eye);
	VECNormalize(&eye,&eye);
	return Roll2UpVec(eye,roll,up);
}

int ghmMath::Roll2UpVec(const Vec &eye,float roll,Vec &up)
{
	Vec		v0,v1;
	Mtx		m;
	
	if( 1.0f - fabsf(eye.y) < 1.0e-4 ){
		v0.x = sqrtf(eye.y * eye.y + eye.z * eye.z);
		v0.y = -eye.x / v0.x * eye.y;
		v0.z = -eye.x / v0.x * eye.z;
	}else{
		v0.y = sqrtf(eye.x * eye.x + eye.z * eye.z);
		v0.x = -eye.y / v0.y * eye.x;
		v0.z = -eye.y / v0.y * eye.z;
	}
	
	MTXRotAxisRad(m,&eye,-roll);
	MTXMultVecSR (m,&v0,&v1);
	VECNormalize (&v1,&up);
	
	return 1;
}

