////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Module
//
// ghmGcModule.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 04/17/2004
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCMODULE_CPP__

#include "ghmGcModule.h"
#include <string.h>

ghmGcListObj		ghmGcModule::mAnchor;
#ifdef	_DEBUG
ghmGcFile*			ghmGcModule::mpDebugFile   = NULL;
int					ghmGcModule::mDebugProc    = -1;
void*				ghmGcModule::mpStringTable = NULL;
#endif

ghmGcModule::ghmGcModule(const char *pFileName,void *pLoadAddr,int size) : ghmGcListObj(&mAnchor)
{
	if( (mWorkSize = size) <= 0 ){
		mpModuleHeader = NULL;
	}else{
		mpModuleHeader = (OSModuleHeader *)pLoadAddr;
	}
	mpFile         = NULL;
	mRefCount      = 0;
	mpBss          = NULL;
	if( pFileName==NULL ){
		mState = IDLE;
		memset(mFileName,0,sizeof(mFileName));
	}else{
		mState = INIT;
		strncpy(mFileName,pFileName,sizeof(mFileName)-1);
	}
}

ghmGcModule::~ghmGcModule()
{
	if( mWorkSize <= 0 ){
		GHMGCMEM_SAFE_FREE(mpModuleHeader);
	}
	GHM_SAFE_DELETE   (mpBss         );
	GHM_SAFE_DELETE   (mpFile        );
}

void ghmGcModule::IncRefCount(void)
{
	mRefCount++;
}

void ghmGcModule::DecRefCount(void)
{
	mRefCount--;
}

ghmGcModule* ghmGcModule::Create(const char *pModuleName,void *pLoadAddr,int size)
{
	if( pModuleName==NULL ){
		return NULL;
	}
	ghmGcModule	*pModule;
	if( (pModule=Find(pModuleName))!=NULL ){
		pModule->IncRefCount();
		return pModule;
	}
	ghmGcMemSetAllocName("Module");
	if( (pModule=new ghmGcModule(pModuleName,pLoadAddr,size))==NULL ){
		return NULL;
	}
	pModule->IncRefCount();
	return pModule;
}

bool ghmGcModule::Release(void)
{
	DecRefCount();
	if( GetRefCount() > 0 ){
		return false;
	}
	if( mpModuleHeader!=NULL ){
		((void (*)(void))mpModuleHeader->epilog)();
		OSUnlink(&mpModuleHeader->info);
	}
	GHM_SAFE_DELETE   (mpBss         );
	if( mWorkSize <= 0 ){
		GHMGCMEM_SAFE_FREE(mpModuleHeader);
	}
	return true;
}

bool ghmGcModule::Release(ghmGcModule *pModule)
{
	if( pModule==NULL ){
		return false;
	}
	if( IsValid(pModule)==false ){
		return false;
	}
	if( pModule->Release()==true ){
		delete pModule;
	}
	return true;
}

bool ghmGcModule::ReleaseAll(void)
{
	ghmGcModule		*pModule,*pNext;
	for(pModule=(ghmGcModule *)mAnchor.GetNext() ; pModule!=NULL ; pModule=pNext){
		pNext = pModule->GetNext();
		while( pModule->GetRefCount() > 1 ){
			pModule->DecRefCount();
		}
		Release(pModule);
	}
	return true;
}

bool ghmGcModule::IsValid(ghmGcModule *pModule)
{
	ghmGcModule	*pModuleTmp;
	for(pModuleTmp=(ghmGcModule *)mAnchor.GetNext() ; pModuleTmp!=NULL ; pModuleTmp=pModuleTmp->GetNext()){
		if( pModule==pModuleTmp ){
			return true;
		}
	}
	return false;
}

ghmGcModule* ghmGcModule::Find(const char *pFileName)
{
	if( pFileName==NULL ){
		return NULL;
	}
	ghmGcModule	*pModule;
	for(pModule=(ghmGcModule *)mAnchor.GetNext() ; pModule!=NULL ; pModule=pModule->GetNext()){
		if( strcmp(pModule->GetFileName(),pFileName)==0 ){
			return pModule;
		}
	}
	return NULL;
}

bool ghmGcModule::FrameProcessAll(void)
{
#ifdef	_DEBUG
	switch( mDebugProc ){
	case	0:
		if( mpDebugFile==NULL ){
			break;
		}
		if( mpDebugFile->GetState()!=ghmGcFile::READFINISH ){
			break;
		}
		mDebugProc = -1;
		OSSetStringTable(mpStringTable);
		break;
	default:
		break;
	}
#endif
	ghmGcModule	*pModule,*pNext;
	for(pModule=(ghmGcModule *)mAnchor.GetNext() ; pModule!=NULL ; pModule=pNext){
		pNext = pModule->GetNext();
		if( pModule->FrameProcess()==false ){
			delete pModule;
		}
	}
	return true;
}

bool ghmGcModule::RenderProcessAll(void)
{
	ghmGcModule	*pModule;
	for(pModule=(ghmGcModule *)mAnchor.GetNext() ; pModule!=NULL ; pModule=pModule->GetNext()){
		pModule->RenderProcess();
	}
	return true;
}

bool ghmGcModule::IsInitialized(void)
{
#ifdef	_DEBUG
	if( mDebugProc >= 0 ){
		return false;
	}
#endif
	return true;
}

void* ghmGcModule::InitializeAll(const char *pStringTblFileName)
{
	if( pStringTblFileName==NULL ){
		return NULL;
	}
#ifdef	_DEBUG
	if( mpDebugFile==NULL ){
		ghmGcMemSetAllocName("ModuleFl");
		if( (mpDebugFile = new ghmGcFile)==NULL ){
			return NULL;
		}
	}
	if( mpDebugFile->Open((char *)pStringTblFileName)==false ){
		return NULL;
	}
	if( (mpStringTable=mpDebugFile->Read())==NULL ){
		return NULL;
	}
	mDebugProc = 0;
	return mpStringTable;
#else
	return NULL;
#endif
}

bool ghmGcModule::TerminateAll(void)
{
#ifdef	_DEBUG
	GHMGCMEM_SAFE_FREE(mpStringTable);
#endif
	OSSetStringTable(NULL);
	return true;
}

bool ghmGcModule::FrameProcess(void)
{
	switch( mState ){
	case	INIT:
//		OSDebugReport("module init\n");
		if( mpFile==NULL ){
			ghmGcMemSetAllocName("ModuleFL");
			if( (mpFile=new ghmGcFile)==NULL ){
				break;
			}
		}
		if( mpFile->Open(mFileName)==false ){
			break;
		}
		mState = LOAD;
	case	LOAD:
//		OSDebugReport("module load : size=%d,addr=%08x\n",mWorkSize,(u32)mpModuleHeader);
		if( mWorkSize <= 0 ){
			if( (mpModuleHeader = (OSModuleHeader *)mpFile->Read())==NULL ){
				break;
			}
		}else{
			if( mWorkSize < mpFile->GetLength() ){
				OSDebugReport("Module Work Size Over !! : %08x(%08x)\n",mWorkSize,mpFile->GetLength());
				break;
			}
			if( mpFile->Read(mpModuleHeader)==NULL ){
				break;
			}
		}
		mState = LOADING;
//		OSDebugReport("module loading init\n");
		break;
	case	LOADING:
//		OSDebugReport("module loading\n");
		if( mpFile->GetState()!=ghmGcFile::READFINISH ){
			break;
		}
		mState = ALLOCBSS;
		break;
	case	ALLOCBSS:
//		OSDebugReport("module allocbss : size=%d\n",mpModuleHeader->bssSize);
//		OSDebugReport("module allocbss : prolog=%08x\n",(u32)mpModuleHeader->prolog);
		ghmGcMemSetAllocName("ModulBSS");
		if( (mpBss = new u8[(int)mpModuleHeader->bssSize])==NULL ){
			break;
		}
//		OSDebugReport("module loaded : bss=%08x\n",(u32)mpBss);
		mState = LOADED;
//		OSDebugReport("oslink\n");
		OSLink(&mpModuleHeader->info,mpBss);
//		OSDebugReport("prolog=%08x\n",(u32)mpModuleHeader->prolog);
		((void (*)(void))mpModuleHeader->prolog)(); 
//		OSDebugReport("prolog end\n");
		break;
	case	LOADED:
		break;
	case	RELEASE:
	case	IDLE:
	default:
		return false;
	}
	return true;
}

bool ghmGcModule::RenderProcess(void)
{
	return true;
}

