////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Performance Meter
//
// ghmGcPerf.h
//
// Creator       : Fujita Yoshikazu
// Modified      : Watanabe Kazuhisa (GHM)
// Creation Date : 11/20/2002
//
////////////////////////////////////////////////////////////////////////////////

#ifndef	__GHMGCPERF_H__
#define	__GHMGCPERF_H__

#include <dolphin/perf.h>
#include "ghmGcTypes.h"

#pragma cpp_extensions	on

class ghmGcSprite{
protected:
	static bool SetRenderMode(void);
public:
	short						mLeft,			// 
								mTop,			// 
								mWidth,			// 
								mHeight;		// 
	short						mLayer;			// C
	static bool RenderStart(void);
	static bool RenderEnd(void);
	
	inline virtual bool Render(void){ return true; }
	
	ghmGcSprite(){}
	virtual ~ghmGcSprite(){}
};

class ghmGcSpriteF : public ghmGcSprite{
protected:
public:
	union{
		GXColor					mColor;
		u32						mColor32;
	};
	
	virtual bool Render(void);
	
	ghmGcSpriteF(){}
	virtual ~ghmGcSpriteF(){}
};

struct ghmGcPerfMetric{
	
//	s32							mTime;
	
	u32							mCpReq;
	u32							mTcReq;
	u32							mCpuRdReq;
	u32							mCpuWrReq;
	u32							mDspReq;
	u32							mIoReq;
	u32							mViReq;
	u32							mPeReq;
	u32							mRfReq;
	u32							mFiReq;
	
	u32							mDiffCpReq;
	u32							mDiffTcReq;
	u32							mDiffCpuRdReq;
	u32							mDiffCpuWrReq;
	u32							mDiffDspReq;
	u32							mDiffIoReq;
	u32							mDiffViReq;
	u32							mDiffPeReq;
	u32							mDiffRfReq;
	u32							mDiffFiReq;
	
	u32							mTopPixelsIn;
	u32							mTopPixelsOut;
	u32							mBotPixelsIn;
	u32							mBotPixelsOut;
	u32							mClrPixelsIn;
	u32							mCopyClks;
	
	u32							mDiffTopPixelsIn;
	u32							mDiffTopPixelsOut;
	u32							mDiffBotPixelsIn;
	u32							mDiffBotPixelsOut;
	u32							mDiffClrPixelsIn;
	u32							mDiffCopyClks;
	
	u32							mGp0Cnt;
	u32							mGp1Cnt;
	
	u32							mDiffGp0Cnt;
	u32							mDiffGp1Cnt;
	
	u32							mPmcInstruction;
	u32							mPmcLoadStore;
	u32							mPmcLlMissCycle;
	u32							mPmcCycle;
	
	u32							mDiffPmcInstruction;
	u32							mDiffPmcLoadStore;
	u32							mDiffPmcLlMissCycle;
	u32							mDiffPmcCycle;
	
};

struct ghmGcPerfEventMetric{
	u32							mStartCycle;
	u32							mEndCycle;
	u32							mCycle;
};

struct ghmGcPerfMetricInfo{
	ghmGcPerfMetric				*mpMetricTop;
	ghmGcPerfMetric				*mpMetricLast;
	int							mMetricCnt;
	ghmGcPerfEventMetric		mProcEvent;
	ghmGcPerfEventMetric		mRenderEvent;
	u32							mTotalTime;
//	u32							mProcTime;
//	u32							mRednerTime;
};

#define	GHMGCPERF_ENABLE		(0x80000000)
#define	GHMGCPERF_PROC			(0x40000000)
#define	GHMGCPERF_RENDER		(0x20000000)
#define	GHMGCPERF_START			(0x10000000)
#define	GHMGCPERF_VISIBLE		(0x08000000)
#define	GHMGCPERF_DUMP			(0x04000000)
#define	GHMGCPERF_SETALARM		(0x02000000)

class ghmGcPerf{
public:
	typedef	void* AllocFunc(u32 size);
	typedef	void FreeFunc(void *ptr);
protected:
	
	union{
		struct{
			u_int						mEnable  : 1;
			u_int						mProc    : 1;
			u_int						mRender  : 1;
			u_int						mStart   : 1;
			u_int						mVisible : 1;
			u_int						mDump    : 1;
			u_int						mSetAlarm: 1;
			u_int						mFlagPad :25;
		};
		u_int							mFlag;
	};
	
	static volatile ghmGcPerfMetricInfo	mMetricInfo[2];
	static volatile int					mMetricIndex;
	static u32							mProcEventCycleMax;
	static u32							mRenderEventCycleMax;
	static OSAlarm						mAlarm;
	static OSTime						mAlarmStartTime;
//	static OSStopwatch					mSwProc;
//	static OSStopwatch					mSwRender;
	
	static float						mSamplingTime;
	
	static AllocFunc					*mpAllocFunc;
	static FreeFunc						*mpFreeFunc;
public:
	ghmGcPerf();
	virtual ~ghmGcPerf();
	
	bool Initialize(void);
	bool Terminate(void);
	
	bool Start(void);
	bool End(void);
	bool StartProc(void);
	bool EndProc(void);
	bool StartRender(void);
	bool EndRender(void);
	
	bool Dump(void);
	bool Draw(void);
	
	void SetEnable(bool flag);
	inline bool GetEnable(void){ return mEnable? true : false; }
	
	void SetVisible(bool flag);
	inline bool GetVisible(void){ return mVisible? true : false; }
	
	inline void SetAllocFunc(AllocFunc *pFunc){ mpAllocFunc=pFunc; }
	inline AllocFunc* GetAllocFunc(void){ return mpAllocFunc; }
	inline void SetFreeFunc(FreeFunc *pFunc){ mpFreeFunc=pFunc; }
	inline FreeFunc* GetFreeFunc(void){ return mpFreeFunc; }
	
	static void* Alloc(u32 size);
	static bool Free(void *ptr);
	
	static bool ResetMetric(void);
	static inline u32 GetProcEventCycleMax(void){ return mProcEventCycleMax; }
	static inline u32 GetRenderEventCycleMax(void){ return mRenderEventCycleMax; }
	static volatile inline ghmGcPerfMetricInfo* GetCurrentMetricInfo(void){ return mMetricInfo + mMetricIndex; }
	static volatile inline int GetCurrentMetricIndex(void){ return mMetricIndex; }
	static inline OSTime GetAlarmStartTime(void){ return mAlarmStartTime; }
	static inline bool SetSamplingTime(float time){ mSamplingTime = time; return true; }
	static inline float GetSamplingTime(void){ return mSamplingTime; }
};

#endif	//__GHMGCPERF_H__

