////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Resource Link
//
// ghmGcResLink.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 04/03/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCRESLINK_CPP__

#include "ghmGcResLink.h"

ghmGcResStrTable*	ghmGcResGroup::mpStrTable = NULL;

void* ghmGcResGroup::operator new(u32,void* p)
{
	u32					*pMagicNo = (u32 *)p;
	ghmGcResGroup		*pGroup   = (ghmGcResGroup *)p;
	mpStrTable = NULL;
	if( *pMagicNo==RESLINK_MAGIC_NO ){
		pGroup->Initialize(false);
		return p;
	}
	if( *pMagicNo==RESLINK_MAGIC_NO_LE ){
		// gGfBA`rbNGfBAɕϊ
		pGroup->Initialize(true);
		return p;
	}
	return NULL;
}

void ghmGcResAttr::Initialize(bool convendian)
{
	if( convendian==true ){
		mOffset        = CONV_U32_ENDIAN(mOffset);
		mSize          = CONV_U32_ENDIAN(mSize  );
		mAttr          = CONV_U32_ENDIAN(mAttr  );
		if( mVersion!=0 ){
			mVersion       = CONV_U32_ENDIAN(mVersion  );
			mpResourceName = (const char *)CONV_U32_ENDIAN(mpResourceName);
		}
	}
	if( ghmGcResGroup::mpStrTable!=NULL && mVersion != 0 ){
		mpResourceName = ghmGcResGroup::mpStrTable->GetString((u32)mpResourceName);
	}
}

const char* ghmGcResAttr::GetResourceName(void) const
{
	if( mVersion==0 ){
		return NULL;
	}
	return mpResourceName;
}

void ghmGcResGroup::Initialize(bool convendian)
{
	int				i;
	ghmGcResAttr	*pAttr;
	ghmGcResGroup	*pGroup;
	if( convendian==true ){
		mMagicNo     = CONV_U32_ENDIAN(mMagicNo    );
		mResourceNum = CONV_U32_ENDIAN(mResourceNum);
		mAttrOffset  = CONV_U32_ENDIAN(mAttrOffset );
		if( mVersion!=0 ){
			mVersion      = CONV_U32_ENDIAN(mVersion    );
			mpStrTablePtr = (ghmGcResStrTable *)CONV_U32_ENDIAN(mpStrTablePtr);
		}
	}
	if( mVersion != 0 ){
		if( mpStrTablePtr != NULL ){
			mpStrTablePtr = (ghmGcResStrTable *)((u32)mpStrTablePtr + (u32)this);
			mpStrTable = mpStrTablePtr;
			if( convendian==true ){
				mpStrTable->mNum       = CONV_U32_ENDIAN(mpStrTable->mNum      );
				mpStrTable->mppStrings = (const char **)CONV_U32_ENDIAN(mpStrTable->mppStrings);
			}
			mpStrTable->mppStrings = (const char **)((u32)mpStrTable->mppStrings + (u32)mpStrTable);
			int			i;
			if( convendian==true ){
				for(i=0 ; i<mpStrTable->mNum ; i++){
					mpStrTable->mppStrings[i] = (const char *)CONV_U32_ENDIAN(mpStrTable->mppStrings[i]);
				}
			}
			for(i=0 ; i<mpStrTable->mNum ; i++){
				mpStrTable->mppStrings[i] = (const char *)((u32)mpStrTable->mppStrings[i] + (u32)mpStrTable);
			}
		}
	}
	int		attrsize;
	pAttr = (ghmGcResAttr *)((u32)this + mAttrOffset);
	switch( mVersion ){
	case	0:
		attrsize = 16;
		break;
	default:
		attrsize = sizeof(ghmGcResAttr);
		break;
	}
	for(i=0 ; i<mResourceNum ; i++,pAttr=(ghmGcResAttr *)((u32)pAttr+attrsize)){
		pAttr->Initialize(convendian);
		if( (pGroup = GetGroup(i))==NULL ){
			continue;
		}
		pGroup->Initialize(convendian);
	}
}

const ghmGcResAttr* ghmGcResGroup::GetFileAttr(int index)
{
	if( index < 0 || index >= mResourceNum ){
		return NULL;
	}
	if( mVersion==0 ){
		return (ghmGcResAttr *)((u32)this + mAttrOffset + index * 16);
	}
	return (ghmGcResAttr *)((u32)this + mAttrOffset) + index;
}

bool ghmGcResGroup::IsGroup(int index)
{
	const ghmGcResAttr	*pAttr;
	if( (pAttr=GetFileAttr(index))==NULL ){
		return false;
	}
	return pAttr->IsGroup();
}

bool ghmGcResGroup::IsNoInstance(int index)
{
	const ghmGcResAttr	*pAttr;
	if( (pAttr=GetFileAttr(index))==NULL ){
		return false;
	}
	return pAttr->IsNoInstance();
}

u32 ghmGcResGroup::GetResourceSize(int index)
{
	const ghmGcResAttr	*pAttr;
	if( (pAttr=GetFileAttr(index))==NULL ){
		return 0;
	}
	return pAttr->GetSize();
}

void* ghmGcResGroup::GetResource(int index)
{
	const ghmGcResAttr	*pAttr;
	if( (pAttr=GetFileAttr(index))==NULL ){
		return NULL;
	}
	if( pAttr->GetOffset()==0 || pAttr->GetSize()==0 ){
		return NULL;
	}
	return (void *)((u32)this + pAttr->GetOffset());
}

ghmGcResGroup* ghmGcResGroup::GetGroup(int index)
{
	const ghmGcResAttr	*pAttr;
	if( (pAttr=GetFileAttr(index))==NULL  ||
	     pAttr->IsNoInstance()    ==true  ||
	     pAttr->IsGroup     ()    ==false ){
		return NULL;
	}
	return (ghmGcResGroup *)((u32)this + pAttr->GetOffset());
}

const char* ghmGcResGroup::GetResourceName(int index)
{
	const ghmGcResAttr	*pAttr;
	if( (pAttr=GetFileAttr(index))==NULL ){
		return NULL;
	}
	return pAttr->GetResourceName();
}

