////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Task
//
// ghmGcTask.h
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 02/04/2003
//
////////////////////////////////////////////////////////////////////////////////

#ifndef	__GHMGCTASK_H__
#define	__GHMGCTASK_H__

#pragma cplusplus			on
#pragma cpp_extensions		on

#include "ghmGcTypes.h"
#include "ghmGcList.h"
#include "ghmGcMem.h"

class ghmGcTask;
typedef	void ghmGcTaskHandler(ghmGcTask *pTask);

////////////////////////////////////////////////////////////////////////////////
// XbhΉ
////////////////////////////////////////////////////////////////////////////////

#ifdef	_DEBUG
#define	GHMGCTASK_DEBUG
#endif

//
// Task
//

// TaskState
#define	GHMGCTASK_IDLE					(0x00000000)
#define	GHMGCTASK_RUN					(0x00000001)
#define	GHMGCTASK_READY					(0x00000002)
#define	GHMGCTASK_SUSPEND				(0x00000004)
#define	GHMGCTASK_WAIT					(0x00000008)
#define	GHMGCTASK_INVALID_STATE			(0xffffffff)

// Save GPR Register
#define	GHMGCTASK_SAVE_GPR14			(0)
#define	GHMGCTASK_SAVE_GPR15			(1)
#define	GHMGCTASK_SAVE_GPR16			(2)
#define	GHMGCTASK_SAVE_GPR17			(3)
#define	GHMGCTASK_SAVE_GPR18			(4)
#define	GHMGCTASK_SAVE_GPR19			(5)
#define	GHMGCTASK_SAVE_GPR20			(6)
#define	GHMGCTASK_SAVE_GPR21			(7)
#define	GHMGCTASK_SAVE_GPR22			(8)
#define	GHMGCTASK_SAVE_GPR23			(9)
#define	GHMGCTASK_SAVE_GPR24			(10)
#define	GHMGCTASK_SAVE_GPR25			(11)
#define	GHMGCTASK_SAVE_GPR26			(12)
#define	GHMGCTASK_SAVE_GPR27			(13)
#define	GHMGCTASK_SAVE_GPR28			(14)
#define	GHMGCTASK_SAVE_GPR29			(15)
#define	GHMGCTASK_SAVE_GPR30			(16)
#define	GHMGCTASK_SAVE_GPR31			(17)
#define	GHMGCTASK_SAVE_GPRNUM			(18)

#define	GHMGCTASK_SAVE_FPR14			(0)
#define	GHMGCTASK_SAVE_FPR15			(1)
#define	GHMGCTASK_SAVE_FPR16			(2)
#define	GHMGCTASK_SAVE_FPR17			(3)
#define	GHMGCTASK_SAVE_FPR18			(4)
#define	GHMGCTASK_SAVE_FPR19			(5)
#define	GHMGCTASK_SAVE_FPR20			(6)
#define	GHMGCTASK_SAVE_FPR21			(7)
#define	GHMGCTASK_SAVE_FPR22			(8)
#define	GHMGCTASK_SAVE_FPR23			(9)
#define	GHMGCTASK_SAVE_FPR24			(10)
#define	GHMGCTASK_SAVE_FPR25			(11)
#define	GHMGCTASK_SAVE_FPR26			(12)
#define	GHMGCTASK_SAVE_FPR27			(13)
#define	GHMGCTASK_SAVE_FPR28			(14)
#define	GHMGCTASK_SAVE_FPR29			(15)
#define	GHMGCTASK_SAVE_FPR30			(16)
#define	GHMGCTASK_SAVE_FPR31			(17)
#define	GHMGCTASK_SAVE_FPRNUM			(18)

#define	GHMGCTASK_DEFAULT_STACK_SIZE	(0x4000)					// ftHgX^bNTCY 16Kbyte
//#define	GHMGCTASK_DEFAULT_STACK_SIZE	(0x1000)					// ftHgX^bNTCY 4Kbyte

#define	GHMGCTASK_NONUNIQUE_ID			(-1)
#define	GHMGCTASK_STACKALLOC_ERRORID	(-2)

#define	GHMGCTASK_FPR_ADDR				(4 * sizeof(u32))
#define	GHMGCTASK_GPR_ADDR				(GHMGCTASK_FPR_ADDR   + GHMGCTASK_SAVE_FPRNUM * sizeof(f64))
#define	GHMGCTASK_CR_ADDR				(GHMGCTASK_GPR_ADDR   + GHMGCTASK_SAVE_GPRNUM * sizeof(u32))
#define	GHMGCTASK_SP_ADDR				(GHMGCTASK_CR_ADDR    + sizeof(u32))
#define	GHMGCTASK_PC_ADDR				(GHMGCTASK_SP_ADDR    + sizeof(u32))
#define	GHMGCTASK_LR_ADDR				(GHMGCTASK_PC_ADDR    + sizeof(u32))

#define	GHMGCTASK_STATE_ADDR			(GHMGCTASK_LR_ADDR    + sizeof(u32))
#define	GHMGCTASK_CONTROLLER_ADDR		(GHMGCTASK_STATE_ADDR + sizeof(u32))

#define	GHMGCTASK_SAVED_TASKPTR_REGID	(GHMGCTASK_SAVE_GPR14)
#define	GHMGCTASK_SAVED_TASKPTR_REG		r14

//                                         T A S K
#define	GHMGCTASK_MAGICNO				(0x5441534B)

class ghmGcTaskController;
class ghmGcTask : public ghmGcListObj{
	friend class ghmGcTaskController;
private:
	// WX^ޔ[N
	f64								mFPR[GHMGCTASK_SAVE_FPRNUM];	// Saved FPRWX^
	u32								mGPR[GHMGCTASK_SAVE_GPRNUM];	// Saved GPRWX^
	u32								mCR;							// CRWX^
	u32								mSP;							// X^bN|C^
	u32								mPC;							// vOJE^
	u32								mLR;							// NWX^
	
	u32								mState;
	ghmGcTaskController				*mpController;
	int								mID;
	ghmGcTaskHandler				*mpHandler;
	ghmGcTaskHandler				*mpTermHandler;
	u32								mParam[2];
	
	union{
		struct{
			u32						mStackAlloc : 1;				// X^bN蓖
			u32						mFlagPad    :31;
		};
		u32							mFlag;							// tO
	};
	
	void							*mpStack;						// X^bN̈
	u32								mStackSize;						// X^bNTCY
#ifdef	GHMGCTASK_DEBUG
	char							mTaskName       [16];
	static char						mCurrentTaskName[16];
#endif
	
	static ghmGcMem					*mpAllocHeap;
	static ghmGcMem					*mpReserveAllocHeap;
	
	inline ghmGcTaskController* GetController(void){ return mpController; }
	
	bool Run(void);
	bool Ready(void);
	
#ifdef	_DEBUG
	bool Check(void);
#endif
	
	bool SetParam(u32 id,u32 param);
public:
	ghmGcTask(ghmGcListObj *pAnchor,
	          ghmGcTaskController *pController,
	          ghmGcTaskHandler *pHandler,
	          ghmGcTaskHandler *pTermHandler=NULL,
	          void *pStack=NULL,
	          int stackSize=-1,
	          int prio=-1);
	virtual ~ghmGcTask();
	
	// AP[Vq[v
	inline static bool SetAllocHeap(ghmGcMem *pHeap){ mpAllocHeap=pHeap; return true; }
	inline static ghmGcMem* GetAllocHeap(void){ return mpAllocHeap; }
	inline static bool SetReserveAllocHeap(ghmGcMem *pHeap){ mpReserveAllocHeap=pHeap; return true; }
	inline static ghmGcMem* GetReserveAllocHeap(void){ return mpReserveAllocHeap; }
	
	inline u32 GetState(void){ return mState; }
	inline int GetID(void){ return mID; }
	inline u32 GetParam(u32 id){ return (id>=sizeof(mParam)/sizeof(u32))? 0:mParam[id]; }
	
	int Start(ghmGcTaskHandler *pHandler,u32 param0=0,u32 param1=0,void *pStack=NULL,int stackSize=-1,int prio=-1);
	int Start(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0=0,u32 param1=0,void *pStack=NULL,int stackSize=-1,int prio=-1);
	bool Stop(int id);
	asm void RetSys(void);
	asm void Exit(void);
	bool Resume(void);
	bool Suspend(void);
	
#ifdef	GHMGCTASK_DEBUG
	const char* GetTaskName(void){ return mTaskName; }
	static void SetCurrentTaskName(const char *pName=NULL);
	static const char* GetCurrentTaskName(void){ return mCurrentTaskName; }
#endif
	
	ghmGcTask* GetNext(void){ return (ghmGcTask *)ghmGcListObj::GetNext(); }
	ghmGcTask* GetPrev(void){ return (ghmGcTask *)ghmGcListObj::GetPrev(); }
}ATTRIBUTE_ALIGN(32);

//
// Task Controller
//

#define	GHMGCTASKCONTROLLER_FPR_ADDR	(0)
#define	GHMGCTASKCONTROLLER_GPR_ADDR	(GHMGCTASKCONTROLLER_FPR_ADDR + GHMGCTASK_SAVE_FPRNUM * sizeof(f64))
#define	GHMGCTASKCONTROLLER_CR_ADDR		(GHMGCTASKCONTROLLER_GPR_ADDR + GHMGCTASK_SAVE_GPRNUM * sizeof(u32))
#define	GHMGCTASKCONTROLLER_SP_ADDR		(GHMGCTASKCONTROLLER_CR_ADDR  + sizeof(u32))
#define	GHMGCTASKCONTROLLER_HI_ADDR		(GHMGCTASKCONTROLLER_SP_ADDR  + sizeof(u32))

struct ghmGcTaskRegister{
	f64								mFPR[GHMGCTASK_SAVE_FPRNUM];	// Saved FPRWX^
	u32								mGPR[GHMGCTASK_SAVE_GPRNUM];	// Saved GPRWX^
	u32								mCR;							// CRWX^
	u32								mSP;							// X^bN|C^
};

class ghmGcTaskController{
	friend class ghmGcTask;
private:
	// WX^ޔ
	f64								mFPR[GHMGCTASK_SAVE_FPRNUM];	// Saved FPRWX^
	u32								mGPR[GHMGCTASK_SAVE_GPRNUM];	// Saved GPRWX^
	u32								mCR;							// CRWX^
	u32								mSP;							// X^bN|C^
	
	ghmGcListObj					mTaskAnchor;
	int								mIDSeed;
	
	static ghmGcTask				*mpCurrentTask;					// Jg^XN
	
	static asm void StoreRegister2(void);
	static asm void LoadRegister2(void);
	asm void CallHandler2(register ghmGcTask *pTask,register ghmGcTaskRegister *pRegister);
	
	asm void StoreMainRegister(void);
	asm void LoadMainRegister(void);
	asm void StoreTaskRegister(void);
	asm void LoadTaskRegister(void);
	asm void CallHandler(register ghmGcTask *pTask);
	
public:
	ghmGcTaskController();
	virtual ~ghmGcTaskController();
	
	bool Initialize(void);
	bool FrameProcess(void);
	bool Terminate(void);
	
	int Start(ghmGcTaskHandler *pHandler,u32 param0=0,u32 param1=0,void *pStack=NULL,int stackSize=-1,int prio=-1);
	int Start(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0=0,u32 param1=0,void *pStack=NULL,int stackSize=-1,int prio=-1);
	int StartIM(ghmGcTaskHandler *pHandler,u32 param0=0,u32 param1=0,void *pStack=NULL,int stackSize=-1,int prio=-1);
	int StartIM(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0=0,u32 param1=0,void *pStack=NULL,int stackSize=-1,int prio=-1);
	bool Stop(int id);
	bool Resume(int id);
	bool Suspend(int id);
	u32 GetState(int id);
	
	ghmGcTask* GetTask(int id);
	
	int GetUniqueID(void);
	
	static inline ghmGcTask* GetCurrentTask(void){ return mpCurrentTask; }
	
}ATTRIBUTE_ALIGN(32);

#define	GHMGCTASK_START \
					ghmGcTaskController::GetCurrentTask()==NULL? -1 : ghmGcTaskController::GetCurrentTask()->Start

#define	GHMGCTASK_STOP(id) \
					(ghmGcTaskController::GetCurrentTask()==NULL? false : ghmGcTaskController::GetCurrentTask()->Stop(id))

#define	GHMGCTASK_RETSYS() \
					{ \
						ghmGcTask	*pTask = ghmGcTaskController::GetCurrentTask(); \
						if( pTask!=NULL ){ \
							pTask->RetSys(); \
						} \
					}
#define	retsys()	GHMGCTASK_RETSYS()

#define	GHMGCTASK_EXIT() \
					{ \
						ghmGcTask	*pTask = ghmGcTaskController::GetCurrentTask(); \
						if( pTask!=NULL ){ \
							pTask->Exit(); \
						} \
					}
#define	task_exit()	GHMGCTASK_EXIT()

#define	GHMGCTASK_GETPARAM(id)	(ghmGcTaskController::GetCurrentTask()==NULL? 0 : ghmGcTaskController::GetCurrentTask()->GetParam(id))
#define	task_get_param(id)	GHMGCTASK_GETPARAM(id)

#define	GHMGCTASK_HANDLER(func)					void func(ghmGcTask *)
#define	GHMGCTASK_HANDLER_USEARG(func,arg)		void func(ghmGcTask *arg)

#ifdef	GHMGCTASK_DEBUG
#define	GHMGCTASK_SETTASKNAME(name)				ghmGcTask::SetCurrentTaskName(name)
#else
#define	GHMGCTASK_SETTASKNAME(name)				
#endif

#endif	//__GHMGCTASK_H__

