////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube Thread
//
// ghmGcThread.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 05/17/2004
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCTHREAD_CPP__

#include "ghmGcThread.h"

ghmGcListObj	ghmGcThread::mAnchor;
ghmGcMem*		ghmGcThread::mpHeap = NULL;

bool ghmGcThread::Initialize(void)
{
	mpHeap = &ghmGcMem::mDefaultHeap;
	return true;
}

bool ghmGcThread::Terminate(void)
{
	ghmGcThread		*pThread;
	for(pThread=(ghmGcThread *)mAnchor.GetNext() ; pThread!=NULL ; pThread=(ghmGcThread *)mAnchor.GetNext()){
		delete pThread;
	}
	return true;
}

ghmGcMem* ghmGcThread::SetAllocHeap(ghmGcMem *pHeap)
{
	ghmGcMem	*pOldHeap = mpHeap;
	mpHeap = pHeap;
	return pOldHeap;
}

ghmGcThread::ghmGcThread(ghmGcThreadHandler *pHandler,u32 param,int prio,int stackSize,void *pStack) : ghmGcListObj(&mAnchor)
{
	if( prio < GHMGCTHREAD_PRIO_MIN ){
		mThreadPriority = GHMGCTHREAD_PRIO_DEFAULT;
	}else if( prio > GHMGCTHREAD_PRIO_MAX ){
		mThreadPriority = GHMGCTHREAD_PRIO_MAX;
	}else{
		mThreadPriority = prio;
	}
	mStackSize = stackSize;
	mParam     = param;
	mpHandler  = pHandler;
	if( pStack!=NULL ){
		mpStack = pStack;
		mStackAlloc = false;
	}else{
		if( stackSize < 0 ){
			stackSize = GHMGCTHREAD_STACKSIZE_DEFAULT;
		}
		mpStack = mpHeap->Alloc(stackSize);
		mStackAlloc = true;
	}
	if( mpStack==NULL ){
		return;
	}
    // Xbh
    OSCreateThread(&mThread,
                   ThreadHandler,
                   (void *)this,
                   (void *)((u32)mpStack + stackSize),
                   (u32)mStackSize,
                   mThreadPriority,
                   OS_THREAD_ATTR_DETACH);
}

ghmGcThread::~ghmGcThread()
{
	Cancel();
	if( mStackAlloc==true ){
		GHMGCMEM_SAFE_FREE(mpStack);
	}
}

bool ghmGcThread::IsCreated(void)
{
	if( mpStack==NULL ){
		return false;
	}
	return true;
}

bool ghmGcThread::SetThreadPriority(int prio)
{
	if( mpStack==NULL ){
		return false;
	}
	if( prio < GHMGCTHREAD_PRIO_MIN ){
		prio = GHMGCTHREAD_PRIO_DEFAULT;
	}else if( prio > GHMGCTHREAD_PRIO_MAX ){
		prio = GHMGCTHREAD_PRIO_MAX;
	}
	OSSetThreadPriority(&mThread,prio);
	return true;
}

bool ghmGcThread::Cancel(void)
{
	if( mpStack==NULL ){
		return false;
	}
	if( IsTerminated()==true ){
		return false;
	}
	OSCancelThread(&mThread);
	return true;
}

bool ghmGcThread::IsTerminated(void)
{
	if( mpStack==NULL ){
		return true;
	}
	if( OSIsThreadTerminated(&mThread)==FALSE ){
		return false;
	}
	return true;
}

bool ghmGcThread::IsSuspended(void)
{
	if( mpStack==NULL ){
		return true;
	}
	if( OSIsThreadSuspended(&mThread)==FALSE ){
		return false;
	}
	return true;
}

bool ghmGcThread::Suspend(void)
{
	if( mpStack==NULL ){
		return false;
	}
	if( IsSuspended()==true ){
		return false;
	}
    OSSuspendThread(&mThread);
	return true;
}

bool ghmGcThread::Resume(void)
{
	if( mpStack==NULL ){
		return false;
	}
	if( IsSuspended()==false ){
		return false;
	}
    OSResumeThread(&mThread);
	return true;
}

void* ghmGcThread::ThreadHandler(void *pParam)
{
	ghmGcThread		*pThread = (ghmGcThread *)pParam;
	void			*pResult = NULL;
	if( pThread->mpHandler!=NULL ){
		pResult = pThread->mpHandler((void *)pThread->mParam);
	}
	return pResult;
}

int ghmGcThread::GetThreadNum(void)
{
	ghmGcThread		*pThread;
	int				count;
	for(pThread=(ghmGcThread *)mAnchor.GetNext(),count=0 ; pThread!=NULL ; pThread=pThread->GetNext(),count++){}
	return count;
}

int ghmGcThread::GetActiveThreadNum(void)
{
	ghmGcThread		*pThread;
	int				count;
	for(pThread=(ghmGcThread *)mAnchor.GetNext(),count=0 ; pThread!=NULL ; pThread=pThread->GetNext()){
		if( pThread->IsTerminated()==true ){
			continue;
		}
		count++;
	}
	return count;
}

