////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube USB Service
//
// ghmGcUsb.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 11/28/2002
//
////////////////////////////////////////////////////////////////////////////////

#define	__GHMGCUSB_CPP__

#include <dolphin.h>
#include <string.h>

#include "ghmGcUsb.h"
#include "ghmGcDefine.h"

bool          ghmGcUsb::mFioInitialized = false;
bool          ghmGcUsb::mTtyInitialized = false;
#ifndef	USE_ISDOLVIEWER
volatile s32  ghmGcUsb::mExiChannel     = -1;
volatile bool ghmGcUsb::mMccInitialized = false;
MCCChannel    ghmGcUsb::mTtyChannel     = (MCCChannel)GHMGCUSB_TTYCHANNEL_DEF;
MCCChannel    ghmGcUsb::mFioChannel     = (MCCChannel)GHMGCUSB_FIOCHANNEL_DEF;
u8            ghmGcUsb::mFioBlocks      =             GHMGCUSB_FIOBLOCKS_DEF;
#endif

#ifdef	USE_ISDOLVIEWER
bool ghmGcUsb::InitializeISDOL(void)
{
	ISFileIsServerExist();
	ISFileInit();
	ISReportInit();
//	while( ISFileIsServerExist()==FALSE ){}
//	while( ISFileInit()==FALSE ){}
//	ISReportInit();
	return true;
}
#endif

ghmGcUsb::ghmGcUsb()
{
#ifdef	USE_ISDOLVIEWER
	mpDir  = NULL;
	mpFile = NULL;
#else
	mHandle = FIO_INVALID_HANDLE;
#endif
	memset(&mFindData,0,sizeof(mFindData));
	mState = CLOSED;
}

ghmGcUsb::~ghmGcUsb()
{
	_Terminate();
}

#ifndef	USE_ISDOLVIEWER
void ghmGcUsb::mccSystemCallback(MCCSysEvent event)
{
	if( event==MCC_SYSEVENT_INITIALIZED ){
		mMccInitialized = true;
	}
}

BOOL ghmGcUsb::hioEnumCallback(s32 chan)
{
	mExiChannel = chan;
	return FALSE;
}
#endif

bool ghmGcUsb::Initialize(void)
{
#ifdef	USE_ISDOLVIEWER
	mFioInitialized = true;
	mTtyInitialized = true;
#else
	if( mExiChannel < 0 ){
		if( MCCEnumDevices(hioEnumCallback)==FALSE ){
			// foCXȂ
			_Terminate();
			return false;
		}
		while( mExiChannel < 0 ){
			// P[uڑĂȂ
			VIWaitForRetrace();
		}
	}
	if( mMccInitialized==false ){
		if( MCCInit((MCCExiChannel)mExiChannel,0,mccSystemCallback)==FALSE ){
			// MCC̏Ɏs
			_Terminate();
			return false;
		}
		while( mMccInitialized==false ){
			VIWaitForRetrace();
		}
	}
	if( TTYInit((MCCExiChannel)mExiChannel,mTtyChannel)==FALSE ){
		// Can't start TTY Server.
		_Terminate();
		return false;
	}
	mTtyInitialized = true;
	if( FIOInit((MCCExiChannel)mExiChannel,mFioChannel,mFioBlocks)==FALSE ){
		_Terminate();
		return false;
	}
	mFioInitialized = true;
#endif
	return true;
}

bool ghmGcUsb::_Terminate(void)
{
	switch( mState ){
	case	FINDOPENED:
		FindClose();
		break;
	case	CLOSED:
		break;
	case	OPENED:
	default:
		Close();
		break;
	}
	return true;
}

bool ghmGcUsb::Terminate(void)
{
#ifdef	USE_ISDOLVIEWER
#else
	if( mFioInitialized==true ){
		FIOExit();
		mFioInitialized = false;
	}
	if( mTtyInitialized==true ){
		TTYExit();
		mTtyInitialized = false;
	}
	if( mMccInitialized==true ){
		MCCExit();
		mMccInitialized = false;
		mExiChannel = -1;
	}
#endif
	return true;
}

u32 ghmGcUsb::GetSize(const char *pfname)
{
	if( mFioInitialized==false || mState!=CLOSED ){
		return 0xffffffff;
	}
	
	u32			filesize;
#ifdef	USE_ISDOLVIEWER
	if( mpFile!=NULL ){
		return 0xffffffff;
	}
	if( (mpFile=ISFileOpen(pfname,"rb"))==NULL ){
		OSDebugReport("FileOpen failed %s\n",pfname);
		return 0xffffffff;
	}
	
	filesize = ISFileLength(mpFile);
	
	ISFileClose(mpFile);
	mpFile = NULL;
#else
	if( mHandle!=FIO_INVALID_HANDLE ){
		return 0xffffffff;
	}
	if( (mHandle=FIOFopen(pfname,FIO_OPEN_RDONLY))==FIO_INVALID_HANDLE ){
		return 0xffffffff;
	}
	
	filesize = FIOFseek(mHandle,0,FIO_SEEK_LAST);
	
	FIOFclose(mHandle);
	mHandle = FIO_INVALID_HANDLE;
#endif
	
	return filesize;
}

void* ghmGcUsb::Read(const char *pfname,void *pMem)
{
	if( mFioInitialized==false || mState!=CLOSED ){
		return NULL;
	}
	
	void		*pBuff;
	
#ifdef	USE_ISDOLVIEWER
	if( mpFile!=NULL ){
		return NULL;
	}
	
	if( (mpFile=ISFileOpen(pfname,"rb"))==NULL ){
		OSDebugReport("FileOpen failed %s\n",pfname);
		return NULL;
	}
	
	if( (mRequestSize = ISFileLength(mpFile))==0xffffffff ){
		ISFileClose(mpFile);
		mpFile = NULL;
		return NULL;
	}
	
	mRequestSize = OSRoundUp32B(mRequestSize);
	
	if( pMem==NULL ){
		if( (pBuff=ghmGcMemAlloc((int)mRequestSize))==NULL ){
			ISFileClose(mpFile);
			mpFile = NULL;
			return NULL;
		}
	}else{
		pBuff = pMem;
	}
	
	mSize = ISFileRead(pBuff,mRequestSize,mpFile);
	
	if( mSize!=mRequestSize ){
		if( pMem==NULL ){
			ghmGcMemFree(pBuff);
		}
		ISFileClose(mpFile);
		mpFile = NULL;
		return NULL;
	}
	
	ISFileClose(mpFile);
	mpFile = NULL;
	
#else
	if( mHandle!=FIO_INVALID_HANDLE ){
		return NULL;
	}
	if( (mHandle=FIOFopen(pfname,FIO_OPEN_RDONLY))==FIO_INVALID_HANDLE ){
		return NULL;
	}
	
	if( (mRequestSize=FIOFseek(mHandle,0,FIO_SEEK_LAST))==0xffffffff ){
		FIOFclose(mHandle);
		return NULL;
	}
	FIOFseek(mHandle,0,FIO_SEEK_TOP);
	
	if( pMem==NULL ){
		if( (pBuff=ghmGcMemAlloc((int)mRequestSize))==NULL ){
			FIOFclose(mHandle);
			mHandle = FIO_INVALID_HANDLE;
			return NULL;
		}
	}else{
		pBuff = pMem;
	}
	
	mSize = FIOFread(mHandle,pBuff,mRequestSize);
	
	if( mSize!=mRequestSize ){
		if( pMem==NULL ){
			ghmGcMemFree(pBuff);
		}
		FIOFclose(mHandle);
		mHandle = FIO_INVALID_HANDLE;
		return NULL;
	}
	
	FIOFclose(mHandle);
	mHandle = FIO_INVALID_HANDLE;
#endif
	
	return pBuff;
}

u32 ghmGcUsb::Write(const char *pfname,void *pMem,u32 size)
{
	if( mFioInitialized==false || mState!=CLOSED || pMem==NULL ){
		return NULL;
	}
	
#ifdef	USE_ISDOLVIEWER
	if( mpFile!=NULL ){
		return NULL;
	}
	
	if( (mpFile=ISFileOpen(pfname,"wb"))==NULL ){
		OSDebugReport("FileOpen failed %s\n",pfname);
		return NULL;
	}
	
	mRequestSize = size;
	mSize        = ISFileWrite(pMem,mRequestSize,mpFile);
	
	ISFileClose(mpFile);
	mpFile = NULL;
#else
	if( mHandle!=FIO_INVALID_HANDLE ){
		return NULL;
	}
	if( (mHandle=FIOFopen(pfname,FIO_OPEN_WRONLY))==FIO_INVALID_HANDLE ){
		return NULL;
	}
	
	mRequestSize = size;
	mSize        = FIOFwrite(mHandle,pMem,mRequestSize);
	
	FIOFclose(mHandle);
	mHandle = FIO_INVALID_HANDLE;
#endif
	
	return mSize;
}

void* ghmGcUsb::ReadAsync(const char *pfname,void *pMem)
{
	if( mFioInitialized==false || mState!=CLOSED ){
		return NULL;
	}
	
	void		*pBuff;
	
#ifdef	USE_ISDOLVIEWER
	if( mpFile!=NULL ){
		return NULL;
	}
	
	if( (mpFile=ISFileOpen(pfname,"rb"))==NULL ){
		OSDebugReport("FileOpen failed %s\n",pfname);
		return NULL;
	}
	
	if( (mRequestSize = ISFileLength(mpFile))==0xffffffff ){
		ISFileClose(mpFile);
		mpFile = NULL;
		return NULL;
	}
	
	mRequestSize = OSRoundUp32B(mRequestSize);
	
	if( pMem==NULL ){
		if( (pBuff=ghmGcMemAlloc((int)mRequestSize))==NULL ){
			ISFileClose(mpFile);
			mpFile = NULL;
			return NULL;
		}
	}else{
		pBuff = pMem;
	}
	
	if( ISFileReadAsync(pBuff,mRequestSize,mpFile,ReadCallback,(u32)this)==FALSE ){
		if( pMem==NULL ){
			ghmGcMemFree(pBuff);
		}
		ISFileClose(mpFile);
		mpFile = NULL;
		return NULL;
	}
	
#else
#pragma unused(pfname)
#pragma unused(pMem)
	return NULL;
//	if( mHandle!=FIO_INVALID_HANDLE ){
//		return NULL;
//	}
//	if( (mHandle=FIOFopen(pfname,FIO_OPEN_RDONLY))==FIO_INVALID_HANDLE ){
//		return NULL;
//	}
//	
//	if( (mRequestSize=FIOFseek(mHandle,0,FIO_SEEK_LAST))==0xffffffff ){
//		FIOFclose(mHandle);
//		return NULL;
//	}
//	FIOFseek(mHandle,0,FIO_SEEK_TOP);
//	
//	if( pMem==NULL ){
//		if( (pBuff=ghmGcMemAlloc((int)mRequestSize))==NULL ){
//			FIOFclose(mHandle);
//			mHandle = FIO_INVALID_HANDLE;
//			return NULL;
//		}
//	}else{
//		pBuff = pMem;
//	}
//	
//	if( FIOFreadAsync(mHandle,pBuff,mRequestSize,ReadCallback,(u32)this)==FALSE ){
//		if( pMem==NULL ){
//			ghmGcMemFree(pBuff);
//		}
//		FIOFclose(mHandle);
//		mHandle = FIO_INVALID_HANDLE;
//		return NULL;
//	}
	
#endif
	
	mState = FILEREAD;
	return pBuff;
}

#ifdef	USE_ISDOLVIEWER
void ghmGcUsb::ReadCallback(u32 result,ISFILE *pFile,u32 dwUserValue)
{
	ghmGcUsb		*pUsb = (ghmGcUsb *)dwUserValue;
	if( pUsb!=NULL ){
		pUsb->_ReadCallback(result,pFile);
	}
}

void ghmGcUsb::_ReadCallback(u32 result,ISFILE *pFile)
{
	if( pFile!=mpFile ){
		return;
	}
	mSize  = result;
	mState = READFINISH;
	ISFileClose(mpFile);
	mpFile = NULL;
}
#endif

bool ghmGcUsb::WriteAsync(const char *pfname,void *pMem,u32 size)
{
	if( mFioInitialized==false || mState!=CLOSED || pMem==NULL ){
		return false;
	}
	
#ifdef	USE_ISDOLVIEWER
	if( mpFile!=NULL ){
		return false;
	}
	
	if( (mpFile=ISFileOpen(pfname,"wb"))==NULL ){
		OSDebugReport("FileOpen failed %s\n",pfname);
		return false;
	}
	
	mRequestSize = size;
	if( ISFileWriteAsync(pMem,mRequestSize,mpFile,WriteCallback,(u32)this)==FALSE ){
		ISFileClose(mpFile);
		mpFile = NULL;
		return false;
	}
	
#else
#pragma unused(pfname)
#pragma unused(pMem)
#pragma unused(size)
	return false;
//	if( mHandle!=FIO_INVALID_HANDLE ){
//		return false;
//	}
//	if( (mHandle=FIOFopen(pfname,FIO_OPEN_WRONLY))==FIO_INVALID_HANDLE ){
//		return false;
//	}
//	
//	mRequestSize = size;
//	if( FIOFwriteAsync(mHandle,pMem,mRequestSize)==FALSE ){
//		FIOFclose(mHandle);
//		mHandle = FIO_INVALID_HANDLE;
//		return false;
//	}
	
#endif
	
	mState = FILEWRITE;
	return true;
}

#ifdef	USE_ISDOLVIEWER
void ghmGcUsb::WriteCallback(u32 result,ISFILE *pFile,u32 dwUserValue)
{
	ghmGcUsb		*pUsb = (ghmGcUsb *)dwUserValue;
	if( pUsb!=NULL ){
		pUsb->_WriteCallback(result,pFile);
	}
}

void ghmGcUsb::_WriteCallback(u32 result,ISFILE *pFile)
{
	if( pFile!=mpFile ){
		return;
	}
	mSize  = result;
	mState = WRITEFINISH;
	ISFileClose(mpFile);
	mpFile = NULL;
}
#endif

bool ghmGcUsb::Open(const char *pfname,u32 mode)
{
	if( mFioInitialized==false || mState!=CLOSED || pfname==NULL ){
		return false;
	}
#ifdef	USE_ISDOLVIEWER
	if( mpFile!=NULL ){
		return false;
	}
	char	*pmodestr;
	switch( mode ){
	case	FIO_OPEN_RDONLY:
		pmodestr = "rb";
		break;
	case	FIO_OPEN_WRONLY:
	case	FIO_OPEN_WRONLY|FIO_OPEN_CREAT:
	case	FIO_OPEN_WRONLY|FIO_OPEN_TRUNC:
	case	FIO_OPEN_WRONLY|FIO_OPEN_CREAT|FIO_OPEN_TRUNC:
		pmodestr = "wb";
		break;
	case	FIO_OPEN_RDWR:
		pmodestr = "r+b";
		break;
	case	FIO_OPEN_RDWR|FIO_OPEN_CREAT:
	case	FIO_OPEN_RDWR|FIO_OPEN_TRUNC:
	case	FIO_OPEN_RDWR|FIO_OPEN_CREAT|FIO_OPEN_TRUNC:
		pmodestr = "w+b";
		break;
	default:
		return false;
	}
	if( (mpFile=ISFileOpen(pfname,pmodestr))==NULL ){
		OSDebugReport("FileOpen failed %s\n",pfname);
		return NULL;
	}
#else
	if( mHandle!=FIO_INVALID_HANDLE ){
		return false;
	}
	char		fname[256];
	if( pfname[0]=='/' ){
		pfname++;
	}
	strncpy(fname,pfname,sizeof(fname)-1);
	if( (mHandle=FIOFopen(fname,mode))==FIO_INVALID_HANDLE ){
		return false;
	}
#endif
	mState = OPENED;
	return true;
}

bool ghmGcUsb::Close(void)
{
	if( mFioInitialized==false ){
		return false;
	}
#ifdef	USE_ISDOLVIEWER
	if( mpFile==NULL ){
		return false;
	}
	ISFileClose(mpFile);
	mpFile = NULL;
#else
	if( mHandle==FIO_INVALID_HANDLE ){
		return false;
	}
	FIOFclose(mHandle);
	mHandle = FIO_INVALID_HANDLE;
#endif
	mState = CLOSED;
	return true;
}

u32 ghmGcUsb::Write(void *pBuff,u32 size)
{
	if( mFioInitialized==false || mState!=OPENED ){
		return 0xffffffff;
	}
#ifdef	USE_ISDOLVIEWER
	if( mpFile==NULL ){
		return 0xffffffff;
	}
	
	size = ISFileWrite(pBuff,size,mpFile);
#else
	if( mHandle==FIO_INVALID_HANDLE ){
		return 0xffffffff;
	}
	
	size = FIOFwrite(mHandle,pBuff,size);
#endif
	
	return size;
}

u32 ghmGcUsb::Read(void *pBuff,u32 size)
{
	if( mFioInitialized==false || mState!=OPENED ){
		return 0xffffffff;
	}
#ifdef	USE_ISDOLVIEWER
	if( mpFile==NULL ){
		return 0xffffffff;
	}
	
	size = ISFileRead(pBuff,size,mpFile);
//	OSReport("size = %d\n",size);
#else
	if( mHandle==FIO_INVALID_HANDLE ){
		return 0xffffffff;
	}
	
	FIOFseek(mHandle,0,FIO_SEEK_TOP);
	size = FIOFread(mHandle,pBuff,size);
#endif
	
	return size;
}

#ifndef	USE_ISDOLVIEWER
bool ghmGcUsb::SetTtyChannel(MCCChannel chan)
{
	if( mTtyInitialized==true ){
		return false;
	}
	mTtyChannel = chan;
	return true;
}

bool ghmGcUsb::SetFioChannel(MCCChannel chan)
{
	if( mFioInitialized==true ){
		return false;
	}
	mFioChannel = chan;
	return true;
}

bool ghmGcUsb::SetFioBlocks(u8 blocks)
{
	if( mFioInitialized==true ){
		return false;
	}
	mFioBlocks = blocks;
	return true;
}
#endif

bool ghmGcUsb::FindFirst(const char *pfilename)
{
//#ifdef	USE_ISDOLVIEWER
//	strcpy(mFindData.m_strName,"toon.dat");
//	return true;
//#endif
	if( mFioInitialized==false || mState!=CLOSED ){
		return false;
	}
#ifdef	USE_ISDOLVIEWER
	char	filename[ISFILE_PATH_MAX+2] = "";
//	strcat(filename,pfilename);
	strcpy(mFindData.m_strName,pfilename);
	OSReport("find %s\n",filename);
	OSReport("attr %x\n",mFindData.m_nAttribute);
	if( (mpDir=ISFileFindFirst(filename,&mFindData))==NULL ){
		OSReport("failed find first\n");
		return false;
	}
	OSReport("success find first\n");
#else
	if( FIOFindFirst(pfilename,&mFindData)==false ){
		return false;
	}
#endif
	mState = FINDOPENED;
	return true;
}

bool ghmGcUsb::FindNext(void)
{
	if( mFioInitialized==false || mState!=FINDOPENED ){
		return false;
	}
#ifdef	USE_ISDOLVIEWER
	if( mpDir==NULL ){
		return false;
	}
	if( ISFileFindNext(mpDir,&mFindData)==FALSE ){
		return false;
	}
#else
	if( FIOFindNext(&mFindData)==false ){
		return false;
	}
#endif
	return true;
}

bool ghmGcUsb::FindClose(void)
{
	if( mFioInitialized==false || mState!=FINDOPENED ){
		return false;
	}
#ifdef	USE_ISDOLVIEWER
	if( mpDir==NULL ){
		return true;
	}
	if( ISFileFindClose(mpDir)==FALSE ){
		return false;
	}
	mpDir = NULL;
#endif
	mState = CLOSED;
	return true;
}

u32 ghmGcUsb::GetFindAttr(void)
{
	if( mFioInitialized==false || mState!=FINDOPENED ){
		return 0;
	}
#ifdef	USE_ISDOLVIEWER
	u32		attr = 0;
	if( mFindData.m_nAttribute & ISFILE_ATRB_READONLY ){
		attr |= FIO_ATTRIBUTE_READONLY;
	}
	if( mFindData.m_nAttribute & ISFILE_ATRB_HIDDEN ){
		attr |= FIO_ATTRIBUTE_HIDDEN;
	}
	if( mFindData.m_nAttribute & ISFILE_ATRB_SYSTEM ){
		attr |= FIO_ATTRIBUTE_SYSTEM;
	}
	if( mFindData.m_nAttribute & ISFILE_ATRB_DIR ){
		attr |= FIO_ATTRIBUTE_DIRECTORY;
	}else{
		attr |= FIO_ATTRIBUTE_NORMAL;
	}
	return attr;
#else
	return mFindData.stat.fileAttributes;
#endif
}

ghmGcUsb::ghmGcUsbState ghmGcUsb::GetState(void)
{
	return mState;
}

u32 ghmGcUsb::GetLength(void)
{
	if( mFioInitialized==false || mState!=OPENED ){
		return 0xffffffff;
	}
	
	u32			filesize;
#ifdef	USE_ISDOLVIEWER
	filesize = ISFileLength(mpFile);
#else
	filesize = FIOFseek(mHandle,0,FIO_SEEK_LAST);
#endif
	
	return filesize;
}

