////////////////////////////////////////////////////////////////////////////////
//
// GrassHopperManufacture GameCube USB Service
//
// ghmGcUsb.h
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 11/28/2002
//
////////////////////////////////////////////////////////////////////////////////

#ifndef	__GHMGCUSB_H__
#define	__GHMGCUSB_H__

#pragma cplusplus			on
#pragma cpp_extensions		on

#include <dolphin.h>
#include <sysdolphin\baselib.h>
#include <size_t.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef	USE_ISDOLVIEWER
#include <isdol.h>
#else
#include <mcc.h>
#include <fio.h>
#include <tty.h>
#endif
#include "ghmGcTypes.h"

#define	GHMGCUSB_TTYCHANNEL_DEF		(8)
#define	GHMGCUSB_FIOCHANNEL_DEF		(1)
#define	GHMGCUSB_FIOBLOCKS_DEF		(10)

class ghmGcUsb{
protected:
	enum ghmGcUsbState{
		READFINISH = 1,
		WRITEFINISH = 1,
		CLOSED = 0,
		OPENED = 1,
		READ,
		WRITE,
		FILEREAD,
		FILEWRITE,
		FINDOPENED,
	};
	volatile ghmGcUsbState		mState;
	u32							mRequestSize;
	u32							mSize;
#ifdef	USE_ISDOLVIEWER
	ISDIR						*mpDir;
	ISDIRDATA					mFindData;
	ISFILE						*mpFile;
	static void ReadCallback(u32 result,ISFILE *pFile,u32 dwUserValue);
	static void WriteCallback(u32 result,ISFILE *pFile,u32 dwUserValue);
#else
	FIOFindData					mFindData;
	FIOHandle					mHandle;
	
	static volatile s32			mExiChannel;
	static volatile bool		mMccInitialized;
	static MCCChannel			mTtyChannel;
	static MCCChannel			mFioChannel;
	static u8					mFioBlocks;
	static void mccSystemCallback(MCCSysEvent event);
	static BOOL hioEnumCallback(s32 chan);
#endif
	static bool					mFioInitialized;
	static bool					mTtyInitialized;
	
	bool _Terminate(void);
public:
	ghmGcUsb();
	virtual ~ghmGcUsb();
	
#ifdef	USE_ISDOLVIEWER
	static bool InitializeISDOL(void);
#endif
	
	bool Initialize(void);
	bool Terminate(void);
	
	u32 GetSize(const char *pfname);
	u32 GetLength(void);
	void* Read(const char *pfname,void *pMem=NULL);
	void* ReadAsync(const char *pfname,void *pMem=NULL);
	u32 Write(const char *pfname,void *pMem,u32 size);
	bool WriteAsync(const char *pfname,void *pMem,u32 size);
	
	bool Open(const char *pfname,u32 mode);
	bool Close(void);
	u32 Read(void *pBuff,u32 size);
	u32 Write(void *pBuff,u32 size);
	
	bool FindFirst(const char *pfilename);
	bool FindNext(void);
#ifdef	USE_ISDOLVIEWER
	const char* GetFindFileName(void){ return mFindData.m_strName; }
	const ISDIRDATA* GetFindData(void){ return &mFindData; }
#else
	const char* GetFindFileName(void){ return mFindData.filename; }
	const FIOFindData* GetFindData(void){ return &mFindData; }
#endif
	bool FindClose(void);
	u32 GetFindAttr(void);
	
#ifdef	USE_ISDOLVIEWER
	void _ReadCallback(u32 result,ISFILE *pFile);
	void _WriteCallback(u32 result,ISFILE *pFile);
#else
	static bool SetTtyChannel(MCCChannel chan);
	static bool SetFioChannel(MCCChannel chan);
	static bool SetFioBlocks(u8 blocks);
	
	inline static MCCChannel GetTtyChannel(void){ return mTtyChannel; }
	inline static MCCChannel GetFioChannel(void){ return mFioChannel; }
	inline static u8 GetFioBlocks(void){ return mFioBlocks; }
#endif
	
	ghmGcUsbState GetState(void);
	inline u32 GetSize(void){ return mSize; }
};

#endif	//__GHMGCUSB_H__

