////////////////////////////////////////////////////////////////////////////////
//
// Killer7 FirstCheck
//
// SK7SaveProcess.cpp
//
// Creator       : Sato Shinya (GHM)
// Creation Date : 02/12/2004
//
////////////////////////////////////////////////////////////////////////////////

#define	__SK7SAVEPROCESS_CPP__

#include <limits.h>
#include "SK7SaveProcess.h"
#include "../matsu/mK7Sound.h"
#include "../wat/K7GlobalFlag.h"
#include "../Yam/YK7Effect.h"
#include "../wat/WK7GameInfo.h"
#include "../Yam/YK7EffectNoise.h"
#include "../yam/YK7CharChange.h"


//========================================================================
//					RXgN^
//========================================================================
SK7SaveProcess::SK7SaveProcess( s8 State )
{
	m_chSaveState	= State;
	m_chSelect 		= 0;
	m_ucMoveNameFlag= 0;
	m_iSaveNameTime = 0;
	m_fSaveNameX	= SAVE_DATA_X;
	m_pSavedName	= NULL;
	m_eProc			= eProcLoad;
	m_bCreateFlag	= false;
	gpSound->mReleasePcSsm();
	gpSound->mReadSsm("seTv.ssm",eBankPlayer);
	m_pEffect = CreateNoiseEffect();
	SetEffectIndeManageFlag( m_pEffect, true );
	if( State==SAVE_STATE_END ){
		SetMessageCheck( CHECK_MES_ENDSAVE_CHECK, 0 );
	}
}

//========================================================================
//					fXgN^
//========================================================================
SK7SaveProcess::~SK7SaveProcess()
{
	gpSound->mReleaseSsm("seTv.ssm");	
	gpSound->mReadPcSsmReq( WK7PLAYERCONTROL->GetMainID() );
	GHM_SAFE_DELETE	( m_pIconData );
	DeleteEffect( m_pEffect );
	GHM_SAFE_DELETE_ARRAY( m_pSavedName );
}

//========================================================================
//					t[vZX
//========================================================================
bool 	SK7SaveProcess::FrameProcess(void)
{
	return true;
}

//========================================================================
//					t[vZXQ
//========================================================================
bool 	SK7SaveProcess::FrameMove()
{		
	if( m_eProc == eProcEnd && m_eProc == eProcWriteEnd )	return true;	

	//----------------	J[h}`FbN	----------------------//
	switch( CheckCardInsert() ){
	case MEMORYCARD_RETCODE_TRUE:	break;
	case MEMORYCARD_RETCODE_FALSE:	if( m_eProc > eProcLoad ) 	m_eProc = eProcMessage;	break;
	case MEMORYCARD_RETCODE_BUSY:	return true;
	case MEMORYCARD_RETCODE_NOCARD:	InitCheck();	return true;
	}

	//	`l^C	--------------------------------------------
	if( !m_bChanFlag && m_eProc > eProcLoad ){
		if( ++m_iChanTime >= CHAN_MOVE_TIME ){
			m_iChanTime = 0;
			m_bChanFlag = true;
			SetChannelNomral();
		}
	}

	if( !m_bCtrl && eProcLoad < m_eProc )	return false;
	//----------------	C	----------------------//
	switch( m_eProc ){
	case eProcLoad:
		if( LoadData(LOAD_ICON) ){
			if( m_chSaveState==SAVE_STATE_TV )		m_eProc = eProcCursolDraw;
			else									m_eProc = eProcAnsCheck;
		}
		break;
	case eProcCursolDraw:
		if( m_bCtrl ){
			gpSound->mPlaySE( SFX_TV_CONTENTS_CHOICE );
			m_eProc = eProcSlotSelect;
		}
		return true;
	case eProcAnsCheck:		FrameEndSaveCheck();	return true;
	case eProcSlotSelect:	CheckSlotSelect();		return true;
	case eProcSetMound:		
		SetMessage	( CERROR_MES_CHECK );
		CreateCard();
		SetCardMount( m_chSlotNo );
		m_eProc = eProcCheckMount;
		return true;
	case eProcMessage:
		switch(FrameMessageKey()){
		case MEMORYCARD_RETCODE_BUSY:	
		case MEMORYCARD_RETCODE_TRUE:	return true;
		case MEMORYCARD_RETCODE_FALSE:	InitCheck(); return true;
		}
		return true;
	case eProcCheckMount:	CheckMount();			return true;
	case eProcFileCheck:	CheckFileCheck();		return true;
	case eProcEreaCheck:	CheckFileErea(false);	return true;
	case eProcDammyDelete:	DeleteDammyFile();		return true;
	case eProcSavedDraw:
		CheckCardSavedData();
		if( m_eProc == eProcSaveSelect ){
			m_bLoadFontFlag = true;
			SetMessage( CERROR_MES_NONE );
		}
		return true;
	case eProcSaveSelect:	CheckSavePoint(eBaseSave);		return true;
	case eProcSaveCheck:	CheckSavePlay(eBaseSave);		return true;
	case eProcWrite:		WriteSaveData();		return true;
	case eProcWriteCheck:	CheckWriteEnd();		return true;
	case eProcInit:			CheckInitCard();		return true;	
	case eProcDelete:		CheckDammyDelete();		return true;
	case eProcRename: 		CheckFileRename();		return true;
	case eProcWriteMes:
		if( HSD_PadGetTrigger(0) & (HSD_PAD_A|HSD_PAD_B) ){
			gpSound->mPlaySE( SFX_TV_CONTENTS_CHOICE );
			m_eProc =  eProcWriteEnd;
		}
		return true;
	case eProcWriteEnd:
	case eProcEnd:			return true;
	}
	return true;
}


//========================================================================
//					_[vZX
//========================================================================
bool 	SK7SaveProcess::RenderProcess(void)
{
	if( m_bRender ){
	
		//	tHg`
		if( m_bLoadFontFlag ){
			RenderSaveComment();
			RenderMoveComment();
		}
		//	XbgZNĝƂ	
		if( m_eProc == eProcSlotSelect || m_eProc == eProcCursolDraw ){
			RenderSlotName( (u8)m_chSlotNo, m_lSlotColor[CARD_SLOT_A], m_lSlotColor[CARD_SLOT_B], false, m_bCtrl );
		}
		//	`FbNpbZ[W
		RenderCheckFont( (u8)m_chSelect );

		//	`l`
		m_spChan.Render();
		
		//	tHg`
		RenderFont( (s8)m_chSlotNo, true );
		
		//	bl̃G[`FbN
		RenderCheckFontError( (u8)m_chSelect );

		DrawSousaLine();
		MoveEffect(m_pEffect);
		DrawEffect(m_pEffect);
		if( (m_eProc != eProcWriteEnd && m_eProc != eProcEnd) || m_chSaveState == SAVE_STATE_END ){
			_DrawTVScreen();
		}

		m_spTV.Render();

		screenImageFullColor.ScreenCopy();
	}
	FrameMove();
	return true;
}

//========================================================================
//					Rg
//========================================================================
void 	SK7SaveProcess::RenderMoveComment(void)
{
	f32 SX;
	for( int i = 0; i < SAVE_DATA_NUM; i++ ){
		if( i == m_chSaveSelectNo && (m_eProc >= eProcDelete || m_eProc <= eProcWriteCheck ) ){
			SX = m_fSaveNameX;
		}else{
			SX = SAVE_DATA_X;
		}
		m_pYFont->DrawSetup();
		if( i == m_chSaveSelectNo )	m_pYFont->SetColor2(0x000000ff);
		else						m_pYFont->SetColor2(SLOT_FONT_COLOR);
		m_pYFont->DrawFSStr( SAVE_NO_X, SAVE_DATA_Y+i*SAVE_DATA_GAPY, false, true, "%d:", (i+1) ); 
		if( m_chSavedFlag & (1<<i) ){
			m_pYFont->DrawFSStr( SX, SAVE_DATA_Y+i*SAVE_DATA_GAPY, false, true, "%s", m_SaveName[i].name ); 
		}else{
			if( i == m_chSaveSelectNo )	m_pYFont->SetColor2(0x000000ff);
			else						m_pYFont->SetColor2(SLOT_FONT_COLOR);
			m_pYFont->DrawFSStr( SX, SAVE_DATA_Y+i*SAVE_DATA_GAPY, false, true, "No Data Saved" ); 
		}	
	}
}


//========================================================================
//					  
//========================================================================
void 			SK7SaveProcess::WriteSaveData()
{
	switch( m_chEnumNo ){
	case 0:
		//	Z[uf[^̍쐬	==========================
		CreateSaveData();	
		CreateCardInfo();	//	쐬
	
		char FileName[32];
	
		//	L[Wƃzbp[V͈ꏏ̖O
		if( K7SCR_GET_GLOBAL_FLAG( K7GLFL_HOPPER7 )==true ||
			K7SCR_GET_GLOBAL_FLAG( K7GLFL_KILLER1 )==true )
		{
			sprintf( FileName, "%s_%d", SAVE_FILE_NAME8, (int)m_chSaveSelectNo );
		}else{
			sprintf( FileName, "%s_%d", SAVE_FILE_NAME, (int)m_chSaveSelectNo );
		}
	
		//	Z[u[h̎	================================
		if( m_chSavedFlag & (1<<(m_chSaveSelectNo)) ){
			//	J[hɗ̈悪ꍇ	---------------------------------
			if( GetCardErea( m_chSlotNo, SAVEDATA_SIZE ) ){
				SetCreateCard( m_chSlotNo, SAVE_DAMMY_NAME, SAVEDATA_SIZE );
				m_bCreateFlag = true;
				m_chSavedFlag |= (1<<m_chSaveSelectNo);						//	tOĂ			
			}else{
				SetCardOpen( m_chSlotNo, FileName );
				m_bCreateFlag = false;
			}
		}else{
		 	int a = SetCreateCard( m_chSlotNo, FileName, SAVEDATA_SIZE );	//	ŏ̓_~[
			m_bCreateFlag = false;
			m_chSavedFlag |= (1<<m_chSaveSelectNo);						//	tOĂ
		}
		m_chEnumNo = 1;
		break;
		
	case 1:
		if( CARDGetResultCode(m_chSlotNo) == CARD_RESULT_BUSY )	return;
		if( ResultCheck( CARDGetResultCode(m_chSlotNo)) ){
			ghmGcMemSetAllocName("SAVELOAD");
			if( (m_pSaveData = (SK7SaveData *)ghmGcMemAlloc(SAVEDATA_SIZE))!=NULL ){
				memset		( m_pSaveData, 	0,	 			sizeof(m_SaveData) );
				memcpy		( m_pSaveData, 	&m_SaveData, 	sizeof(m_SaveData)  );
				SetWriteCard( m_pSaveData, 	SAVEDATA_SIZE,	0 );
			}
			m_ucMoveNameFlag 	= 0;
			m_iSaveNameTime 	= 0;
			m_chEnumNo			= 0;
			m_eProc = eProcWriteCheck;
		}else{
			m_eProc = eProcMessage;
		}
		break;
	}
}


//========================================================================
//					݃f[^쐬
//========================================================================
void 		SK7SaveProcess::CreateSaveData()
{
	char	comment[SAVE_FONT_SIZE];
	char Time[32];
	memset( comment, 0, sizeof( char )*SAVE_FONT_SIZE );
	GHM_SAFE_DELETE_ARRAY( m_pSavedName );
	ghmGcMemSetAllocName("SAVELOAD");
	m_pSavedName	= new char[32];

	//	Ԏ擾	======================================================
	OSCalendarTime td;
	OSTicksToCalendarTime(OSGetTime(), &td);
	int Year = td.year%2000;

	//	filename	======================================================
	if( K7SCR_GET_GLOBAL_FLAG( K7GLFL_HOPPER7 )==true ){
		sprintf( comment, "%s  %s", KILLER7_COMMENT, KILLERH_DATANAME );		
	}else if( K7SCR_GET_GLOBAL_FLAG( K7GLFL_KILLER1 )==true ){
		sprintf( comment, "%s  ViI%d %s", KILLER7_COMMENT, (int)GameInfo.GetScenarioID(), tblSaveScenarioName[GameInfo.GetScenarioID()+7] );
	}else{
		sprintf( comment, "%s  ViI%d %s", KILLER7_COMMENT, (int)GameInfo.GetScenarioID(), tblSaveScenarioName[GameInfo.GetScenarioID()+7] );
	}
	sprintf( Time, "%02d/%02d/%02d %02d:%02d", Year, td.mon+1, td.mday, td.hour, td.min );
	if( K7SCR_GET_GLOBAL_FLAG( K7GLFL_HOPPER7 )==true ){
		sprintf( m_pSavedName, "[%s] %s", KILLERH_FILENAME, Time );
	}else{
		sprintf( m_pSavedName, "[%s] %s", tblSaveScenarioName[GameInfo.GetScenarioID()],Time );
	}
	memcpy( &comment[32], Time, sizeof(Time) );
	
	//	Z[uf[^Zbg	
	memcpy( &m_SaveData.Comment, 	comment, 			SAVE_FONT_SIZE );			//	Rg
	memcpy( &m_SaveData.Banner, 	m_pBanner->image, 	BANNER_SIZE );	//	oi[  
	memcpy( &m_SaveData.Icon, 		m_pIconData, 		ICON_SIZE );	//	ACR
	m_SaveData.Version 			= SAVELOAD_VERSION;
	if( K7SCR_GET_GLOBAL_FLAG(K7GLFL_GAME_CLEAR_K8)==true ){
		memcpy( m_SaveData.GameClearWord, KILLERH_COMMENT, DATACHECK_WORD_NUM );
	}else if( K7SCR_GET_GLOBAL_FLAG(K7GLFL_GAME_CLEAR)==true ){	// Killer8p
		memcpy( m_SaveData.GameClearWord, KILLER8_COMMENT, DATACHECK_WORD_NUM );
	}else{
		memcpy( m_SaveData.GameClearWord, KILLER7_COMMENT, DATACHECK_WORD_NUM );
	}
//	m_SaveData.GameClearFlag 	= (K7SCR_GET_GLOBAL_FLAG(K7GLFL_GAME_CLEAR)==true) ? true : false;
//	m_SaveData.Hopper7Flag 		= (K7SCR_GET_GLOBAL_FLAG(K7GLFL_GAME_CLEAR_K8)==true) ? true : false;
	m_SaveData.CrcNo 			= GetCRC( &m_SaveData );
}

//========================================================================
//					ݏI
//========================================================================
void			SK7SaveProcess::CheckWriteEnd()
{
	//	ړ	-----------------------------------
	if( m_ucMoveNameFlag == 0 ){
		m_fSaveNameX += (SAVE_NAME_RIGHT_X - SAVE_DATA_X)/SAVE_NAME_TIME;
	}
	else if( m_ucMoveNameFlag == 1 ){
		if( m_iSaveNameTime < SAVE_NAME_TIME ){
			m_fSaveNameX -= (SAVE_NAME_RIGHT_X - SAVE_DATA_X)/SAVE_NAME_TIME;		
		}
	}
	++m_iSaveNameTime;
	s32 State = CARDGetResultCode(m_chSlotNo);

	if( State == CARD_RESULT_BUSY ) return;	
	if( State == CARD_RESULT_READY ){
		switch( m_chEnumNo ){
		case 0:						//	˃ChEăJ[h̃Zbg
			if( m_iSaveNameTime < SAVE_NAME_TIME )	return;
			memcpy( m_SaveName[m_chSaveSelectNo].name, m_pSavedName, 32 );
			GHM_SAFE_DELETE_ARRAY( m_pSavedName );
			m_ucMoveNameFlag= 2;		//	Xgbv
			//	J[hZbg	-----------------------------
			SetCardData( m_chSlotNo );
			m_chEnumNo = 1;
			break;
			
		case 1:						//	J[hZbgI`FbN
			m_iSaveNameTime 	= 0;
			m_ucMoveNameFlag 	= 1;
			m_chEnumNo			= 2;
			m_fSaveNameX		= SAVE_NAME_RIGHT_X;
			//	J[hl[ݒ	--------------------
			if( m_bCreateFlag ){
				char FileName[32];
				if( K7SCR_GET_GLOBAL_FLAG( K7GLFL_KILLER1 )==true ||
					K7SCR_GET_GLOBAL_FLAG( K7GLFL_HOPPER7 )==true )
				{
					sprintf( FileName, "%s_%d", SAVE_FILE_NAME8, (int)m_chSaveSelectNo );
				}else{
					sprintf( FileName, "%s_%d", SAVE_FILE_NAME, (int)m_chSaveSelectNo );
				}
				SetDeleteCard( m_chSlotNo, FileName );
				m_eProc = eProcDelete;
			}
			break;
			
		case 2:
			if( m_iSaveNameTime < SAVE_NAME_TIME )	return;
			m_fSaveNameX = SAVE_DATA_X;
			//	N[Y	---------------------------------
			SetCardClose();
			SetCardUnMount	(m_chSlotNo);
			DeleteCard();
			//	I܂B
			SetMessageCheck( CHECK_MES_SAVEEND, 0 );
			m_ucMoveNameFlag 	= 0;
			m_iSaveNameTime 	= 0;
			m_chEnumNo			= 0;
			gpSound->mPlaySE( SFX_TV_SAVEOK );
			k7app.SetResetProcessCallback( SK7ResetProcessCallBack, true );
			m_eProc = eProcWriteMes;
			break;
		}
	}else{
		if( !ResultCheck( State )){
			m_eProc = eProcMessage;
		}		
	}
}

//========================================================================
//					폜
//========================================================================
void 		SK7SaveProcess::CheckDammyDelete(void)
{
	switch( CARDGetResultCode(m_chSlotNo) ){
	case CARD_RESULT_BUSY:	return;
	case CARD_RESULT_READY:
		char FileName[32];
		if( K7SCR_GET_GLOBAL_FLAG( K7GLFL_HOPPER7 )==true ||
			K7SCR_GET_GLOBAL_FLAG( K7GLFL_KILLER1 )==true )
		{
			sprintf( FileName, "%s_%d", SAVE_FILE_NAME8, (int)m_chSaveSelectNo );
		}else{
			sprintf( FileName, "%s_%d", SAVE_FILE_NAME, (int)m_chSaveSelectNo );
		}
		//	_~[t@Cl[
		SetRenameCard( m_chSlotNo, SAVE_DAMMY_NAME, FileName );
		m_eProc = eProcRename;
		return;
	default:
		ResultCheck( CARDGetResultCode(m_chSlotNo) );
		m_eProc = eProcMessage;
		return;
	}
}

//========================================================================
//					l[
//========================================================================
void 		SK7SaveProcess::CheckFileRename(void)
{
	switch( CARDGetResultCode(m_chSlotNo) ){
	case CARD_RESULT_BUSY:	return;
	case CARD_RESULT_READY:	m_eProc = eProcWriteCheck;	return;
	default:	
		ResultCheck( CARDGetResultCode(m_chSlotNo) );
		m_eProc = eProcMessage;
		return;
	}
}

//========================================================================
//					Z[umF
//========================================================================
void		SK7SaveProcess::FrameEndSaveCheck(void)
{
	if( ghmGcFile::IsDiskStandby()==false ){
		return;
	}
	//		-------------------------------------
	if( CheckKeyT( SK7_KEY_LEFT ) ){
		if( m_chSelect == 1 )	gpSound->mPlaySE( SFX_TV_MOVE_CURSOR );
		m_chSelect = 0;
		return;
	}
	//	E	-------------------------------------
	if( CheckKeyT( SK7_KEY_RIGHT ) ){
		if( m_chSelect == 0 )	gpSound->mPlaySE( SFX_TV_MOVE_CURSOR );
		m_chSelect = 1;
		return;
	}
	//	I	-------------------------------------
	if( HSD_PadGetTrigger(0) & HSD_PAD_A ){
		gpSound->mPlaySE( SFX_TV_CONTENTS_CHOICE );
		if( m_chSelect == 0 ){
			m_eProc = eProcSlotSelect;
		}else{
			m_eProc = eProcEnd;
		}
		SetMessageCheck ( CHECK_MES_NONE, 0 );
		return;
	}
}


//========================================================================
//					
//========================================================================
bool 		SK7SaveProcess::GetMenuEnd(void)
{
	if( m_eProc == eProcWriteEnd || m_eProc == eProcEnd ){
		return true;
	}
	return false;
}

//========================================================================
//					
//========================================================================
void 		SK7SaveProcess::CheckInitCard(void)
{
	switch( CARDGetResultCode(m_chSlotNo) ){
	case CARD_RESULT_BUSY:			return;
	case CARD_RESULT_READY:			
		InitCheck();
		SetMessage( CERROR_MES_INIT_OK );
		m_eProc 		= eProcMessage;
		return;
	case CARD_RESULT_NOCARD:		InitCheck();	return;
	case CARD_RESULT_FATAL_ERROR:
	case CARD_RESULT_IOERROR:		SetMessage( CERROR_MES_INIT_MISS );	break;
	}
	SetCardUnMount( m_chSlotNo );
	DeleteCard();
	m_eProc = eProcMessage;
}

//========================================================================
//					`FbN
//========================================================================
void 		SK7SaveProcess::InitCheck(void)
{
	InitBase();
	k7app.SetResetProcessCallback( SK7ResetProcessCallBack, true );
	GHM_SAFE_DELETE_ARRAY( m_pSavedName );
	m_eProc 		= eProcSlotSelect;
	m_bLoadFontFlag	= false;
	m_bCreateFlag	= false;
	m_fSaveNameX	= SAVE_DATA_X;
}
