////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Archive
//
// WK7Archive.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 02/08/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7ARCHIVE_CPP__

//#include <dolphin.h>
//#include <sysdolphin/baselib.h>
//#include "..\Lib\ghmGcFile.h"
#include ".\WK7Archive.h"
#include ".\WK7Define.h"
#include "..\Fujita\C_SceneDesc.h"
#include ".\WK7SceneDesc.h"
#include ".\WK7Model.h"
#include ".\WK7Path.h"
#include ".\WK7Sight.h"

ghmGcListObj	WK7Archive::mAnchor;
#ifdef	_DEBUG
f32				WK7Archive::mDebugPrintSec = 4.0f;
#endif

WK7Archive::WK7Archive(ghmGcListObj *pAnchor,int prio) : WK7Obj(pAnchor,prio)
{
	InitMember();
}

void WK7Archive::InitMember(void)
{
	mpData         = NULL;
	mpArchive      = NULL;
	mpFile         = NULL;
	mpData         = NULL;
	mFlag          = 0;
	mRefCount      = 0;
	mType          = 0xffffffff;
	mID            = 0xffffffff;
	mState         = WK7ARCHIVE_VACANT;
}

WK7Archive::~WK7Archive()
{
	if( mRefCount ){
		OSDebugReport("Archive Free Error : Remain RefCount = %d\n",mRefCount);
		mRefCount = 0;
	}
	FreeArchive();
}

bool WK7Archive::Initialize(void)
{
//	if( mRefCount ){
//		return false;
//	}
//	FreeArchive();
//	mType     = 0xffffffff;
//	mID       = 0xffffffff;
	return true;
}

bool WK7Archive::Terminate(void)
{
	if( mRefCount || mStatic ){
		return false;
	}
	FreeArchive();
	mType     = 0xffffffff;
	mID       = 0xffffffff;
	delete this;
	return true;
}

bool WK7Archive::FrameProcess(void)
{
	switch( mState ){
	case	WK7ARCHIVE_LOADING:
		if( mpFile==NULL ){
			break;
		}
		if( mpFile->GetState()!=ghmGcFile::READFINISH ){
			break;
		}
		mpFile->Close();
		delete mpFile;
		mpFile = NULL;
		mState = WK7ARCHIVE_VACANT;
		if( LoadArchive(mpData,mpArchive)==false ){
			FreeArchive();
			break;
		}
		break;
	case	WK7ARCHIVE_VACANT:
	case	WK7ARCHIVE_LOADED:
	default:
		break;
	}
	return true;
}

bool WK7Archive::FreeArchive(void)
{
	if( mRefCount ){
		return false;
	}
	WK7Sight::WaitSyncToken();
//	GXSetDrawSync(0x9876);
//	while( 0x9876!=GXReadDrawSync() ){}
//	GXSetDrawSync(0x1234);
	GHM_SAFE_DELETE(mpFile);
#ifdef	_DEBUG
	if( mState==WK7ARCHIVE_LOADED ){
		WK7SdExpInfo	*pSdExpInfo;
		if( (pSdExpInfo=GetSdExpInfo())==NULL ){
			C_ModelSet	*pModelSet;
			HSD_Joint	*pJoint;
			int			i,modelSetNum = GetModelSetNum();
			for(i=0 ; i<modelSetNum ; i++){
				pModelSet = GetModelSet(i);
				pJoint    = pModelSet->GetJoint();
				ReviseFormatTerm(pJoint);
			}
		}
		if( mPathAlloc ){
			ghmScenePathSet		**ppPathSet;
			if( (ppPathSet = (ghmScenePathSet **)GetPublicAddress("scene_path"))!=NULL ){
				while( *ppPathSet!=NULL ){
					WK7PathDataDesc		*pPathDataDesc = (WK7PathDataDesc *)(*ppPathSet)->mpData;
					GHM_SAFE_DELETE(pPathDataDesc);
					(*ppPathSet)->mpData = NULL;
					ppPathSet++;
				}
			}
		}
	}
#endif
	if( mArchiveAlloc ){
		if( mpArchive!=NULL ){
			HSD_Free(mpArchive);
		}
		mArchiveAlloc = 0;
	}
	mpArchive = NULL;
	if( mDataAlloc ){
		GHMGCMEM_SAFE_FREE(mpData);
		mDataAlloc = 0;
	}
//	mType     = 0xffffffff;
//	mID       = 0xffffffff;
	mState    = WK7ARCHIVE_VACANT;
	return true;
}

HSD_Archive* WK7Archive::AllocArchive(void)
{
	return (HSD_Archive *)HSD_Alloc(sizeof(HSD_Archive));
}

bool WK7Archive::LoadArchiveB(const char *pfilename,HSD_Archive *pArchive)
{
	if( mState!=WK7ARCHIVE_VACANT ){
		return false;
	}
//	if( FreeArchive()==false ){
//		return false;
//	}
	ghmGcMemSetAllocName("GcFile");
	if( (mpFile = new ghmGcFile)==NULL ){
		return false;
	}
#ifdef	_DEBUG
	const char	*plastslash;
	char		allocname[16];
	if( (plastslash=strrchr(pfilename,'/'))!=NULL ){
		plastslash++;
		strcpy (allocname,"A");
		strncat(allocname,plastslash,sizeof(allocname)-2);
	}else{
		strcpy(allocname,"ArcFile");
	}
	ghmGcMemSetAllocName(allocname);
#else
	ghmGcMemSetAllocName("ArcFile");
#endif
	if( mpFile->Open((char *)pfilename)==false ){
		return false;
	}
	if( (mpData=mpFile->ReadB())==NULL ){
		delete mpFile;
		mpFile = NULL;
		return false;
	}
	mpFile->Close();
	delete mpFile;
	mpFile = NULL;
	if( LoadArchive(mpData,pArchive)==false ){
		GHMGCMEM_SAFE_FREE(mpData);
		return false;
	}
	mDataAlloc = 1;
	mState = WK7ARCHIVE_LOADED;
	return true;
}

bool WK7Archive::LoadArchive(const char *pfilename,HSD_Archive *pArchive)
{
	if( mState!=WK7ARCHIVE_VACANT ){
		return false;
	}
//	if( FreeArchive()==false ){
//		return false;
//	}
	ghmGcMemSetAllocName("GcFile");
	if( (mpFile = new ghmGcFile)==NULL ){
		return false;
	}
#ifdef	_DEBUG
	const char	*plastslash;
	char		allocname[16];
	if( (plastslash=strrchr(pfilename,'/'))!=NULL ){
		plastslash++;
		strcpy (allocname,"A");
		strncat(allocname,plastslash,sizeof(allocname)-2);
	}else{
		strcpy(allocname,"ArcFile");
	}
	ghmGcMemSetAllocName(allocname);
#else
	ghmGcMemSetAllocName("ArcFile");
#endif
	if( mpFile->Open((char *)pfilename)==false ){
		return false;
	}
	if( (mpData=mpFile->Read())==NULL ){
		delete mpFile;
		mpFile = NULL;
		return false;
	}
	mDataAlloc = 1;
	mpArchive = pArchive;
	mState = WK7ARCHIVE_LOADING;
	return true;
}

bool WK7Archive::LoadArchiveB(ghmGcUsb &serv,const char *pfilename,HSD_Archive *pArchive)
{
	if( mState!=WK7ARCHIVE_VACANT ){
		return false;
	}
//	if( FreeArchive()==false ){
//		return false;
//	}
	OSDebugReport("WK7Archive Load Archive ... : %s\n",pfilename);
	if( (mpData = serv.Read(pfilename))==NULL ){
		OSDebugReport(" File Read Failed !\n");
		return false;
	}
	if( LoadArchive(mpData,pArchive)==false ){
		GHMGCMEM_SAFE_FREE(mpData);
		return false;
	}
	mDataAlloc = 1;
	mState = WK7ARCHIVE_LOADED;
	return true;
}

bool WK7Archive::LoadArchive(void *pData,HSD_Archive *pArchive)
{
	if( mState!=WK7ARCHIVE_VACANT ){
		return false;
	}
	if( pData==NULL ){
		return false;
	}
//	if( FreeArchive()==false ){
//		return false;
//	}
	if( pArchive==NULL ){
		if( (mpArchive = AllocArchive())==NULL ){
			OSDebugReport("WK7Archive : Archive Alloc Failed !\n");
			return false;
		}
		mArchiveAlloc = 1;
	}else{
		mpArchive = pArchive;
	}
	HSD_ArchiveHeader		*pHeader = (HSD_ArchiveHeader *)pData;
#ifdef	_DEBUG
	if( pHeader->file_size==0 || pHeader->file_size >= WK7ARCHIVE_ILLEGAL_FILESIZE ){
		OSDebugReport("WK7Archive : Illegal FileSize 0x%08x (0x%08x,0x%08x)\n",pHeader->file_size,mType,mID);
		return false;
	}
#endif
	if( HSD_ArchiveParse(mpArchive,(u8 *)pData,pHeader->file_size)!=0 ){
		OSDebugReport("WK7Archive ArchiveParse Failed !\n");
		return false;
	}
	
	mState = WK7ARCHIVE_LOADED;
	mpData = pData;
	
#ifdef	_DEBUG
	WK7SdExpInfo	*pSdExpInfo;
	if( (pSdExpInfo=GetSdExpInfo())==NULL ){
		C_ModelSet	*pModelSet;
		HSD_Joint	*pJoint;
		int			i,modelSetNum = GetModelSetNum();
		bool		olddata = false;
		for(i=0 ; i<modelSetNum ; i++){
			pModelSet = GetModelSet(i);
			pJoint    = pModelSet->GetJoint();
			if( pJoint==NULL ){
				continue;
			}
			ReviseFormat(pJoint);
			olddata = true;
		}
		if( olddata==true ){
			char	str[256];
//			sprintf(str,"Old Version Archive !! - Type %x ID %x\n",mType,mID);
			strcpy(str,"Old Version ModelData !!");
			if( mType!=0xffffffff || mID!=0xffffffff ){
				strcat(str," - ");
				if( mType!=0xffffffff ){
					char	type[64];
					sprintf(type,"Type %x ",mType);
					strcat(str,type);
				}
				if( mID!=0xffffffff ){
					char	id[64];
					sprintf(id,"ID %x",mID);
					strcat(str,id);
				}
			}
			K7AppDebugPrintSec(str,mDebugPrintSec);
		}
	}
	if( pSdExpInfo==NULL || pSdExpInfo->mVersion < 153 ){
		ghmScenePathSet		**ppPathSet;
		if( (ppPathSet = (ghmScenePathSet **)GetPublicAddress("scene_path"))!=NULL ){
			mPathAlloc = 1;
			while( *ppPathSet!=NULL ){
				WK7PathDataDesc		*pPathDataDesc = (WK7PathDataDesc *)(*ppPathSet)->mpData;
				ghmGcMemSetAllocName("PathData");
				WK7PathDataDesc		*pNewPathDataDesc = new WK7PathDataDesc;
				*pNewPathDataDesc = *pPathDataDesc;
				if( pSdExpInfo->mVersion < 140 ){
					pNewPathDataDesc->mpRouteName = NULL;
					pNewPathDataDesc->mFlag       = 0;
					pNewPathDataDesc->mPass       = 1;
				}
				if( pSdExpInfo->mVersion < 153 ){
					pNewPathDataDesc->mCamPosV     = 0xffff;
					pNewPathDataDesc->mCamPosH     = 0xffff;
					pNewPathDataDesc->mCamType     = 0xffff;
					pNewPathDataDesc->mFootSE      = WK7PATH_FOOTSE_DEFAULT;
					pNewPathDataDesc->mpCameraDesc = NULL;
				}
				(*ppPathSet)->mpData = pNewPathDataDesc;
				ppPathSet++;
			}
		}
	}
#endif
	
	return true;
}

#ifdef	_DEBUG
void WK7Archive::ReviseFormat(HSD_DObjDesc *pDObjDesc)
{
	if( pDObjDesc==NULL ){
		return;
	}
	WK7RenderDesc		*pOldRenderDesc;
	if( pDObjDesc->mobjdesc!=NULL && (pOldRenderDesc=(WK7RenderDesc *)pDObjDesc->mobjdesc->renderdesc)!=NULL ){
		ghmGcMemSetAllocName("RndrDesc");
		WK7RenderDesc		*pNewRenderDesc = new WK7RenderDesc;
		if( pNewRenderDesc!=NULL ){
			pNewRenderDesc->mpToonTObjDesc = pOldRenderDesc->mpToonTObjDesc;
			pNewRenderDesc->mpGradTObjDesc = NULL;
			pNewRenderDesc->mTerminator    = -1;
		}
		pDObjDesc->mobjdesc->renderdesc = (HSD_RenderDesc *)pNewRenderDesc;
	}
	ReviseFormat(pDObjDesc->next);
}

void WK7Archive::ReviseFormat(HSD_Joint *pJoint)
{
	if( pJoint==NULL ){
		return;
	}
	if( union_type_dobj(pJoint)==TRUE ){
		ReviseFormat(pJoint->u.dobjdesc);
	}
	ReviseFormat(pJoint->child);
	ReviseFormat(pJoint->next);
}

void WK7Archive::ReviseFormatTerm(HSD_DObjDesc *pDObjDesc)
{
	if( pDObjDesc==NULL ){
		return;
	}
	WK7RenderDesc		*pRenderDesc;
	if( pDObjDesc->mobjdesc!=NULL && (pRenderDesc=(WK7RenderDesc *)pDObjDesc->mobjdesc->renderdesc)!=NULL ){
		delete pRenderDesc;
		pDObjDesc->mobjdesc->renderdesc = NULL;
	}
	ReviseFormat(pDObjDesc->next);
}

void WK7Archive::ReviseFormatTerm(HSD_Joint *pJoint)
{
	if( pJoint==NULL ){
		return;
	}
	if( union_type_dobj(pJoint)==TRUE ){
		ReviseFormatTerm(pJoint->u.dobjdesc);
	}
	ReviseFormatTerm(pJoint->child);
	ReviseFormatTerm(pJoint->next);
}
#endif

WK7Model* WK7Archive::CreateModel(u32 mirrorType)
{
	// ff[^
	WK7SceneDesc	*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL || pSceneDesc->GetModelSetNum()<=0 ){
		return NULL;
	}
	ghmGcMemSetAllocName("WK7Model");
	return new WK7Model(this,mirrorType);
}

WK7ModelMotion* WK7Archive::CreateModelMotion(void)
{
	// f[Vf[^
	WK7SceneDesc	*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL || pSceneDesc->GetModelAnimExist()==false ){
		return NULL;
	}
	ghmGcMemSetAllocName("WK7MdlMt");
	return new WK7ModelMotion(this);
}

WK7CameraMotion* WK7Archive::CreateCameraMotion(void)
{
	// J[Vf[^
	WK7SceneDesc	*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL || pSceneDesc->GetCameraAnimExist()==false ){
		return NULL;
	}
	ghmGcMemSetAllocName("WK7CamMt");
	return new WK7CameraMotion(this);
}

WK7LightMotion* WK7Archive::CreateLightMotion(void)
{
	// Cg[Vf[^
	WK7SceneDesc	*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL || pSceneDesc->GetLightAnimExist()==false ){
		return NULL;
	}
	ghmGcMemSetAllocName("WK7LgtMt");
	return new WK7LightMotion(this);
}

WK7Path* WK7Archive::CreatePath(u32 stgid)
{
	if( GetPublicAddress("scene_path")==NULL ){
		return NULL;
	}
	ghmGcMemSetAllocName("WK7Path");
	return new WK7Path(this,stgid);
}

HSD_Archive* WK7Archive::ReferenceArchive(void)
{
	if( mState != WK7ARCHIVE_LOADED ){
		return NULL;
	}
	mRefCount++;
	return mpArchive;
}

bool WK7Archive::ReleaseReferenceArchive(void)
{
	if( mState != WK7ARCHIVE_LOADED ){
		return false;
	}
	if( mRefCount ){
		mRefCount--;
	}
	Terminate();
	return true;
}

int WK7Archive::GetExternNum(void)
{
	if( mState != WK7ARCHIVE_LOADED ){
		return -1;
	}
	return (int)mpArchive->header.nb_extern;
}

const char* WK7Archive::GetExtern(int index)
{
	if( mState != WK7ARCHIVE_LOADED ){
		return NULL;
	}
	return HSD_ArchiveGetExtern(mpArchive,index);
}

int WK7Archive::GetPublicNum(void)
{
	if( mState != WK7ARCHIVE_LOADED ){
		return -1;
	}
	return (int)mpArchive->header.nb_public;
}

const char* WK7Archive::GetPublic(int index)
{
	if( mState != WK7ARCHIVE_LOADED ){
		return NULL;
	}
	return HSD_ArchiveGetPublic(mpArchive,index);
}

void* WK7Archive::GetPublicAddress(const char *pSymbol)
{
	if( mState != WK7ARCHIVE_LOADED || pSymbol==NULL ){
		return NULL;
	}
	return HSD_ArchiveGetPublicAddress(mpArchive,(char *)pSymbol);
}

WK7SceneDesc* WK7Archive::GetSceneDesc(void)
{
	if( mState != WK7ARCHIVE_LOADED ){
		return NULL;
	}
	return new (mpArchive)WK7SceneDesc;
}

bool WK7Archive::IncludeModelData(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return false;
	}
	u32				i,num;
	C_ModelSet		*pModelSet;
	num = pSceneDesc->GetModelSetNum();
	for(i=0 ; i<num ; i++){
		pModelSet = pSceneDesc->GetModelSet(i);
		if( pModelSet->joint != NULL ){
			return true;
		}
	}
	return false;
}

bool WK7Archive::IncludeModelAnim(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return false;
	}
	u32				i,num;
	C_ModelSet		*pModelSet;
	num = pSceneDesc->GetModelSetNum();
	for(i=0 ; i<num ; i++){
		pModelSet = pSceneDesc->GetModelSet(i);
		if( pModelSet->adjointlist != NULL ){
			return true;
		}
	}
	return false;
}

bool WK7Archive::IncludeModelMatAnim(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return false;
	}
	u32				i,num;
	C_ModelSet		*pModelSet;
	num = pSceneDesc->GetModelSetNum();
	for(i=0 ; i<num ; i++){
		pModelSet = pSceneDesc->GetModelSet(i);
		if( pModelSet->matanimjointlist != NULL ){
			return true;
		}
	}
	return false;
}

bool WK7Archive::IncludeModelShapeAnim(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return false;
	}
	u32				i,num;
	C_ModelSet		*pModelSet;
	num = pSceneDesc->GetModelSetNum();
	for(i=0 ; i<num ; i++){
		pModelSet = pSceneDesc->GetModelSet(i);
		if( pModelSet->shapeanimjointlist != NULL ){
			return true;
		}
	}
	return false;
}

int WK7Archive::GetModelSetNum(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return 0;
	}
	return (int)pSceneDesc->GetModelSetNum();
}

C_ModelSet* WK7Archive::GetModelSet(int index)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	return pSceneDesc->GetModelSet((u32)index);
}

C_ModelSet* WK7Archive::GetModelSet(const char *pName)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	return pSceneDesc->GetModelSet(pName);
}

C_CameraSet* WK7Archive::GetCameraSet(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	return pSceneDesc->GetCameraSet();
}

int WK7Archive::GetLightSetNum(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return 0;
	}
	return (int)pSceneDesc->GetLightSetNum();
}

C_LightSet* WK7Archive::GetLightSet(int index)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	return pSceneDesc->GetLightSet((u32)index);
}

C_LightSet* WK7Archive::GetLightSet(const char *pName)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	return pSceneDesc->GetLightSet(pName);
}

C_FogSet* WK7Archive::GetFogSet(void)
{
	WK7SceneDesc		*pSceneDesc = GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	return pSceneDesc->GetFogSet();
}

HSD_CameraDesc** WK7Archive::GetCameraDescList(void)
{
	return (HSD_CameraDesc **)GetPublicAddress("scene_camera");
}

int WK7Archive::GetCameraAnimNum(void)
{
	return WK7CameraMotion::GetAnimNum(this);
}

HSD_CameraAnim* WK7Archive::GetCameraAnim(int index)
{
	return WK7CameraMotion::GetAnim(this,index);
}

HSD_CameraAnim* WK7Archive::GetCameraAnim(const char *pName)
{
	return WK7CameraMotion::GetAnim(this,pName);
}

int WK7Archive::GetCameraDescNum(void)
{
	HSD_CameraDesc	**ppCameraDesc;
	if( (ppCameraDesc = GetCameraDescList())==NULL ){
		return 0;
	}
	int				i;
	for(i=0 ; ppCameraDesc[i]!=NULL ; i++){}
	return i;
}

HSD_CameraDesc* WK7Archive::GetCameraDesc(int index)
{
	HSD_CameraDesc	**ppCameraDesc;
	if( (ppCameraDesc = GetCameraDescList())==NULL ){
		return NULL;
	}
	int				i;
	for(i=0 ; i<index && ppCameraDesc[i]!=NULL ; i++){}
	return ppCameraDesc[i];
}

HSD_CameraDesc* WK7Archive::GetCameraDesc(const char *pName)
{
	HSD_CameraDesc	**ppCameraDesc;
	if( (ppCameraDesc = GetCameraDescList())==NULL ){
		return NULL;
	}
	int				i;
	for(i=0 ; ppCameraDesc[i]!=NULL ; i++){
		if( strcmp(GetDescName(ppCameraDesc[i]),pName)==0 ){
			return ppCameraDesc[i];
		}
	}
	return NULL;
}

HSD_CameraDesc* WK7Archive::GetCameraDescPartialName(const char *pPartialName)
{
	HSD_CameraDesc	**ppCameraDesc;
	if( (ppCameraDesc = GetCameraDescList())==NULL ){
		return NULL;
	}
	int				i;
	for(i=0 ; ppCameraDesc[i]!=NULL ; i++){
		if( strstr(GetDescName(ppCameraDesc[i]),pPartialName)!=NULL ){
			return ppCameraDesc[i];
		}
	}
	return NULL;
}

HSD_CameraDesc* WK7Archive::GetCameraDescPartialCameraName(const char *pName)
{
	HSD_CameraDesc	**ppCameraDesc;
	if( (ppCameraDesc = GetCameraDescList())==NULL ){
		return NULL;
	}
	int				i;
	for(i=0 ; ppCameraDesc[i]!=NULL ; i++){
		if( strstr(pName,GetDescName(ppCameraDesc[i]))!=NULL ){
			return ppCameraDesc[i];
		}
	}
	return NULL;
}

bool WK7Archive::ReleaseAll(void)
{
	WK7Archive		*pArchive = (WK7Archive *)mAnchor.GetNext();
	WK7Archive		*pNextArchive;
	while( pArchive!=NULL ){
		pNextArchive = pArchive->GetNext();
		delete pArchive;
		pArchive = pNextArchive;
	}
	return true;
}

bool WK7Archive::ReleaseDynamicAll(void)
{
	WK7Archive		*pArchive = (WK7Archive *)mAnchor.GetNext();
	WK7Archive		*pNextArchive;
	while( pArchive!=NULL ){
		pNextArchive = pArchive->GetNext();
		if( pArchive->GetStatic()==false ){
			delete pArchive;
		}
		pArchive = pNextArchive;
	}
	return true;
}

bool WK7Archive::ReleaseStaticAll(void)
{
	WK7Archive		*pArchive = (WK7Archive *)mAnchor.GetNext();
	WK7Archive		*pNextArchive;
	while( pArchive!=NULL ){
		pNextArchive = pArchive->GetNext();
		if( pArchive->GetStatic()==true ){
			delete pArchive;
		}
		pArchive = pNextArchive;
	}
	return true;
}

WK7Archive* WK7Archive::Create(int prio)
{
	ghmGcMemSetAllocName("WK7Arc");
	return new WK7Archive(&mAnchor,prio);
}

WK7Archive* WK7Archive::sysFind(u32 type,u32 id,WK7Archive *pArchive)
{
	if( pArchive==NULL ){
		pArchive = (WK7Archive *)mAnchor.GetNext();
	}else{
		pArchive = pArchive->GetNext();
	}
	while( pArchive != NULL ){
		if( pArchive->GetType()==type &&
		    pArchive->GetID  ()==id   ){
			return pArchive;
		}
		pArchive = pArchive->GetNext();
	}
	return NULL;
}

WK7Archive* WK7Archive::Find(u32 type,u32 id,WK7Archive *pArchive)
{
	if( (pArchive = sysFind(type,id,pArchive))!=NULL ){
//		pArchive->ReferenceArchive();
		return pArchive;
	}
	return NULL;
}

bool WK7Archive::FrameProcessAll(void)
{
	WK7Archive	*pArchive = (WK7Archive *)mAnchor.GetNext();
	WK7Archive	*pNext;
	while( pArchive!=NULL ){
		pNext    = pArchive->GetNext();
		pArchive->FrameProcess();
		pArchive = pNext;
	}
	return true;
}

WK7SdExpInfo* WK7Archive::GetSdExpInfo(void)
{
	if( mpArchive==NULL ){
		return NULL;
	}
	WK7SdExpInfo		*pInfo = (WK7SdExpInfo *)HSD_ArchiveGetPublicAddress(mpArchive,"SdExpInfo");
	if( pInfo!=NULL ){
		return pInfo;
	}
	pInfo = (WK7SdExpInfo *)HSD_ArchiveGetPublicAddress(mpArchive,"scene_data");
	if( pInfo==NULL ){
		return NULL;
	}
	pInfo--;
	if( pInfo->mVersion==113 ){
		return pInfo;
	}
	return NULL;
}

bool WK7Archive::IsValid(WK7Archive *pArchive)
{
	WK7Archive	*pTmpArchive;
	for(pTmpArchive=(WK7Archive *)mAnchor.GetNext() ; pTmpArchive!=NULL ; pTmpArchive=pTmpArchive->GetNext()){
		if( pTmpArchive==pArchive ){
			return true;
		}
	}
	return false;
}

