////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Chara Object
//
// WK7Chara.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 11/09/2002
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7CHARA_CPP__

//#include <dolphin.h>
#include ".\WK7Chara.h"
#include ".\WK7Define.h"
#include ".\K7ID.h"
#include ".\WK7Event.h"
#include ".\WK7Shadow.h"
#include ".\WK7Light.h"
#include ".\WK7Archive.h"

bool	WK7Chara::mRenderChara = true;

WK7Chara::WK7Chara(ghmGcListObj *pAnchor,int prio) : WK7RenderObj(pAnchor,prio)
{
	mStageLight = true;
}

WK7Chara::~WK7Chara()
{
}

u32 WK7Chara::GetMainID(void)
{
	return K7CHR_MAINID(mID);
}

u32 WK7Chara::GetSuffix(void)
{
	return K7CHR_SUFFIX(mID);
}

struct WK7CharaType{
	u32				mMainID;
	const char		*mpName;
};

static	WK7CharaType	charaTypeTable[] = {
	{ K7DAN, K7DAN_INITIAL },
	{ K7GAL, K7GAL_INITIAL },
	{ K7CON, K7CON_INITIAL },
	{ K7HAC, K7HAC_INITIAL },
	{ K7HAS, K7HAS_INITIAL },
	{ K7HAY, K7HAY_INITIAL },
	{ K7CYO, K7CYO_INITIAL },
	{ K7KED, K7KED_INITIAL },
	{ K7KVN, K7KVN_INITIAL },
	{ K7MSK, K7MSK_INITIAL },
	{ K7KUN, K7KUN_INITIAL },
	{ K7ENM, K7ENM_INITIAL },
	{ K7ANG, K7ANG_INITIAL },
	{ K7KRR, K7KRR_INITIAL },
	{ K7KIB, K7KIB_INITIAL },
	{ K7KIS, K7KIS_INITIAL },
	{ K7HAT, K7HAT_INITIAL },
	{ K7DE2, K7DE2_INITIAL },
	{ K7UR2, K7UR2_INITIAL },
	{ K7SMA, K7SMA_INITIAL },
	{ K7AKN, K7AKN_INITIAL },
	{ K7AK2, K7AK2_INITIAL },
	{ K7CUR, K7CUR_INITIAL },
	{ K7ZA1, K7ZA1_INITIAL },
	{ K7ZA2, K7ZA2_INITIAL },
	{ K7ZA3, K7ZA3_INITIAL },
	{ K7ZA4, K7ZA4_INITIAL },
	{ K7ZA5, K7ZA5_INITIAL },
	{ K7ZA6, K7ZA6_INITIAL },
	{ K7ZA7, K7ZA7_INITIAL },
	{ K7ZA8, K7ZA8_INITIAL },
	{ K7ZA9, K7ZA9_INITIAL },
	{ K7ZAA, K7ZAA_INITIAL },
	{ K7ZAB, K7ZAB_INITIAL },
	{ K7ZAC, K7ZAC_INITIAL },
	{ K7ZAD, K7ZAD_INITIAL },
	{ K7ZAE, K7ZAE_INITIAL },
	{ K7ZAF, K7ZAF_INITIAL },
	{ K7ZAG, K7ZAG_INITIAL },
	{ K7ZAH, K7ZAH_INITIAL },
	{ K7ZAI, K7ZAI_INITIAL },
	{ K7ZAJ, K7ZAJ_INITIAL },
	{ K7ZAK, K7ZAK_INITIAL },
	{ K7ZAL, K7ZAL_INITIAL },
	{ K7ZAM, K7ZAM_INITIAL },
	{ K7ZAN, K7ZAN_INITIAL },
	{ K7ZAO, K7ZAO_INITIAL },
	{ K7ZAP, K7ZAP_INITIAL },
	{ K7ZAQ, K7ZAQ_INITIAL },
	{ K7ZAR, K7ZAR_INITIAL },
	{ K7ZAS, K7ZAS_INITIAL },
	{ K7ZAT, K7ZAT_INITIAL },
	{ K7ZAU, K7ZAU_INITIAL },
	{ K7ZAV, K7ZAV_INITIAL },
	{ K7ZAW, K7ZAW_INITIAL },
	{ K7ZAX, K7ZAX_INITIAL },
	{ K7ZAY, K7ZAY_INITIAL },
	{ K7ZAZ, K7ZAZ_INITIAL },
	{ K7BAN, K7BAN_INITIAL },
	{ K7TEN, K7TEN_INITIAL },
	{ K7FAN, K7FAN_INITIAL },
	{ K7GLY, K7GLY_INITIAL },
	{ K7MAT, K7MAT_INITIAL },
	{ K7HM1, K7HM1_INITIAL },
	{ K7HN2, K7HN2_INITIAL },
	{ K7HN3, K7HN3_INITIAL },
	{ K7HN4, K7HN4_INITIAL },
	{ K7HN5, K7HN5_INITIAL },
	{ K7HN6, K7HN6_INITIAL },
	{ K7HN8, K7HN8_INITIAL },
	{ K7HN9, K7HN9_INITIAL },
	{ K7GRY, K7GRY_INITIAL },
	{ K7HNG, K7HNG_INITIAL },
	{ K7SNG, K7SNG_INITIAL },
	{ K7HNG, K7HNG_INITIAL },
	{ K7FER, K7FER_INITIAL },
	{ 0xffffffff, NULL },
};

u32 WK7Chara::GetMainID(char *pname)
{
	if( pname==NULL ){
		return 0xffffffff;
	}
	WK7CharaType	*pType;
	for(pType=charaTypeTable ; pType->mMainID!=0xffffffff ; pType++){
		if( strcmp(pType->mpName,pname)==0 ){
			return pType->mMainID;
		}
	}
	u32		mainid;
	switch( pname[0] ){
	case	'E':
		mainid = K7E00;
		break;
	case	'S':
		mainid = K7S00;
		break;
	default:
		return 0xffffffff;
	}
	int		i;
	u32		id = 0;
	for(i=1 ; i<3 ; i++){
		if( pname[i] < '0' || pname[i] > '9' ){
			return 0xffffffff;
		}
		id *= 10;
		id += (pname[i] - '0');
	}
	return mainid + id;
}

bool WK7Chara::GetInitial(u32 mainid,char *pname)
{
	if( pname==NULL ){
		return false;
	}
	if( mainid > K7E00 ){
		if( mainid < K7E00+K7E_NUM ){
			sprintf(pname,"E%02d",mainid-K7E00);
			return true;
		}else if( mainid < K7S00+K7S_NUM ){
			sprintf(pname,"S%02d",mainid-K7S00);
			return true;
		}
	}
	WK7CharaType	*pType;
	for(pType=charaTypeTable ; pType->mMainID!=0xffffffff ; pType++){
		if( pType->mMainID==mainid ){
			strcpy(pname,pType->mpName);
			return true;
		}
	}
	return false;
}

bool WK7Chara::GetFileName(char *pFileName,u32 id)
{
	
	u32			mainid = K7STG_MAINID(id);
	u32			suffix = K7STG_SUFFIX(id);
	char		initial[16];
	
	OSDebugReportExp(mainid > K7CHRID_MAX,"Chara Main ID Over 1000 : %d\n" ,mainid);
	OSDebugReportExp(suffix > 100        ,"Chara Suffix ID Over 100 : %d\n",suffix);
	
	GetInitial(mainid,initial);
	
	sprintf(pFileName,WK7_3DDATA_FOLDER WK7_MODEL_FOLDER WK7_CHARA_FOLDER "%s/%s_M%02d.dat",initial,initial,suffix);
	return true;
}

bool WK7Chara::Initialize(void)
{
	if( WK7RenderObj::Initialize()==false ){
		return false;
	}
	WK7JObj		*pJObj;
	if( mpModel!=NULL ){
		WK7CastShadowObj	*pCastShadowObj;
		const char			*pObjName;
		if( K7CHR_MAINID(mID)==K7HAC ){
			pObjName = "Wheelchair";
		}else{
			pObjName = K7CHRBONE_WAIST;
		}
		if( (pJObj=mpModel->GetObj(pObjName))!=NULL ){
			if( (pCastShadowObj = WK7SHADOW->AddCastShadowObj(pJObj))!=NULL ){
				pCastShadowObj->SetRenderObj(this);
			}
		}
	}
	return true;
}

bool WK7Chara::RenderProcess(void)
{
	if( mType==K7OBJTYPE_CHR && mRenderChara==false ){
		return false;
	}
	if( WK7RenderObj::RenderProcess()==false ){
		return false;
	}
	return true;
}

bool WK7Chara::DispLight(u32 mirrorType,Vec *pPos)
{
	WK7Light			*pLight;
	WK7LightFindData	data;
//	WK7Light			*pPlayerLight = WK7PLAYERCONTROL->GetLight();
	data.Initialize();
	while( (pLight=WK7Light::Find(data))!=NULL ){
		if( mObjLight==false ){
			if( pLight->GetObjectType()==mType && pLight->GetObjectID()==mID ){
				continue;
			}
		}
		if( mStageLight==false ){
			if( pLight->GetObjectType()==K7OBJTYPE_STG ){
				continue;
			}
		}
		if( pLight->GetPrivate()==true ){
			DispPrivateLight(pLight,mirrorType,pPos);
		}else{
//			if( mPlayerLight==false && pLight==pPlayerLight ){
//				continue;
//			}
			if( mDispLight & pLight->GetLightIDBit() ){
				pLight->Disp(mirrorType,pPos);
			}
		}
	}
	return true;
}

//bool WK7Chara::ReleaseArchive(u32 id)
//{
//	WK7Archive	*pArchive;
//	if( (pArchive = WK7Archive::Find(K7OBJTYPE_CHR,id))==NULL ){
//		return false;
//	}
//	delete pArchive;
//	return true;
//}

bool WK7Chara::CreateStaticArchive(u32 id)
{
	WK7Archive	*pArchive;
	if( (pArchive = WK7Archive::Find(K7OBJTYPE_CHR,id))==NULL ){
		if( (pArchive = WK7Archive::Create())==NULL ){
			return false;
		}
		// t@C擾
		char	filename[WK7_FNAME_MAX];
		GetFileName(filename,id);
		// ACe^Cvɐݒ
		pArchive->SetType(K7OBJTYPE_CHR);
		// hcݒ
		pArchive->SetID  (id);
		// A[JCu[h
		if( pArchive->LoadArchive(filename)==false ){
			delete pArchive;
			return false;
		}
	}
	pArchive->SetStatic(true);
	return true;
}

WK7CharaMotion::WK7CharaMotion(ghmGcListObj *pAnchor,int prio) : WK7MotionObj(pAnchor,prio)
{
}

WK7CharaMotion::~WK7CharaMotion()
{
}

bool WK7CharaMotion::GetFileName(char *pFileName,u32 id)
{
	u32			mainid = K7MOT_MAINID(id);
	u32			suffix = K7MOT_SUFFIX(id);
	char		initial[16];
	
	OSDebugReportExp(mainid > K7CHRID_MAX,"Chara Motion Main ID Over 1000 : %d\n"   ,mainid);
	OSDebugReportExp(suffix > 10000      ,"Chara Motion Suffix ID Over 10000 : %d\n",suffix);
	
	WK7Chara::GetInitial(mainid,initial);
	
	sprintf(pFileName,WK7_3DDATA_FOLDER WK7_MOTION_FOLDER WK7_CHARA_FOLDER "%s/%s_%04d.dat",initial,initial,suffix);
	return true;
}

bool WK7CharaMotion::CreateStaticArchive(u32 id)
{
	WK7Archive	*pArchive;
	if( (pArchive = WK7Archive::Find(K7MOT_ATTR|K7OBJTYPE_CHR,id))==NULL ){
		if( (pArchive = WK7Archive::Create())==NULL ){
			return false;
		}
		// t@C擾
		char	filename[WK7_FNAME_MAX];
		GetFileName(filename,id);
		// ACe^Cvɐݒ
		pArchive->SetType(K7MOT_ATTR|K7OBJTYPE_CHR);
		// hcݒ
		pArchive->SetID  (id);
		// A[JCu[h
		if( pArchive->LoadArchive(filename)==false ){
			delete pArchive;
			return false;
		}
	}
	pArchive->SetStatic(true);
	return true;
}

