////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Event
//
// WK7Event.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 02/02/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7EVENT_CPP__

//#include <dolphin.h>
//#include <sysdolphin\baselib.h>
//#include "..\Lib\ghmGcDefine.h"
#include ".\WK7Event.h"
#include ".\WK7EventItem.h"
#include ".\K7ID.h"

#include ".\WK7Player.h"
#include ".\WK7Npc.h"
#include ".\WK7HitObj.h"
#include ".\WK7Obj.h"
#include ".\WK7DeadBodySign.h"
#include ".\WK7EventScript.h"
#include ".\WK7Stage.h"
#include ".\WK7GameInfo.h"
//#include ".\WK7MainMenu.h"
//#include ".\WK7Memo.h"

#include "..\Matsu\MK7Proc.h"
#include "..\Yam\YK7Effect.h"
#include "..\Sato\SK7Pack.h"
#include "..\Yam\YK7EffectMessage2.h"
#include "..\Yam\YK7EffectItemGet.h"
#include "..\Kaw\KK7BattleMainTask.h"
#include "..\Kaw\KK7BattleCtrl.h"
#include "..\Yam\YK7StageTitle.h"

#include "..\Lib\ghmGcModule.h"

//#define	DEBUG_STOP
#define	DEBUG_PRINT
//#define	DEBUG_OSREPORT

#if defined(ORBITAL) || !defined(_DEBUG)
#undef	DEBUG_STOP
#undef	DEBUG_PRINT
#undef	DEBUG_OSREPORT
#endif

#ifdef	DEBUG_OSREPORT
#define	OSReportEV(...)			OSReport(__VA_ARGS__)
#else
#define	OSReportEV(...)			
#endif

//#define	K7EVENT_MEMORY_DEBUG
#ifndef	_DEBUG
#undef	K7EVENT_MEMORY_DEBUG
#endif

#define	WK7EVENT_TASK_STACK_SIZE		(0x8000)

WK7EventIDTable*	WK7Event::mpEventIDTable = NULL;
#ifdef	_DEBUG
bool				WK7Event::mDebugPrintEnable = true;
#endif

WK7Event::WK7Event(u32 id,ghmGcListObj *pAnchor,int prio) : ghmGcListObj(pAnchor,prio)
{
	mID           = id;
	mNo           = 0xffffffff;
	mpTask        = NULL;
	mpHandler     = NULL;
	mpTermHandler = NULL;
}

WK7Event::~WK7Event()
{
	CallTermHandler();
}

void WK7Event::TermHandler(ghmGcTask *pTask)
{
	WK7Event		*pEvent = (WK7Event *)pTask->GetParam(0);
	if( pEvent->mpTask==pTask ){
		pEvent->CallTermHandler();
		delete pEvent;
	}else{
		OSDebugReport("Illegal Event TerminateHandler Call\n");
	}
}

void WK7Event::Handler(ghmGcTask *pTask)
{
	WK7Event		*pEvent = (WK7Event *)pTask->GetParam(0);
	pEvent->mpTask = pTask;
	pEvent->CallHandler();
}

void WK7Event::CallHandler(void)
{
	if( mpHandler != NULL ){
		mpHandler(mpTask);
	}
}

void WK7Event::CallTermHandler(void)
{
	if( mpTermHandler != NULL ){
		mpTermHandler(mpTask);
	}
	mpHandler     = NULL;
	mpTermHandler = NULL;
	mpTask        = NULL;
}

int WK7Event::SetEventEnableAll(u32 eventid,bool flag)
{
	int		count;
	count  = WK7RenderObj::SetEventEnableAll(eventid,flag);
	count += WK7HitObj::SetEventEnable(eventid,flag);
	return count;
}

int WK7Event::SetEventNoEnableAll(u32 eventno,bool flag)
{
	int		count;
	count  = WK7RenderObj::SetEventNoEnableAll(eventno,flag);
	count += WK7HitObj::SetEventNoEnable(eventno,flag);
	return count;
}

WK7EventIDTable* WK7Event::FindIDTableFromEventID(u32 evid)
{
	WK7EventIDTable		*pTable;
	if( mpEventIDTable==NULL ){
		return NULL;
	}
	for(pTable=mpEventIDTable ; pTable->mEventNo!=WK7EVENTIDTBL_TERM_EVNO ; pTable++){
		if( pTable->mEventID==evid ){
			return pTable;
		}
	}
	return NULL;
}

WK7EventIDTable* WK7Event::FindIDTable(u32 evno)
{
	WK7EventIDTable		*pTable;
	if( mpEventIDTable==NULL ){
		return NULL;
	}
	for(pTable=mpEventIDTable ; pTable->mEventNo!=WK7EVENTIDTBL_TERM_EVNO ; pTable++){
		if( pTable->mEventNo==evno ){
			return pTable;
		}
	}
	return NULL;
}

bool WK7Event::CallStageCondChange(u32 stgid)
{
	return WK7Stage::CallStageCondChange(stgid);
}

bool WK7Event::CallCurrentStageCondChange(void)
{
	return CallStageCondChange(WK7Stage::GetCurrentStageID());
}

bool WK7Event::CallCondChangeAll(void)
{
	WK7RenderObj::CallCondChangeAll();
	return true;
}

#ifdef	_DEBUG
bool WK7Event::DebugPrint(const char *pstr)
{
	if( pstr==NULL ){
		return false;
	}
	if( WK7Event::mDebugPrintEnable==false ){
		return true;
	}
	K7AppDebugPrint(pstr);
	return true;
}
#endif

WK7EventController*	WK7EventController::mpEventControllerInstance = NULL;

WK7EventController* WK7EventController::GetInstance(void)
{
	if( mpEventControllerInstance==NULL ){
		ghmGcMemSetAllocName("WK7EVCON");
		mpEventControllerInstance = new WK7EventController;
	}
	return mpEventControllerInstance;
}

WK7EventController::WK7EventController()
{
	mpGlobalEventHandler     = NULL;
	mpScenarioEventHandler   = NULL;
	mppGlobalEventDesc       = NULL;
	mppScenarioEventDesc     = NULL;
	mGlobalEventHandlerNum   = 0;
	mScenarioEventHandlerNum = 0;
	mGlobalEventDescNum      = 0;
	mScenarioEventDescNum    = 0;
	mState                   = WK7EVENTCONTROLLERSTATE_IDLE;
	mMessageHandle           = -1;
	mpKeepMemory             = NULL;
	mStageMemorySize         = -1;
	mRestMemorySize          = -1;
	mMatMemorySize           = -1;
	mYamMemorySize           = -1;
	mSatMemorySize           = -1;
	mKawMemorySize           = -1;
	mWatMemorySize           = -1;
	mDynamicMemorySize       = -1;
	mMemoryCheck             = false;
	mpModule                 = NULL;
	mpModuleWork             = NULL;
	mStaticDataState         = DATASTATE_IDLE;
	mStaticDataProc          = DATAPROC_IDLE;
	mTempRelease             = 0;
}

WK7EventController::~WK7EventController()
{
	Terminate();
}

int WK7EventController::GetHandlerTableNum(K7EventHandler *pHandler)
{
	if( pHandler==NULL ){
		return 0;
	}
	int		count = 0;
	while( 1 ){
		if( pHandler->GetID       ()==0xffffffff        &&
		    pHandler->GetCondition()==K7EV_COND_UNKNOWN &&
		    pHandler->mpHandler     ==NULL              &&
		    pHandler->mpTermHandler ==NULL              ){
			break;
		}
		pHandler++;
		count++;
	}
	return count;
}

int WK7EventController::GetDescTableNum(K7EventDesc **ppDesc)
{
	if( ppDesc==NULL ){
		return 0;
	}
	int		count = 0;
	while( *ppDesc!=NULL ){
		count++;
		ppDesc++;
	}
	return count;
}

//#define	STATIC_MEMORY_SIZE		(0x2b2000)
#define	STATIC_MEMORY_SIZE		(0x280000)
//#define	STATIC_MEMORY_SIZE		(0x290000)
bool WK7EventController::FreeKeepMemory(void)
{
	GHMGCMEM_SAFE_FREE(mpKeepMemory);
	return true;
}

bool WK7EventController::AllocKeepMemory(void)
{
	void		*pStaticMemorySize;
	int			size;
	FreeKeepMemory();
	size = STATIC_MEMORY_SIZE;
	if( mMatMemorySize > 0 ){
		size -= mMatMemorySize;
	}
	if( mYamMemorySize > 0 ){
		size -= mYamMemorySize;
	}
	if( mSatMemorySize > 0 ){
		size -= mSatMemorySize;
	}
	if( mKawMemorySize > 0 ){
		size -= mKawMemorySize;
	}
	if( mWatMemorySize > 0 ){
		size -= mWatMemorySize;
	}
	if( mDynamicMemorySize > 0 ){
		size -= mDynamicMemorySize;
	}
	if( mMemoryCheck==true ){
		mRestMemorySize  = size;
		return true;
	}
//	K7AppOSDumpMemoryType(1);
	if( (pStaticMemorySize = ghmGcMemAlloc(size))==NULL ){
//	if( (pStaticMemorySize = ghmGcMemAlloc(STATIC_MEMORY_SIZE))==NULL ){
#ifdef	K7EVENT_MEMORY_DEBUG
		K7AppDumpMemoryType(1);
#endif
		return false;
	}
//	K7AppOSDumpMemoryType(1);
	long 	freesize = ghmGcMemGetFreeSize();
	mStageMemorySize = freesize;
	mpKeepMemory = ghmGcMemAlloc(freesize - 64);
//	mpKeepMemory = ghmGcMemAlloc(freesize);
	OSDebugReport("Free Memory Size = 0x%08x\n",freesize);
	GHMGCMEM_SAFE_FREE(pStaticMemorySize);
	if( mpKeepMemory==NULL ){
//		K7AppOSDumpMemoryType(1);
#ifdef	K7EVENT_MEMORY_DEBUG
		K7AppDumpMemoryType(1);
#endif
		return false;
	}
	mRestMemorySize  = ghmGcMemGetFreeSize();
	return true;
}

bool WK7EventController::AllocRestMemory(void)
{
	if( mRestMemorySize < 0 ){
		return false;
	}
	FreeKeepMemory();
	if( (mpKeepMemory = ghmGcMemAlloc(mRestMemorySize))==NULL ){
		return false;
	}
	return true;
}

bool WK7EventController::SetScenarioEventHandler(K7EventHandler *pHandler)
{
	mpScenarioEventHandler = pHandler;
	return true;
}

bool WK7EventController::SetScenarioEventDesc(K7EventDesc **ppDesc)
{
	mppScenarioEventDesc   = ppDesc;
	return true;
}

bool WK7EventController::Initialize(void)
{
	
	if( mState!=WK7EVENTCONTROLLERSTATE_IDLE ){
		// ViIN
		return false;
	}
	
	if( GameInfo.GetContinue()==false ){
		WK7EventItem::Initialize();
	}
	
	K7EV_HANDLER_TABLE_DECL
	K7EV_DESC_TABLE_DECL
	mpGlobalEventHandler   = K7EV_HANDLER_TABLE;
	mppGlobalEventDesc     = K7EV_DESC_TABLE;
	
	if( mTaskController.Initialize()==false ){
		return false;
	}
	
	mState = WK7EVENTCONTROLLERSTATE_INIT;
	
	return true;
}

bool WK7EventController::Terminate(void)
{
	
	// og֘Ã^XNI@
	KK7BattleMainTask::BattleMainTaskShutDown();
	WK7PlayerControl::SpAtkTaskShutDown();
	
	
	if( mTaskController.Terminate()==false ){
		return false;
	}
	
	DeleteEffectAll();
	
	if( mMessageHandle >= 0 ){
		ReleaseMessagePack((s8)mMessageHandle);
		mMessageHandle = -1;
	}
	
	WK7HitObj::SetTable   (NULL);
	WK7EventItem::SetTable(NULL);
	WK7Event::SetIDTable  (NULL);
	
	mState                   = WK7EVENTCONTROLLERSTATE_IDLE;
	mpGlobalEventHandler     = NULL;
	mpScenarioEventHandler   = NULL;
	mppGlobalEventDesc       = NULL;
	mppScenarioEventDesc     = NULL;
	mGlobalEventHandlerNum   = 0;
	mScenarioEventHandlerNum = 0;
	mGlobalEventDescNum      = 0;
	mScenarioEventDescNum    = 0;
	
	mMemoryCheck             = false;
	mRestMemorySize          = -1;
	mMatMemorySize           = -1;
	mYamMemorySize           = -1;
	mSatMemorySize           = -1;
	mKawMemorySize           = -1;
	mWatMemorySize           = -1;
	mDynamicMemorySize       = -1;
	mStaticDataState         = DATASTATE_IDLE;
	mStaticDataProc          = DATAPROC_IDLE;
	mTempRelease             = 0;
	
	ghmGcModule::Release(mpModule);
	mpModule = NULL;
	GHMGCMEM_SAFE_FREE(mpModuleWork);
	
	FreeKeepMemory();
	
	return true;
}

#if	defined(USE_ISDOLVIEWER)
#define	MODULE_SUFFIX		"v"
#elif defined(_DEBUG)
#define	MODULE_SUFFIX		"d"
#else
#define	MODULE_SUFFIX		"r"
#endif
#define	MODULE_SIZE_MAX		(0x59000)		// 352K
//#define	MODULE_SIZE_MAX		(0x52000)		// 328K
//#define	MODULE_SIZE_MAX		(0x50000)		// 320K
//#define	MODULE_SIZE_MAX		(0x4b000)		// 300K
bool WK7EventController::LoadModule(void)
{
	const char		*pModuleName;
	int				size = MODULE_SIZE_MAX;
	switch( GameInfo.GetScenarioID() ){
	case	K7SCN01: pModuleName="/Module/scn01" MODULE_SUFFIX ".rel"; break;
	case	K7SCN02: pModuleName="/Module/scn02" MODULE_SUFFIX ".rel"; break;
#ifdef	_DEBUG
	case	K7SCN03: pModuleName="/Module/scn03" MODULE_SUFFIX ".rel"; size = 0x5b240; break;
#else
	case	K7SCN03: pModuleName="/Module/scn03" MODULE_SUFFIX ".rel"; size = 0x5b000; break;
#endif
	case	K7SCN04: pModuleName="/Module/scn04" MODULE_SUFFIX ".rel"; break;
	case	K7SCN05: pModuleName="/Module/scn05" MODULE_SUFFIX ".rel"; break;
	case	K7SCN06: pModuleName="/Module/scn06" MODULE_SUFFIX ".rel"; break;
	case	K7SCN07: pModuleName="/Module/scn07" MODULE_SUFFIX ".rel"; break;
	case	K7SCN99: pModuleName="/Module/scn99" MODULE_SUFFIX ".rel"; break;
	default:         pModuleName=NULL;                                 break;
	}
	
	if( pModuleName!=NULL ){
		if( (mpModuleWork = ghmGcMemAlloc(size))==NULL ){
			return false;
		}
		mpModule = ghmGcModule::Create(pModuleName,mpModuleWork,size);
	}
	return true;
}

bool WK7EventController::LoadStaticData(void)
{
//	if( mState!=WK7EVENTCONTROLLERSTATE_INIT ){
//		return false;
//	}
	
	// GtFNg
	EffectInit(mpRenderProcessTaskController,WK7MAINTASK_SCREENCAPTURE_TASK_PRIO);
	
	// JIuWFNg̃Zbg
	SetEffectCamera(WK7CAM->GetCObj());
	
	//쐬̏
	mK7InitProc();
	
	ghmGcMemSetAllocName("SKSPDATA");
	if( (g_pSpData = new SK7SpecialData)!=NULL ){
		g_pSpData->Initialize();
	}
	
	WK7DeadBodySign::LoadData();
		
	mState = WK7EVENTCONTROLLERSTATE_LOAD;
	return true;
}

bool WK7EventController::IsLoadedStaticData(void)
{
	// I҂
	if( mK7CheckInitEnd()==false ){
		return false;
	}
	if( EffectSetupCheck()==FALSE ){
		return false;
	}
	if( g_pSpData!=NULL && g_pSpData->GetFileReadEnd()==false ){
		return false;
	}
	if( WK7DeadBodySign::IsLoadedData()==false ){
		return false;
	}
	if( WK7RenderObj::IsInitialized()==false ){
		return false;
	}
	return true;
}

bool WK7EventController::ReleaseStaticData(void)
{
	
	WK7DeadBodySign::ReleaseData();
	
	if( g_pSpData!=NULL ){
		g_pSpData->Terminate();
		GHM_SAFE_DELETE(g_pSpData);
	}
	
	//쐬̏I
	mK7TermProc();
	
	EffectEnd();
	
	return true;
}

bool WK7EventController::ReloadDynamicData(void)
{
	ReleaseDynamicData();
	return LoadDynamicData();
}

bool WK7EventController::LoadDynamicData(void)
{
	ghmGcMemSetAllocName("SKSTDATA");
	if( (g_pStageData = new SK7StageData)!=NULL ){
		g_pStageData->Initialize();
	}
	
	mState = WK7EVENTCONTROLLERSTATE_DYNAMICLOAD;
	return true;
}

bool WK7EventController::IsLoadedDynamicData(void)
{
	// I҂
	if( g_pStageData!=NULL && g_pStageData->GetFileReadEnd()==false ){
		return false;
	}
	return true;
}

bool WK7EventController::ReleaseDynamicData(void)
{
	if( g_pStageData!=NULL ){
		g_pStageData->Terminate();
		GHM_SAFE_DELETE(g_pStageData);
	}
	
	return true;
}

bool WK7EventController::InitProcess(void)
{
	if( GameInfo.GetNewGame()==true ){
		// j[Q[ɃO[otONA
		GameInfo.ClearGlobalFlag();
	}
	// Xe[WtONA
	WK7Stage::ClearFlag();
	// ViItONA
	GameInfo.ClearScenarioFlag();
	// LN^X^[gn_Zbg
	WK7PLAYERCONTROL->ChangeStartChkPoint();
	// ViICxg
	K7EventDescData		data;
	K7EventDesc			*pDesc;
	data.Initialize();
	while( (pDesc=GetEventDesc(K7EV_DESC_TYPE_SCNINIT,data))!=NULL ){
		Call(pDesc->GetID(),(u32)GameInfo.GetScenarioID());
	}
#ifdef	DEBUG_PRINT
	K7AppDebugPrint("ScenarioInitTask Start !!");
#endif
	return true;
}

bool WK7EventController::ContinueProcess(void)
{
	// SLĎ^XN
//	BattleCtrl.RestoreDeadTask();
	// ViIReBj[Cxg
	K7EventDescData		data;
	K7EventDesc			*pDesc;
	data.Initialize();
	while( (pDesc=GetEventDesc(K7EV_DESC_TYPE_SCNCONTINUE,data))!=NULL ){
		Call(pDesc->GetID(),(u32)GameInfo.GetScenarioID());
	}
	if( mTaskController.Start(ContinueTask,0,0,NULL,WK7EVENT_TASK_STACK_SIZE) < 0 ){
		return false;
	}
#ifdef	DEBUG_PRINT
	K7AppDebugPrint("ScenarioContinueTask Start !!");
#endif
	return true;
}

void WK7EventController::ContinueTask(ghmGcTask *)
{
	K7SCR_ENTER_PLAYMODE();
	K7SCR_ENTER_DEMOPAUSE();
	
	K7SCR_MOON_START();
	K7SCR_MOON_WAIT();
	
//	while( gpNoise->mSandRequest(eSandLongNoise,eNoiseColorBlue)==false ){
//		K7SCR_RETSYS();
//	}
	
	u32		stgid;
	u32		chrid;
	
	stgid = GameInfo.GetCurrentStageID();
	chrid = GameInfo.GetCharaID();
	
	K7SCR_STG_CREATE_B   (stgid);
	K7SCR_PC_CREATE      (chrid);				// fǂݍ
	K7SCR_PC_MOTIONCREATE(chrid);				// [Vǂݍ
	K7SCR_PC_CREATEWAIT  ();					// ǂݍݑ҂
	K7SCR_STG_SET_CURRENT(stgid);
	
	WK7Stage	*pStage = WK7RenderObj::GetStage(stgid);
	if( pStage!=NULL ){
		pStage->SetCurrentPath(GameInfo.GetCurrentStagePathName (),
		                       GameInfo.GetCurrentStagePathParam(),
		                       GameInfo.GetCurrentStagePathDir  (),
		                       WK7PATHINFO_INITDIR_TGTDIR_REV);
		WK7PLAYERCONTROL->SetPathSelectState();
		if( stgid==K7STGID(121,0) && strcmp(GameInfo.GetCurrentStagePathName(),WK7STAGE_TVSTAGE_BPNAME)==0 ){
			pStage->SetCameraStartPosition(WK7STAGE_TVSTAGE_BPNAME);
		}
		pStage->RenewPath();
		
	}
	
	// X^[gJ̐ݒ
	WK7Stage::SetCameraStartPosition(stgid);
	K7SCR_RETSYS();
	
//	while( gpNoise->mGetSandFrameProc()!=eSandExec ){
//		K7SCR_RETSYS();
//	}
	
	K7SCR_PC_VISIBLE     (true);
	K7SCR_RETSYS();
	K7SCR_STG_VISIBLE    (stgid,true);
	
//	WK7Stage::SandRelease(stgid);
//	while( gpNoise->mGetSandFrameProc()!=eSandStandby ){
//		K7SCR_RETSYS();
//	}
	
	k7app.FadeStart(0,WK7_TICK_PER_SEC);
	while( k7app.IsFade()==true ){
		GHMGCTASK_RETSYS();
	}
	
	K7SCR_EXIT_PLAYMODE();
	K7SCR_EXIT_DEMOPAUSE();
}

bool WK7EventController::LoadData(void)
{
	if( mStaticDataState!=DATASTATE_IDLE ){
		return false;
	}
	if( mStaticDataState!=DATASTATE_LOADING ){
		switch( mStaticDataProc ){
		case	DATAPROC_DYNAMICRELEASED:
			if( AllocRestMemory()==false ){
				return false;
			}
			break;
		default:
			if( AllocKeepMemory()==false ){
				return false;
			}
			break;
		}
	}
	mStaticDataState = DATASTATE_LOADING;
	return true;
}

bool WK7EventController::IsLoadedData(void)
{
	if( mStaticDataState!=DATASTATE_LOADED ){
		return false;
	}
	return true;
}

bool WK7EventController::ReleaseData(void)
{
	if( mStaticDataState!=DATASTATE_LOADED ){
		return false;
	}
	FreeKeepMemory();
	mTempRelease     = 0;
	mStaticDataState = DATASTATE_RELEASE;
	return true;
}

bool WK7EventController::IsReleasedData(void)
{
	if( mStaticDataState!=DATASTATE_RELEASED ){
		return false;
	}
	return true;
}

bool WK7EventController::ReleaseTempData(void)
{
	if( mStaticDataState!=DATASTATE_LOADED ){
		return false;
	}
	FreeKeepMemory();
	mTempRelease     = 1;
	mStaticDataState = DATASTATE_RELEASE;
	return true;
}

bool WK7EventController::IsReleasedTempData(void)
{
	return IsReleasedData();
}

bool WK7EventController::ReleaseTempData2(void)
{
	if( mStaticDataState!=DATASTATE_LOADED ){
		return false;
	}
	FreeKeepMemory();
	mTempRelease     = 2;
	mStaticDataState = DATASTATE_RELEASE;
	return true;
}

bool WK7EventController::IsReleasedTempData2(void)
{
	return IsReleasedData();
}

bool WK7EventController::ReleaseTempData3(void)
{
	if( mStaticDataState!=DATASTATE_LOADED ){
		return false;
	}
	FreeKeepMemory();
	mTempRelease     = 3;
	mStaticDataState = DATASTATE_RELEASE;
	return true;
}

bool WK7EventController::IsReleasedTempData3(void)
{
	return IsReleasedData();
}

bool WK7EventController::FrameProcessData(void)
{
#ifdef	DEBUG_STOP
static	bool	stop = false;
	if( stop==true ){
		if( HSD_PadGetTrigger(1) & HSD_PAD_B ){
			stop = false;
		}
		return true;
	}
#endif
#ifdef	DEBUG_PRINT
	char	dbgstr[256];
#endif
//	if( HSD_PadGetTrigger(1) & HSD_PAD_A ){
//		K7AppOSDumpMemory();
//	}
	
	switch( mStaticDataState ){
	case	DATASTATE_IDLE:
		break;
	case	DATASTATE_LOADING:
		switch( mStaticDataProc ){
		case	DATAPROC_IDLE:
//		case	DATAPROC_MATRELEASED:
			mK7InitProc();
			mStaticDataProc = DATAPROC_MATLOAD;
#ifdef	DEBUG_STOP
			stop = true; break;
#endif
		case	DATAPROC_MATLOAD:
			if( mK7CheckInitEnd()==false ){
				break;
			}
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Mat Mem Loaded >>>\n");
			K7AppDumpMemoryType(1);
#endif
			if( mMemoryCheck==false ){
				mMatMemorySize     = mRestMemorySize - ghmGcMemGetFreeSize();
				mRestMemorySize    = ghmGcMemGetFreeSize();
			}
#ifdef	DEBUG_PRINT
			sprintf(dbgstr,"M Memory Size : %d",mMatMemorySize);
			OSReportEV("%s\n",dbgstr);
			if( GameInfo.GetScenarioID()!=K7SCN_UNKNOWN ){
				K7AppDebugPrint(dbgstr);
			}
#endif
			mStaticDataProc = DATAPROC_MATLOADED;
		case	DATAPROC_MATLOADED:
		case	DATAPROC_YAMRELEASED:
//			BattleCtrl.RestoreDeadTask();
			// GtFNg
			EffectInit(mpRenderProcessTaskController,WK7MAINTASK_SCREENCAPTURE_TASK_PRIO);
			// JIuWFNg̃Zbg
			SetEffectCamera(WK7CAM->GetCObj());
			mStaticDataProc = DATAPROC_YAMLOAD;
#ifdef	DEBUG_STOP
			stop = true; break;
#endif
		case	DATAPROC_YAMLOAD:
			if( EffectSetupCheck()==FALSE ){
				break;
			}
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Yam Mem Loaded >>>\n");
			K7AppDumpMemoryType(1);
#endif
			if( mMemoryCheck==false ){
				mYamMemorySize     = mRestMemorySize - ghmGcMemGetFreeSize();
				mRestMemorySize    = ghmGcMemGetFreeSize();
			}
#ifdef	DEBUG_PRINT
			sprintf(dbgstr,"Y Memory Size : %d",mYamMemorySize);
			OSReportEV("%s\n",dbgstr);
			if( GameInfo.GetScenarioID()!=K7SCN_UNKNOWN ){
				K7AppDebugPrint(dbgstr);
			}
#endif
			mStaticDataProc = DATAPROC_YAMLOADED;
		case	DATAPROC_YAMLOADED:
		case	DATAPROC_SATRELEASED:
			ghmGcMemSetAllocName("SKSPDATA");
			if( g_pSpData==NULL && (g_pSpData = new SK7SpecialData)==NULL ){
				break;
			}
			g_pSpData->Initialize();
			mStaticDataProc = DATAPROC_SATLOAD;
#ifdef	DEBUG_STOP
			stop = true; break;
#endif
		case	DATAPROC_SATLOAD:
			if( g_pSpData!=NULL && g_pSpData->GetFileReadEnd()==false ){
				break;
			}
			GameInfo.CreateSpOption();
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Sat Mem Loaded >>>\n");
			K7AppDumpMemoryType(1);
#endif
			if( mMemoryCheck==false ){
				mSatMemorySize     = mRestMemorySize - ghmGcMemGetFreeSize();
				mRestMemorySize    = ghmGcMemGetFreeSize();
			}
#ifdef	DEBUG_PRINT
			sprintf(dbgstr,"S Memory Size : %d",mSatMemorySize);
			OSReportEV("%s\n",dbgstr);
			if( GameInfo.GetScenarioID()!=K7SCN_UNKNOWN ){
				K7AppDebugPrint(dbgstr);
			}
#endif
			mStaticDataProc = DATAPROC_SATLOADED;
		case	DATAPROC_SATLOADED:
		case	DATAPROC_KAWRELEASED:
			mStaticDataProc = DATAPROC_KAWLOAD;
		case	DATAPROC_KAWLOAD:
			if( mMemoryCheck==false ){
				mKawMemorySize     = mRestMemorySize - ghmGcMemGetFreeSize();
				mRestMemorySize    = ghmGcMemGetFreeSize();
			}
#ifdef	DEBUG_PRINT
			sprintf(dbgstr,"K Memory Size : %d",mKawMemorySize);
			OSReportEV("%s\n",dbgstr);
			if( GameInfo.GetScenarioID()!=K7SCN_UNKNOWN ){
				K7AppDebugPrint(dbgstr);
			}
#endif
			mStaticDataProc = DATAPROC_KAWLOADED;
		case	DATAPROC_KAWLOADED:
		case	DATAPROC_WATRELEASED:
			if( (mMessageHandle = (int)LoadMessagePack("Message/System/System.jmb")) < 0 ){
				break;
			}
			if( WK7DeadBodySign::LoadData()==false ){
				break;
			}
//			WK7Memo::LoadStart();
			mStaticDataProc = DATAPROC_WATLOAD;;
#ifdef	DEBUG_STOP
			stop = true; break;
#endif
		case	DATAPROC_WATLOAD:
			if( mMessageHandle >= 0 && CheckMessagePackLoadComp((s8)mMessageHandle)==0 ){
				break;
			}
			if( WK7DeadBodySign::IsLoadedData()==false ){
				break;
			}
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Wat Mem Loaded >>>\n");
			K7AppDumpMemoryType(1);
#endif
//			if( WK7Memo::IsLoaded()==false ){
//				break;
//			}
			if( mMemoryCheck==false ){
				if( mWatMemorySize < 0 ){
					mWatMemorySize  = mRestMemorySize - ghmGcMemGetFreeSize();
					mRestMemorySize = ghmGcMemGetFreeSize();
					OSDebugReport("Rest Memory : 0x%08x\n",mRestMemorySize);
				}
			}
			if( AllocRestMemory()==false ){
				break;
			}
#ifdef	DEBUG_PRINT
			sprintf(dbgstr,"W Memory Size : %d",mWatMemorySize);
			OSReportEV("%s\n",dbgstr);
			if( GameInfo.GetScenarioID()!=K7SCN_UNKNOWN ){
				K7AppDebugPrint(dbgstr);
			}
			sprintf(dbgstr,"Rest Mem Size : %d",mRestMemorySize);
			OSReportEV("%s\n",dbgstr);
			if( GameInfo.GetScenarioID()!=K7SCN_UNKNOWN ){
				K7AppDebugPrint(dbgstr);
			}
#endif
			BattleCtrl.RestoreDeadTask();
			mStaticDataProc = DATAPROC_WATLOADED;
		case	DATAPROC_WATLOADED:
		case	DATAPROC_DYNAMICRELEASED:
			ghmGcMemSetAllocName("SKSTDATA");
			if( g_pStageData==NULL && (g_pStageData = new SK7StageData)==NULL ){
				break;
			}
			g_pStageData->Initialize();
			mStaticDataProc  = DATAPROC_DYNAMICLOAD;
#ifdef	DEBUG_STOP
			stop = true; break;
#endif
		case	DATAPROC_DYNAMICLOAD:
			if( g_pStageData!=NULL && g_pStageData->GetFileReadEnd()==false ){
				break;
			}
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Dyn Mem Loaded >>>\n");
			K7AppDumpMemoryType(1);
#endif
			if( mMemoryCheck==false ){
				mDynamicMemorySize = mStageMemorySize - ghmGcMemGetFreeSize();
			}
#ifdef	DEBUG_PRINT
			sprintf(dbgstr,"D Memory Size : %d",mDynamicMemorySize);
			OSReportEV("%s\n",dbgstr);
			if( GameInfo.GetScenarioID()!=K7SCN_UNKNOWN ){
				K7AppDebugPrint(dbgstr);
			}
#endif
			mStaticDataProc  = DATAPROC_DYNAMICLOADED;
		case	DATAPROC_DYNAMICLOADED:
		case	DATAPROC_LOADED:
			mStaticDataProc  = DATAPROC_LOADED;
			mStaticDataState = DATASTATE_LOADED;
			mMemoryCheck = true;
			break;
		default:
			return false;
		}
		break;
	case	DATASTATE_LOADED:
		break;
	case	DATASTATE_RELEASE:
		switch( mStaticDataProc ){
		case	DATAPROC_LOADED:
		case	DATAPROC_DYNAMICLOADED:
			if( g_pStageData!=NULL ){
				g_pStageData->Terminate();
				GHM_SAFE_DELETE(g_pStageData);
			}
//			mDynamicMemorySize = -1;
			mStaticDataProc = DATAPROC_DYNAMICRELEASE;
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Dyn Mem Released >>>\n");
			K7AppDumpMemoryType(1);
#endif
			break;
		case	DATAPROC_DYNAMICRELEASE:
			mStaticDataProc = DATAPROC_DYNAMICRELEASED;
			if( mTempRelease==3 ){
				mStaticDataState = DATASTATE_RELEASED;
				break;
			}
			break;
		case	DATAPROC_WATLOADED:
		case	DATAPROC_DYNAMICRELEASED:
			BattleCtrl.ReleaseDeadTask();
			WK7DeadBodySign::ReleaseData();
			if( mMessageHandle >= 0 ){
				ReleaseMessagePack((s8)mMessageHandle);
				mMessageHandle = -1;
			}
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Wat Mem Released >>>\n");
			K7AppDumpMemoryType(1);
#endif
//			mWatMemorySize = -1;
			mStaticDataProc = DATAPROC_WATRELEASE;
			break;
		case	DATAPROC_WATRELEASE:
			mStaticDataProc = DATAPROC_WATRELEASED;
			break;
		case	DATAPROC_KAWLOADED:
		case	DATAPROC_WATRELEASED:
//			mKawMemorySize = -1;
			mStaticDataProc = DATAPROC_KAWRELEASE;
			break;
		case	DATAPROC_KAWRELEASE:
			mStaticDataProc = DATAPROC_KAWRELEASED;
			break;
		case	DATAPROC_SATLOADED:
		case	DATAPROC_KAWRELEASED:
			GameInfo.ReleaseSpOption();
			if( g_pSpData!=NULL ){
				g_pSpData->Terminate();
				GHM_SAFE_DELETE(g_pSpData);
			}
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Sat Mem Released >>>\n");
			K7AppDumpMemoryType(1);
#endif
//			mSatMemorySize = -1;
			mStaticDataProc = DATAPROC_SATRELEASE;
			break;
		case	DATAPROC_SATRELEASE:
			mStaticDataProc = DATAPROC_SATRELEASED;
			if( mTempRelease==2 ){
				mStaticDataState = DATASTATE_RELEASED;
				break;
			}
			break;
		case	DATAPROC_YAMLOADED:
		case	DATAPROC_SATRELEASED:
			EffectEnd();
//			BattleCtrl.ReleaseDeadTask();
//			mYamMemorySize = -1;
			mStaticDataProc = DATAPROC_YAMRELEASE;
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Yam Mem Released >>>\n");
			K7AppDumpMemoryType(1);
#endif
			break;
		case	DATAPROC_YAMRELEASE:
			mStaticDataProc = DATAPROC_YAMRELEASED;
			if( mTempRelease==1 ){
				mStaticDataState = DATASTATE_RELEASED;
				break;
			}
			break;
		case	DATAPROC_MATLOADED:
		case	DATAPROC_YAMRELEASED:
			mK7TermProc();
#ifdef	K7EVENT_MEMORY_DEBUG
			OSReport("<<< Mat Mem Released >>>\n");
			K7AppDumpMemoryType(1);
#endif
//			mMatMemorySize = -1;
			mStaticDataProc = DATAPROC_MATRELEASE;
			break;
		case	DATAPROC_MATRELEASE:
			mStaticDataProc = DATAPROC_MATRELEASED;
			break;
		case	DATAPROC_IDLE:
			mStaticDataState = DATASTATE_RELEASED;
			break;
		default:
			return false;
		}
		break;
	default:
		return false;
	}
	return true;
}

bool WK7EventController::FrameProcess(void)
{
#if 0
static	int	proc = 0;
	switch( proc ){
	case	0:
		if( HSD_PadGetTrigger(1) & HSD_PAD_A ){
			proc++;
		}
		break;
	case	1:
		if( WK7Stage::ReleaseStageArchive()==false ){
			break;
		}
		proc++;
	case	2:
		if( WK7Stage::IsReleasedStageArchive()==false ){
			break;
		}
		proc++;
	case	3:
		if( IsLoadedData()==true ){
			ReleaseTempData();
//			ReleaseData();
			proc++;
		}
		break;
	case	4:
		if( IsReleasedData()==true ){
			proc = 100;
		}
		break;
	case	100:
#ifdef	MOVIE_THP_FORMAT
		if( gpMovie->mGetStatus()==THP_PLAYER_STOP ){
#endif
#ifdef	MOVIE_HVQM4_FORMAT
		if( gpMovie->mGetStatus()==eMvProcStandby ){
#endif		
			if( HSD_PadGetTrigger(1) & HSD_PAD_A ){
//			if( HSD_PadGetTrigger(1) & HSD_PAD_LEFT ){
				if( gpMovie!=NULL && gpMovie->mRequest("/Movie/cloudman.h4m")==true ){
//				if( gpMovie!=NULL && gpMovie->mRequest("/Mat/gurad.h4m")==true ){
//				if( gpMovie!=NULL && gpMovie->mRequest("/Movie/q100gurad.h4m")==true ){
					
					break;
				}
			}
			if( HSD_PadGetTrigger(1) & HSD_PAD_B ){
//			if( HSD_PadGetTrigger(1) & HSD_PAD_A ){
				proc++;
			}
		}
		break;
	case	101:
		if( IsReleasedData()==true ){
			LoadData();
			proc++;
		}
		break;
	case	102:
		if( IsLoadedData()==true ){
			proc++;
		}
		break;
	case	103:
		if( WK7Stage::RestoreStageArchive()==false ){
			break;
		}
		proc++;
		break;
	case	104:
		if( WK7Stage::IsRestoredStageArchive()==false ){
			break;
		}
		proc = 0;
		break;
	}
#endif
	
	// f[^
	FrameProcessData();
	
//static	bool	streamEnable;
	switch( mState ){
	case	WK7EVENTCONTROLLERSTATE_INIT:
		switch( GameInfo.GetScenarioID() ){
		case	K7SCN04:
		case	K7SCN05:
		case	K7SCN06:
			ghmGcFile::ChangeDisk(1);
			break;
		case	K7SCN07:
		case	K7SCN01:
		case	K7SCN02:
		case	K7SCN03:
			ghmGcFile::ChangeDisk(0);
			break;
		default:
			break;
		}
//		streamEnable = mainMenu.IsStreamPlayEnable();
//		mainMenu.SetStreamPlayEnable(false);
		mState = WK7EVENTCONTROLLERSTATE_MODULELOAD;
	case	WK7EVENTCONTROLLERSTATE_MODULELOAD:
		int		diskid;
		switch( GameInfo.GetScenarioID() ){
		case	K7SCN04:
		case	K7SCN05:
		case	K7SCN06:
			diskid = 1;
			break;
		case	K7SCN07:
		case	K7SCN01:
		case	K7SCN02:
		case	K7SCN03:
			diskid = 0;
			break;
		default:
			diskid = -1;
			break;
		}
		if( diskid >= 0 && ghmGcFile::GetCurrentDiskID()!=diskid ){
			break;
		}
//		mainMenu.SetStreamPlayEnable(streamEnable);
//		OSDebugReport("e1\n");
		if( LoadModule()==false ){
			break;
		}
//		OSDebugReport("e2\n");
		mState = WK7EVENTCONTROLLERSTATE_MODULELOADWAIT;
	case	WK7EVENTCONTROLLERSTATE_MODULELOADWAIT:
//		OSDebugReport("module loading\n");
		if( mpModule!=NULL && mpModule->GetState()!=ghmGcModule::AVAILABLE ){
			break;
		}
//		OSDebugReport("e3\n");
		if( LoadData()==false ){
			break;
		}
//		OSDebugReport("e4\n");
//		LoadStaticData();
	case	WK7EVENTCONTROLLERSTATE_LOAD:
		mState = WK7EVENTCONTROLLERSTATE_LOADWAIT;
	case	WK7EVENTCONTROLLERSTATE_LOADWAIT:
		// ǂݍݏI҂
		if( IsLoadedData()==false ){
//		if( IsLoadedStaticData()==false ){
			break;
		}
//		OSDebugReport("e5\n");
		mGlobalEventHandlerNum   = GetHandlerTableNum(mpGlobalEventHandler  );
		mGlobalEventDescNum      = GetDescTableNum   (mppGlobalEventDesc    );
		mScenarioEventHandlerNum = GetHandlerTableNum(mpScenarioEventHandler);
		mScenarioEventDescNum    = GetDescTableNum   (mppScenarioEventDesc  );
		WK7Stage::InitDefEvent();
		mState = WK7EVENTCONTROLLERSTATE_RUN;
		// og֘Ã^XNJn 
		if ( false == KK7BattleMainTask::BattleMainTaskSetUp(mpFrameProcessTaskController) ) {
			return false;
		}
		if ( false == WK7PlayerControl::SpAtkTaskSetUp(mpFrameProcessTaskController) ) {
			return false;
		}
		InitStageTitleData();
		if( GameInfo.GetContinue()==true ){
			ContinueProcess();
		}else{
			InitProcess();
		}
		GameInfo.InitAfterLoad();
	case	WK7EVENTCONTROLLERSTATE_RUN:
		break;
		mState = WK7EVENTCONTROLLERSTATE_RUN;
	default:
		return false;
	}
	FrameProcessEffect();
	if( mTaskController.FrameProcess()==false ){
		return false;
	}
	return true;
}

WK7EventEffect::WK7EventEffect(ghmGcListObj *pAnchor,EFFECT *pEffect) : ghmGcListObj(pAnchor,WK7OBJ_DEFAULT_PRIO)
{
	if( (mpEffect = pEffect)!=NULL ){
		mEffectID = GetEffectID(pEffect);
	}else{
		mEffectID = 0xffffffff;
	}
	mType = NORMAL;
}

WK7EventEffect::~WK7EventEffect()
{
	if( mpEffect != NULL && CheckEffectValidState(mpEffect,mEffectID)==TRUE ){
		::DeleteEffect(mpEffect);
	}
}

bool WK7EventEffect::IsValid(void)
{
	if( mpEffect==NULL ){
		return false;
	}
	if( CheckEffectValidState(mpEffect,mEffectID)==FALSE ){
		return false;
	}
	return true;
}

bool WK7EventEffect::FrameProcess(void)
{
	if( IsValid()==false ){
		return false;
	}
	return true;
}

bool WK7EventController::FrameProcessEffect(void)
{
	WK7EventEffect		*pEffect;
	WK7EventEffect		*pNext;
	for(pEffect=(WK7EventEffect *)mEffectAnchor.GetNext() ; pEffect!=NULL ; pEffect=pNext){
		pNext = pEffect->GetNext();
		if( pEffect->FrameProcess()==false ){
			delete pEffect;
		}
	}
	return true;
}

bool WK7EventController::DeleteEffectAll(void)
{
	WK7EventEffect		*pEffect;
	WK7EventEffect		*pNext;
	for(pEffect=(WK7EventEffect *)mEffectAnchor.GetNext() ; pEffect!=NULL ; pEffect=pNext){
		pNext = pEffect->GetNext();
		delete pEffect;
	}
	return true;
}

bool WK7EventController::IsValidEffect(WK7EventEffect *pEffect)
{
	WK7EventEffect		*pTempEffect;
	for(pTempEffect=(WK7EventEffect *)mEffectAnchor.GetNext() ; pTempEffect!=NULL ; pTempEffect=pTempEffect->GetNext()){
		if( pEffect==pTempEffect ){
			return true;
		}
	}
	return false;
}

bool WK7EventController::IsValidMessage(u32 messageID)
{
	WK7EventEffect		*pEffect;
	for(pEffect=(WK7EventEffect *)mEffectAnchor.GetNext() ; pEffect!=NULL ; pEffect=pEffect->GetNext()){
		if( pEffect->GetType()!=WK7EventEffect::MESSAGE ){
			continue;
		}
		WK7EventEffectMessage	*pEffectMessage = (WK7EventEffectMessage *)pEffect;
		if( pEffectMessage->GetKind()!=WK7EVENTEFFECTMESSAGE_KIND_MESSAGE ){
			continue;
		}
		if( pEffectMessage->GetMessageID()!=messageID ){
			continue;
		}
		return true;
	}
	return false;
}

bool WK7EventController::IsDisplayedMessage(u32 messageID)
{
	WK7EventEffect		*pEffect;
	for(pEffect=(WK7EventEffect *)mEffectAnchor.GetNext() ; pEffect!=NULL ; pEffect=pEffect->GetNext()){
		if( pEffect->GetType()!=WK7EventEffect::MESSAGE ){
			continue;
		}
		WK7EventEffectMessage	*pEffectMessage = (WK7EventEffectMessage *)pEffect;
		if( pEffectMessage->GetKind()!=WK7EVENTEFFECTMESSAGE_KIND_MESSAGE ){
			continue;
		}
		if( pEffectMessage->GetMessageID()!=messageID ){
			continue;
		}
		EFFECT	*pEffect = pEffectMessage->GetEffect();
		if( pEffect==NULL ){
			continue;
		}
		if( ::GetStopStateMessage2Effect(pEffect)==FALSE ){
			return false;
		}
		return true;
	}
	return false;
}

bool WK7EventController::IsValidItemGet(u32 messageID)
{
	WK7EventEffect		*pEffect;
	for(pEffect=(WK7EventEffect *)mEffectAnchor.GetNext() ; pEffect!=NULL ; pEffect=pEffect->GetNext()){
		if( pEffect->GetType()!=WK7EventEffect::MESSAGE ){
			continue;
		}
		WK7EventEffectMessage	*pEffectMessage = (WK7EventEffectMessage *)pEffect;
		if( pEffectMessage->GetKind()!=WK7EVENTEFFECTMESSAGE_KIND_ITEMGET ){
			continue;
		}
		if( pEffectMessage->GetMessageID()!=messageID ){
			continue;
		}
		return true;
	}
	return false;
}

bool WK7EventController::IsDisplayedItemGet(u32 messageID)
{
	WK7EventEffect		*pEffect;
	for(pEffect=(WK7EventEffect *)mEffectAnchor.GetNext() ; pEffect!=NULL ; pEffect=pEffect->GetNext()){
		if( pEffect->GetType()!=WK7EventEffect::MESSAGE ){
			continue;
		}
		WK7EventEffectMessage	*pEffectMessage = (WK7EventEffectMessage *)pEffect;
		if( pEffectMessage->GetKind()!=WK7EVENTEFFECTMESSAGE_KIND_ITEMGET ){
			continue;
		}
		if( pEffectMessage->GetMessageID()!=messageID ){
			continue;
		}
		EFFECT	*pEffect = pEffectMessage->GetEffect();
		if( pEffect==NULL ){
			continue;
		}
		if( ::GetItemGetState(pEffect)==0 ){
			return false;
		}
		return true;
	}
	return false;
}

WK7EventEffectMessage* WK7EventController::CreateMessage(u32 messageID,s32 iconID,s32 soundID)	
{
	if( mMessageHandle < 0 ){
		return NULL;
	}
	EFFECT		*pMessageEffect;
	if( (pMessageEffect=CreateMessage2Effect((s8)mMessageHandle,(u16)messageID, TRUE, iconID,soundID))==NULL ){
		return NULL;
	}
	WK7EventEffectMessage	*pEventEffect;
	ghmGcMemSetAllocName("WK7EVEFF");
	if( (pEventEffect=new WK7EventEffectMessage(messageID,
	                                            iconID,
	                                            soundID,
	                                            WK7EVENTEFFECTMESSAGE_KIND_MESSAGE,
	                                            &mEffectAnchor,
	                                            pMessageEffect))==NULL ){
		::DeleteEffect(pMessageEffect);
		return NULL;
	}
	
	return pEventEffect;
}

WK7EventEffectMessage* WK7EventController::CallMessage(u32 messageID,s32 iconID,s32 soundID)
{
	return CreateMessage(messageID,iconID,soundID);
}

bool WK7EventController::CallMessageB(u32 messageID,s32 iconID,s32 soundID)
{
	bool SetOperateEnableSave;
	bool NpcPauseAllSave;
	
	WK7EventEffect* pEffect;
	
	while( (pEffect = CreateMessage(messageID, iconID, soundID))==NULL ){
		retsys();
	}
	
	{
		SetOperateEnableSave = WK7PLAYERCONTROL->GetOperateEnable();
		NpcPauseAllSave = WK7Npc::GetNpcPauseAll();
		
		WK7PLAYERCONTROL->SetOperateEnable(false);
		WK7Npc::NpcPauseAll();
		
		while( IsValidEffect(pEffect)==true ){
			retsys();
		}
//		while( false == GetStopStateMessageEffect( pMessageEffect ) ){
//			retsys();
//		}
		
#if 1
		WK7PLAYERCONTROL->SetOperateEnable(SetOperateEnableSave);
		if ( NpcPauseAllSave == false ){
			WK7Npc::NpcResetPauseAll();
		}
#else
//		WK7PLAYERCONTROL->SetOperateEnable(true);
//		WK7Npc::NpcResetPauseAll();
#endif
		
		return true;
	}
}

bool WK7EventController::CallMessageB2(u32 messageID,s32 iconID,s32 soundID)
{
	
	WK7EventEffect* pEffect;
	
	while( (pEffect = CreateMessage(messageID, iconID, soundID))==NULL ){
		retsys();
	}
	
	while( IsValidEffect(pEffect)==true ){
		retsys();
	}
	
	return true;
}

WK7EventEffectMessage* WK7EventController::CreateItemGet(u32 messageID,bool buttonCancel,s32 iconID,s32 soundID)	
{
	if( mMessageHandle < 0 ){
		return NULL;
	}
	EFFECT		*pEffect;
	if( (pEffect=CreateItemGetEffect((s8)mMessageHandle,(u16)messageID,buttonCancel,iconID,soundID))==NULL ){
		return NULL;
	}
	WK7EventEffectMessage		*pEventEffect;
	ghmGcMemSetAllocName("EVEFITGT");
	if( (pEventEffect=new WK7EventEffectMessage(messageID,
	                                            iconID,
	                                            soundID,
	                                            WK7EVENTEFFECTMESSAGE_KIND_ITEMGET,
	                                            &mEffectAnchor,
	                                            pEffect))==NULL ){
		::DeleteEffect(pEffect);
		return NULL;
	}
	return pEventEffect;
}

WK7EventEffectMessage* WK7EventController::CallItemGet(u32 messageID,bool buttonCancel,s32 iconID,s32 soundID)
{
	return CreateItemGet(messageID,buttonCancel,iconID,soundID);
}

bool WK7EventController::CallItemGetB(u32 messageID,bool buttonCancel,s32 iconID,s32 soundID)
{
	bool		SetOperateEnableSave;
	bool		NpcPauseAllSave;
	WK7EventEffect	*pEffect;
	
	while( (pEffect = CreateItemGet(messageID,buttonCancel,iconID,soundID))==NULL ){
		retsys();
	}
	
	SetOperateEnableSave = WK7PLAYERCONTROL->GetOperateEnable();
	NpcPauseAllSave = WK7Npc::GetNpcPauseAll();
	
	WK7PLAYERCONTROL->SetOperateEnable(false);
	WK7Npc::NpcPauseAll();
	
	while( IsValidEffect(pEffect)==true ){
		retsys();
	}
	WK7PLAYERCONTROL->SetOperateEnable(SetOperateEnableSave);
	if ( NpcPauseAllSave == false ){
		WK7Npc::NpcResetPauseAll();
	}
	return true;
}

bool WK7EventController::CallItemGetB2(u32 messageID,bool buttonCancel,s32 iconID,s32 soundID)
{
	WK7EventEffect	*pEffect;
	
	while( (pEffect = CreateItemGet(messageID,buttonCancel,iconID,soundID))==NULL ){
		retsys();
	}
	while( IsValidEffect(pEffect)==true ){
		retsys();
	}
	return true;
}

K7EventHandler* WK7EventController::GetHandler(u32 id)
{
	K7EventHandler	*pHandler;
	int				i;
	if( mpGlobalEventHandler != NULL ){
		pHandler = mpGlobalEventHandler;
		for(i=0 ; i<mGlobalEventHandlerNum ; i++,pHandler++){
			if( pHandler->GetID()==id ){
				return pHandler;
			}
		}
	}
	if( mpScenarioEventHandler != NULL ){
		pHandler = mpScenarioEventHandler;
		for(i=0 ; i<mScenarioEventHandlerNum ; i++,pHandler++){
			if( pHandler->GetID()==id ){
				return pHandler;
			}
		}
	}
	return NULL;
}

WK7Event* WK7EventController::GetEvent(u32 id)
{
	WK7Event	*pEvent;
	for(pEvent=(WK7Event *)mAnchor.GetNext() ; pEvent!=NULL ; pEvent=pEvent->GetNext()){
		if( pEvent->GetID()==id ){
			return pEvent;
		}
	}
	return NULL;
}

WK7Event* WK7EventController::GetEvent(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0,u32 param1)
{
	WK7Event	*pEvent;
	for(pEvent=(WK7Event *)mAnchor.GetNext() ; pEvent!=NULL ; pEvent=pEvent->GetNext()){
		if( pEvent->GetHandler    ()==pHandler     &&
		    pEvent->GetTermHandler()==pTermHandler &&
		    pEvent->GetParam(0)     ==param0       &&
		    pEvent->GetParam(1)     ==param1       ){
			return pEvent;
		}
	}
	return NULL;
}

K7EventDesc* WK7EventController::GetEventDesc(int type,K7EventDescData &data)
{
	if( data.mStep==-1 ){
		data.mStep  = 0;
		data.mCount = 0;
	}else{
		data.mCount++;
	}
	K7EventDesc		**ppDesc;
	while( data.mStep < 2 ){
		switch( data.mStep ){
		case	0:
			ppDesc = &mppGlobalEventDesc[data.mCount];
			for( ; data.mCount<mGlobalEventDescNum ; data.mCount++,ppDesc++){
				if( (*ppDesc)->GetType()==type ){
					return *ppDesc;
				}
			}
			data.mCount = 0;
			data.mStep++;
			break;
		case	1:
			ppDesc = &mppScenarioEventDesc[data.mCount];
			for( ; data.mCount<mScenarioEventDescNum ; data.mCount++,ppDesc++){
				if( (*ppDesc)->GetType()==type ){
					return *ppDesc;
				}
			}
			data.mCount = 0;
			data.mStep++;
			break;
		}
		
	}
	return NULL;
}

int WK7EventController::Call(u32 id,u32 param0,u32 param1,int prio)
{
	if( mState != WK7EVENTCONTROLLERSTATE_RUN ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	K7EventHandler	*pHandler = GetHandler(id);
	if( pHandler==NULL ){
		return WK7EVENTCONTROLLER_CALL_NOTFOUND;
	}
	if( pHandler->GetCondition()==K7EV_COND_NOTRUN ){
		if( GetEvent(id) != NULL ){
			return WK7EVENTCONTROLLER_CALL_DISABLE;
		}
	}
	ghmGcMemSetAllocName("WK7EV");
	WK7Event		*pEvent = new WK7Event(id,&mAnchor);
	if( pEvent==NULL ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	pEvent->mpTask        = NULL;
	pEvent->mpHandler     = pHandler->mpHandler;
	pEvent->mpTermHandler = pHandler->mpTermHandler;
	pEvent->SetParam(0,param0);
	pEvent->SetParam(1,param1);
	if( (pEvent->mTaskID = mTaskController.Start(WK7Event::Handler,WK7Event::TermHandler,(u32)pEvent,0,NULL,WK7EVENT_TASK_STACK_SIZE,prio)) < 0 ){
		delete pEvent;
		return WK7EVENTCONTROLLER_CALL_TASKSTART_FAILED;
	}
	return WK7EVENTCONTROLLER_CALL_SUCCESS;
}

int WK7EventController::Call(K7EventDesc *pDesc,u32 param0,u32 param1,int prio)
{
	if( pDesc==NULL ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	return Call(pDesc->GetID(),param0,param1,prio);
}

int WK7EventController::Call(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0,u32 param1,int prio)
{
	if( mState != WK7EVENTCONTROLLERSTATE_RUN || pHandler==NULL ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	ghmGcMemSetAllocName("WK7EVT");
	WK7Event		*pEvent = new WK7Event(WK7EVENT_NOID,&mAnchor);
	if( pEvent==NULL ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	pEvent->mpTask        = NULL;
	pEvent->mpHandler     = pHandler;
	pEvent->mpTermHandler = pTermHandler;
	pEvent->SetParam(0,param0);
	pEvent->SetParam(1,param1);
	if( (pEvent->mTaskID = mTaskController.Start(WK7Event::Handler,WK7Event::TermHandler,(u32)pEvent,0,NULL,WK7EVENT_TASK_STACK_SIZE,prio)) < 0 ){
		delete pEvent;
		return WK7EVENTCONTROLLER_CALL_TASKSTART_FAILED;
	}
	return WK7EVENTCONTROLLER_CALL_SUCCESS;
}

int WK7EventController::Call2(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0,u32 param1,int prio)
{
	WK7Event		*pEvent = GetEvent(pHandler,pTermHandler,param0,param1);
	if( pEvent!=NULL ){
		return WK7EVENTCONTROLLER_CALL_DISABLE;
	}
	return Call(pHandler,pTermHandler,param0,param1,prio);
}

int WK7EventController::CallIM(u32 id,u32 param0,u32 param1,int prio)
{
	if( mState != WK7EVENTCONTROLLERSTATE_RUN ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	K7EventHandler	*pHandler = GetHandler(id);
	if( pHandler==NULL ){
		return WK7EVENTCONTROLLER_CALL_NOTFOUND;
	}
	if( pHandler->GetCondition()==K7EV_COND_NOTRUN ){
		if( GetEvent(id) != NULL ){
			return WK7EVENTCONTROLLER_CALL_DISABLE;
		}
	}
	ghmGcMemSetAllocName("WK7EV2");
	WK7Event		*pEvent = new WK7Event(id,&mAnchor);
	if( pEvent==NULL ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	pEvent->mpTask        = NULL;
	pEvent->mpHandler     = pHandler->mpHandler;
	pEvent->mpTermHandler = pHandler->mpTermHandler;
	pEvent->SetParam(0,param0);
	pEvent->SetParam(1,param1);
	if( (pEvent->mTaskID = mTaskController.StartIM(WK7Event::Handler,WK7Event::TermHandler,(u32)pEvent,0,NULL,WK7EVENT_TASK_STACK_SIZE,prio)) < 0 ){
		delete pEvent;
		return WK7EVENTCONTROLLER_CALL_TASKSTART_FAILED;
	}
	return WK7EVENTCONTROLLER_CALL_SUCCESS;
}

int WK7EventController::CallIM(K7EventDesc *pDesc,u32 param0,u32 param1,int prio)
{
	if( pDesc==NULL ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	return CallIM(pDesc->GetID(),param0,param1,prio);
}

int WK7EventController::CallIM(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0,u32 param1,int prio)
{
	if( mState != WK7EVENTCONTROLLERSTATE_RUN ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	ghmGcMemSetAllocName("WK7EV2T");
	WK7Event		*pEvent = new WK7Event(WK7EVENT_NOID,&mAnchor);
	if( pEvent==NULL ){
		return WK7EVENTCONTROLLER_CALL_FAILED;
	}
	pEvent->mpTask        = NULL;
	pEvent->mpHandler     = pHandler;
	pEvent->mpTermHandler = pTermHandler;
	pEvent->SetParam(0,param0);
	pEvent->SetParam(1,param1);
	if( (pEvent->mTaskID = mTaskController.StartIM(WK7Event::Handler,WK7Event::TermHandler,(u32)pEvent,0,NULL,WK7EVENT_TASK_STACK_SIZE,prio)) < 0 ){
		delete pEvent;
		return WK7EVENTCONTROLLER_CALL_TASKSTART_FAILED;
	}
	return WK7EVENTCONTROLLER_CALL_SUCCESS;
}

int WK7EventController::CallIM2(ghmGcTaskHandler *pHandler,ghmGcTaskHandler *pTermHandler,u32 param0,u32 param1,int prio)
{
	WK7Event		*pEvent = GetEvent(pHandler,pTermHandler,param0,param1);
	if( pEvent!=NULL ){
		return WK7EVENTCONTROLLER_CALL_DISABLE;
	}
	return CallIM(pHandler,pTermHandler,param0,param1,prio);
}

bool WK7EventController::CompareBranchPointEventDesc(K7EventDesc *pDesc,const char *pBpName)
{
	if( pDesc==NULL || pBpName==NULL ){
		return false;
	}
	switch( pDesc->GetType() ){
//	case	K7EV_DESC_TYPE_STGBPARRIVE:{
//		K7EventDescStgBpArrive	*pStgBpArrive = (K7EventDescStgBpArrive *)pDesc;
//		if( strcmp(pStgBpArrive->mpBpName,pBpName)==0 ){
//			return true;
//		}
//		}break;
	case	K7EV_DESC_TYPE_STGBPLEAVE:{
		K7EventDescStgBpLeave	*pStgBpLeave = (K7EventDescStgBpLeave *)pDesc;
		if( strcmp(pStgBpLeave->mpBpName,pBpName)==0 ){
			return true;
		}
		}break;
//	case	K7EV_DESC_TYPE_STGBPEXTRACONTACT:{
//		K7EventDescStgBpExtraContact	*pStgBpExtraContact = (K7EventDescStgBpExtraContact *)pDesc;
//		if( strcmp(pStgBpExtraContact->mpBpName,pBpName)==0 ){
//			return true;
//		}
//		}break;
//	case	K7EV_DESC_TYPE_STGBPSELECTOBJCONTACT:{
//		K7EventDescStgBpSelectObjContact	*pStgBpSelectObjContact = (K7EventDescStgBpSelectObjContact *)pDesc;
//		if( strcmp(pStgBpSelectObjContact->mpBpName,pBpName)==0 ){
//			return true;
//		}
//		}break;
//		break;
	default:
		break;
	}
	return false;
}

K7EventDesc* WK7EventController::GetBranchPointEventDesc(const char *pBpName,K7EventDescData &data)
{
	if( data.mStep==-1 ){
		data.mStep  = 0;
		data.mCount = 0;
	}else{
		data.mCount++;
	}
	K7EventDesc		**ppDesc;
	while( data.mStep < 2 ){
		switch( data.mStep ){
		case	0:
			ppDesc = &mppGlobalEventDesc[data.mCount];
			for( ; data.mCount<mGlobalEventDescNum ; data.mCount++,ppDesc++){
				if( CompareBranchPointEventDesc(*ppDesc,pBpName)==true ){
					return *ppDesc;
				}
			}
			data.mCount = 0;
			data.mStep++;
			break;
		case	1:
			ppDesc = &mppScenarioEventDesc[data.mCount];
			for( ; data.mCount<mScenarioEventDescNum ; data.mCount++,ppDesc++){
				if( CompareBranchPointEventDesc(*ppDesc,pBpName)==true ){
					return *ppDesc;
				}
			}
			data.mCount = 0;
			data.mStep++;
			break;
		}
		
	}
	return NULL;
}

int WK7EventController::CallSpecialAttackEvent(u32 chrid)
{
	K7EventDescStgSpAttack	*pDesc;
	K7EventDescData			data;
	int						count = 0;
	u32						stgid = WK7Stage::GetCurrentStageID();
	data.Initialize();
	while( (pDesc=(K7EventDescStgSpAttack *)GetEventDesc(K7EV_DESC_TYPE_STGSPATTACK,data))!=NULL ){
		if( pDesc->mChrMainID==chrid && pDesc->mStgID==stgid ){
			CallIM(&pDesc->mBase,chrid);
		}
	}
	return count;
}

const char* GetTokenPointer(const char *pStr,char token)
{
	for( ; *pStr!='\0' ; pStr++){
		if( *pStr==token ){
			return pStr;
		}
	}
	return NULL;
}

const char* GetTokenPointer2(const char *pStr,char token)
{
	for( ; *pStr!='\0' ; pStr++){
		if( *pStr==token ){
			if( pStr[1]!=token ){
				return pStr;
			}
			pStr++;
		}
	}
	return NULL;
}

const char* GetPartialStr(char *pDstStr,int maxlen,const char *pStr,char token)
{
	for( ; *pStr!='\0' ; pStr++){
		if( *pStr==token ){
			if( maxlen ){
				*pDstStr = '\0';
			}
			return pStr;
		}
		if( maxlen > 0 ){
			*pDstStr++ = *pStr;
			maxlen--;
			if( maxlen==0 ){
				*pDstStr = '\0';
			}
		}
	}
	if( maxlen ){
		*pDstStr = '\0';
	}
	return NULL;
}

const char* GetPartialStr2(char *pDstStr,int maxlen,const char *pStr,char token)
{
	for( ; *pStr!='\0' ; pStr++){
		if( *pStr==token ){
			if( pStr[1]!=token ){
				if( maxlen ){
					*pDstStr = '\0';
				}
				return pStr;
			}else{
				pStr++;
			}
		}
		if( maxlen > 0 ){
			*pDstStr++ = *pStr;
			maxlen--;
			if( maxlen==0 ){
				*pDstStr = '\0';
			}
		}
	}
	if( maxlen ){
		*pDstStr = '\0';
	}
	return NULL;
}

const char* GetPartialStr(char* &pDstStr,const char *pStr,char token)
{
	const char	*pLast = GetTokenPointer(pStr,token);
	char		*pCurDstStr;
	int			maxlen;
	if( pLast==NULL ){
		maxlen = (int)strlen(pStr);
	}else{
		maxlen = (int)((u32)pLast - (u32)pStr);
	}
	ghmGcMemSetAllocName("GetPSTR");
	pCurDstStr = pDstStr = new char[maxlen + 1];
	for( ; *pStr!='\0' ; pStr++){
		if( *pStr==token ){
			if( maxlen ){
				*pCurDstStr = '\0';
			}
			return pStr;
		}
		if( maxlen > 0 ){
			*pCurDstStr++ = *pStr;
			maxlen--;
			if( maxlen==0 ){
				*pCurDstStr = '\0';
			}
		}
	}
	if( maxlen ){
		*pDstStr = '\0';
	}
	return NULL;
}

const char* GetPartialStr2(char *&pDstStr,const char *pStr,char token)
{
	const char	*pLast = GetTokenPointer2(pStr,token);
	char		*pCurDstStr;
	int			maxlen;
	if( pLast==NULL ){
		maxlen = (int)strlen(pStr);
	}else{
		maxlen = (int)((u32)pLast - (u32)pStr);
	}
	ghmGcMemSetAllocName("GetPSTR2");
	pCurDstStr = pDstStr = new char[maxlen + 1];
	for( ; *pStr!='\0' ; pStr++){
		if( *pStr==token ){
			if( pStr[1]!=token ){
				if( maxlen ){
					*pCurDstStr = '\0';
				}
				return pStr;
			}else{
				pStr++;
			}
		}
		if( maxlen > 0 ){
			*pCurDstStr++ = *pStr;
			maxlen--;
			if( maxlen==0 ){
				*pCurDstStr = '\0';
			}
		}
	}
	if( maxlen ){
		*pCurDstStr = '\0';
	}
	return NULL;
}

