////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Fog
//
// WK7Fog.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 05/07/2004
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7FOG_CPP__

#include ".\WK7Fog.h"

WK7Fog*			WK7Fog::mpTop    = NULL;
WK7Fog*			WK7Fog::mpAllFog = NULL;

HSD_FogDesc		allFogDesc = {
	GX_FOG_LIN,					// fog type
	NULL,						// adjdesc
	85.0f,						// startz
	1024.0f,					// endz
//	{255, 255, 255, 255},		// color
	{192, 192, 255, 255},		// color
};

WK7Fog::WK7Fog(HSD_FogDesc *pFogDesc)
{
	mpFog   = NULL;
	mpNext  = NULL;
//	mEnable = true;
	mEnable = false;
	LoadDesc(pFogDesc);
}

WK7Fog::~WK7Fog()
{
	if( mpFog!=NULL ){
		HSD_FogUnref(mpFog);
		mpFog = NULL;
	}
}

bool WK7Fog::Clear(void)
{
	while( mpTop!=NULL ){
		mpTop->EndCurrent();
	}
	return true;
}

bool WK7Fog::Initialize(void)
{
	ghmGcMemSetAllocName("WKFOGA");
	mpAllFog = new WK7Fog(&allFogDesc);
	return true;
}

bool WK7Fog::Terminate(void)
{
	GHM_SAFE_DELETE(mpAllFog);
	return true;
}

bool WK7Fog::SetCurrentAll(void)
{
	if( mpAllFog!=NULL ){
		mpAllFog->SetCurrent();
	}
	return true;
}

bool WK7Fog::EndCurrentAll(void)
{
	if( mpAllFog!=NULL ){
		mpAllFog->EndCurrent();
	}
	return true;
}

bool WK7Fog::LoadDesc(HSD_FogDesc *pFogDesc)
{
	if( mpFog!=NULL ){
		HSD_FogUnref(mpFog);
		mpFog = NULL;
	}
	if( pFogDesc==NULL ){
		return true;
	}
	mpFog  = HSD_FogLoadDesc(pFogDesc);
	return true;
}

bool WK7Fog::SetCurrent(void)
{
	if( mpNext!=NULL ){
		return false;
	}
	mpNext = mpTop;
	mpTop  = this;
	if( mEnable==true ){
		HSD_FogSet(mpFog);
	}else{
		HSD_FogSet(NULL);
	}
	return true;
}

bool WK7Fog::EndCurrent(void)
{
	if( mpTop!=this ){
		return false;
	}
	mpTop  = mpNext;
	mpNext = NULL;
	if( mpTop==NULL || mpTop->GetEnable()==false ){
		HSD_FogSet(NULL);
	}else{
		HSD_FogSet(mpTop->GetFog());
	}
	return true;
}

bool WK7Fog::SetType(GXFogType type)
{
	if( mpFog==NULL ){
		return false;
	}
	HSD_FogSetType(mpFog,type);
	return true;
}

bool WK7Fog::SetStartZ(f32 z)
{
	if( mpFog==NULL ){
		return false;
	}
	HSD_FogSetStartZ(mpFog,z);
	return true;
}

bool WK7Fog::SetEndZ(f32 z)
{
	if( mpFog==NULL ){
		return false;
	}
	HSD_FogSetEndZ(mpFog,z);
	return true;
}

bool WK7Fog::SetColor(GXColor color)
{
	if( mpFog==NULL ){
		return false;
	}
	HSD_FogSetColor(mpFog,color);
	return true;
}

GXFogType WK7Fog::GetType(void)
{
	return HSD_FogGetType(mpFog);
}

f32 WK7Fog::GetStartZ(void)
{
	return HSD_FogGetStartZ(mpFog);
}

f32 WK7Fog::GetEndZ(void)
{
	return HSD_FogGetEndZ(mpFog);
}

GXColor WK7Fog::GetColor(void)
{
	return HSD_FogGetColor(mpFog);
}

bool WK7Fog::IsFogging(void)
{
	if( mpTop==NULL ){
		return false;
	}
	if( mpTop->GetEnable()==false ){
		return false;
	}
	if( mpTop->GetFog()==NULL ){
		return false;
	}
	return true;
}

