////////////////////////////////////////////////////////////////////////////////
//
// Killer7 HitObj
//
// WK7HitObj.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 09/26/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7HITOBJ_CPP__

#include ".\WK7HitObj.h"
#include ".\WK7Event.h"
#include ".\WK7EventScript.h"
#include ".\WK7Stage.h"
#include ".\WK7GameInfo.h"
#include ".\WK7Path.h"
#include ".\WK7Model.h"

ghmGcListObj			WK7HitObj::mAnchor;
WK7HitObj*				WK7HitObj::mpFindObj = NULL;
ghmGcTaskController*	WK7HitObj::mpDefaultTaskController = NULL;
WK7HitObjTable*			WK7HitObj::mpTable = NULL;

WK7HitObj::WK7HitObj(int prio) : ghmGcListObj(&mAnchor,prio)
{
	mType            = BASE_TYPE;
	mFlag            = 0;
	mEnable          = 1;
	mRenderID        = 0xffffffff;
	mEventID         = 0xffffffff;
	mpHandler        = NULL;
	mpTermHandler    = NULL;
	mpTaskController = NULL;
	mEventParam[0]   = 
	mEventParam[1]   = 
	mTaskParam [0]   = 
	mTaskParam [1]   = (u32)this;
	mpStage          = NULL;
}

WK7HitObj::~WK7HitObj()
{
}

bool WK7HitObj::InitFind(void)
{
	mpFindObj = (WK7HitObj *)mAnchor.GetNext();
	return true;
}

WK7HitObj* WK7HitObj::FindEvent(u32 eventid)
{
	WK7HitObj	*pObj;
	for(pObj=mpFindObj ; pObj!=NULL ; pObj=pObj->GetNext()){
		if( pObj->GetEventID()!=eventid ){
			continue;
		}
		mpFindObj = pObj->GetNext();
		return pObj;
	}
	return NULL;
}

WK7HitObjJObj* WK7HitObj::FindEventNo(u32 eventno)
{
	WK7HitObj	*pObj;
	for(pObj=mpFindObj ; pObj!=NULL ; pObj=pObj->GetNext()){
		if( pObj->GetType()!=JOBJ_TYPE ){
			continue;
		}
		WK7HitObjJObj	*pObjJObj = (WK7HitObjJObj *)pObj;
		if( pObjJObj->GetEventNo()!=eventno ){
			continue;
		}
		mpFindObj = pObj->GetNext();
		return pObjJObj;
	}
	return NULL;
}

WK7HitObj* WK7HitObj::Find(u32 renderid)
{
	WK7HitObj	*pObj;
	for(pObj=mpFindObj ; pObj!=NULL ; pObj=pObj->GetNext()){
		if( pObj->GetRenderID()!=renderid ){
			continue;
		}
		mpFindObj = pObj->GetNext();
		return pObj;
	}
	return NULL;
}

WK7HitObj* WK7HitObj::Find(WK7JObj *pJObj)
{
	WK7HitObj	*pObj = mpFindObj;
	for(pObj=mpFindObj ; pObj!=NULL ; pObj=pObj->GetNext()){
		if( pObj->GetType()!=JOBJ_TYPE ){
			continue;
		}
		WK7HitObjJObj	*pObjJObj = (WK7HitObjJObj *)pObj;
		if( pObjJObj->GetJObj()!=pJObj ){
			continue;
		}
		mpFindObj = pObj->GetNext();
		return pObj;
		
	}
	return NULL;
}

int WK7HitObj::Release(u32 renderid)
{
	WK7HitObj	*pObj;
	int			count = 0;
	WK7HitObj::InitFind();
	while( (pObj=WK7HitObj::Find(renderid))!=NULL ){
		delete pObj;
		count++;
	}
	return count;
}

int WK7HitObj::Release(WK7JObj *pJObj)
{
	WK7HitObj	*pObj;
	int			count = 0;
	WK7HitObj::InitFind();
	while( (pObj=WK7HitObj::Find(pJObj))!=NULL ){
		delete pObj;
		count++;
	}
	return count;
}

bool WK7HitObj::SetEventParam(int index,u32 param)
{
	if( index < 0 || index >= 2 ){
		return false;
	}
	mEventParam[index] = param;
	return true;
}

u32 WK7HitObj::GetEventParam(int index)
{
	if( index < 0 || index >= 2 ){
		return 0xffffffff;
	}
	return mEventParam[index];
}

bool WK7HitObj::SetTaskParam(int index,u32 param)
{
	if( index < 0 || index >= 2 ){
		return false;
	}
	mTaskParam[index] = param;
	return true;
}

u32 WK7HitObj::GetTaskParam(int index)
{
	if( index < 0 || index >= 2 ){
		return 0xffffffff;
	}
	return mTaskParam[index];
}

bool WK7HitObj::CallEvent(void)
{
	if( mEventID==0xffffffff ){
		return true;
	}
	if( WK7EVENTCONTROLLER->Call(mEventID,mEventParam[0],mEventParam[1]) <= 0 ){
		return false;
	}
	return true;
}

bool WK7HitObj::TaskStart(void)
{
	ghmGcTaskController	*pTaskController;
	if( mpHandler==NULL || (pTaskController=GetCurrentTaskController())==NULL ){
		return true;
	}
	if( pTaskController->Start(mpHandler,
	                           mpTermHandler,
	                           mTaskParam[0],
	                           mTaskParam[1]) < 0 ){
		return false;
	}
	return true;
}

int WK7HitObj::HitCheck(u32 renderid, WK7HitObjInfoType hittypeid, s16 inSightPx, s16 inSightPy, Vec inHitPos)
{
	WK7HitObj	*pObj;
	int			count = 0;
	for(pObj=(WK7HitObj *)mAnchor.GetNext() ; pObj!=NULL ; pObj=pObj->GetNext()){
		if( pObj->GetRenderID()!=renderid ){
			continue;
		}
		if( pObj->GetEnable()==false ){
			continue;
		}
		if( pObj->CallEvent()==false ){
			continue;
		}
		if( pObj->TaskStart()==false ){
			continue;
		}
		pObj->SetEnable(false);
		WK7Stage	*pStage = pObj->GetStage();
		if( pStage!=NULL ){
			WK7HitObjInfo		info;
			
			info.Initialize();
			info.SetHitObj(pObj);
//			info.mpHitObj		= pObj;
			info.mTypeID		= hittypeid;
			info.mSightPx		= inSightPx;
			info.mSightPy		= inSightPy;
			info.mHitPos		= inHitPos;
			
			if( pObj->GetType()==JOBJ_TYPE ){
				WK7HitObjJObj	*pObjJObj = (WK7HitObjJObj *)pObj;
				info.mEventNo = pObjJObj->GetEventNo();
				info.mpJObj   = pObjJObj->GetJObj   ();
			}
			switch( info.mEventNo & K7EV_TYPE_MASK ){
			case	K7EV_TYPE_HANA:
				pStage->CallHanaEventCallBackTask(info);
				break;
			default:
				pStage->CallHitObjCallBackTask(info);
				break;
			}
		}
		count++;
	}
	return count;
}

int WK7HitObj::SetEventEnable(u32 eventid,bool flag)
{
	WK7HitObj	*pObj;
	int			count = 0;
	for(pObj=(WK7HitObj *)mAnchor.GetNext() ; pObj!=NULL ; pObj=pObj->GetNext()){
		if( pObj->GetEventID()!=eventid ){
			continue;
		}
		pObj->SetEnable(flag);
		count++;
	}
	return count;
}

int WK7HitObj::SetEventNoEnable(u32 eventno,bool flag)
{
	WK7HitObj		*pObj;
	int				count = 0;
	for(pObj=(WK7HitObj *)mAnchor.GetNext() ; pObj!=NULL ; pObj=pObj->GetNext()){
		if( pObj->GetType()!=JOBJ_TYPE ){
			continue;
		}
		WK7HitObjJObj	*pObjJObj = (WK7HitObjJObj *)pObj;
		if( pObjJObj->GetEventNo()!=eventno ){
			continue;
		}
		pObj->SetEnable(flag);
		count++;
	}
	return count;
}

bool WK7HitObj::SetTable(WK7HitObjTable *pTable)
{
	WK7HitObj::mpTable = pTable;
	return true;
}

WK7HitObjTable* WK7HitObj::GetTable(u32 eventNo)
{
	WK7HitObjTable	*pTable;
	if( (pTable = WK7HitObj::mpTable)==NULL ){
		return NULL;
	}
	while( pTable->mEventNo!=WK7HITOBJTBL_TERM_EVENTNO ){
		if( pTable->mEventNo==eventNo ){
			return pTable;
		}
		pTable++;
	}
	return NULL;
}

bool WK7HitObj::IsValid(WK7HitObj *pObj)
{
	if( pObj==NULL ){
		return false;
	}
	WK7HitObj	*pTmpObj;
	for(pTmpObj=(WK7HitObj *)mAnchor.GetNext() ; pTmpObj!=NULL ; pTmpObj=pTmpObj->GetNext()){
		if( pTmpObj==pObj ){
			return true;
		}
	}
	return false;
}

bool WK7HitObj::InitEventEnable(void)
{
	WK7HitObjTable	*pTable;
	if( (pTable = WK7HitObj::mpTable)==NULL ){
		return false;
	}
	while( pTable->mEventNo!=WK7HITOBJTBL_TERM_EVENTNO ){
		WK7HitObj	*pObj;
		int			count = 0;
		for(pObj=(WK7HitObj *)mAnchor.GetNext() ; pObj!=NULL ; pObj=pObj->GetNext()){
			if( pObj->GetType()!=JOBJ_TYPE ){
				continue;
			}
			WK7HitObjJObj	*pObjJObj = (WK7HitObjJObj *)pObj;
			WK7JObj			*pJObj    = pObjJObj->GetJObj();
			if( pObjJObj->GetEventNo()!=pTable->mEventNo ){
				continue;
			}
			if( pTable->mFlagID < 0 ){
				continue;
			}
			bool	flag;
			if( GameInfo.GetScenarioFlag(pTable->mFlagID) > 0 ){
				flag = true;
			}else{
				flag = false;
			}
			if( flag==false ){
				pObj->SetEnable(true);
				switch( pTable->mType ){
				case	HITOBJTYPE_NOACTION:
					break;
				case	HITOBJTYPE_HIDE_PATHOPEN:
					if( pJObj!=NULL ){
						WK7PathNurbs	*pPath = WK7HitObjJObj::GetPath(pJObj);
						if( pPath!=NULL ){
							pPath->SetEnable  (flag);
							pPath->SetEnableEx(flag);
						}
					}
				case	HITOBJTYPE_HIDE:
					pJObj->Disp(false,true);
					break;
				case	HITOBJTYPE_DISP:
					pJObj->Hide(false,true);
					break;
				default:
					break;
				}
			}else{
				pObj->SetEnable(false);
				switch( pTable->mType ){
				case	HITOBJTYPE_NOACTION:
					break;
				case	HITOBJTYPE_HIDE_PATHOPEN:
					if( pJObj!=NULL ){
						WK7PathNurbs	*pPath = WK7HitObjJObj::GetPath(pJObj);
						if( pPath!=NULL ){
							pPath->SetEnable  (flag);
							pPath->SetEnableEx(flag);
						}
					}
				case	HITOBJTYPE_HIDE:
					pJObj->Hide(false,true);
					break;
				case	HITOBJTYPE_DISP:
					pJObj->Disp(false,true);
					break;
				default:
					break;
				}
			}
			count++;
		}
		pTable++;
	}
	return true;
}

WK7HitObjJObj::WK7HitObjJObj(int prio) : WK7HitObj(prio)
{
	mType = JOBJ_TYPE;
	mEventNo = 0xffffffff;
}

WK7HitObjJObj::~WK7HitObjJObj()
{
}

WK7PathNurbs* WK7HitObjJObj::GetPath(WK7JObj *pJObj)
{
	const char		*pJObjName;
	if( pJObj==NULL || (pJObjName=pJObj->GetName())==NULL ){
		return NULL;
	}
	if( (pJObjName = GetTokenPointer(pJObjName,'_'))==NULL ){
		return NULL;
	}
	pJObjName++;
	if( (pJObjName = GetTokenPointer(pJObjName,'_'))==NULL ){
		return NULL;
	}
	pJObjName++;
	return WK7Stage::GetPathNurbsAll(pJObjName);
}

bool WK7HitObjJObj::CallEvent(void)
{
	if( mEventID==0xffffffff ){
		return true;
	}
	WK7HitObjTable	*pTable = GetTable(mEventNo);
	if( pTable!=NULL ){
		switch( pTable->mType ){
		case	HITOBJTYPE_NOACTION:
			break;
		case	HITOBJTYPE_HIDE_PATHOPEN:
			{
				WK7PathNurbs	*pPath = GetPath(mpJObj);
				if( pPath!=NULL ){
					pPath->SetEnable  (true);
					pPath->SetEnableEx(true);
				}
			}
		case	HITOBJTYPE_HIDE:
			if( mpJObj!=NULL ){
				mpJObj->Hide(false,true);
			}
			break;
		case	HITOBJTYPE_DISP:
			if( mpJObj!=NULL ){
				mpJObj->Disp(false,true);
			}
			break;
		}
		if( pTable->mFlagID >= 0 ){
			GameInfo.SetScenarioFlag(pTable->mFlagID);
		}
	}
	if( WK7EVENTCONTROLLER->Call(mEventID,mEventParam[0],mEventParam[1]) <= 0 ){
		return false;
	}
	return true;
}

WK7PathNurbs* WK7HitObjJObj::GetClosePath(void)
{
	if( (mEventNo & K7EV_TYPE_MASK)!=K7EV_TYPE_DOOR || mpJObj==NULL ){
		return NULL;
	}
	WK7PathNurbs	*pPathNurbs;
	pPathNurbs = GetPath(mpJObj);
	return pPathNurbs;
}

