////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Motion
//
// WK7Motion.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 02/08/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7MOTION_CPP__

//#include <dolphin.h>
//#include <sysdolphin/baselib.h>
//#include "..\Lib\ghmGcFile.h"
#include ".\WK7Motion.h"
#include ".\WK7Archive.h"
#include ".\WK7Cam.h"
#include ".\WK7Obj.h"
#include ".\WK7Light.h"

// [Ve[u
WK7MotionAttr	WK7MotionAttrTable[] = {
	// _
	WK7MOTIONATTR_DECL(K7MOTID(K7DAN, 0),WK7MOTIONATTR_LOOP),
	WK7MOTIONATTR_DECL(K7MOTID(K7DAN,10),WK7MOTIONATTR_LOOP),
	WK7MOTIONATTR_DECL(K7MOTID(K7DAN,20),WK7MOTIONATTR_LOOP),
	WK7MOTIONATTR_DECL(K7MOTID(K7DAN,40),WK7MOTIONATTR_LOOP),
	WK7MOTIONATTR_DECL(K7MOTID(K7DAN,50),WK7MOTIONATTR_LOOP),
	
	// Gl~[
	WK7MOTIONATTR_DECL(K7MOTID(K7ENM,20),WK7MOTIONATTR_LOOP),
	WK7MOTIONATTR_DECL(K7MOTID(K7ENM,31),WK7MOTIONATTR_LOOP),
	WK7MOTIONATTR_DECL(K7MOTID(K7ENM,40),WK7MOTIONATTR_LOOP),
};

WK7MotionAttr* WK7ModelMotion::GetAttr(u32 motid)
{
	WK7MotionAttr	*pAttr;
	int				i;
	for(pAttr=WK7MotionAttrTable,i=0 ; i<sizeof(WK7MotionAttrTable)/sizeof(WK7MotionAttr) ; i++,pAttr++){
		if( pAttr->mMotionID > motid ){
			break;
		}
		if( pAttr->mMotionID==motid ){
			return pAttr;
		}
	}
	return NULL;
}

// Ŝ̃[V[g
f32		WK7Motion::mWholeMotionRate = 1.0f;

WK7Motion::WK7Motion(WK7Archive *pArchive)
{
	if( (mpArchive = pArchive)!=NULL ){
		mpArchive->ReferenceArchive();
	}
}

WK7Motion::~WK7Motion()
{
	if( mpArchive!=NULL ){
		mpArchive->ReleaseReferenceArchive();
	}
}

f32 WK7Motion::ConvMotionRate(f32 rate)
{
	return rate * mWholeMotionRate * WK7MOTION_DEFAULT_RATE;
}

bool WK7Motion::SetWholeMotionRate(f32 rate)
{
	mWholeMotionRate = rate;
	WK7RenderObj::RenewMotionRateAll();
	WK7Light::RenewMotionRateAll();
	WK7CAM->RenewMotionRate();
	return true;
}

bool WK7Motion::SetMotionRate(HSD_AObj *pAObj,f32 rate,bool wholeRateEnable)
{
	if( pAObj==NULL ){
		return true;
	}
	if( wholeRateEnable==true ){
		rate *= mWholeMotionRate;
	}
	HSD_AObjSetRate(pAObj,rate * WK7MOTION_DEFAULT_RATE);
	return true;
}

WK7SdExpInfo* WK7Motion::GetSdExpInfo(void)
{
	if( mpArchive==NULL ){
		return NULL;
	}
	return mpArchive->GetSdExpInfo();
}

WK7ModelMotion::WK7ModelMotion(WK7Archive *pArchive) : WK7Motion(pArchive)
{
	mType = 0;
	mAttr = 0;
	if( mpArchive != NULL ){
		if( GetAnimJointNum() > 0 ){
			mType |= WK7MODELMOTIONTYPE_NORM;
		}
		if( GetMatAnimJointNum() > 0 ){
			mType |= WK7MODELMOTIONTYPE_MAT;
		}
		if( GetShapeAnimJointNum() > 0 ){
			mType |= WK7MODELMOTIONTYPE_SHAPE;
		}
		WK7MotionAttr	*pAttr;
		if( (pAttr=WK7ModelMotion::GetAttr(pArchive->GetID()))!=NULL ){
			mAttr = pAttr->mAttr;
		}
	}
}

WK7ModelMotion::~WK7ModelMotion()
{
}

C_ModelSet* WK7ModelMotion::GetAnimModelSet(const char *pName)
{
	if( mpArchive==NULL ){
		return NULL;
	}
	WK7SceneDesc		*pSceneDesc = mpArchive->GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	HSD_SceneModelSet		**ms = pSceneDesc->models;
	for( ; *ms ; ms++){
		if( (*ms)->adjointlist==NULL ){
			continue;
		}
		if( strcmp(GetDescName(*((*ms)->adjointlist)),pName)==0 ){
			return new (*ms) C_ModelSet;
		}
	}
	return NULL;
}

HSD_AnimJoint* WK7ModelMotion::GetAnimJoint(const char *pName,HSD_AnimJoint *pAnimJoint,bool next,bool child)
{
	if( pAnimJoint==NULL ){
		return NULL;
	}
	const char	*pAnimJointName = GetDescName(pAnimJoint);
	if( pAnimJointName!=NULL && strcmp(pAnimJointName,pName)==0 ){
		return pAnimJoint;
	}
	HSD_AnimJoint	*pResult;
	if( next==true ){
		if( (pResult=GetAnimJoint(pName,pAnimJoint->next,true,child))!=NULL ){
			return pResult;
		}
	}
	if( child==true ){
		if( (pResult=GetAnimJoint(pName,pAnimJoint->child,true,true))!=NULL ){
			return pResult;
		}
	}
	return NULL;
}

HSD_MatAnimJoint* WK7ModelMotion::GetMatAnimJoint(const char *pName,HSD_MatAnimJoint *pMatAnimJoint,bool next,bool child)
{
	if( pMatAnimJoint==NULL ){
		return NULL;
	}
	const char	*pMatAnimJointName = GetDescName(pMatAnimJoint);
	if( strcmp(pMatAnimJointName,pName)==0 ){
		return pMatAnimJoint;
	}
	HSD_MatAnimJoint	*pResult;
	if( next==true ){
		if( (pResult=GetMatAnimJoint(pName,pMatAnimJoint->next,true,child))!=NULL ){
			return pResult;
		}
	}
	if( child==true ){
		if( (pResult=GetMatAnimJoint(pName,pMatAnimJoint->child,true,true))!=NULL ){
			return pResult;
		}
	}
	return NULL;
}

HSD_ShapeAnimJoint* WK7ModelMotion::GetShapeAnimJoint(const char *pName,HSD_ShapeAnimJoint *pShapeAnimJoint,bool next,bool child)
{
	if( pShapeAnimJoint==NULL ){
		return NULL;
	}
	const char	*pShapeAnimJointName = GetDescName(pShapeAnimJoint);
	if( strcmp(pShapeAnimJointName,pName)==0 ){
		return pShapeAnimJoint;
	}
	HSD_ShapeAnimJoint	*pResult;
	if( next==true ){
		if( (pResult=GetShapeAnimJoint(pName,pShapeAnimJoint->next,true,child))!=NULL ){
			return pResult;
		}
	}
	if( child==true ){
		if( (pResult=GetShapeAnimJoint(pName,pShapeAnimJoint->child,true,true))!=NULL ){
			return pResult;
		}
	}
	return NULL;
}

bool WK7ModelMotion::GetAnimJoint(const char *pName,HSD_AnimJoint      **ppAnimJoint,
                                                    HSD_MatAnimJoint   **ppMatAnimJoint,
                                                    HSD_ShapeAnimJoint **ppShapeAnimJoint)
{
	if( mpArchive==NULL ){
		return false;
	}
//	C_ModelSet	*pModelSet = GetAnimModelSet(pName);
//	return GetAnimJoint(pModelSet,ppAnimJoint,ppMatAnimJoint,ppShapeAnimJoint);
	WK7SceneDesc		*pSceneDesc = mpArchive->GetSceneDesc();
	if( pSceneDesc==NULL ){
		return NULL;
	}
	HSD_SceneModelSet		**ms = pSceneDesc->models;
	if( ppAnimJoint!=NULL ){
		for(ms=pSceneDesc->models ; *ms ; ms++){
			if( (*ms)->adjointlist==NULL ){
				continue;
			}
			if( (*ppAnimJoint=GetAnimJoint(pName,(*ms)->adjointlist[0]))!=NULL ){
				break;
			}
		}
	}
	if( ppMatAnimJoint!=NULL ){
		for(ms=pSceneDesc->models ; *ms ; ms++){
			if( (*ms)->matanimjointlist==NULL ){
				continue;
			}
			if( (*ppMatAnimJoint=GetMatAnimJoint(pName,(*ms)->matanimjointlist[0]))!=NULL ){
				break;
			}
		}
	}
	if( ppShapeAnimJoint!=NULL ){
		for(ms=pSceneDesc->models ; *ms ; ms++){
			if( (*ms)->shapeanimjointlist==NULL ){
				continue;
			}
			if( (*ppShapeAnimJoint=GetShapeAnimJoint(pName,(*ms)->shapeanimjointlist[0]))!=NULL ){
				break;
			}
		}
	}
	return true;
}

bool WK7ModelMotion::GetAnimJoint(int index        ,HSD_AnimJoint      **ppAnimJoint,
                                                    HSD_MatAnimJoint   **ppMatAnimJoint,
                                                    HSD_ShapeAnimJoint **ppShapeAnimJoint)
{
	if( mpArchive==NULL ){
		return false;
	}
	return GetAnimJoint(mpArchive->GetModelSet(index),ppAnimJoint,ppMatAnimJoint,ppShapeAnimJoint);
}

bool WK7ModelMotion::GetAnimJoint(C_ModelSet *pModelSet,HSD_AnimJoint      **ppAnimJoint,
                                                        HSD_MatAnimJoint   **ppMatAnimJoint,
                                                        HSD_ShapeAnimJoint **ppShapeAnimJoint)
{
	if( pModelSet==NULL ){
		return false;
	}
	if( ppAnimJoint!=NULL ){
		*ppAnimJoint      = pModelSet->adjointlist       ==NULL? NULL : pModelSet->adjointlist       [0];
	}
	if( ppMatAnimJoint!=NULL ){
		*ppMatAnimJoint   = pModelSet->matanimjointlist  ==NULL? NULL : pModelSet->matanimjointlist  [0];
	}
	if( ppShapeAnimJoint!=NULL ){
		*ppShapeAnimJoint = pModelSet->shapeanimjointlist==NULL? NULL : pModelSet->shapeanimjointlist[0];
	}
	return true;
}

bool WK7ModelMotion::GetAnimExist(C_ModelSet *pModelSet)
{
	if( pModelSet==NULL ){
		return false;
	}
	if( pModelSet->adjointlist       !=NULL && pModelSet->adjointlist       [0]!=NULL ){
		return true;
	}
	if( pModelSet->matanimjointlist  !=NULL && pModelSet->matanimjointlist  [0]!=NULL ){
		return true;
	}
	if( pModelSet->shapeanimjointlist!=NULL && pModelSet->shapeanimjointlist[0]!=NULL ){
		return true;
	}
	return false;
}

int WK7ModelMotion::GetModelSetNum(void)
{
	return mpArchive==NULL? 0 : mpArchive->GetModelSetNum();
}

int WK7ModelMotion::GetAnimJointNum(void)
{
	if( mpArchive==NULL ){
		return 0;
	}
	int		modelSetNum = mpArchive->GetModelSetNum();
	int		i,num = 0;
	for(i=0 ; i<modelSetNum ; i++){
		C_ModelSet	*pModelSet = mpArchive->GetModelSet(i);
		if( pModelSet->adjointlist       !=NULL && pModelSet->adjointlist       [0]!=NULL ){
			num++;
		}
	}
	return num;
}

int WK7ModelMotion::GetMatAnimJointNum(void)
{
	if( mpArchive==NULL ){
		return 0;
	}
	int		modelSetNum = mpArchive->GetModelSetNum();
	int		i,num = 0;
	for(i=0 ; i<modelSetNum ; i++){
		C_ModelSet	*pModelSet = mpArchive->GetModelSet(i);
		if( pModelSet->matanimjointlist  !=NULL && pModelSet->matanimjointlist  [0]!=NULL ){
			num++;
		}
	}
	return num;
}

int WK7ModelMotion::GetShapeAnimJointNum(void)
{
	if( mpArchive==NULL ){
		return 0;
	}
	int		modelSetNum = mpArchive->GetModelSetNum();
	int		i,num = 0;
	for(i=0 ; i<modelSetNum ; i++){
		C_ModelSet	*pModelSet = mpArchive->GetModelSet(i);
		if( pModelSet->shapeanimjointlist!=NULL && pModelSet->shapeanimjointlist[0]!=NULL ){
			num++;
		}
	}
	return num;
}

f32 WK7ModelMotion::GetTotalMotionFrame(HSD_AnimJoint *pAnimJoint,bool next,bool child)
{
	if( pAnimJoint==NULL ){
		return -1.0f;
	}
	f32		frame;
	if( pAnimJoint->aobjdesc != NULL ){
		return pAnimJoint->aobjdesc->endframe;
	}
	if( child==true && pAnimJoint->child!=NULL ){
		if( (frame = GetTotalMotionFrame(pAnimJoint->child,true,true)) >= 0.0f ){
			return frame;
		}
	}
	if( next==true && pAnimJoint->next!=NULL ){
		if( (frame = GetTotalMotionFrame(pAnimJoint->next,true,child)) >= 0.0f ){
			return frame;
		}
	}
	return -1.0f;
}

f32 WK7ModelMotion::GetTotalAllMotionFrame(int index,bool next,bool child)
{
	f32		frame;
	if( (frame=GetTotalMotionFrame(index,next,child)) >= 0.0f ){
		return frame;
	}
	return -1.0f;
}

f32 WK7ModelMotion::GetTotalMotionFrame(int index,bool next,bool child)
{
	if( mpArchive==NULL ){
		return -1.0f;
	}
	C_ModelSet	*pModelSet;
	if( index >= 0 ){
		if( (pModelSet = mpArchive->GetModelSet(index))==NULL ){
			return -1.0f;
		}
		if( pModelSet->adjointlist==NULL || pModelSet->adjointlist[0]==NULL ){
			return -1.0f;
		}
		return GetTotalMotionFrame(pModelSet->adjointlist[0],next,child);
	}
	int		modelSetNum = mpArchive->GetModelSetNum();
	int		i;
	f32		frame;
	for(i=0 ; i<modelSetNum ; i++){
		pModelSet = mpArchive->GetModelSet(i);
		if( pModelSet->adjointlist==NULL || pModelSet->adjointlist[0]==NULL ){
			continue;
		}
		if( (frame = GetTotalMotionFrame(pModelSet->adjointlist[0],next,child)) >= 0.0f ){
			return frame;
		}
	}
	return -1.0f;
}

WK7CameraMotion::WK7CameraMotion(WK7Archive *pArchive) : WK7Motion(pArchive)
{
}

WK7CameraMotion::~WK7CameraMotion()
{
}

int WK7CameraMotion::GetAnimNum(WK7Archive *pArchive)
{
	if( pArchive==NULL ){
		return 0;
	}
	C_CameraSet	*pCameraSet = pArchive->GetCameraSet();
	if( pCameraSet==NULL || pCameraSet->camanimlist==NULL ){
		return 0;
	}
	int				i;
	for(i=0 ; pCameraSet->camanimlist[i]!=NULL ; i++){}
	return i;
}

HSD_CameraAnim* WK7CameraMotion::GetAnim(WK7Archive *pArchive,int index)
{
	if( pArchive==NULL ){
		return NULL;
	}
	C_CameraSet	*pCameraSet = pArchive->GetCameraSet();
	if( pCameraSet==NULL || pCameraSet->camanimlist==NULL ){
		return NULL;
	}
	int				i;
	HSD_CameraAnim	*pAnim;
	if( index < 0 ){
		for(i=0 ; (pAnim=pCameraSet->camanimlist[i]) != NULL ; i++){
			if( strcmp(GetDescName(pAnim),"persp")!=0 ){
				break;
			}
		}
	}else{
		for(i=0 ; (pAnim=pCameraSet->camanimlist[i]) != NULL && i!=index; i++){}
	}
	return pAnim;
}

HSD_CameraAnim* WK7CameraMotion::GetAnim(WK7Archive *pArchive,const char *pName)
{
	if( pArchive==NULL || pName==NULL ){
		return NULL;
	}
	C_CameraSet	*pCameraSet = pArchive->GetCameraSet();
	if( pCameraSet==NULL || pCameraSet->camanimlist==NULL ){
		return NULL;
	}
	int				i;
	HSD_CameraAnim	*pAnim;
	for(i=0 ; (pAnim=pCameraSet->camanimlist[i]) != NULL ; i++){
		if( strcmp(GetDescName(pAnim),pName)==0 ){
			break;
		}
	}
	return pAnim;
}

const char* WK7CameraMotion::GetAnimName(WK7Archive *pArchive,int index)
{
	if( pArchive==NULL ){
		return NULL;
	}
	C_CameraSet	*pCameraSet = pArchive->GetCameraSet();
	if( pCameraSet==NULL || pCameraSet->camanimlist==NULL ){
		return NULL;
	}
	int				i;
	HSD_CameraAnim	*pAnim;
	for(i=0 ; (pAnim=pCameraSet->camanimlist[i]) != NULL && i!=index; i++){}
	return GetDescName(pAnim);
}

WK7LightMotion::WK7LightMotion(WK7Archive *pArchive) : WK7Motion(pArchive)
{
}

WK7LightMotion::~WK7LightMotion()
{
}

int WK7LightMotion::GetAnimNum(void)
{
	if( mpArchive==NULL ){
		return 0;
	}
	int		lightSetNum = mpArchive->GetLightSetNum();
	int		i,num = 0;
	for(i=0 ; i<lightSetNum ; i++){
		C_LightSet	*pLightSet = mpArchive->GetLightSet(i);
		if( pLightSet->lightanimlist!=NULL && pLightSet->lightanimlist[0]!=NULL ){
			num++;
		}
	}
	return num;
}

HSD_LightAnim* WK7LightMotion::GetAnim(int index)
{
	if( mpArchive==NULL ){
		return NULL;
	}
	return GetAnim(mpArchive->GetLightSet(index));
}

HSD_LightAnim* WK7LightMotion::GetAnim(const char *pName)
{
	if( mpArchive==NULL ){
		return NULL;
	}
	return GetAnim(mpArchive->GetLightSet(pName));
}

HSD_LightAnim* WK7LightMotion::GetAnim(C_LightSet *pLightSet)
{
	if( pLightSet==NULL ){
		return NULL;
	}
	if( pLightSet->lightanimlist==NULL || pLightSet->lightanimlist[0]==NULL ){
		return NULL;
	}
	return pLightSet->lightanimlist[0];
}

const char* WK7LightMotion::GetAnimName(int index)
{
	HSD_LightAnim	*pAnim = WK7LightMotion::GetAnim(index);
	if( pAnim==NULL ){
		return NULL;
	}
	return GetDescName(pAnim);
}

