////////////////////////////////////////////////////////////////////////////////
//
// Killer7 SceneBlur
//
// WK7SceneBlur.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 07/16/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7SCENEBLUR_CPP__

//#include <dolphin.h>
#include ".\WK7Define.h"
#include ".\WK7SceneBlur.h"
#include ".\WK7Cam.h"

WK7SceneBlur	SceneBlur;

WK7SceneBlurScreen::WK7SceneBlurScreen(ghmGcListObj *pAnchor) : ghmGcListObj(pAnchor)
{
	mVisible = true;
	mOffsetX=2.0f; mOffsetY=2.0f; mZDist=100.0f; mAlpha=0x40; mColor=0xffffff00;
//	mOffsetX=10.0f; mOffsetY=10.0f; mZDist=100.0f; mAlpha=0x80; mColor=0xff000000;
}

bool WK7SceneBlurScreen::RenderProcess(WK7SceneBlur *pBlur)
{
	if( mVisible==false ){
		return true;
	}
	f32		z = pBlur->TransformCamZDisttoScreenZ(mZDist);
	u32		color = (mColor & 0xffffff00)|(mAlpha & 0x000000ff);
	
	GXBegin(GX_QUADS,GX_VTXFMT0,4);
	
		GXPosition3f32(mOffsetX + 0.0f            ,mOffsetY + 0.0f             ,z);
		GXColor1u32   (color);
		GXTexCoord2f32(0.0f,0.0f);
		
		GXPosition3f32(mOffsetX + WK7_SCREEN_WIDTH,mOffsetY + 0.0f             ,z);
		GXColor1u32   (color);
		GXTexCoord2f32(1.0f,0.0f);
		
		GXPosition3f32(mOffsetX + WK7_SCREEN_WIDTH,mOffsetY + WK7_SCREEN_HEIGHT,z);
		GXColor1u32   (color);
		GXTexCoord2f32(1.0f,1.0f);
		
		GXPosition3f32(mOffsetX + 0.0f            ,mOffsetY + WK7_SCREEN_HEIGHT,z);
		GXColor1u32   (color);
		GXTexCoord2f32(0.0f,1.0f);
	
	GXEnd();
	
	return true;
	
}

WK7SceneBlur::WK7SceneBlur()
{
	Vec				vPos;
	Vec				vUp;
	Vec				vAt;
	mScreenNum = 0;
	memset(mProjection,0,sizeof(mProjection));
	memset(mViewport  ,0,sizeof(mViewport  ));
	vUp .x =    0.0f;
	vUp .y =    1.0f;
	vUp .z =    0.0f;
	vPos.x =    0.0f;
	vPos.y =    0.0f;
	vPos.z =    0.0f;
	vAt .x =  vPos.x;
	vAt .y =  vPos.y;
	vAt .z =    1.0f;
	MTXLookAt(mViewMatrix  ,&vPos,&vUp,&vAt);
	vUp .y = -  1.0f;
	MTXLookAt(mView2DMatrix,&vPos,&vUp,&vAt);
	MTXOrtho (mProj2DMatrix,0,-WK7_SCREEN_HEIGHT,0,WK7_SCREEN_WIDTH,0.0f,1.0f);
	mVisible = true;
}

WK7SceneBlur::~WK7SceneBlur()
{
}

f32 WK7SceneBlur::TransformCamZDisttoScreenZ(f32 camzdist)
{
	Vec		pos;
	GXProject(0.0f,0.0f,camzdist,
	          mViewMatrix,
	          mProjection,
	          mViewport,
	          &pos.x,&pos.y,&pos.z);
	return pos.z;
}

bool WK7SceneBlur::Setup(GXTexObj *pTexObj)
{
	
	HSD_RectU16		sc_rect;
	
	WK7CAM->SetCurrent();
	WK7CAM->GetScissor(sc_rect);
	
	GXGetProjectionv(mProjection);
	GXGetViewportv  (mViewport);
	
	GXSetScissor    (sc_rect.xmin,
	                 sc_rect.ymin,
	                 (u32)(sc_rect.xmax-sc_rect.xmin),
	                 (u32)(sc_rect.ymax-sc_rect.ymin));
	GXSetViewport   (0.0f,0.0f,WK7_SCREEN_WIDTH,WK7_SCREEN_HEIGHT,0.0f,1.0f);
	
	GXSetDstAlpha   (GX_FALSE,0x00);
	GXSetDither     (GX_FALSE);
	
	GXSetAlphaCompare(GX_ALWAYS,0xff,GX_AOP_AND,GX_ALWAYS,0xff);
	
	GXSetNumChans     (1);
	GXSetChanCtrl     (GX_COLOR0A0,GX_FALSE,GX_SRC_VTX,GX_SRC_VTX,GX_LIGHT_NULL,GX_DF_NONE,GX_AF_NONE);
	
	GXSetCullMode     (GX_CULL_NONE);
	
	GXSetZMode      (GX_TRUE,GX_LEQUAL,GX_FALSE);
	GXSetZCompLoc   (GX_TRUE);
	GXSetColorUpdate(GX_TRUE);
	GXSetAlphaUpdate(GX_FALSE);
	
	GXSetBlendMode  (GX_BM_BLEND,GX_BL_SRCALPHA,GX_BL_INVSRCALPHA,GX_LO_CLEAR);
	
	GXLoadTexObj      (pTexObj,GX_TEXMAP0);
	
	GXClearVtxDesc    ();
	GXSetVtxDesc      (GX_VA_POS ,GX_DIRECT);
	GXSetVtxDesc      (GX_VA_NRM ,GX_NONE  );
	GXSetVtxDesc      (GX_VA_CLR0,GX_DIRECT);
	GXSetVtxDesc      (GX_VA_TEX0,GX_DIRECT);
	
	GXSetVtxAttrFmt   (GX_VTXFMT0,GX_VA_POS ,GX_POS_XYZ ,GX_F32  ,0);
	GXSetVtxAttrFmt   (GX_VTXFMT0,GX_VA_CLR0,GX_CLR_RGBA,GX_RGBA8,0);
	GXSetVtxAttrFmt   (GX_VTXFMT0,GX_VA_TEX0,GX_TEX_ST  ,GX_F32  ,0);
	
	GXSetNumTexGens   (1);
	GXSetTexCoordGen  (GX_TEXCOORD0,GX_TG_MTX2x4,GX_TG_TEX0,GX_IDENTITY);
	
	GXSetNumTevStages (1);
	GXSetTevColorIn   (GX_TEVSTAGE0,GX_CC_ZERO,GX_CC_RASC,GX_CC_TEXC,GX_CC_ZERO);
//	GXSetTevColorIn   (GX_TEVSTAGE0,GX_CC_ZERO,GX_CC_ZERO,GX_CC_ZERO,GX_CC_TEXC);
	GXSetTevColorOp   (GX_TEVSTAGE0,GX_TEV_ADD,GX_TB_ZERO,GX_CS_SCALE_1,GX_TRUE,GX_TEVPREV);
	GXSetTevAlphaIn   (GX_TEVSTAGE0,GX_CA_ZERO,GX_CA_ZERO,GX_CA_ZERO,GX_CA_RASA);
	GXSetTevAlphaOp   (GX_TEVSTAGE0,GX_TEV_ADD,GX_TB_ZERO,GX_CS_SCALE_1,GX_TRUE,GX_TEVPREV);
	GXSetTevOrder     (GX_TEVSTAGE0,GX_TEXCOORD0,GX_TEXMAP0,GX_COLOR0A0);
	
	GXLoadPosMtxImm   (mView2DMatrix,GX_PNMTX0);
	GXLoadNrmMtxImm   (mView2DMatrix,GX_PNMTX0);
	GXSetCurrentMtx   (GX_PNMTX0);
	GXSetProjection   (mProj2DMatrix,GX_ORTHOGRAPHIC);
	
	return true;
}

bool WK7SceneBlur::RenderProcess(GXTexObj *pTexObj)
{
	
	if( pTexObj==NULL ){
		return false;
	}
	
	if( mVisible==false || mScreenNum <= 0 ){
		return true;
	}
	
	Setup(pTexObj);
	
	WK7SceneBlurScreen	*pScreen = (WK7SceneBlurScreen *)mAnchor.GetNext();
	
	while( pScreen!=NULL ){
		pScreen->RenderProcess(this);
		pScreen = pScreen->GetNext();
	}
	
	GXSetDither(GX_TRUE);
	
	return true;
}

WK7SceneBlurScreen* WK7SceneBlur::CreateScreen(f32 ofsx,f32 ofsy,f32 z,u32 alpha)
{
	ghmGcMemSetAllocName("ScrBlur");
	WK7SceneBlurScreen	*pScreen = new WK7SceneBlurScreen(&mAnchor);
	if( pScreen!=NULL ){
		pScreen->SetOffset(ofsx,ofsy);
		pScreen->SetZDist (z);
		pScreen->SetAlpha (alpha);
		mScreenNum++;
	}
	return pScreen;
}

bool WK7SceneBlur::ReleaseScreenAll(void)
{
	WK7SceneBlurScreen	*pScreen = (WK7SceneBlurScreen *)mAnchor.GetNext();
	WK7SceneBlurScreen	*pNext;
	while( pScreen!=NULL ){
		pNext   = pScreen->GetNext();
		delete pScreen;
		pScreen = pNext;
	}
	mScreenNum = 0;
	return true;
}

bool WK7SceneBlur::ReleaseScreen(WK7SceneBlurScreen* pScreen)
{
	if( pScreen==NULL ){
		return false;
	}
	WK7SceneBlurScreen	*pTmpScreen = (WK7SceneBlurScreen *)mAnchor.GetNext();
	while( pTmpScreen!=NULL ){
		if( pTmpScreen==pScreen ){
			delete pScreen;
			mScreenNum--;
			return true;
		}
		pTmpScreen = pTmpScreen->GetNext();
	}
	return false;
	
}

bool WK7SceneBlur::ReleaseScreen(int index)
{
	WK7SceneBlurScreen	*pScreen = GetScreen(index);
	if( pScreen==NULL ){
		return false;
	}
	mScreenNum--;
	delete pScreen;
	return true;
}

WK7SceneBlurScreen* WK7SceneBlur::GetScreen(int index)
{
	if( index < 0 || index >= mScreenNum ){
		return NULL;
	}
	WK7SceneBlurScreen	*pScreen = (WK7SceneBlurScreen *)mAnchor.GetNext();
	while( pScreen!=NULL ){
		if( !index ){
			return pScreen;
		}
		pScreen = pScreen->GetNext();
		index--;
	}
	return NULL;
}

