////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Select InterFace
//
// WK7SelectIF.h
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 02/18/2003
//
////////////////////////////////////////////////////////////////////////////////

#ifndef	__WK7SELECTIF_H__
#define	__WK7SELECTIF_H__

#include ".\WK7Obj.h"
#include ".\WK7Path.h"
#include "..\Matsu\MK7Typo.h"
//#include "..\Lib\ghmGcColl.h"

#define	WK7SELECTITEM_EXPAND_RATE_MAX			(1024)
#define	WK7SELECTITEM_DEFAULT_FADETICK			(2400)

class WK7SelectObj;
typedef	int WK7SelectObjCallback(WK7SelectObj *pObj);

class WK7SelectObj : public WK7Obj{
	friend class WK7SelectIF;
public:
	enum{
		SELECT_CANCEL,
		SELECT_OK,
		SELECT_EVENT_CANCEL,
	};
protected:
	char							*mpObjName[2];
	union{
		struct{
			u32						mEnable     :  1;
			u32						mPlayEnable :  1;
			u32						mFlagPad    : 30;
		};
		u32							mFlag;
	};
	WK7SelectObjCallback			*mpSelectCallback;
	u32								mParam[2];
	u32								mSelectEventID;
public:
	WK7SelectObj();
	virtual ~WK7SelectObj();
	
	// Xg|C^
	inline WK7SelectObj* GetPrev(void){ return (WK7SelectObj *)ghmGcListObj::GetPrev(); }
	inline WK7SelectObj* GetNext(void){ return (WK7SelectObj *)ghmGcListObj::GetNext(); }
	
	virtual bool SetObjName(int index,const char *pName);
	virtual const char* GetObjName(int index);
	
	virtual bool SetPosition(const Vec &pos) = 0;
	virtual bool SetPosition(f32 x,f32 y,f32 z){
		Vec	pos;
		pos.x = x;
		pos.y = y;
		pos.z = z;
		return SetPosition(pos);
	}
	virtual bool GetPosition(Vec &pos) = 0;
	virtual int GetCrossState(ghmGcViewVolume &vv) = 0;
	
	inline void SetEnable(bool flag){ mEnable = (flag==true? 1 : 0); }
	inline bool GetEnable(void){ return mEnable? true : false; }
	inline void SetPlayEnable(bool flag){ mPlayEnable = (flag==true? 1 : 0); }
	inline bool GetPlayEnable(void){ return mPlayEnable? true : false; }
	
	bool SetParam(int index,u32 param);
	u32 GetParam(int index);
	
	bool CallSelectCallback(void);
	WK7SelectObjCallback* SetSelectCallback(WK7SelectObjCallback *pSelectCallback);
	inline WK7SelectObjCallback* GetSelectCallback(void){ return mpSelectCallback; }
	
	inline void SetSelectEventID(u32 eventid){ mSelectEventID=eventid; }
	inline u32 GetSelectEventID(void){ return mSelectEventID; }
};

class WK7SelectObjSphere : public WK7SelectObj{
	friend class WK7SelectIF;
protected:
	ghmGcCollSphere					mSphere;
public:
	WK7SelectObjSphere();
	virtual ~WK7SelectObjSphere();
	
	bool SetRadius(f32 radius);
	virtual bool SetPosition(const Vec &pos);
	virtual bool GetPosition(Vec &pos);
	virtual int GetCrossState(ghmGcViewVolume &vv);
};

class WK7SelectObjSphereJObj : public WK7SelectObjSphere{
	friend class WK7SelectIF;
protected:
	WK7JObj							*mpJObj;
public:
	WK7SelectObjSphereJObj(WK7JObj *pJObj);
	virtual ~WK7SelectObjSphereJObj();
	
	virtual bool FrameProcess(void);
	virtual bool SetPosition(const Vec &pos);
	virtual bool GetPosition(Vec &pos);
};

class WK7SelectItemFrameLineSize{
	friend class WK7SelectItemFrameLine;
	friend class WK7SelectItem;
protected:
	WK7AnimLinearF					mSize;
	f32								mMin;
	f32								mMax;
	f32								mColorRate;
	inline bool SetColorRate(f32 color){ mColorRate = color; return true; }
	inline f32 GetColorRate(void){ return mColorRate; }
public:
	WK7SelectItemFrameLineSize();
	~WK7SelectItemFrameLineSize();
	
	bool SetRange(f32 max,f32 min);
	bool SetSize(f32 size);
	bool FrameProcess(void);
};

class WK7SelectItemFrameLinePoint{
	friend class WK7SelectItemFrameLine;
	friend class WK7SelectItem;
protected:
	WK7AnimLinearF					mRate;
	f32								mMin;
	f32								mMax;
public:
	WK7SelectItemFrameLinePoint();
	~WK7SelectItemFrameLinePoint();
	
	bool SetRange(f32 max,f32 min);
	bool SetRate(f32 rate);
	bool FrameProcess(void);
};

class WK7SelectItemFrameLine{
	friend class WK7SelectItem;
protected:
	static f32						mRenderZ;
	Vec								*mpPosition [2];
	Vec								*mpPositionP[2];
	u32								mColor[2];
	WK7SelectItemFrameLinePoint		*mpPoint;
	int								mPointCount;
	WK7SelectItemFrameLineSize		*mpSize;
public:
	WK7SelectItemFrameLine();
	~WK7SelectItemFrameLine();
	
	inline static void SetRenderZ(f32 z){ mRenderZ = z; }
	inline static f32 GetRenderZ(void){ return mRenderZ; }
	
	bool SetDivCount(int count);
	
	void SetPosition(Vec *pStartPos,Vec *pEndPos);
	void SetPositionP(Vec *pStartPos,Vec *pEndPos);
	inline const Vec* GetStartPosition(void){ return mpPosition[0]; }
	inline const Vec* GetEndPosition(void){ return mpPosition[1]; }
	inline const Vec* GetStartPositionP(void){ return mpPositionP[0]; }
	inline const Vec* GetEndPositionP(void){ return mpPositionP[1]; }
	
	void SetColor(u32 scol,u32 ecol);
	inline u32 GetStartColor(void){ return mColor[0]; }
	inline u32 GetEndColor(void){ return mColor[1]; }
	
	bool FrameProcess(void);
	bool RenderProcess(f32 linesize);
};

class WK7SelectItemFrame{
public:
	Vec						mCutFramePosition     [6];		// Jbg̃t[ʒu
//	Vec						mCutFramePosition     [4];		// Jbg̃t[ʒu
	Vec						mExpandFramePosition  [4];		// g厞̃t[ʒu
	Vec						mCurrentFramePosition [4];		// ݂̃t[ʒu
	Vec						mParallelFramePosition[4];		// t[ʒuɕsȈʒu
};

class WK7SelectItemFrameEx : public WK7SelectItemFrame{
public:
	WK7AnimLinearF			mDeltaX[4];						// wψڒl
	WK7AnimLinearF			mDeltaY[4];						// xψڒl
	WK7SelectItemFrameEx();
};

class WK7SelectItemLockOn{
public:
	enum State{
		IDLE = 0,
		LOCKING,
		LOCKED,
	};
protected:
	f32						mPosX,mPosY;
	f32						mZ;
	State					mState;
	u32						mColor;
	WK7AnimLinear			mCount;
public:
	WK7SelectItemLockOn();
	~WK7SelectItemLockOn();
	
	bool SetPosition(f32 x,f32 y);
	inline bool SetZ(f32 z){ mZ = z; return true; }
	inline bool SetColor(u32 col){ mColor=col; return true; }
	
	bool FrameProcess(void);
	bool RenderProcess(void);
};

enum{
	FONTTYPE_ARIAL = 0,
	FONTTYPE_ARIALBLACK,
};

stFontParam* GetFontParam(char letter,int type=FONTTYPE_ARIAL);

class WK7Letter{
protected:
	mK7Sprite				mPrim;
	f32						mCurPosX,mCurPosY;
	f32						mDstPosX,mDstPosY;
	u32						mRGBA;
	int						mType;
	s8						mLetter;
	u8						mDisp;
	bool SetPrim(void);
	bool SetTexture(void);
	stFontParam* GetFontParam(void);
public:
	WK7Letter();
//	~WK7Letter(){}
	
	inline int GetType(void){ return mType; }
	inline u32 GetRGBA(void){ return mRGBA; }
	inline void GetCurXY(f32 &x,f32 &y){ x=mCurPosX;y=mCurPosY; }
	inline void GetDstXY(f32 &x,f32 &y){ x=mDstPosX;y=mDstPosY; }
	inline void GetXY(f32 &x,f32 &y){ GetCurXY(x,y); }
	inline s8 GetLetter(void){ return mLetter; }
	inline bool GetDisp(void){ return mDisp? true : false; }
	inline s16 GetW(void){ return mPrim.mGetW(); }
	inline s16 GetH(void){ return mPrim.mGetH(); }
	
	bool FrameProcess(void);
	
	void SetDisp(bool flag);
	void SetLetter(s8 letter);
	bool SetType(int type);
	void SetXY(f32 x,f32 y,bool im=false);
	void SetRGBA(u32 rgba);
	void SetRGB(u32 rgb);
	void SetR(u32 r);
	void SetG(u32 g);
	void SetB(u32 b);
	void SetA(u32 a);
};

class WK7SelectItemFrameBlur : public WK7Obj{
	friend class WK7SelectItem;
protected:
	WK7SelectItemFrame			mFrame;
	WK7AnimLinear				mAlpha;
	u32							mColor;
	bool						mTerminate;
	f32							mRenderZ;
public:
	WK7SelectItemFrameBlur(ghmGcListObj *pAnchor,WK7SelectItemFrame &frame,u32 color,f32 z);
	
	WK7SelectItemFrameBlur* GetNext(void){ return (WK7SelectItemFrameBlur *)WK7Obj::GetNext(); }
	WK7SelectItemFrameBlur* GetPrev(void){ return (WK7SelectItemFrameBlur *)WK7Obj::GetPrev(); }
	
	virtual bool FrameProcess(void);
	virtual bool RenderProcess(void);
	inline void SetTerminate(void){ mTerminate=true; }
	bool IsEqual(WK7SelectItemFrame &frame);
};

class WK7SelectItem : public WK7Obj{
	friend class WK7SelectIF;
public:
	enum{
		DIR_N = 0,
		DIR_NNW,
		DIR_NW,
		DIR_WNW,
		DIR_W,
		DIR_WSW,
		DIR_SW,
		DIR_SSW,
		DIR_S,
		DIR_SSE,
		DIR_SE,
		DIR_ESE,
		DIR_E,
		DIR_ENE,
		DIR_NE,
		DIR_NNE,
		DIR_MAX,
	};
	enum Result{
		RESULT_OK = 0,
	};
	typedef Result Callback(WK7SelectItem *pItem,u32 param0,u32 param1);
	enum{
		CUTFRAME = 0,
		FADEIN,
		FADEOUT,
		VANISHED,
		EXPAND,
		EXPANDED,
	};
	enum{
		TYPE_CUT = 0,
		TYPE_LINE,
	};
protected:
	// \n
	static bool				mTargetCircleVisible;
	static Vec				mWorldItemCenter;
	static Vec				mScreenItemCenter;
	static f32				mProjection[GX_PROJECTION_SZ];
	static f32				mViewport  [GX_VIEWPORT_SZ  ];
	static Mtx				mViewMatrix;
	static u32				mIDSeed;
	
	u32						mID;
	
	Vec						*mpTargetPoints;
	int						mTargetPointNum;
	Vec						mCutInPosition;
	Vec						mCutInDirection;
	
	Vec						*mpScreenTargetPoints;
	Vec						mScreenTargetCenter;
	f32						mScreenTargetRadius;
//	Vec						tp[2];
	
	Vec						mPrevScreenTargetCenter;
	
	Vec						mTypoCenterCurrent;
	Vec						mTypoCenter;
	Vec						mTypoCenterVelocity;
	Vec						mTypoCenterAccel;
	
	WK7SelectItemFrame		mFrame;
	WK7SelectItemFrameEx	mFrameEx[2];
//	Vec						mCutFramePosition    [4];		// Jbg̃t[ʒu
//	Vec						mExpandFramePosition [4];		// g厞̃t[ʒu
//	Vec						mCurrentFramePosition[4];		// ݂̃t[ʒu
//	Vec						mParallelFramePosition[4];		// t[ʒuɕsȈʒu
	WK7SelectItemFrameLine	*mpFrameLine;
	
	Vec						mRevPos;
	Vec						mFocusPos;
	
	f32						mCutRate;
	
	f32						mRenderZ;
	
	union{
		struct{
			u8				mLineWidth;
			u8				mLineWidthPad[3];
		};
		struct{
			u32				mLineWidthPart   : 8;
			u32				mVisible         : 1;
			u32				mRender          : 1;
			u32				mEnable          : 1;
			u32				mTargetCenterSet : 1;
			u32				mHighLighted     : 1;
			u32				mSemiHighLighted : 1;
			u32				mEnableSelect    : 1;
			u32				mLetterDisp      : 1;
			u32				mResetTypoPos    : 1;
			u32				mFlagPad         :15;
		};
		u32					mFlag;
	};
	
	int						mState;
	
	MK7Typos				mTypos;							// ^C|
	
	WK7AnimLinear			mAlpha;							// l
	WK7AnimLinear			mExpandRate;					// g劄
//	WK7AnimLinear			mTypoAlpha;						// ^C|l
	WK7AnimLinear			mTypoPosition;					// ^C|ʒu
	
	WK7AnimLinear			mColorR;						// J[ql
	WK7AnimLinear			mColorG;						// J[fl
	WK7AnimLinear			mColorB;						// J[al
	
	static GXColor			mNormalColor;
	static GXColor			mSelectedColor;
	
	// f[^n
	static int				mPadDirection;
	
	int						mDirection;
	bool					mSelected;
	u32						mParam0,mParam1;
	
	WK7PathBranch			*mpBranch;
	WK7PathBranchPointObj	*mpBranchPointObj;
	u32						mCallEventID;
	
	WK7SelectObj			*mpObj;
	
	u32						mCurrentSelectedColor;
	u32						mCurrentNormalColor;
	u32						mCurrentMaskColor;
	u32						mCurrentBackColor;
	
	int						mType;
	WK7SelectItemLockOn		*mpLockOn;
	u32						mIcon;
	mK7Sprite				*mpRatePrim;
	f32						mRate;
	
	WK7Letter				*mpLetter;
	WK7AnimLinearF			mLetterDirRate;
	
	WK7AnimLinearF			mDeformRate;
	
	ghmGcListObj			mBlurAnchor;
	WK7AnimLinear			mBlurCounter;
	
	// 
	virtual bool Initialize(void);
	
	// I
	virtual bool Terminate(void);
	
	// t[
	virtual bool FrameProcess(void);
	static bool CalcCenterPosition(void);
	bool CalcFramePosition(void);
	
	bool CalcScreenTargetPoint(void);
	bool CalcCutInPosition(void);
	bool CalcCutFramePosition(void);
	bool CalcExpandPosition(void);
	bool CalcFrameExPosition(WK7SelectItemFrameEx *pFrameEx);
	bool CalcTypoPosition(void);
	bool CalcTypoOverlap(void);
	bool CalcTypoOverlap(WK7SelectItem *pItem);
	
	// _O
	virtual bool RenderProcess(void);
	bool RenderMask(void);
	bool RenderLine(f32 backz);
	bool RenderItemLine(void);
	bool RenderRate(void);
	
	f32 GetSelectRate(const Vec &dir);
	bool SetSelected(bool flag);
	
	bool CalcParallelFramePosition(f32 dist);
	
	bool SetupRenderProcess(void);
	bool RenderBack(void);
	bool RenderBackColor(void);
	bool RenderFrame(void);
	bool RenderTargetCircle(void);
	
	u32 GetBackAlpha(void);
	
	WK7SelectItem(ghmGcListObj *pAnchor,int prio=WK7OBJ_DEFAULT_PRIO);
	virtual ~WK7SelectItem();
	
	inline void SetEnable(bool flag){ mEnable = (flag==true)? 1:0; }
	inline void SetEnableSelect(bool flag){ mEnableSelect = (flag==true)? 1:0; }
public:
	
	u32 GetID(void);
	
	inline static void SetTargetCircleVisible(bool flag){ mTargetCircleVisible = flag; }
	inline static bool GetTargetCircleVisible(void){ return mTargetCircleVisible; }
	
	void SetDeformRate(f32 rate=1.0f);
	inline f32 GetDeformRate(void){ return mDeformRate.GetValue(); }
	
	static bool CalcPadDirection(void);
	static int GetPadDirection(void);
	
	inline f32 GetRate(void){ return mRate; }
	bool SetType(int type);
	inline int GetType(void){ return mType; }
	
	inline void SetHighLighted(bool flag){ mHighLighted = (flag==true)? 1 : 0; }
	inline bool GetHighLighted(void){ return mHighLighted? true : false; }
	inline void SetSemiHighLighted(bool flag){ mSemiHighLighted = (flag==true)? 1 : 0; }
	inline bool GetSemiHighLighted(void){ return mSemiHighLighted? true : false; }
	GXColor GetSelectedColor(void);
	GXColor GetNormalColor(void);
	
	// Xg|C^
	inline WK7SelectItem* GetPrev(void){ return (WK7SelectItem *)ghmGcListObj::GetPrev(); }
	inline WK7SelectItem* GetNext(void){ return (WK7SelectItem *)ghmGcListObj::GetNext(); }
	
	bool SetTargetPoints(WK7JObj *pObj);
	bool SetTargetPoints(HSD_JObj *pJObj);
	bool SetTargetPoints(const Vec *pPoints,int pointNum,MtxPtr pLocalWorldMatrix=NULL);
	void SetCutInParam(const Vec *pCutInPosition,const Vec *pCutInDirection);
	
	void SetVisible(bool flag);
	inline bool GetVisible(void){ return mVisible? true:false; }
	
	inline void SetLineWidth(u8 width){ mLineWidth = width; }
	inline u8 GetLineWidth(void){ return mLineWidth; };
	
	inline int GetState(void){ return mState; }
	
	bool GetTargetPointAverage(Vec &ave);
	inline const Vec* GetTargetPoint(void){ return mpTargetPoints; }
	inline int GetTargetPointNum(void){ return mTargetPointNum; }
	
	// p[^ݒ聕擾
	inline void SetParam0(u32 param){ mParam0=param; }
	inline u32 GetParam0(void){ return mParam0; }
	inline void SetParam1(u32 param){ mParam1=param; }
	inline u32 GetParam1(void){ return mParam1; }
	
	// Iݒ聕擾
	bool SetDirection(int dir);
	inline int GetDirection(void){ return mDirection; }
	
	inline bool GetSelected(void){ return mSelected; }
	inline bool GetEnable(void){ return (mEnable? true : false); }
	inline bool GetEnableSelect(void){ return (mEnableSelect? true : false); }
	
	bool Appear(u32 tick=WK7SELECTITEM_DEFAULT_FADETICK);
	bool AppearTypo(u32 tick=WK7SELECTITEM_DEFAULT_FADETICK);
	bool Disappear(u32 tick=WK7SELECTITEM_DEFAULT_FADETICK);
	bool Select(u32 tick=WK7SELECTITEM_DEFAULT_FADETICK);
	
	inline void SetPathBranch(WK7PathBranch *pBranch){ mpBranch=pBranch; }
	inline WK7PathBranch* GetPathBranch(void){ return mpBranch; }
	inline void SetPathBranchPointObj(WK7PathBranchPointObj *pObj){ mpBranchPointObj=pObj; }
	inline WK7PathBranchPointObj* GetPathBranchPointObj(void){ return mpBranchPointObj; }
	inline void SetObj(WK7SelectObj* pObj){ mpObj=pObj; }
	inline WK7SelectObj* GetObj(void){ return mpObj; }
	
	inline void SetCallEventID(u32 eventid){ mCallEventID=eventid; }
	inline u32 GetCallEventID(void){ return mCallEventID; }
	
	bool SetTypo(const char *pName,const char *pDir,u32 icon=0xffffffff);
};

class WK7SelectDirControl{
protected:
	static	Vec			mDirection[8];
	static	Vec			mPosition [8];
	
	u8					mFlag[8];
	int					mUseNum;
	
	
public:
	WK7SelectDirControl();
	~WK7SelectDirControl();
	void Initialize(void);
	int AllocDirection(const Vec *pTargetPoint,const Vec *pBasePoint);
	bool FreeDirection(int dir);
	static bool GetPosition(int dir,Vec *pPosition);
	static bool GetDirection(int dir,Vec *pDirection);
};

class WK7SelectIF{
	friend class WK7SelectObj;
public:
	enum Result{
		RESULT_OK = 0,
	};
	enum State{
		IDLE,				// ~
		APPEAR,				// o
		EXECUTE,			// s
		DISAPPEAR,			// 
	};
	typedef	Result Callback(WK7SelectIF *pIF,u32 param0,u32 param1);
protected:
	static WK7SelectIF				*mpSelectIFInstance;
	
	// ItemAJ[
	ghmGcListObj					mAnchor;
	int								mItemNum;
	
	// ObjAJ[
	ghmGcListObj					mObjAnchor;
	
	union{
		struct{
			u32						mEnable        :  1;
			u32						mVisible       :  1;
			u32						mOperateEnable :  1;
			u32						mBackAlphaMax  :  1;
			u32						mRender        :  1;
			u32						mFlagPad       : 27;
		};
		u32							mFlag;
	};
	
	State							mState;						// 
	bool							mTypoVisible;				// ^C|\
	
	WK7SelectDirControl				mSelectDirControl;			// IRg[
	
	bool							mBackVisible;
	f32								mBackScale;
	f32								mBackZ;
	u32								mBackColor;
	u32								mBackTexColor;
	WK7AnimLinear					mBackAlpha;
	u32								mCinemaScopeColor;
	u32								mSelectBackColor;
	u32								mSelectBackDeltaColor;
	u32								mHighLightedBackColor;
	u32								mHighLightedBackDeltaColor;
	u32								mSemiHighLightedBackColor;
	u32								mSemiHighLightedBackDeltaColor;
	u32								mSelectedFrameColor;
	
	f32								mWindowZ;
	f32								mWindowMaskZ;
	f32								mMaskX,mMaskY,mMaskW,mMaskH;
	f32								mLineSize;
	GXTexObj						*mpFullColorScreenTexObj;
	GXTexObj						*mpMonoColorScreenTexObj;
	
	static f32						mAppearDist;
	
	WK7SelectIF();
	
	bool CalcSelectedItem(void);
	
	int CalcDir(WK7PathBranch *pBranch);
	const char* GetRouteName(char *pRouteName,const char *pOrgName,WK7PathBranch *pBranch);
	bool CreateBranchSelectItem(WK7PathBranchPoint *pBranchPoint,WK7PathBranch *pBranch);
	bool CreateBranchSelectItem(WK7PathBranchPointObj *pBranchPointObj);
	
	bool CheckSelectObj(void);
	
	bool RenderBack(void);
	
	bool RenderWindowMask(void);
	bool RenderWindowFrame(void);
	bool RenderWindowBack(void);
	bool SetupWindowViewport(void);
	bool RenderWindow(void);
	
	bool RenderCinemaScope(void);
	
	bool SetupRenderProcess(void);
	bool TermRenderProcess(void);
	bool AdjustColor(void);
public:
	virtual ~WK7SelectIF();
	
	bool IsValid(WK7SelectItem *pTgtItem,u32 id=0xffffffff);
	
	bool SetSelectedItemAll(bool flag,WK7SelectItem *pExceptItem=NULL);
	void SetDeformRateAll(f32 rate=1.0f);
	
	inline static f32 GetAppearDist(void){ return mAppearDist; }
	static bool SetupContrastTev(void);
	
	inline void SetEnable(bool flag){ mEnable = (flag==true)? 1:0; }
	inline bool GetEnable(void){ return mEnable? true : false; }
	void SetVisible(bool flag);
	inline bool GetVisible(void){ return mVisible? true : false; }
	inline void SetOperateEnable(bool flag){ mOperateEnable = (flag==true)? 1:0; }
	inline bool GetOperateEnable(void){ return mOperateEnable? true : false; }
	
	inline void SetBackVisible(bool flag){ mBackVisible=flag; }
	inline bool GetBackVisible(void){ return mBackVisible; }
	inline void SetBackScale(f32 scale){ mBackScale=scale; }
	inline f32 GetBackScale(void){ return mBackScale; }
	inline void SetBackColor(u32 col){ mBackColor=col; }
	inline u32 GetBackColor(void){ return mBackColor; }
	inline void SetBackTexColor(u32 col){ mBackTexColor=col; }
	inline u32 GetBackTexColor(void){ return mBackTexColor; }
	WK7AnimLinear* GetBackAlpha(void){ return &mBackAlpha; }
	u32 GetBackAlphaValue(void){ return mBackAlpha.GetValue(); }
	void SetCinemaScopeColor(u32 col){ mCinemaScopeColor=col; }
	u32 GetCinemaScopeColor(void){ return mCinemaScopeColor; }
	void SetSelectBackColor(u32 col){ mSelectBackColor=col; }
	u32 GetSelectBackColor(void){ return mSelectBackColor; }
	void SetSelectBackDeltaColor(u32 col){ mSelectBackDeltaColor=col; }
	u32 GetSelectBackDeltaColor(void){ return mSelectBackDeltaColor; }
	void SetHighLightedBackColor(u32 col){ mHighLightedBackColor=col; }
	u32 GetHighLightedBackColor(void){ return mHighLightedBackColor; }
	void SetHighLightedBackDeltaColor(u32 col){ mHighLightedBackDeltaColor=col; }
	u32 GetHighLightedBackDeltaColor(void){ return mHighLightedBackDeltaColor; }
	void SetSemiHighLightedBackColor(u32 col){ mSemiHighLightedBackColor=col; }
	u32 GetSemiHighLightedBackColor(void){ return mSemiHighLightedBackColor; }
	void SetSemiHighLightedBackDeltaColor(u32 col){ mSemiHighLightedBackDeltaColor=col; }
	u32 GetSemiHighLightedBackDeltaColor(void){ return mSemiHighLightedBackDeltaColor; }
	void SetSelectedFrameColor(u32 col){ mSelectedFrameColor=col; }
	u32 GetSelectedFrameColor(void){ return mSelectedFrameColor; }
	
	inline f32 GetMaskX(void){ return mMaskX; }
	inline f32 GetMaskY(void){ return mMaskY; }
	inline f32 GetMaskW(void){ return mMaskW; }
	inline f32 GetMaskH(void){ return mMaskH; }
	inline f32 GetWindowMaskZ(void){ return mWindowMaskZ; }
	inline f32 GetLineSize(void){ return mLineSize; }
	
	// 
	virtual bool Initialize(void);
	
	// I
	virtual bool Terminate(void);
	
	// t[
	virtual bool FrameProcess(void);
	
	// _O
	virtual bool RenderProcess(void);
	bool RenderBackProcess(void);
	
	inline State GetState(void){ return mState; }
	inline bool GetTypoVisible(void){ return mTypoVisible; }
	
	bool AllocInit(void);
	
	bool AppearTypo(u32 tick=WK7SELECTITEM_DEFAULT_FADETICK);
	bool Appear(u32 tick=WK7SELECTITEM_DEFAULT_FADETICK);
	bool Disappear(u32 tick=WK7SELECTITEM_DEFAULT_FADETICK,WK7SelectItem *pExceptItem=NULL);
	bool SelectItem(WK7SelectItem *pSelectItem,u32 tick=WK7SELECTITEM_DEFAULT_FADETICK);
	bool ReleaseItemAll(void);
	
	// ̕
	bool CreateNextBranchSelectItem(WK7PathInfo *pInfo,bool pathselect=false);
	
	WK7SelectItem* GetSelectedItem(void);
	
	WK7SelectItem* GetItem(WK7PathBranch *pBranch);
	WK7SelectItem* GetItem(WK7SelectObj *pObj);
	int ReleaseBranch(WK7PathBranch *pBranch);
	int ReleaseObj(WK7SelectObj *pObj);
	int ReleasePathBranchPointObj(WK7PathBranchPointObj *pObj);
	
	static WK7SelectIF* GetInstance(void);
	
	int GetEnableItemNum(void);
	inline int GetItemNum(void){ return mItemNum; }
	int GetBranchItemNum(int enable=-1);
	int GetObjItemNum(int enable=-1);
	
	int GetItemNum(int type);
	
	static f32 GetStageDirOffset(void);
};

#define	WK7SELECTIF		WK7SelectIF::GetInstance()

bool CalcParallelFramePosition(Vec *pParallelFramePosition,Vec *pFramePosition,int framePositionNum,f32 dist);
bool CalcParallelFramePosition2(Vec *pParallelFramePosition,Vec *pFramePosition,int framePositionNum,f32 dist);

#endif	//__WK7SELECTIF_H__

