////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Sight
//
// WK7Sight.h
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 04/20/2003
//
////////////////////////////////////////////////////////////////////////////////

#ifndef	__WK7SIGHT_H__
#define	__WK7SIGHT_H__

//#include <sysdolphin\baselib.h>
//#include <dolphin.h>
//#include "..\Lib\ghmGcTex.h"

//                                     S P
#define	WK7SIGHT_PEEKTOKEN			(0x5350)

struct WK7SightPeekData{
	union{
		struct{
			u8				A;
			u8				R;
			u8				G;
			u8				B;
		};
		u32					ARGB;
	};
	u32						Z;
	u32						State;
	u32						Pad;
}ATTRIBUTE_ALIGN(32);

class WK7Sight{
protected:
	static WK7Sight					*mpSightInstance;
	static GhmGcTexHeader			*mpTexture;
	static GXTexObj					mTexObj;
	static int						mPeekIndex;
	static volatile bool			mCallBackCall;
	static GXDrawSyncCallback		mpOldCallback;
	static HSD_VIGXDrawDoneCallback	mpOldDrawDoneCallback;
	
	f32								mPosX,mPosY;
	f32								mSize;
	bool							mVisible;
	GXColor							mHotColor;
	GXColor							mColdColor;
	GXColor							mCrossColor;
	f32								mCrossSize;
	u8								mCrossWidth;
	
	WK7SightPeekData				mPeekData[2];
	
	static void DrawDoneCallback(int idx);
	static void DrawSyncCallback(u16 token);
public:
	WK7Sight();
	~WK7Sight();
	
	bool FrameProcess(void);
	bool RenderProcess(void);
	
	void SetPosition(f32 x,f32 y);
	void GetPosition(f32 &x,f32 &y);
	
	inline void SetVisible(bool flag){ mVisible=flag; }
	inline bool GetVisible(void){ return mVisible; }
	
	inline void SetSize(f32 size){ mSize=size; }
	inline f32 GetSize(void){ return mSize; }
	inline void SetHotColor(GXColor col){ mHotColor=col; }
	inline GXColor GetHotColor(void){ return mHotColor; }
	inline void SetColdColor(GXColor col){ mColdColor=col; }
	inline GXColor GetColdColor(void){ return mColdColor; }
	inline void SetCrossColor(GXColor col){ mCrossColor=col; }
	inline GXColor GetCrossColor(void){ return mCrossColor; }
	inline void SetCrossSize(f32 size){ mCrossSize=size; }
	inline f32 GetCrossSize(void){ return mCrossSize; }
	inline void SetCrossWidth(u8 w){ mCrossWidth=w; }
	inline u8 GetCrossWidth(void){ return mCrossWidth; }
	
	static bool LoadTexture(const char *pfname);
	static WK7Sight* GetInstance(void);
	
	static bool GetDrawDone(void);
	
	static bool Initialize(void);
	static bool Terminate(void);
	static bool DrawSyncToken(void);
	static bool WaitSyncToken(void);
	bool PeekData(void);
	bool IsValidPeekData(void);
	u32 GetPeekARGB(void);
	u32 GetPeekZ(void);
	f32 GetPeekDist(void);
	bool ScreenToCamCoord(f32 &x,f32 &y,f32 &z);
};

#define	WK7SIGHT	WK7Sight::GetInstance()

#ifdef	_DEBUG
extern	"C"{
extern	void SetDrawSync(u16 token,void *pObj);
extern	u16 GetDrawSyncToken(void);
extern	void ClearDrawSync(void);
extern	int SearchDrawSync(u16 token);
extern	void SetDrawSyncCounter(u16 cnt);
extern	u16 GetDrawSyncCounter(void);
extern	void ResetDrawSyncCounter(void);
extern	void SetDrawSyncTokenEnable(int flag);
extern	int GetDrawSyncTokenEnable(void);
}
#define	SETDS(token,obj)		SetDrawSync((u16)(token),(obj))
#define	SETDSNO(token)			SetDrawSync((u16)(token),NULL)
#define	GETDS()					GetDrawSyncToken()
#define	CLRDS()					ClearDrawSync()
#define	SEARCHDS(token)			SearchDrawSync((u16)(token))
#define	SETDSCNT(cnt)			SetDrawSyncCounter((u16)(cnt))
#define	GETDSCNT()				GetDrawSyncCounter()
#define	RESETDSCNT()			ResetDrawSyncCounter()
#define	SETDSENABLE(flag)		SetDrawSyncTokenEnable((flag))
#define	GETDSENABLE()			GetDrawSyncTokenEnable()
#else
#define	SETDS(token,obj)		
#define	SETDSNO(token)			
#define	GETDS()					
#define	CLRDS()					
#define	SEARCHDS(token)			
#define	SETDSCNT(cnt)			
#define	GETDSCNT()				
#define	RESETDSCNT()			
#define	SETDSENABLE(flag)		
#define	GETDSENABLE()			
#endif

#define	MAINTASK_TH				(0xff00)
#define	RENDEROBJ_TH			(0xfe00)
#define	JOBJ_TH					(0xfd00)
#define	DOBJ_TH					(0xfc00)
#define	EFFECT_TH				(0xe000)

#endif	//__WK7SIGHT_H__

