////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Viewer Task
//
// WK7ViewerTask.h
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 11/28/2002
//
////////////////////////////////////////////////////////////////////////////////

#ifndef	__WK7VIEWERTASK_H__
#define	__WK7VIEWERTASK_H__

//#include "..\Lib\ghmGcAppTask.h"
//#include "..\Lib\ghmGcMem.h"
//#include "..\Lib\ghmGcFile.h"
//#include "..\Lib\ghmGcUsb.h"
//#include "..\Lib\ghmGcTex.h"

#include ".\WK7Archive.h"
#include ".\WK7AfterImage.h"
#include ".\WK7Cam.h"
#include ".\WK7Chara.h"
#include ".\WK7Play.h"
#include ".\WK7SceneBlur.h"
#include "..\Fujita\C_CObj.h"
#include "..\Fujita\C_LObj.h"
#include "..\Fujita\C_Shadow.h"
#include "..\Matsu\MK7Jimaku.h"

struct Pos2D{
	short						mX,mY;
};

class ghmGcLine{
protected:
public:
	short						mPosNum;
	short						mLayer;			// C
	
	inline virtual bool Render(void){ return true; }
	
	ghmGcLine(){}
	virtual ~ghmGcLine(){}
};

class ghmGcLineF2 : public ghmGcLine{
protected:
public:
	Pos2D						mPos[2];
	union{
		GXColor					mColor;
		u32						mColor32;
	};
	
	bool Render(void);
	
	ghmGcLineF2(){ mPosNum=2; }
	virtual ~ghmGcLineF2(){}
};

class WK7ViewerFileListItem{
public:
	char				*mpFileName;
	bool				mIsDir;
	union{
		struct{
			u32			mParent : 1;
			u32			mFlagPad:31;
		};
		u32				mFlag;
	};
	WK7ViewerFileListItem();
	~WK7ViewerFileListItem();
};

#define	WK7VIEWERFILELIST_CURSOR_STACKNUM		(16)
class WK7ViewerFileList{
protected:
	char					*mpDirName;
	char					mExt[256];
	int						mDirLevel;
	WK7ViewerFileListItem	*mpFiles;
	int						mFileNum;
	int						mCursorID;
	int						mCursorStack[WK7VIEWERFILELIST_CURSOR_STACKNUM];
	static bool CompareExt(const char *pfilename,const char *pext);
public:
	WK7ViewerFileList();
	~WK7ViewerFileList();
	
	inline int GetCursorID(void){ return mCursorID; }
	bool SetCursorID(int id);
	
	inline const char* GetFileName(int index){
		if( index < 0 || index >=mFileNum ){
			return NULL;
		}
		return mpFiles[index].mpFileName;
//		return mppFileName[index];
	}
	bool GetFullFileName(int index,char *pFileName);
	inline int IsDir(int index){
		if( index < 0 || index >=mFileNum ){
			return -1;
		}
		return mpFiles[index].mIsDir==true? 1:0;
	}
	inline int IsParent(int index){
		if( index < 0 || index >=mFileNum ){
			return -1;
		}
		return mpFiles[index].mParent? 1:0;
	}
	inline int IsFile(int index){
		if( index < 0 || index >=mFileNum ){
			return -1;
		}
		return (mpFiles[index].mIsDir==false && mpFiles[index].mParent==false)? 1:0;
	}
	inline int GetFileNum(void){ return mFileNum; }
	
	void SetFilter(const char *pExt);
	bool CreateFileList(ghmGcUsb &serv,const char *pdir=NULL);
	bool ReleaseFileList(void);
	
	const char* GetDirName(void){return mpDirName; }
	
	bool ChangeDir(ghmGcUsb &serv,int index);
	bool MoveParentDir(ghmGcUsb &serv);
	int GetDirLevel(void){ return mDirLevel; }
};

class WK7ViewerMotion : public ghmGcListObj{
public:
	WK7Archive					*mpArchive;
	WK7ModelMotion				*mpModelMotion;
	WK7CameraMotion				*mpCameraMotion;
	char						*mpFileName;
	bool						mInterpolate;
	f32							mInterpolateFrame;
	
	inline WK7ViewerMotion* GetPrev(void){ return (WK7ViewerMotion *)ghmGcListObj::GetPrev(); }
	inline WK7ViewerMotion* GetNext(void){ return (WK7ViewerMotion *)ghmGcListObj::GetNext(); }
	
	bool ReleaseMotion(void);
	const char* GetArchiveVersionStr(void);
	int GetArchiveVersion(void);
	
	WK7ViewerMotion(ghmGcListObj *pAnchor,int prio);
	~WK7ViewerMotion();
};

class WK7ViewerPlay : public ghmGcListObj{
protected:
	char						*mpFileName;
	char						*mpDirName;
public:
	inline WK7ViewerPlay* GetPrev(void){ return (WK7ViewerPlay *)ghmGcListObj::GetPrev(); }
	inline WK7ViewerPlay* GetNext(void){ return (WK7ViewerPlay *)ghmGcListObj::GetNext(); }
	
	WK7ViewerPlay(ghmGcListObj *pAnchor,int prio);
	~WK7ViewerPlay();
	
	bool SetDirName(const char *pdirname);
	bool SetFileName(const char *pfname);
	inline const char* GetDirName(void){ return mpDirName; }
	inline const char* GetFileName(void){ return mpFileName; }
	
	bool GetFullFileName(char *pfname);
};

class WK7ViewerModel : public ghmGcListObj{
protected:
	static bool					mMotionPlayAll;
	static bool					mMotionPauseAll;
	static bool					mMotionLoopAll;
	static u32					mModelIDSeed;
	static u32					mMotionIDSeed;
	
	// f
	WK7Archive					*mpModelArchive;
	WK7RenderObj				*mpCharaObj;
	WK7Model					*mpModel;
	char						*mpModelFileName;
	WK7ModelMotion				*mpModelMotion;
	WK7CameraMotion				*mpCameraMotion;
	
	// [V
	ghmGcListObj				mMotionAnchor;
	int							mCurrentMotion;
	int							mMotionNum;
	bool						mMotionPlay;
	
	WK7AnimLinear				mDispCounter;
	
	WK7ViewerMotion* CreateMotion(void);
public:
	inline WK7ViewerModel* GetPrev(void){ return (WK7ViewerModel *)ghmGcListObj::GetPrev(); }
	inline WK7ViewerModel* GetNext(void){ return (WK7ViewerModel *)ghmGcListObj::GetNext(); }
	
	WK7ViewerModel(ghmGcListObj *pAnchor,int prio);
	~WK7ViewerModel();
	
	const char* GetArchiveVersionStr(void);
	int GetArchiveVersion(void);
	const char* GetCurrentCamMotionFileName(void);
	
	static bool SetMotionPlayAll(bool flag);
	inline static bool GetMotionPlayAll(void){ return mMotionPlayAll; }
	static bool SetMotionLoopAll(bool flag);
	inline static bool GetMotionLoopAll(void){ return mMotionLoopAll; }
	static bool SetMotionPauseAll(bool flag);
	inline static bool GetMotionPauseAll(void){ return mMotionPauseAll; }
	
	inline bool SetMotionPlay(bool flag){ mMotionPlay=flag; return true; }
	inline bool GetMotionPlay(void){ return mMotionPlay; }
	
	const char* GetFileName(void){ return mpModelFileName; }
	
	bool LoadModel(const char *pfilename);
	bool LoadMotion(int index,const char *pfilename);
	bool ReleaseModel(void);
	bool ReleaseMotion(int index,bool workRelease=false);
	bool ReleaseMotionAll(void);
	
	WK7CameraMotion* GetCurrentCamMotion(void);
	WK7ViewerMotion* GetMotion(int index);
	bool PlayMotion(void);
	bool StopMotion(void);
	bool PauseMotion(void);
	bool ReplayMotion(void);
	
	bool MotionControl(void);
	bool RenderInfo(bool checked=false);
	bool SetPosition(Vec &pos);
	bool SetRotY(f32 roty);
	bool RenderVersionCheck(bool anim);
	
	bool SetDispLight(WK7Light *pLight,bool flag);
	inline WK7Model* GetModel(void){ return mpModel; }
	inline int GetMotionNum(void){ return mMotionNum; }
};

struct stSentence;
#define	EDITCAM_DELAYRATE_MAX			(10000)
class WK7ViewerTask : public ghmGcAppTask{
public:
	enum WK7ViewerMode{
		WK7VIEWERMODE_OPERATE,
		WK7VIEWERMODE_EXIT,
		WK7VIEWERMODE_MENU,
		WK7VIEWERMODE_MODELSELECT,
		WK7VIEWERMODE_MODELLOAD,
		WK7VIEWERMODE_STAGELOAD,
		WK7VIEWERMODE_SCREENTONE,
		WK7VIEWERMODE_PERFMETER,
		WK7VIEWERMODE_SHADOW,
		WK7VIEWERMODE_TEXTURE,
		WK7VIEWERMODE_INFODISP,
		WK7VIEWERMODE_MOTIONRATE,
		WK7VIEWERMODE_MOTIONLOAD_MODELSELECT,
		WK7VIEWERMODE_MOTIONLOAD,
		WK7VIEWERMODE_MOTIONLOADSUB,
		WK7VIEWERMODE_PLAYLOAD,
		WK7VIEWERMODE_PLAY,
		WK7VIEWERMODE_SEQPLAY,
		WK7VIEWERMODE_ORIGDISP,
		WK7VIEWERMODE_CAMTGTDISP,
		WK7VIEWERMODE_CAMMOTIONLOAD,
		WK7VIEWERMODE_CAMMOTIONLOOP,
		WK7VIEWERMODE_MOTIONLOOP,
		WK7VIEWERMODE_DEFLIGHT,
		WK7VIEWERMODE_RENDERMEMORY,
		WK7VIEWERMODE_CAMWOBBLE,
		WK7VIEWERMODE_CAMWOBBLESCALE,
		WK7VIEWERMODE_MOVIELOAD,
		WK7VIEWERMODE_MOVIEPLAY,
		WK7VIEWERMODE_WALKAREA,
		WK7VIEWERMODE_HITOBJCHECK,
		WK7VIEWERMODE_CINEMASCOPE,
		WK7VIEWERMODE_STREAM,
		WK7VIEWERMODE_ZANLOAD,
		WK7VIEWERMODE_ZANPLAY,
		WK7VIEWERMODE_SYSMESLOAD,
		WK7VIEWERMODE_SYSMESPLAY,
		WK7VIEWERMODE_CHANGEDISK,
		WK7VIEWERMODE_GXWARNING,
		WK7VIEWERMODE_MODEINFO,
		WK7VIEWERMODE_PLAYENDSTOP,
		WK7VIEWERMODE_ZANDEBUGINFO,
	};
	enum EditMode{
		VIEWER = 0,
		MOTIONPLAY,
		CAMEDIT,
		HIERARCHY,
		BLUREDIT,
		LIGHTEDIT,
		EDITMODE_MAX,
	};
	enum EditCamType{
		LINIP_LINPATH = 0,
		DELAYIP_LINPATH,
		LINIP_CIRPATH,
		EDITCAMTYPE_MAX,
	};
protected:
	// J
	Vec							mCamDefaultPosition;
	Vec							mCamDefaultTargetPosition;
	
	Vec							mCamStorePosition;
	Vec							mCamStoreTargetPosition;
	
	Vec							mEditCamStartPosition;
	Vec							mEditCamEndPosition;
	Vec							mEditCamStartTargetPosition;
	Vec							mEditCamEndTargetPosition;
	f32							mEditCamStartFov;
	f32							mEditCamEndFov;
	int							mEditCamDelayRate;
	int							mEditCamInterpolateTime;
	EditCamType					mEditCamType;
	int							mEditCamProc;
	int							mEditCamProcMax;
	
	int							mCurrentModel;
	
	WK7JObj						*mpCurrentJObj;
	
	bool						mCamWobble;
	int							mCamWobbleScale;
	bool						mCinemaScope;
	
	ghmGcListObj				mModelAnchor;
	int							mModelNum;
	int							mSelectModel;
	
	int							mSelectMotion;
	
	// Xe[Wf
	WK7Archive					*mpStageArchive;
	WK7Stage					*mpStageObj;
//	WK7PlayRenderObj			*mpStageObj;
	WK7Model					*mpStage;
	char						*mpStageFileName;
	
	// J[V
	WK7Archive					*mpCameraMotionArchive;
	WK7CameraMotion				*mpCameraMotion;
	char						*mpCameraMotionFileName;
//	bool						mCameraMotionLoop;
	bool						mCameraMotionPlay;
	
	bool						mInfoDisp;
	bool						mOrigDisp;
	bool						mCamTargetDisp;
	
	bool						mPlayInfoDisp;
	int							mPlayInfoDispCounter;
	
	int							mFrameRate;
	
	// Cg
	C_LObj						*mpLObj;
	WK7Light					*mpDefLight;
	bool						mDefLight;
	
	// eNX`mFp
	GhmGcTexHeader				*mpTexture;
	GXTexObj					mTexObj;
	bool						mTextureVisible;
	
	// _\p
	void						*mpOriginArchive;			// 
	C_JObj						*mpOriginJObj;				// ^[QbgpIuWFNg
	
	WK7ViewerMode				mMode;
	WK7ViewerMode				mReqMode;
	int							mProc;
	int							mProcSub;
	int							mCursor;
	int							mPrevCursor;
	
	// Lp
	WK7Play						*mpPlay;
	void						*mpPlayResource;
	WK7Play						*mpPrevPlay;
	void						*mpPrevPlayResource;
	ghmGcListObj				mPlayAnchor;
	int							mPlayNum;
	int							mCurrentPlay;
	int							mPlayTaskID;
	u32							mPlayProc;
	u32							mPlayCount;
	f32							mPlayStartTime;
	f32							mPlayWaitTime;
	bool						mPlayEndStop;
	int							mCaptureFrame;
	int							mCaptureFrameIndex;
	
	int							mSysMesProc;
	int							mSysMesHandle;
	int							mSysMesSelect;
	stSentence					*mpSysMesResource;
	EFFECT						*mpSysMesEffect;
	u32							mSysMesEffectID;
	int							mSysMesCount;
	int							mSysMesFlag;
	
	WK7ViewerFileList			mFileList;
	
	bool						mScreenTone;
	bool						mPerfMeter;
	bool						mShadowTrigger;
	bool						mRenderMemory;
	
	EditMode					mEditMode;
	
	// i{P
//	WK7SceneBlur				mSceneBlur;
	int							mSceneBlurID;
	int							mSceneBlurProc;
	bool						mSceneBlurVisible;
	
	int							mLightEditProc;
	int							mLightID;
	
	bool						mHitObjCheck;
	bool						mStreamFlag;
	
	// ^XNRg[
	ghmGcTaskController			mFrameProcessTaskController;
	ghmGcTaskController			mRenderProcessTaskController;
	
	int							mPreScreenCaptureTaskID;
	int							mScreenCaptureTaskID;
	int							mPostScreenCaptureTaskID;
	int							mPostScreenCapture2TaskID;
	int							mLastFrameProcessTaskID;
	
	int							mMovieID;
	
	u32							mZanID;
	u32							mZanMesID;
	u32							mZanProc;
	u32							mZanCount;
	f32							mZanStartTime;
	f32							mZanWaitTime;
	
	int							mReqDiskID;
	
	WK7Light* GetLight(int index);
	WK7CameraMotion* GetCurrentCamMotion(void);
	WK7ViewerModel* GetCurrentCamMotionModel(void);
	const char* GetCurrentCamMotionArchiveVersionStr(void);
	int GetCurrentCamMotionArchiveVersion(void);
	
	bool SetCinemaScope(bool flag);
	bool SetCamWobble(bool flag);
	bool SetWalkArea(bool flag);
	bool SetHitObjCheck(bool flag);
	bool SetStream(bool flag);
	bool SetGXWarning(bool flag);
	bool SetModeInfo(bool flag);
	bool SetZanDebugInfo(bool flag);
	bool SetMenuItem(WK7ViewerTask::WK7ViewerMode mode,bool flag);
	bool SetOrigDisp(bool flag);
	bool SetInfoDisp(bool flag);
	bool SetCamTargetDisp(bool flag);
	bool SetRenderMemory(bool flag);
	bool SetPerfMeter(bool flag);
	
	bool SetMode(WK7ViewerMode mode);
	bool SetCamMode(WK7ViewerMode mode);
	
	bool CreateLObj(HSD_SceneLightSet *pLightSet=NULL);
	bool ReleaseLObj(void);
	bool SetShadow(void);
	
	bool LoadModel(const char *pfilename);
	bool LoadStage(const char *pfilename);
	bool LoadMotion(const char *pfilename);
	bool LoadCamMotion(const char *pfilename);
	
//	bool ReleaseModelArchive(void);
	bool ReleaseModel(int id,bool workRelease=true);
	bool ReleaseModelAll(void);
	bool ReleaseStageArchive(void);
	bool ReleaseMotionArchive(int mdlid,int motid,bool workRelease=true);
//	bool ReleaseMotionArchiveAll(void);
	bool ReleaseCamMotionArchive(void);
	
	bool MotionTrigger(void);
	bool MotionPauseTrigger(void);
	bool PlayMotion(void);
	bool StopMotion(void);
	bool PauseMotion(void);
	bool ReplayMotion(void);
	bool PlayCamMotion(void);
	bool StopCamMotion(void);
	bool PauseCamMotion(void);
	bool ReplayCamMotion(void);
	
	bool SelectFile(void);
	bool FrameProcessOperate(void);
	bool FrameProcessMenu(void);
	bool FrameProcessModelSelect(bool newmodel=true);
	bool FrameProcessModelLoad(void);
	bool FrameProcessStageLoad(void);
	bool FrameProcessMotionLoadModelSelect(void);
	bool FrameProcessMotionLoad(void);
	bool FrameProcessMotionLoadSub(void);
	bool FrameProcessTextureLoad(void);
	bool FrameProcessPlayLoad(void);
	bool FrameProcessPlay(void);
	bool FrameProcessSeqPlay(void);
	bool FrameProcessCamMotionLoad(void);
	bool FrameProcessMovieLoad(void);
	bool FrameProcessMoviePlay(void);
	bool FrameProcessZanLoad(void);
	bool FrameProcessZanPlay(void);
	bool FrameProcessSysMesLoad(void);
	bool FrameProcessSysMesPlay(void);
	bool FrameProcessChangeDisk(void);
	bool MotionControl(void);
	
	bool RenderFileList(const char *pTitle,int dispmax=-1);
	bool RenderOperate(void);
	bool RenderMenu(void);
	bool RenderModelSelect(bool newmodel=true);
	bool RenderModelLoad(void);
	bool RenderStageLoad(void);
	bool RenderMotionLoadModelSelect(void);
	bool RenderMotionLoad(void);
	bool RenderMotionLoadSub(void);
	bool RenderTextureLoad(void);
	bool RenderPlayLoad(void);
	bool RenderPlay(void);
	bool RenderSeqPlay(void);
	bool RenderCamMotionLoad(void);
	bool RenderMovieLoad(void);
	bool RenderMoviePlay(void);
	bool RenderZanLoad(void);
	bool RenderZanPlay(void);
	bool RenderSysMesLoad(void);
	bool RenderSysMesPlay(void);
	bool RenderChangeDisk(void);
	
	bool ReleasePlay(void);
	bool ReleaseSysMes(void);
	
	bool RenderInfo(void);
	bool RenderTexture(void);
	bool SetCamera2D(float nearz = 0.0f,float farz = 1024.0f);
	
	WK7ViewerMotion* GetMotion(int index);
//	WK7ViewerMotion* CreateMotion(void);
	
	WK7ViewerModel* GetModel(int index);
	WK7ViewerModel* CreateModel(void);
	
	WK7ViewerPlay* GetPlay(int index);
	WK7ViewerPlay* CreatePlay(void);
	bool RemovePlay(int index);
	void ClearPlay(void);
	
	bool ChangeEditMode(void);
	bool FrameProcessEditModeViewer(void);
	bool FrameProcessEditModeMotionPlay(void);
	bool FrameProcessEditModeCamEdit(void);
	bool FrameProcessEditModeHierarchy(void);
	bool FrameProcessEditModeBlurEdit(void);
	bool FrameProcessEditModeLightEdit(void);
	bool FrameProcessEditMode(void);
	bool RenderProcessEditModeLightEdit(void);
	bool RenderProcessEditMode(void);
	bool RenderHierarchy(WK7JObj *pJObj,ghmGcSpriteF *psprite);
	bool CalcTopHierarchy(WK7JObj *pJObj,ghmGcSpriteF *psprite);
	
	bool FrameProcessFileList(int itemnum=-1,int dispnum=-1);
	bool RenewCursorRange(int itemnum,int dispnum);
	
	bool SetCurrentModel(int index);
	int GetCurrentModel(void);
public:
	WK7ViewerTask(ghmGcApp *pApp,int prio=GHMGCLIST_DEFAULT_PRIO);
	virtual ~WK7ViewerTask();
	
	// R[obN
	virtual bool InitializeCallBack(void);
	
	// t[
	virtual bool FrameProcess(void);
	
	// IR[obN
	virtual bool TerminateCallBack(void);
	
	// _O
	virtual bool PreRenderProcess(void);
	virtual bool RenderProcess(void);
	
	void _SeqPlayTask(ghmGcTask *pTask);
	static void SeqPlayTask(ghmGcTask *pTask);
	
	// t[Xg
	bool LastFrameProcessTask(ghmGcTask *pTask);
	static void _LastFrameProcessTask(ghmGcTask *pTask);
	
	// XN[Lv`[OiɂRcnj
	bool PreScreenCaptureTask(ghmGcTask *pTask);
	static void _PreScreenCaptureTask(ghmGcTask *pTask);
	// XN[Lv`[
	bool ScreenCaptureTask(ghmGcTask *pTask);
	static void _ScreenCaptureTask(ghmGcTask *pTask);
	// XN[Lv`[㏈iɂQcnj
	bool PostScreenCaptureTask(ghmGcTask *pTask);
	static void _PostScreenCaptureTask(ghmGcTask *pTask);
	bool PostScreenCapture2Task(ghmGcTask *pTask);
	static void _PostScreenCapture2Task(ghmGcTask *pTask);
};

#endif	//__WK7VIEWERTASK_H__

