#include "common.h"
#include "utility.h"
#include "c_edge.h"

void C_Edge::Disp( u32 col, u32 flg )
{
#pragma unused( flg )
  GXBegin(GX_LINES, GX_VTXFMT0, 2);
  GXPosition3f32(p[0].x, p[0].y, p[0].z);
  GXColor1u32(col);
  GXPosition3f32(p[1].x, p[1].y, p[1].z);
  GXColor1u32(col);
  GXEnd();
}

C_EdgeArray::C_EdgeArray( C_TrigonArray& t )
  : trigon(t)
{
  Init();
}

C_EdgeArray::C_EdgeArray( HSD_PObj *pobj, HSD_JObj *j )
  : trigon_body( pobj,j ), trigon(trigon_body)
{
  Init();
}

void C_EdgeArray::Print( void )
{
  for( u32 i=0; i<edge_num; i++ ){
	s16* t = edge[i].TriIdx();
	s16* v = edge[i].VtxIdx();
	OSReport( "--no %3d--\n", i );
	OSReport( "  %d %d \n", t[0], t[1] );
	OSReport( "  %d %d \n", v[0], v[1] );
  }
}
void C_EdgeArray::Init()
{
  int trigon_num = (int)trigon.GetNum();

  //SOp`B
  for( u32 i=0; i<trigon_num; i++ ){
	C_Trigon& tri = trigon.GetTrigon(i);
	s16* idx = tri.GetIdx();

	//SŐB
	for( u32 j=0; j<3; j++ ){
	  s16 i0,i1;
	  i0 = idx[j]; i1 = idx[(j+1)%3];

	  //łɓo^Ă邩`FbNB
	  u32 k;
	  for( k=0; k<edge.size(); k++ ){
		if( ((edge[k].VtxIdx())[0] == i0 && (edge[k].VtxIdx())[1] == i1) ||
			((edge[k].VtxIdx())[0] == i1 && (edge[k].VtxIdx())[1] == i0) ){
		  edge[k].SetShareTrigon( 1, (s16)i );
		  break;
		}
	  }
	  //VKo^
	  if( k == edge.size() ){
		C_Edge e( tri.GetVtx()[j], tri.GetVtx()[(j+1)%3], i0, i1 );
		e.SetShareTrigon( 0, (s16)i );
		edge.push_back( e );
	  }
	}
  }

  edge_num = edge.size();
  //Print();
}

C_EdgeArray::~C_EdgeArray()
{
}


void C_EdgeArray::DispStart( void )
{
  HSD_CObj *cobj = HSD_CObjGetCurrent();

  GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD_NULL, GX_TEXMAP_NULL, GX_COLOR0A0);
  GXSetTevOp(GX_TEVSTAGE0, GX_PASSCLR);
  GXSetNumTevStages(1);
  GXSetNumTexGens(0);
  GXSetBlendMode(GX_BM_NONE, GX_BL_SRCALPHA, GX_BL_INVSRCALPHA, GX_LO_CLEAR);
  GXSetChanCtrl(GX_COLOR0A0,
				GX_ENABLE,		// enable channel
				GX_SRC_VTX,		// amb source
				GX_SRC_VTX,		// mat source
				GX_LIGHT_NULL,	// light mask
				GX_DF_CLAMP,	// diffuse function
				GX_AF_NONE);
  GXSetNumChans(1);

  GXLoadPosMtxImm(HSD_CObjGetViewingMtxPtr(cobj), GX_PNMTX0);
  GXSetCurrentMtx( GX_PNMTX0 );

  GXClearVtxDesc();
  GXSetVtxDesc(GX_VA_POS,  GX_DIRECT);
  GXSetVtxDesc(GX_VA_CLR0,  GX_DIRECT);
  GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS,  GX_POS_XYZ, GX_F32, 0);
  GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGBA8, 0);
  GXSetLineWidth( (u8)16,GX_TO_ZERO );
}

void C_EdgeArray::DispEnd( void )
{
  HSD_StateInvalidate( HSD_STATE_ALL );
}

void C_EdgeArray::Disp( u32 idx, u32 flg, u32 col )
{
  //|SƂΑǁA
  //ǂfobOpȂ̂ŃdB
  Vec vtx[2];
  MTXMultVec(trigon.GetMtx(), &edge[idx].GetVtx(0), &vtx[0]);
  MTXMultVec(trigon.GetMtx(), &edge[idx].GetVtx(1), &vtx[1]);
  C_Edge e( C_Vec(vtx[0]), C_Vec(vtx[1]) );
  e.Disp( col, flg );
}

void C_EdgeArray::Proc( void )
{
  if( trigon_body.GetNum() ){
	//gSOŎĂꍇA
	trigon.Proc();
  }else{
	//Փ˔Ƃ̃gSLŁA
	//lĂꍇAlƐMB
	//trigon.Proc();
  }
}


void C_EdgeArray::RecalcFlg( C_Vec& pos  )
{
  int trigon_num = (int)trigon.GetNum();
  u8 trigon_flg[ TRIGON_MAX ];
  ASSERT2( (trigon_num < TRIGON_MAX), "Too many Shadow Polygon." ) ;

  //̎Op`̕\B
  for( u32 i=0; i<trigon_num; i++ ){
	C_Trigon tri = trigon.GetTrigon( i );
	C_Vec vtx[3];
	for( int j=0; j<3; j++ )
	  MTXMultVec(trigon.GetMtx(), &(tri.GetVtx()[j]), &vtx[j]);
	C_Trigon tnew( vtx, &trigon );
	C_Vec n = tnew.GetNrm();
	C_Vec c = tnew.GetCtr();
	C_Vec dir = c - pos;
	if( (dir | n) < 0 ) 
	  trigon_flg[i] = 0;
	else
	  trigon_flg[i] = 1;
  }

  for( u32 i=0; i<edge_num; i++ ){
	s16 *idx = edge[i].TriIdx();
	s32 flg = 0;
	if( idx[1] >= 0 ){
	  //Op`ɋLꂽŐB
	  if( trigon_flg[idx[0]] && !trigon_flg[idx[1]] ) flg = 1;
	  if( !trigon_flg[idx[0]] && trigon_flg[idx[1]] ) flg = 2;
	}else{
	  //[B
	  if( trigon_flg[idx[0]] ) flg = 1;
	}
	edge[i].SetFlg(flg);
  }
}


