#include "common.h"
#include "c_mirror.h"
#include "utility.h"

C_Mirror::C_Mirror( HSD_JObj *j, HSD_CObj *c, HSD_DObj *d, HSD_JObj *size_j )
{
  ASSERT2( j, "No JObj\n" );
  ASSERT2( c, "No CObj\n" );
  ASSERT2( d, "No DObj\n" );
  jobj = j;
  cobj = c;
	
  //IWĩC[WfXN
  HSD_MObj *mobj = HSD_DObjGetMObj(d);
  ASSERT2( mobj, "There is no material" );
  tobj = HSD_MObjGetTObj(mobj);
  ASSERT2( tobj, "There is no texture" );
  img_org = tobj->imagedesc;


  //J
  mirror_cobj = HSD_CObjAlloc();
  f32 W=256.f;
  f32 H=256.f;
  static HSD_RectF32 viewport = { 0, W, 0, H };
  static HSD_RectU16 scissor  = { 0, (u16)W, 0, (u16)H };
	
  HSD_CObjSetViewportf     (mirror_cobj, &viewport);
  HSD_CObjSetScissor       (mirror_cobj, &scissor);
  HSD_CObjSetProjectionType(mirror_cobj, COBJ_PROJECTION_FRUSTUM);
	
  //eNX`
  img = HSD_ImageDescAlloc();
  img->width  = (u16)W;
  img->height = (u16)H;
  //img->format = GX_TF_RGB565;
  img->format = GX_TF_RGB5A3;
	
  size = GXGetTexBufferSize(img->width, img->height, img->format, GX_FALSE, 0);
  img->image_ptr = new u8[OSRoundUp32B(size)];
  tobj->imagedesc = img;
#if 1
  tobj->rotate.x = 0.f;
  tobj->rotate.y = 0.f;
  tobj->rotate.z = 0.f;
  tobj->scale.x = -1.f;
  tobj->scale.y = 1.f;
  tobj->scale.z = 1.f;
  tobj->translate.x = 1.f;
  tobj->translate.y = 0.f;
  tobj->translate.z = 0.f;
  tobj->flag |= TEX_MTX_DIRTY;
#endif
  //ftHg̑傫B
  if( size_j ){
	//Vec s;
	//HSD_JObjSetupMatrix( size_j );
	//HSD_MtxGetScale( size_j->mtx, &s );
	MirrorSize( size_j->scale.x, size_j->scale.y );
	//HSD_DmpVector( "scl", &size_j->scale );
	//HSD_DmpVector( "trs", &size_j->translate );
  }else	MirrorSize( 1.f, 1.f );
  
}

C_Mirror::~C_Mirror()
{
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //A
  //~[̊J̓f̑OɁB
  tobj->imagedesc = img_org;
  delete img->image_ptr;
  HSD_CObjUnref( mirror_cobj );
  HSD_ImageDescRemove(img);
}

void C_Mirror::Proc( void )
{
}

//Tv̂́B
// iQƁj/HAL/sysdolphin/src/sample/baselib/mirror/
bool C_Mirror::DispStart( void )
{
  Vec cpos, jpos, mpos;
  Vec xaxis, yaxis, zaxis;
  Mtx mtx, imtx;
  float dot;

  // ݂̃JƋ̈ʒuAʗpeNX`
  // `悷邽߂̃Jʒu肷B

  //xɂĂPWOx]B
  HSD_JObjSetupMatrix( jobj );
  MTXCopy(HSD_JObjGetMtxPtr(jobj), mtx);
#if 0
  static bool flg = false;
  if( !flg ) {
	Vec x,y,z,pos;
	flg = true;
	HSD_MtxColVec(mtx, 0, &x);
	HSD_MtxColVec(mtx, 1, &y);
	HSD_MtxColVec(mtx, 2, &z);
	HSD_MtxGetTranslate(mtx, &pos);
	HSD_DmpVector( "pos", &pos );
	HSD_DmpVector( "x", &x );
	HSD_DmpVector( "y", &y );
	HSD_DmpVector( "z", &z );
  }
#endif
  HSD_MtxColVec(mtx, 0, &xaxis);
  HSD_MtxColVec(mtx, 1, &yaxis);
  VECScale(&xaxis, &xaxis, -1.0F);
  VECCrossProduct(&xaxis, &yaxis, &zaxis);
  VECNormalize(&zaxis, &zaxis);
#if 1
  HSD_MtxSetColVec(mtx, 0, &xaxis);
  HSD_MtxSetColVec(mtx, 2, &zaxis);
#else
  VECNormalize(&zaxis, &zaxis);
  VECNormalize(&xaxis, &xaxis);
  VECNormalize(&yaxis, &yaxis);
  HSD_MtxSetColVec(mtx, 0, &xaxis);
  HSD_MtxSetColVec(mtx, 1, &yaxis);
  HSD_MtxSetColVec(mtx, 2, &zaxis);
#endif
	
  HSD_CObjGetEyePosition( cobj, &cpos );
  HSD_MtxGetTranslate(mtx, &jpos);
  VECSubtract(&jpos, &cpos, &mpos);
  dot = VECDotProduct(&zaxis, &mpos);
  if (dot <= 0.0F) {
	// J~[̌둤
	return false;
  }
  
  //ʂɃJ̈ʒu𔽓]B
  VECScale(&zaxis, &mpos, dot * 2);
  VECAdd(&mpos, &cpos, &mpos);
  HSD_MtxSetColVec(mtx, 3, &mpos);


  MTXInverse(mtx, imtx);	
	
  // ~[J̍Wnł̃~[̈ʒu
  MTXMultVec(imtx, &jpos, &jpos);
  HSD_CObjSetNear(mirror_cobj, -jpos.z + 0.01f);
  HSD_CObjSetFar (mirror_cobj, -jpos.z + 0.01f + 32768.0f);
  HSD_CObjSetFrustum(mirror_cobj,
					 mirror_h + jpos.y,
					 -mirror_h + jpos.y,
					 -mirror_w + jpos.x,
					 mirror_w + jpos.x);
  HSD_CObjSetEyePosition(mirror_cobj, &mpos);
  HSD_CObjSetUserDefinedMtx(mirror_cobj, imtx);


#if 0
  if( 1 ) {
	Vec x,y,z,pos;
	HSD_MtxColVec(imtx, 0, &x);
	HSD_MtxColVec(imtx, 1, &y);
	HSD_MtxColVec(imtx, 2, &z);
	HSD_MtxGetTranslate(imtx, &pos);
	HSD_DmpVector( "pos", &pos );
	HSD_DmpVector( "x", &x );
	HSD_DmpVector( "y", &y );
	HSD_DmpVector( "z", &z );
  }
#endif

  //GXColor		clearclr = { 59,59,59,255*(100-70)/100 };
  static GXColor		clearclr = { 0,0,0,0 };
  
  HSD_VISetClearClr(clearclr);
  HSD_ImageDescCopyFromEFB(img, 0, 0, GX_TRUE, 1);

  HSD_StartRender(HSD_RP_OFFSCREEN);
  if( HSD_CObjSetCurrent(mirror_cobj) ) return true;
  else return false;
}

void C_Mirror::DispEnd( void )
{
  HSD_CObjEndCurrent();
  HSD_EndRender();
  HSD_ImageDescCopyFromEFB(img, 0, 0, GX_TRUE, 1);
  GXInvalidateTexAll();
  DCStoreRangeNoSync( img->image_ptr, size );
  GXColor		clearclr = { 0,0,0,0x00 };
  HSD_VISetClearClr(clearclr);
}

void  C_Mirror::DrawTex( void )
{

  HSD_TObj *t = tobj;
  if( !t ) return;

  //eNX`[h
  GXTexObj r;
  GXInitTexObj(&r, t->imagedesc->image_ptr,
				   t->imagedesc->width, t->imagedesc->height, 
				   t->imagedesc->format, 
				   t->wrap_s, t->wrap_t,
				   t->imagedesc->mipmap? GX_TRUE : GX_FALSE);
  GXInitTexObjLOD(&r, GX_NEAR, GX_NEAR, 0, 0, 0, 0, 0, GX_ANISO_1);
  GXLoadTexObj(&r, GX_TEXMAP0);

  f32 w = t->imagedesc->width;
  f32 h = t->imagedesc->height;

  //JݒB
  GXRenderModeObj *rmode = HSD_VIGetConfigure();
  HSD_CObj *cobj = HSD_CObjAlloc();
  HSD_CObjSetProjectionType(cobj, COBJ_PROJECTION_ORTHO);
  HSD_CObjSetViewportfx4(cobj, 0, (u16)rmode->viWidth, 0, (u16)rmode->viHeight);
  HSD_CObjSetScissorx4  (cobj, 0, (u16)rmode->viWidth, 0, (u16)rmode->viHeight);
  HSD_CObjSetNear (cobj,   0.0F);
  HSD_CObjSetFar  (cobj, 200.0F);
  HSD_CObjSetOrtho(cobj, 0.f, 480.f, 0.f, 640.f);
  Vec	eyepos   = {0.0f, 0.0f, 100.0f};
  Vec	interest = {0.0f, 0.0f,   0.0f};
  Vec	up       = {0.0f, 1.0f,   0.0f};
  Mtx	mtxView;
  MTXLookAt(mtxView, &eyepos, &up, &interest);
  HSD_CObjSetUserDefinedMtx(cobj, mtxView);
	
  //Tev
  GXColor col={255,0,0,0};
  GXSetTevColor(GX_TEVREG1,col);

  GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
  GXSetTevOp(GX_TEVSTAGE0, GX_MODULATE);
  GXSetNumTevStages(1);
  //TexCoord
  Mtx	mat;
  MTXIdentity(mat);
  GXLoadTexMtxImm(mat, GX_TEXMTX0, GX_MTX2x4);
  GXSetTexCoordGen( GX_TEXCOORD0, GX_TG_MTX2x4, GX_TG_TEX0, GX_TEXMTX0 );
  GXSetNumTexGens(1);

  //PE
  GXSetBlendMode(GX_BM_NONE, GX_BL_SRCALPHA, GX_BL_INVSRCALPHA, GX_LO_CLEAR);
	
  //`lB
  GXSetChanCtrl(GX_COLOR0A0,
				GX_DISABLE,		// enable channel
				GX_SRC_VTX,		// amb source
				GX_SRC_VTX,		// mat source
				GX_LIGHT_NULL,	// light mask
				GX_DF_CLAMP,	// diffuse function
				GX_AF_NONE);
  GXSetNumChans(1);
	
  if (HSD_CObjSetCurrent(cobj)) {
	MTXTrans( mat, 320.f-w*0.5f, 240.f-h*0.5f, 0 );
    GXLoadPosMtxImm(mat, GX_PNMTX0);
	GXClearVtxDesc();
	GXSetVtxDesc(GX_VA_POS,  GX_DIRECT);
	GXSetVtxDesc(GX_VA_CLR0, GX_DIRECT);
	GXSetVtxDesc(GX_VA_TEX0, GX_DIRECT);
	GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS,  GX_POS_XYZ, GX_F32, 0);
	GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGBA8, 0);
	GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_F32, 0);

	#define	ZMID -0.5f
	u32 col=(u32)(0x88<<24 | 0x88<<16 | 0x88<<8 |0x88);
	GXBegin(GX_QUADS, GX_VTXFMT0, 4);
		GXPosition3f32(0, 0, ZMID);
		GXColor1u32(col);
		GXTexCoord2f32(0,0);
		
		GXPosition3f32(w, 0, ZMID);
		GXColor1u32(col);
		GXTexCoord2f32(1,0);
		
		GXPosition3f32(w, h, ZMID);
		GXColor1u32(col);
		GXTexCoord2f32(1,1);
		
		GXPosition3f32(0, h, ZMID);
		GXColor1u32(col);
		GXTexCoord2f32(0,1);
	GXEnd();
	HSD_CObjEndCurrent();
  }
  HSD_CObjUnref( cobj );
  HSD_StateInvalidate(HSD_STATE_ALL);
}










