#include "common.h"
#include "c_scene.h"
#include "utility.h"

C_Scene::C_Scene(const C_SceneDesc* sd)
   : jobj(0), lobj(0), cobj(0), fog(0)
{
  u32		i, num;

  jobj.reserve(num = sd->GetModelSetNum());
  for(i = 0; i < num; i++){
	jobj.push_back(new C_JObj(sd->GetModelSet(i)));
	jobj[i] = new C_JObj(sd->GetModelSet(i));
	if(jobj[jobj.size()-1]->GetActionNum() > 0){
	jobj[jobj.size()-1]->SetAction((u32)0);
	}
  }
  lobj.reserve(num = sd->GetLightSetNum());
  for(i = 0; i < num; i++){
	lobj.push_back(new C_LObj(sd->GetLightSet(i)));
	lobj[i] = new C_LObj(sd->GetLightSet(i));
	if(lobj[lobj.size()-1]->GetActionNum() > 0){
	  lobj[lobj.size()-1]->SetAction((u32)0);
	}
  }
  if(sd->GetCameraSet()){
	cobj = new C_CObj(sd->GetCameraSet());
	if(cobj->GetActionNum() > 0)	cobj->SetAction((u32)0);
  }
  if(sd->GetFogSet()){
	fog = new C_Fog(sd->GetFogSet());
	if(fog->GetActionNum() > 0)	fog->SetAction((u32)0);
  }
}

C_Scene::~C_Scene()
{
  u32		i;
  if(cobj)	delete cobj;
  for(i = 0; i < jobj.size(); i++)	delete jobj[i];
  for(i = 0; i < lobj.size(); i++)	delete lobj[i];
  if(fog)	delete fog;
}

void C_Scene::Proc()
{
  int		i;
  for(i = 0; i < jobj.size(); i++){
	jobj[(u32)i]->Proc();
  }
  for(i = 0; i < lobj.size(); i++){
	lobj[(u32)i]->Proc();
  }
  if(cobj){
	cobj->Proc();
  }
}


void C_Scene::DispLight( void )
{
  HSD_WObjDesc infinite_light_pos_wobjdesc = {
	NULL,                                    /* class name */
	{ 3.0000f, 5.0000f, 3.0000f },	   	   /* wobj position */
	//{ 0.0000f, 0.0000f, -1.0000f },	   	   /* wobj position */
	NULL,	                                   /* robj Ή */
  };
  float shininess = 16.000000f;
  HSD_LightDesc infinite_light = {
	NULL, 				   /* class name */
	NULL, 				   /* N */
	LOBJ_SPECULAR|LOBJ_DIFFUSE|LOBJ_INFINITE,/* flags */
	0x00,					   /* attnflags */
	{ 255, 255, 255, 255 },                  /* ligh color */
	&infinite_light_pos_wobjdesc,		   /* position */
	NULL,					   /* interest */
	&shininess,                              /* shininess */
  };

  /*
   * ArGgCg
   */
  HSD_LightDesc amb_light = {
	NULL,                                    /* class name */
	&infinite_light,                         /* N */
	LOBJ_DIFFUSE,                            /* flags */
	0,                                       /* attnflags */
	{ 76, 76, 76, 0xff },                    /* color */
	NULL,                                    /* pos */
	NULL,                                    /* interest */
	NULL,
  };
  if( lobj.size() ){
	C_LObj::DispStart();
	for(int i = 0; i < lobj.size(); i++){
	  lobj[(u32)i]->Disp();
	}
	C_LObj::DispEnd();
  }else{
	static HSD_LObj *l = (HSD_LObj*)0;
	if( !l ){
	  l = HSD_LObjLoadDesc(&amb_light);
	}
	HSD_LObjSetCurrentAll(l);
	HSD_LObjSetupInit(cobj->GetCObj());
	//HSD_LObjUnref( l );
  }
}

void C_Scene::DispModel( void )
{
  for(int i = 0; i < jobj.size(); i++){
	jobj[(u32)i]->Disp();
  }
}
void C_Scene::Disp( void )
{
  if( StartCam() ){
	DispLight();
	DispModel();
	EndCam();
  }
}







