#include "common.h"
#include "utility.h"
#include "c_scenedesc.h"

u32 C_ModelSet::GetActionNum() const
{
  HSD_AnimJoint**	a = adjointlist;
  u32 i=0;

  for( i=0; a && a[i]; i++ ){}
  return i;
}

void C_ModelSet::GetAction(u32 no, HSD_AnimJoint*& aj
						   , HSD_MatAnimJoint*& mj, HSD_ShapeAnimJoint*& sj) const{
  aj = (adjointlist)?adjointlist[no]:0;
  mj = (matanimjointlist)?matanimjointlist[no]:0;
  sj = (shapeanimjointlist)?shapeanimjointlist[no]:0;
}
void C_ModelSet::GetAction(const char* name, HSD_AnimJoint*& aj
						   , HSD_MatAnimJoint*& mj, HSD_ShapeAnimJoint*& sj) const
{
  HSD_AnimJoint**	a = adjointlist;
  u32 i=0;
  aj = 0; mj = 0; sj = 0;
  for( i=0; a && a[i]; i++ ){
	if(!strcmp(GetDescName(a[i]), name)){
	  GetAction(i, aj, mj, sj);
	  break;
	}
  }
  return;
}

HSD_DObjDesc* C_ModelSet::GetJointDObjDesc(HSD_Joint* joint
										   , const char* name) const
{
  HSD_DObjDesc*	d;
  for(d = joint->u.dobjdesc; d; d = d->next){
	if(!strcmp(GetDescName(d), name))	return d;
  }
  if(joint->child){
	if( (d = GetJointDObjDesc(joint->child, name)) ) return d;
  }
  if(joint->next){
	if( (d = GetJointDObjDesc(joint->next, name)) ) return d;
  }
  return 0;
}


u32 C_CameraSet::GetActionNum() const
{
  if(camanimlist == 0)	return 0;
  u32		i;
  for(i = 0; camanimlist[i]; i++){
	if( camanimlist[i]->aobjdesc ||
		camanimlist[i]->eyepos_anim ||
		camanimlist[i]->interest_anim ){
	  //AjB
	}else{
	  //AjȂB
	  break;
	}
  }
  return i;
}
HSD_CameraAnim* C_CameraSet::GetAction(const char* name) const
{
  HSD_CameraAnim**	a = camanimlist;
  if(a){
	u32		i;
	for(i = 0; a[i]; i++){
	  if(!strcmp(GetDescName(a[i]), name))	break;
	}
	if(a[i])	return GetAction(i);
  }
  return 0;
}


u32 C_FogSet::GetActionNum() const
{
  if(animlist == 0)	return 0;
  u32		i;
  for(i = 0; animlist[i]; i++){
	if( animlist[i]->fog_aobjdesc || animlist[i]->adj_aobjdesc ){
	}else{
	  break;
	}
  }
  return i;
}

HSD_FogAnim* C_FogSet::GetAction(const char* name) const
{
  HSD_FogAnim**	a = animlist;
  if(a){
	u32		i;
	for(i = 0; a[i]; i++){
	  if(!strcmp(GetDescName(a[i]), name))	break;
	}
	if(a[i])	return GetAction(i);
  }
  return 0;
}


u32 C_LightSet::GetActionNum() const
{
  if(lightanimlist == 0)	return 0;
  u32		i;
  for(i = 0; lightanimlist[i]; i++){
	if( lightanimlist[i]->aobjdesc ||
		lightanimlist[i]->position_anim ||
		lightanimlist[i]->interest_anim ){
	  //AjB
	}else{
	  //AjȂB
	  break;
	}
  }
  return i;
}

HSD_LightAnim* C_LightSet::GetAction(const char* name) const
{
  HSD_LightAnim**	a = lightanimlist;
  if(a){
	u32		i;
	for(i = 0; a[i]; i++){
	  if(!strcmp(GetDescName(a[i]), name))	break;
	}
	if(a[i])	return GetAction(i);
  }
  return 0;
}



void* C_SceneDesc::operator new(u32, void* p)
{
  HSD_Archive	arc;
  u32	offset = 0;

  memset(&arc, 0, sizeof(HSD_Archive));
  memcpy(&arc, p, sizeof(HSD_ArchiveHeader));
  ASSERT2(arc.header.file_size == *(u32*)p, "Wrong archive.\n");
  offset += (sizeof(HSD_ArchiveHeader) + 31)&~31;
  if(arc.header.data_size > 0){
	arc.data = (u8*)p + offset;
	offset += arc.header.data_size;
  }
  if(arc.header.nb_reloc > 0){
	arc.reloc_info = (HSD_ArchiveRelocationInfo*)((u8*)p + offset);
	offset += sizeof(HSD_ArchiveRelocationInfo)*arc.header.nb_reloc;
  }
  if(arc.header.nb_public > 0){
	arc.public_info = (HSD_ArchivePublicInfo*)((u8*)p + offset);
	offset += sizeof(HSD_ArchivePublicInfo)*arc.header.nb_public;
  }
  if(arc.header.nb_extern > 0){
	arc.extern_info = (HSD_ArchiveExternInfo*)((u8*)p + offset);
	offset += sizeof(HSD_ArchiveExternInfo)*arc.header.nb_extern;
  }
  if(offset < arc.header.file_size){
	arc.symbols = (char*)p + offset;
  }
  arc.top_ptr = p;
  for(int i = 0; i < arc.header.nb_reloc; i++){
	u32*	p = (u32*)(arc.data + arc.reloc_info[i].offset);
	if(!(*p & 0x80000000))	*p += (u32)arc.data;
  }
  return (void*)HSD_ArchiveGetPublicAddress(&arc, "scene_data");
}

u32 C_SceneDesc::GetModelSetNum( void ) const
{
  if(models == 0)	return 0;
  u32		i;
  for(i = 0; models[i]; i++){}
  return i;
}

u32 C_SceneDesc::GetLightSetNum() const
{
  if(lights == 0)	return 0;
  u32		i;
  for(i = 0; lights[i]; i++){}
  return i;
}

C_LightSet* C_SceneDesc::GetLightSet(const char* name) const
{
  HSD_SceneLightSet** ls = lights;
  for(; *ls && strcmp(GetDescName((*ls)->lightdesc), name); ls++){}
  ASSERT2(*ls,"No light set\n");
  return new (*ls) C_LightSet;
}

C_ModelSet* C_SceneDesc::GetModelSet(const char* name) const
{
  HSD_SceneModelSet** ms = models;
  for(; *ms && strcmp(GetDescName((*ms)->joint), name); ms++){}
  ASSERT2(*ms,"No model set\n");
  return new (*ms) C_ModelSet;
}


//V[̕\B
void C_SceneDesc::PrintJoint(HSD_Joint* joint, int nest, bool* node) const
{
  u32		i;
  
  //--- Joint\ ---
  node[nest] = (joint->next != 0);
  for(i = 1; i < nest; i++)	OSReport((char*)(node[i]?"":"@"));
  if(nest > 0)	OSReport((char*)(node[i]?"":((nest > 0)?"":"")));
  OSReport("%s : ", GetDescName(joint));
  
  //--- DObjDesc\ ---
  if(union_type_dobj(joint)){
	for(HSD_DObjDesc* dd = joint->u.dobjdesc; dd; dd = dd->next){
	  OSReport("%s ", GetDescName(dd));
	}
  }
  OSReport("\n");
  
  //--- Joint ---
  if(joint->child)	PrintJoint(joint->child, nest+1, node);
  if(joint->next)	PrintJoint(joint->next, nest, node);
}
void C_SceneDesc::Print() const
{
  u32		i, j;
  bool	node[64];

  OSReport("=== Scene Info ===\n");

  //--- Joint ---
  for(i = 0; i < GetModelSetNum(); i++){
	OSReport("--- Joint %d ---\n", i);
	C_ModelSet*	ms = GetModelSet(i);

	//--- Tree\ ---
	PrintJoint(ms->GetJoint(), 0, node);

	//--- Action\ ---
	for(j = 0; j < ms->GetActionNum(); j++){
	  HSD_AnimJoint*		aj;
	  HSD_MatAnimJoint*	mj;
	  HSD_ShapeAnimJoint*	sj;
	  ms->GetAction(j, aj, mj, sj);
	  OSReport("%s\n", GetDescName(aj));
	}
	OSReport("\n");
  }
  //--- Light ---
  for(i = 0; i < GetLightSetNum(); i++){
	OSReport("--- Light %d ---\n", i);
	C_LightSet*	ls = GetLightSet(i);
	OSReport("%s\n", GetDescName(ls->GetLightDesc()));
	for(j = 0; j < ls->GetActionNum(); j++){
	  OSReport("  %s\n", GetDescName(ls->GetAction(j)));
	}
  }
  //--- Camera ---
  OSReport("--- Camera ---\n");
  C_CameraSet*	cs = GetCameraSet();
  if(cs)	for(i = 0; i < cs->GetActionNum(); i++){
	OSReport("%s\n", GetDescName(cs->GetAction(i)));
  }
  //--- Fog ---
  OSReport("--- Fog ---\n");
  C_FogSet*	fs = GetFogSet();
  if(fs)	for(i = 0; i < fs->GetActionNum(); i++){
	OSReport("%s\n", GetDescName(fs->GetAction(i)));
  }
}

