#include "common.h"
#include "c_shader.h"
C_ToonShader::C_ToonShader(void)
{
  img = new u16[TOON_TEXTURE_SIZE*4];

  img_desc.image_ptr = img;
  img_desc.width     = TOON_TEXTURE_SIZE;	
  img_desc.height    = 1;				
  img_desc.format    = (GXTexFmt)GX_TF_RGB565;
  img_desc.mipmap    = GX_FALSE;				
  img_desc.minLOD    = 0.f;
  img_desc.maxLOD    = 0.f;
	
#define	TOON_STEP		 3

#define ORG 1
#if ORG
#define	TOON_RGB_LOW	0
#define	TOON_RGB_HIGH	31
#else
#define TOON_THR_0 4
#define TOON_THR_1 8
#define TOON_THR_2 16
#define TOON_THR_3 24

#define TOON_VAL_0 0
#define TOON_VAL_1 0
#define TOON_VAL_2 27
#define TOON_VAL_3 31
#endif

  u8	rgb;
#if ORG
  u16	color = (TOON_RGB_LOW << 11) | (TOON_RGB_LOW << 6) | TOON_RGB_LOW;
#else
  u16	color = (TOON_VAL_3 << 11) | (TOON_VAL_3 << 6) | TOON_VAL_3;
#endif
  img[0] = color;
  for (int i = 1; i < TOON_TEXTURE_SIZE; i++) {
	if (i % (TOON_TEXTURE_SIZE / TOON_STEP) == 0) {
	  rgb = (u8)(((i / (TOON_TEXTURE_SIZE / TOON_STEP)) * 
				  (TOON_TEXTURE_SIZE / (TOON_STEP - 1)) - 1) >> 3);

#if ORG
	  if (rgb > TOON_RGB_HIGH) rgb = TOON_RGB_HIGH;
	  else if (rgb < TOON_RGB_LOW ) rgb = TOON_RGB_LOW;
#else
	  if(rgb > TOON_THR_3 ) rgb = TOON_VAL_3;
	  else if(rgb > TOON_THR_2 ) rgb = TOON_VAL_2;
	  else if(rgb > TOON_THR_1 ) rgb = TOON_VAL_1;
	  else if(rgb > TOON_THR_0 ) rgb = TOON_VAL_0;
#endif
	  color = (u16)((rgb << 11) | (rgb << 6) | rgb);
	}
	img[(i/4)*16 + (i%4)] = color;
  }
  //  DCFlushRange(img, sizeof(unsigned short)*TOON_TEXTURE_SIZE*4);
  DCStoreRange(img, sizeof(unsigned short)*TOON_TEXTURE_SIZE*4);
  GXInvalidateTexAll();
}


C_ToonShader::~C_ToonShader()
{
  delete img;
}


void C_ScreenTone::Proc( void )
{
}

void C_ScreenTone::Disp( void )
{
  //JݒB
  GXRenderModeObj *rmode = HSD_VIGetConfigure();
  HSD_CObj *cobj = HSD_CObjAlloc();
  HSD_CObjSetProjectionType(cobj, COBJ_PROJECTION_ORTHO);
  HSD_CObjSetViewportfx4(cobj, 0, (u16)rmode->viWidth, 0, (u16)rmode->viHeight);
  HSD_CObjSetScissorx4  (cobj, 0, (u16)rmode->viWidth, 0, (u16)rmode->viHeight);
  HSD_CObjSetNear (cobj,   0.0F);
  HSD_CObjSetFar  (cobj, 200.0F);
  HSD_CObjSetOrtho(cobj, 0.f, 480.f, 0.f, 640.f);
  Vec	eyepos   = {0.0f, 0.0f, 100.0f};
  Vec	interest = {0.0f, 0.0f,   0.0f};
  Vec	up       = {0.0f, 1.0f,   0.0f};
  Mtx	mtxView;
  MTXLookAt(mtxView, &eyepos, &up, &interest);
  HSD_CObjSetUserDefinedMtx(cobj, mtxView);
	
  //`lB
  GXSetChanCtrl(GX_COLOR0A0,
				GX_DISABLE,		// enable channel
				GX_SRC_VTX,		// amb source
				GX_SRC_VTX,		// mat source
				GX_LIGHT_NULL,	// light mask
				GX_DF_CLAMP,	// diffuse function
				GX_AF_NONE);
  GXSetNumChans(1);
	
  GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD_NULL, GX_TEXMAP_NULL, GX_COLOR0A0);
  GXSetTevOp(GX_TEVSTAGE0, GX_PASSCLR);
  GXSetNumTevStages(1);
  GXSetNumTexGens(0);

  //PE
  GXSetBlendMode(GX_BM_BLEND, GX_BL_SRCALPHA, GX_BL_INVSRCALPHA, GX_LO_CLEAR);

  GXSetZMode( GX_DISABLE, GX_NEVER, GX_DISABLE );
  GXSetAlphaUpdate( GX_DISABLE );
  GXSetColorUpdate( GX_ENABLE  );
  GXSetCullMode( GX_CULL_NONE );
	
  if (HSD_CObjSetCurrent(cobj)) {
	Mtx	mat;
	MTXIdentity(mat);
    GXLoadPosMtxImm(mat, GX_PNMTX0);
	GXSetCurrentMtx( GX_PNMTX0 );
	GXClearVtxDesc();
	GXSetVtxDesc(GX_VA_POS,  GX_DIRECT);
	GXSetVtxDesc(GX_VA_CLR0, GX_DIRECT);
	GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS,  GX_POS_XYZ, GX_F32, 0);
	GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGBA8, 0);

	#define	ZMID -0.5f
	GXBegin(GX_QUADS, GX_VTXFMT0, 4);
		GXPosition3f32(rect.GetX(), rect.GetY(), ZMID);
		GXColor4u8( 0, 0, 0, s );

		GXPosition3f32(rect.GetX()+rect.GetW(), rect.GetY(), ZMID);
		GXColor4u8( 0, 0, 0, s );
		
		GXPosition3f32(rect.GetX()+rect.GetW(), rect.GetY()+rect.GetH(), ZMID);
		GXColor4u8( 0, 0, 0, e );
		
		GXPosition3f32(rect.GetX(), rect.GetY()+rect.GetH(), ZMID);
		GXColor4u8( 0, 0, 0, e );
	GXEnd();
	HSD_CObjEndCurrent();
  }
  HSD_CObjUnref( cobj );

  GXSetZMode( GX_ENABLE, GX_LEQUAL, GX_ENABLE );
  GXSetAlphaUpdate( GX_DISABLE );
	

  HSD_StateInvalidate(HSD_STATE_ALL);
}








