/*
 * Copyright 1999-2002 HAL Laboratory, Inc. All rights reserved.
 *
 * $Id: lobj.c,v 1.5 2004/07/27 02:14:29 TypeR Exp $
 */

#ifdef WATANABE
#include "../../Lib/ghmGcDefine.h"
#endif
#ifdef FUJITA
#pragma cplusplus			off
#pragma cpp_extensions		off
#pragma warn_implicitconv	off
#endif

#include <string.h>
#include "sysdolphin/baselib/aobj.h"
#include "sysdolphin/baselib/cobj.h"
#include "sysdolphin/baselib/debug.h"
#include "sysdolphin/baselib/jobj.h"
#include "sysdolphin/baselib/lobj.h"
#include "sysdolphin/baselib/mtx.h"
#include "sysdolphin/baselib/spline.h"
#include "sysdolphin/baselib/state.h"
#include "sysdolphin/baselib/list.h"
#include "private.h"

#if DOLPHIN_API_AGE < 1
#define GXInitLightDirv(o, ha)	GXInitLightDirv((o), *(ha));
#define GXInitLightPosv(o, ha)	GXInitLightPosv((o), *(ha));
#endif

#define BIG_NUMBER      (f32)(1024 * 1024)

static void LObjInfoInit(void);

HSD_LObjInfo hsdLObj = { LObjInfoInit };

static HSD_LObjInfo *default_class = NULL;

static HSD_SList *current_lights = NULL;
static HSD_LObj  *active_lights[MAX_GXLIGHT];
static int        nb_active_lights = 0;

static GXLightID lightmask_diffuse  = GX_LIGHT_NULL;
static GXLightID lightmask_specular = GX_LIGHT_NULL;
static GXLightID lightmask_attnfunc = GX_LIGHT_NULL;
static GXLightID lightmask_alpha    = GX_LIGHT_NULL;

/** @package LObj */

/**
 * tO擾
 * 
 * @see HSD_LObjSetFlags
 */
u32
HSD_LObjGetFlags(HSD_LObj *lobj)
{
  return (lobj)? lobj->flags : 0;
}

/**
 * tOݒ肷
 * 
 * @see HSD_LObjClearFlags
 * @see HSD_LObjModifyFlags
 */
void
HSD_LObjSetFlags(HSD_LObj *lobj, u32 flags)
{
  if (lobj) lobj->flags |= flags;
}

/**
 * tONA
 * 
 * @see HSD_LObjSetFlags
 * @see HSD_LObjModifyFlags
 */
void
HSD_LObjClearFlags(HSD_LObj *lobj, u32 flags)
{
  if (lobj) lobj->flags &= ~flags;
}

/**
 * tOύX
 * 
 *  mask Ŏw肵tO flags ̒lɕύXB
 * 
 * @see HSD_LObjSetFlags
 * @see HSD_LObjClearFlags
 */
void
HSD_LObjModifyFlags(HSD_LObj *lobj, u32 flags, u32 mask)
{
  if (lobj) lobj->flags = (lobj->flags & ~mask) | (flags & mask);
}

/*
 * (J֐)
 * 
 * ˌp̃Cg}XN擾
 * 
 * HSD_LObjSetupInit()ŃANeBuɂꂽ lobj 
 * ˌgpn[hEFACg̃rbg}XN
 * 擾B
 * 
 * ̒l͂̂܂܁AJ[`l̃Cg}XNƂ
 * gpłB
 * 
 * @see HSD_LObjSetupInit
 */
GXLightID
HSD_LObjGetLightMaskDiffuse(void)
{
  return lightmask_diffuse;
}

/*
 * (J֐)
 * 
 * pxE̕KvȃCg̃}XN擾
 */
GXLightID
HSD_LObjGetLightMaskAttnFunc(void)
{
  return lightmask_attnfunc;
}

/*
 * (J֐)
 * 
 * At@`l̕KvȃCg̃}XN擾
 * 
 * HSD_LObjSetupInit()ŃANeBuɂꂽ lobj 
 * At@gpn[hEFACg̃rbg}XN
 * 擾B
 * 
 * ̒l͂̂܂܁AAt@`l̃Cg}XNƂ
 * gpłB
 * 
 * @see HSD_LObjSetupInit
 */
GXLightID
HSD_LObjGetLightMaskAlpha(void)
{
  return lightmask_alpha;
}

/*
 * (J֐)
 * 
 * ʔ˗p̃Cg}XN擾
 * 
 * HSD_LObjSetupInit()ŃANeBuɂꂽ lobj 
 * ʔˌgpn[hEFACg̃rbg}XN
 * 擾B
 * 
 * ̒l͂̂܂܁AJ[`l1̃Cg}XNƂ
 * gpłB
 * 
 * @see HSD_LObjSetupInit
 */
GXLightID
HSD_LObjGetLightMaskSpecular(void)
{
  return lightmask_specular;
}

/*
 * (J֐)
 * 
 * lobj \̂LȃCgƂēo^
 */
void
HSD_LObjSetActive(HSD_LObj *lobj)
{
  int idx;

  if (HSD_LObjGetType(lobj) == LOBJ_AMBIENT) {
    idx = MAX_GXLIGHT - 1;
    if (active_lights[idx])
      return;
  } else {
    idx = nb_active_lights ++;
  }
  active_lights[idx] = lobj;
  lobj->id = HSD_Index2LightID(idx);
#if 0
  HSD_Report("_HSD_SetActive lobj (%x) id (%x)\n", lobj, lobj->id);
#endif
}

/*
 * (J֐)
 * 
 * ANeBuȃCg擾B
 * 
 * @warning
 * ArGgCg͊܂܂ȂB
 */
int
HSD_LObjGetNbActive(void)
{
  return nb_active_lights;
}

/*
 * (J֐)
 * 
 * Light ID w肵ăANeBu lobj 擾
 */
HSD_LObj *
HSD_LObjGetActiveByID(GXLightID id)
{
  int idx = HSD_LightID2Index(id);
  if (0 <= idx && idx < MAX_GXLIGHT) {
    return active_lights[idx];
  } else {
    return NULL;
  }
}

/*
 * (J֐)
 * 
 * CfbNXw肵ăANeBu lobj 擾
 */
HSD_LObj *
HSD_LObjGetActiveByIndex(int idx)
{
  if (0 <= idx && idx < MAX_GXLIGHT - 1) {
    return active_lights[idx];
  } else {
    return NULL;
  }
}

/*
 * (J֐)
 * 
 * Cg̃ANeBuݒׂď
 */
void
HSD_LObjClearActive(void)
{
  int i;

  for (i = 0; i < MAX_GXLIGHT; i++) {
    active_lights[i] = NULL;
  }
  nb_active_lights = 0;
}


/**
 * 1 lobj  flags Ŏw肳ꂽAj[V폜B
 *
 * @see HSD_LObjRemoveAnimAllByFlags
 * @see HSD_LObjRemoveAnim
 * @see HSD_LObjRemoveAnimAll
 */
void
HSD_LObjRemoveAnimByFlags(HSD_LObj *lobj, u32 flags)
{
  if (lobj == NULL)
    return;

  if (flags & LOBJ_ANIM) {
    HSD_AObjRemove(lobj->aobj);
    lobj->aobj = NULL;
    HSD_WObjRemoveAnim(HSD_LObjGetPositionWObj(lobj));
    HSD_WObjRemoveAnim(HSD_LObjGetInterestWObj(lobj));
  }
}

/**
 * AꂽׂĂ lobj  flags Ŏw肳ꂽAj[V폜B
 *
 * @see HSD_LObjRemoveAnimByFlags
 * @see HSD_LObjRemoveAnim
 * @see HSD_LObjRemoveAnimAll
 */
void
HSD_LObjRemoveAnimAllByFlags(HSD_LObj *lobj, u32 flags)
{
  HSD_LObj *lp;

  if (lobj == NULL)
    return;

  for (lp = lobj; lp; lp = lp->next) {  
    HSD_LObjRemoveAnimByFlags(lp, flags);
  }
}

/**
 * 1 lobj ̑SẴAj[V폜B
 *
 * @see HSD_LObjRemoveAnimByFlags
 * @see HSD_LObjRemoveAnimAllByFlags
 * @see HSD_LObjRemoveAnimAll
 */
void
HSD_LObjRemoveAnim(HSD_LObj *lobj)
{
  HSD_LObjRemoveAnimByFlags(lobj, ALL_ANIM);
}

/**
 * AꂽׂĂ lobj ̑SẴAj[V폜B
 *
 * @see HSD_LObjRemoveAnimByFlags
 * @see HSD_LObjRemoveAnimAllByFlags
 * @see HSD_LObjRemoveAnim
 */
void
HSD_LObjRemoveAnimAll(HSD_LObj *lobj)
{
  HSD_LObj *lp;

  if (lobj == NULL)
    return;

  for (lp = lobj; lp; lp = lp->next) {
    HSD_LObjRemoveAnimAllByFlags(lp, ALL_ANIM);
  }
}

static void
LObjUpdateFunc(void *obj, u32 type, FObjData *val)
{
  HSD_LObj *lobj = obj;

  if (lobj == NULL)
    return;

  switch (type) {
  case HSD_A_L_VIS:
    if (val->fv >=  0.5) 
      lobj->flags &= ~LOBJ_HIDDEN;
    else
      lobj->flags |= LOBJ_HIDDEN;
    break;
  case HSD_A_L_A0:
  case HSD_A_L_CUTOFF:
    if (lobj->flags & LOBJ_RAW_PARAM) {
      lobj->u.attn.a0 = val->fv;
    } else {
      lobj->u.spot.cutoff = val->fv;
    }
    break;
  case HSD_A_L_A1:
  case HSD_A_L_REFDIST:
    if (lobj->flags & LOBJ_RAW_PARAM) {
      lobj->u.attn.a1 = val->fv;
    } else {
      lobj->u.spot.ref_dist = val->fv;
    }
    break;
  case HSD_A_L_A2:
  case HSD_A_L_REFBRIGHT:
    if (lobj->flags & LOBJ_RAW_PARAM) {
      lobj->u.attn.a2 = val->fv;
    } else {
      lobj->u.spot.ref_br = val->fv;
    }
    break;
  case HSD_A_L_K0:
    if (lobj->flags & LOBJ_RAW_PARAM) {
      lobj->u.attn.k0 = val->fv;
    }
    break;
  case HSD_A_L_K1:
    if (lobj->flags & LOBJ_RAW_PARAM) {
      lobj->u.attn.k1 = val->fv;
    }
    break;
  case HSD_A_L_K2:
    if (lobj->flags & LOBJ_RAW_PARAM) {
      lobj->u.attn.k2 = val->fv;
    }
    break;
  case HSD_A_L_LITC_R:
    lobj->color.r = (u8)(255.0F * HSD_ClampFloat(val->fv, 0, 1));
    break;
  case HSD_A_L_LITC_G:
    lobj->color.g = (u8)(255.0F * HSD_ClampFloat(val->fv, 0, 1));
    break;
  case HSD_A_L_LITC_B:
    lobj->color.b = (u8)(255.0F * HSD_ClampFloat(val->fv, 0, 1));
    break;
  case HSD_A_L_LITC_A:
    lobj->color.a = (u8)(255.0F * HSD_ClampFloat(val->fv, 0, 1));
    break;
  }
}

/**
 * P lobj ̃Aj[VߎsB
 *
 * @see HSD_LObjAnimAll
 */
void
HSD_LObjAnim(HSD_LObj *lobj)
{
  if (lobj == NULL)
    return;

  /* Aj[Vf[^ߎs */
#if HSD_ANIM_METHOD_LEVEL <= 3
  HSD_AObjInterpretAnim(lobj->aobj, lobj, HSD_LOBJ_METHOD(lobj)->update);
#else
  HSD_AObjInterpretAnim(lobj->aobj, lobj, LObjUpdateFunc);
#endif
  HSD_WObjInterpretAnim(HSD_LObjGetPositionWObj(lobj));
  HSD_WObjInterpretAnim(HSD_LObjGetInterestWObj(lobj));
}

/**
 * AꂽׂĂ lobj ̃Aj[VߎsB
 *
 * @see HSD_LObjAnim
 */
void
HSD_LObjAnimAll(HSD_LObj *lobj)
{
  HSD_LObj *lp;

  if (lobj == NULL)
    return;

  for (lp = lobj; lp; lp = lp->next) {
    HSD_LObjAnim(lp);
  }
}

/**
 * 1 lobj ɑ΂flags Ŏw肳ꂽAj[V̊JnwB
 *
 * @see HSD_LObjReqAnimAllByFlags
 * @see HSD_LObjReqAnim
 * @see HSD_LObjReqAnimAll
 */
void
HSD_LObjReqAnimByFlags(HSD_LObj *lobj, float startframe, u32 flags)
{
  if (lobj == NULL)
    return;

  if (flags & LOBJ_ANIM) {
    HSD_AObjReqAnim(lobj->aobj, startframe);
    HSD_WObjReqAnim(HSD_LObjGetPositionWObj(lobj), startframe);
    HSD_WObjReqAnim(HSD_LObjGetInterestWObj(lobj), startframe);
  }
}

/**
 * AꂽׂĂ lobj ɑ΂ flags Ŏw肳ꂽAj[V
 * JnwB
 *
 * @see HSD_LObjReqAnimByFlags
 * @see HSD_LObjReqAnim
 * @see HSD_LObjReqAnimAll
 */
void
HSD_LObjReqAnimAllByFlags(HSD_LObj *lobj, float startframe, u32 flags)
{
  HSD_LObj *lp;

  if (lobj == NULL)
    return;
  for (lp = lobj; lp; lp = lp->next) {
    HSD_LObjReqAnimByFlags(lp, startframe, flags);
  }
}

/**
 * 1 lobj ɑ΂ăAj[V̊JnwB
 *
 * @see HSD_LObjReqAnimByFlags
 * @see HSD_LObjReqAnimAllByFlags
 * @see HSD_LObjReqAnimAll
 */
void
HSD_LObjReqAnim(HSD_LObj *lobj, float startframe)
{
  HSD_LObjReqAnimByFlags(lobj, startframe, LOBJ_ANIM);
}
  
/**
 * AꂽׂĂ lobj ɑ΂ăAj[V̊JnwB
 *
 * @see HSD_LObjReqAnimByFlags
 * @see HSD_LObjReqAnimAllByFlags
 * @see HSD_LObjReqAnim
 */
void
HSD_LObjReqAnimAll(HSD_LObj *lobj, float startframe)
{
  HSD_LObjReqAnimAllByFlags(lobj, startframe, LOBJ_ANIM);
}

/**
 * lobj ̃CgxNg擾B
 * 
 * Cg̈ʒu璍_֌KꂽxNgԂB
 * 
 * InfiniteCgPointCg̏ꍇ́A_̑
 * [hWň_֌xNgԂB
 * 
 * @see HSD_LObjSetPosition
 * @see HSD_LObjSetInterest
 */
void
HSD_LObjGetLightVector(HSD_LObj *lobj, VecPtr dir)
{
  Vec position = {0.0F, 0.0F, 0.0F};
  Vec interest = {0.0F, 0.0F, 0.0F};

  if (lobj == NULL)
    return;

  HSD_LObjGetPosition(lobj, &position);
  HSD_LObjGetInterest(lobj, &interest);
  VECSubtract(&interest, &position, dir);
  if (vec_normalize_check(dir, dir) != 0) {
    dir->x = 0;
    dir->y = 0;
    dir->z = 1;
  }
}

/*
 * (J֐)
 * 
 * Cg̐ݒB
 *
 * n[hEFACgݒ肷B
 *
 * @param color		diffuseFBCgƃ}eA̍ς݃J[w
 * @param shininess	ʔ˂̎w
 */
void
HSD_LObjSetup(HSD_LObj *lobj, GXColor color, float shininess)
{
#ifdef PRINT_DEBUG
  HSD_Report("HSD_LObjSetup in flags (%x)\n", lobj->flags);
#endif
  
  if (lobj->flags & LOBJ_HIDDEN || HSD_LObjGetType(lobj) == LOBJ_AMBIENT)
    return;
    
  if (lobj->flags & (LOBJ_DIFFUSE|LOBJ_ALPHA)) {
    if (lobj->hw_color.r != color.r
	|| lobj->hw_color.g != color.g
	|| lobj->hw_color.b != color.b
	|| lobj->hw_color.a != color.a) {
      GXInitLightColor(&lobj->lightobj, color);
      lobj->hw_color = color;
      lobj->flags |= LOBJ_DIFF_DIRTY;
    }

    if (lobj->flags & LOBJ_DIFF_DIRTY) {
      GXLoadLightObjImm(&lobj->lightobj, lobj->id);
      lobj->flags &= ~LOBJ_DIFF_DIRTY;
    }
  }

  if (lobj->spec_id != GX_LIGHT_NULL) {
    if (lobj->shininess != shininess) {
      lobj->shininess = shininess;
      GXInitLightShininess(&lobj->spec_lightobj, shininess);
      lobj->flags |= LOBJ_SPEC_DIRTY;
    }

    if (lobj->flags & LOBJ_SPEC_DIRTY) {
      GXLoadLightObjImm(&lobj->spec_lightobj, lobj->spec_id);
      lobj->flags &= ~LOBJ_SPEC_DIRTY;
    }
  }
}

/*
 * (J֐)
 * 
 * GXInitSpecularDir() ̓IuWFNg̈ʒulȂ̂
 * O half angle vector ݒ肷B
 *
 * @param pmtx	`撆jobj̃JWnւ̍Wϊ}gNX
 */
void
HSD_LObjSetupSpecularInit(MtxPtr pmtx)
{
  int i, num;
  Vec cdir, jpos;

  /* JWnłjobj̒Sʒu(jpos)A
   * jobj ̒Sւ̕xNgicdir:Kj𓾂
   */
  HSD_MtxColVec(pmtx, 3, &jpos);
  if (vec_normalize_check(&jpos, &cdir) != 0) {
    cdir.x = 0.0F;
    cdir.y = 0.0F;
    cdir.z = -1.0F;
  }

  num = HSD_LObjGetNbActive();
  for (i = 0; i < num; i++) {
    Vec half, ldir;	
    HSD_LObj *lobj = HSD_LObjGetActiveByIndex(i);

    if (lobj->spec_id == GX_LIGHT_NULL) {
      /* XyLCg̏ɒBAn[hEFACg
       * Ȃ߁AXyLʂȂ
       */
      continue;
    }
    /* make half-angle */
    switch (HSD_LObjGetType(lobj)) {
    case LOBJ_POINT:
    case LOBJ_SPOT:
      /* IuWFNg̒SɌxNg */
      VECSubtract(&jpos, &lobj->lvec, &ldir);
      VECNormalize(&ldir, &ldir);
// modified by watanabe 05/18/2004
//      VECAdd(&ldir, &cdir, &half);
      GXInitSpecularDirv(&lobj->spec_lightobj, &ldir);				// modified by watanabe 05/18/2004
      break;

    case LOBJ_INFINITE:
      /* sȂ̂ŁAIuWFNg̈ʒuɂ炸͈̌ */
      VECAdd(&lobj->lvec, &cdir, &half);
      GXInitSpecularDirv(&lobj->spec_lightobj, &lobj->lvec);		// modified by watanabe 05/18/2004
      break;

    default:
      assert(0);
    }
// modified by watanabe 05/18/2004
//    VECNormalize(&half, &half);
//    GXInitLightDirv(&lobj->spec_lightobj, &half);

    lobj->flags |= LOBJ_SPEC_DIRTY;
  }
}

static void
setup_diffuse_lightobj(HSD_LObj *lobj)
{
  GXInitLightColor(&lobj->lightobj, lobj->color);
  lobj->hw_color = lobj->color;
  lobj->flags |= LOBJ_DIFF_DIRTY;

  switch (HSD_LObjGetType(lobj)) {
  case LOBJ_SPOT:
  case LOBJ_POINT:
    lightmask_attnfunc |= lobj->id;
    break;
  case LOBJ_INFINITE:
    break;
  default:
    assert(0);
  }
  if (lobj->flags & LOBJ_DIFFUSE) {
    lightmask_diffuse |= lobj->id;
  }
  if (lobj->flags & LOBJ_ALPHA) {
    lightmask_alpha |= lobj->id;
  }
}

static void
setup_spec_lightobj(HSD_LObj *lobj, Mtx vmtx, GXLightID spec_id)
{
  lobj->spec_id = spec_id;
  if (spec_id != GX_LIGHT_NULL) {
    GXInitLightColor(&lobj->spec_lightobj, lobj->color);
    lobj->shininess = 50.0F;
    GXInitLightShininess(&lobj->spec_lightobj, lobj->shininess);

    switch (HSD_LObjGetType(lobj)) {
    case LOBJ_POINT:
    case LOBJ_SPOT:
      /*
       * JWnɂʂւ̃xNg𓾂邽߁A
       * 炩߁AJWnɂʒu𓾂ĂB
       */
      HSD_LObjGetPosition(lobj, &lobj->lvec);
      MTXMultVec(vmtx, &lobj->lvec, &lobj->lvec);
      break;

    case LOBJ_INFINITE:
      /*
       * JWnɂ̕炩ߓĂB
       */
      HSD_LObjGetLightVector(lobj, &lobj->lvec);
      MTXMultVecSR(vmtx, &lobj->lvec, &lobj->lvec);
      VECNormalize(&lobj->lvec, &lobj->lvec);
      break;

    default:
      assert(0);
    }
    lobj->flags |= LOBJ_SPEC_DIRTY;
    lightmask_specular |= spec_id;
  }
}

static void
setup_infinite_lightobj(HSD_LObj *lobj, MtxPtr vmtx)
{
  Vec lpos;

  HSD_LObjGetPosition(lobj, &lpos);
  lpos.x *= BIG_NUMBER;
  lpos.y *= BIG_NUMBER;
  lpos.z *= BIG_NUMBER;
  MTXMultVec(vmtx, &lpos, &lpos);
  if (lobj->flags & LOBJ_DIFFUSE) {
    GXInitLightPosv(&lobj->lightobj, &lpos);
    GXInitLightAttn(&lobj->lightobj, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0);
  }
  if (lobj->flags & LOBJ_SPECULAR) {
    GXInitLightPosv(&lobj->spec_lightobj, &lpos);
  }
}

static void
setup_point_lightobj(HSD_LObj *l, MtxPtr vmtx)
{
  Vec lpos;

  GXInitLightColor(&l->lightobj, l->color);
  l->hw_color = l->color;
  HSD_LObjGetPosition(l, &lpos);
  MTXMultVec(vmtx, &lpos, &lpos);
  GXInitLightPosv(&l->lightobj, &lpos);
  GXInitLightPosv(&l->spec_lightobj, &lpos);
#if DOLPHIN_API_AGE < 1
  GXInitLightDirv(&l->lightobj, &lpos); /* ĂȂƂ */
#endif
  if (l->flags & LOBJ_RAW_PARAM) {
    GXInitLightAttn(&l->lightobj,
		    1.0F, 0.0F, 0.0F,
		    l->u.attn.k0,
		    l->u.attn.k1,
		    l->u.attn.k2);
  } else {
    float ref_br = l->u.spot.ref_br;
    float ref_dist = l->u.spot.ref_dist;
    GXDistAttnFn dist_func = l->u.spot.dist_func;
    
    if (ref_dist < FLT_EPSILON) {
      ref_dist = FLT_EPSILON;
    }
    if (ref_br < FLT_EPSILON) {
      ref_br = FLT_EPSILON;
    }
    GXInitLightDistAttn(&l->lightobj, ref_dist, ref_br, dist_func);
    GXInitLightSpot(&l->lightobj, 0.0F, GX_SP_OFF);
    GXInitLightDistAttn(&l->spec_lightobj, ref_dist, ref_br, dist_func);
  }
}

static void
setup_spot_lightobj(HSD_LObj *l, MtxPtr vmtx)
{
  Vec lpos, ldir;

  HSD_LObjGetPosition(l, &lpos);
  MTXMultVec(vmtx, &lpos, &lpos);
  HSD_LObjGetLightVector(l, &ldir);
  MTXMultVecSR(vmtx, &ldir, &ldir);
  VECNormalize(&ldir, &ldir);
  GXInitLightPosv(&l->lightobj, &lpos);
  GXInitLightPosv(&l->spec_lightobj, &lpos);
  GXInitLightDirv(&l->lightobj, &ldir);
  if (l->flags & LOBJ_RAW_PARAM) {
    GXInitLightAttn(&l->lightobj,
		    l->u.attn.a0,
		    l->u.attn.a1,
		    l->u.attn.a2,
		    l->u.attn.k0,
		    l->u.attn.k1,
		    l->u.attn.k2);
  } else {
    float ref_br = l->u.spot.ref_br;
    float ref_dist = l->u.spot.ref_dist;
    float cutoff = l->u.spot.cutoff;
    GXSpotFn spot_func = l->u.spot.spot_func;
    GXDistAttnFn dist_func = l->u.spot.dist_func;
    
    if (ref_dist < FLT_EPSILON) {
      ref_dist = FLT_EPSILON;
    }
    if (ref_br < FLT_EPSILON) {
      ref_br = FLT_EPSILON;
    }
    GXInitLightDistAttn(&l->lightobj, ref_dist, ref_br, dist_func);
    GXInitLightDistAttn(&l->spec_lightobj, ref_dist, ref_br, dist_func);
    /*
     * Cu̖炵Bcutoff 60xʂɂƌB
     * 2000/06/15 C.Itoh at N.T.D
     */
    GXInitLightSpot(&l->lightobj, cutoff, spot_func);
  }
}

/**
 * Cgݒ̏B
 * 
 * JgCgQvCIeB̍Ƀn[hEFA
 * 蓖ĉ\ȃCgIANeBuB
 * 
 * @warning
 * LȃCg̍\ς邽т1xĂԎB
 * 
 * @param cobj	JgJ
 */
void
HSD_LObjSetupInit(HSD_CObj *cobj)
{
  MtxPtr vmtx;
  int i, num, idx;
  HSD_SList *list;

#ifdef PRINT_DEBUG
  HSD_Report("HSD_LObjSetupInit:\n");
#endif

  lightmask_diffuse  = GX_LIGHT_NULL;
  lightmask_specular = GX_LIGHT_NULL;
  lightmask_attnfunc = GX_LIGHT_NULL;
  lightmask_alpha    = GX_LIGHT_NULL;
  vmtx = HSD_CObjGetViewingMtxPtrDirect(cobj);

  idx = 0;

  HSD_LObjClearActive();

  /* ^CvɉĊeLOBJLightObjݒ肷BɁA
   * Iɐspecular lightƂĎgp\Ȍ𒲂ׂ.
   */
  for (list=current_lights; idx<MAX_GXLIGHT-1 && list; list=list->next) {
    HSD_LObj *lobj = list->data;
    u32 type;
    u32 ref_type;

    if (!lobj || lobj->flags & LOBJ_HIDDEN) {
      continue;
    }

    type = HSD_LObjGetType(lobj);

    ref_type = lobj->flags & (LOBJ_DIFFUSE|LOBJ_ALPHA|LOBJ_SPECULAR);

    if (!ref_type) {
      continue;
    }

    HSD_LObjSetActive(lobj);
    idx = HSD_LObjGetNbActive();

    lobj->spec_id = GX_LIGHT_NULL;

    switch (type) {
    case LOBJ_INFINITE:
      setup_infinite_lightobj(lobj, vmtx);
      break;
    case LOBJ_POINT:
      setup_point_lightobj(lobj, vmtx);
      break;
    case LOBJ_SPOT:
      setup_spot_lightobj(lobj, vmtx);
      break;
    case LOBJ_AMBIENT:
      continue;
    }

    if (ref_type & (LOBJ_DIFFUSE|LOBJ_ALPHA)) {
      setup_diffuse_lightobj(lobj);
    } else if (ref_type & LOBJ_SPECULAR) {
      setup_spec_lightobj(lobj, vmtx, lobj->id);
    }
  }

  /* ArGgCg܂o^ĂȂ猩܂ŒTB*/
  if (!HSD_LObjGetActiveByID(GX_LIGHT_AMBIENT)) {
    for (; list; list=list->next) {
      HSD_LObj *lobj = list->data;
      if (!lobj || lobj->flags & LOBJ_HIDDEN) {
	continue;
      }
      if (HSD_LObjGetType(lobj) == LOBJ_AMBIENT &&
	  (lobj->flags & (LOBJ_DIFFUSE|LOBJ_ALPHA)))
      {
	HSD_LObjSetActive(lobj);
	break;
      }
    }
  }

  /*
   * ̎_ł́AANeBuɂȂĂ郉CĝADIFFUSE܂ALPHA
   * SPECULARɎ̂SPECULARɂ̓n[hEFA蓖ĂĂȂB
   * 
   * n[hEFACgɗ]T΁ÃCgSPECULARʂ
   * n[hEFA蓖ĂB
   */
  
  num = HSD_LObjGetNbActive();
  for (i=0; idx<MAX_GXLIGHT-1 && i<num; i++) {
    HSD_LObj *lobj;
    u32 flags;
    
    if (!(lobj = HSD_LObjGetActiveByIndex(i))) {
      continue;
    }

    flags = lobj->flags;

    if ((flags & LOBJ_SPECULAR) && (flags & (LOBJ_DIFFUSE|LOBJ_ALPHA))) {
      setup_spec_lightobj(lobj, vmtx, HSD_Index2LightID(idx++));
    }
  }

  /*
   * ANeBuȃCg̃p[^n[hEFAɐݒ肷B
   */
  for (i=0; i<num; i++) {
    HSD_LObj *lobj;
    
    if ((lobj = HSD_LObjGetActiveByIndex(i))) {
      HSD_LObjSetup(lobj, lobj->color, lobj->shininess);
    }
  }
}

/**
 * CgJgCgɒǉ
 * 
 * JgCgƂēo^ꂽ̂̂ȂDx̍
 * n[hEFAɊ蓖ĂB
 * 
 * eCĝvCIeBl(HSD_LObjSetPriority())ق
 * DxB
 * 
 * ̃vCIeBlCgo^ꂽꍇA
 * ɓo^ꂽ̂荂DxB
 *
 * @see HSD_LObjAddCurrentAll
 * @see HSD_LObjDeleteCurrent
 * @see HSD_LObjSetPriority
 */
void
HSD_LObjAddCurrent(HSD_LObj *lobj)
{
  HSD_SList *list, **listp;

  if (lobj == NULL)
    return;

  for (list=current_lights; list; list=list->next) {
    if (list->data == lobj) {
      HSD_LObjDeleteCurrent(lobj);
      break;
    }
  }

  HSD_LObjRef(lobj);
  for (listp=&current_lights; *listp; listp=&(*listp)->next) {
    if (HSD_LObjGetPriority((*listp)->data) > HSD_LObjGetPriority(lobj)) {
      break;
    }
  }
  *listp = HSD_SListPrepend(*listp, lobj);
}


/**
 * AꂽׂẴCgJgCgɒǉ
 *
 * @see HSD_LObjAddCurrent
 * @see HSD_LObjDeleteCurrentAll
 */
void
HSD_LObjAddCurrentAll(HSD_LObj *lobj)
{
  HSD_LObj *l;

  for(l=lobj; l; l=l->next) {
    HSD_LObjAddCurrent(l);
  }
}

/**
 * CgJgCg폜
 *
 * @see HSD_LObjAddCurrent
 * @see HSD_LObjDeleteCurrentAll
 */
void
HSD_LObjDeleteCurrent(HSD_LObj *lobj)
{
  HSD_SList **p;
  int i;

  if (!lobj)
    return;
  
  for (p=&current_lights; *p; p=&(*p)->next) {
    if ((*p)->data == lobj) {
      for (i=0; i<MAX_GXLIGHT; i++) {
	if (lobj == active_lights[i]) {
	  active_lights[i] = NULL;
	}
      }
      (*p) = HSD_SListRemove(*p);
      HSD_LObjUnref(lobj);
      return;
    }
  }
}

/**
 * AꂽׂẴCgJgCg폜
 *
 * @see HSD_LObjDeleteCurrent
 * @see HSD_LObjAddCurrentAll
 */
void
HSD_LObjDeleteCurrentAll(HSD_LObj *lobj)
{
  if (lobj) {
    for (; lobj; lobj=lobj->next) {
      HSD_LObjDeleteCurrent(lobj);
    }
  } else {
    HSD_LObjClearActive();
    while (current_lights) {
      HSD_LObjUnref((HSD_LObj *)(current_lights->data));
      current_lights = HSD_SListRemove(current_lights);
    }
  }
}

/**
 * CgJgCgɐݒ
 * 
 * ȑÕJgCgׂ͂ĖAVɎw肳ꂽCg
 * JgCgɐݒ肷B
 * 
 * ȉ̏Ɠ
 * <pre>
 * HSD_LObjDeleteCurrentAll(NULL);
 * HSD_LObjAddCurrent(lobj);
 * </pre>
 * 
 * @see HSD_LObjDeleteCurrentAll
 * @see HSD_LObjAddCurrent
 */
void
HSD_LObjSetCurrent(HSD_LObj *lobj)
{
  HSD_LObjDeleteCurrentAll(NULL);
  HSD_LObjAddCurrent(lobj);
}

/**
 * AꂽׂẴCgJgCgɐݒ
 * 
 * ȑÕJgCgׂ͂ĖAVɎw肳ꂽCgȉA
 * AꂽׂẴCgJgCgɐݒ肷B
 * 
 * ȉ̏Ɠ
 * <pre>
 * HSD_LObjDeleteCurrentAll(NULL);
 * HSD_LObjAddCurrentAll(lobj);
 * </pre>
 * 
 * @see HSD_LObjSetCurrentAll
 * @see HSD_LObjAddCurrentAll
 */
void
HSD_LObjSetCurrentAll(HSD_LObj *lobj)
{
  HSD_LObjDeleteCurrentAll(NULL);
  HSD_LObjAddCurrentAll(lobj);
}

/**
 * JgCg̐𓾂
 * 
 * @see HSD_LObjAddCurrent
 */
int
HSD_LObjGetNbCurrent(void)
{
  HSD_SList *list;
  int num = 0;

  for (list=current_lights; list; list=list->next) {
    num++;
  }
  return num;
}

/**
 * idx ԖڂɗDx̍JgCg擾
 * 
 * @param idx	DxiOƂDxAȉ1, 2, 3 Ƒj
 * 
 * @see HSD_LObjAddCurrent
 * @see HSD_LObjGetCurrentByType
 */
HSD_LObj *
HSD_LObjGetCurrentByIndex(int idx)
{
  HSD_SList *list;
  int i = 0;

  for (list=current_lights; list; list=list->next) {
    if (i == idx) {
      return list->data;
    }
    i++;
  }
  return NULL;
}

/**
 * w肵^Cv̍łDx̍JgCg擾
 * 
 * wł^CvHSD_LObjGetTypeQƁB
 * 
 * @see HSD_LObjAddCurrent
 * @see HSD_LObjGetCurrentByIndex
 * @see HSD_LObjGetType
 */
HSD_LObj *
HSD_LObjGetCurrentByType(u32 type)
{
  HSD_SList *list;

  type &= LOBJ_TYPE_MASK;
  for (list=current_lights; list; list=list->next) {
    if (HSD_LObjGetType(list->data) == type) {
      return list->data;
    }
  }
  return NULL;
}

/**
 * LightID ΉCfbNXԂ
 * 
 * <pre>
 *   idx | id
 * ------+-----------
 *    0  | GX_LIGHT0
 *    1  | GX_LIGHT1
 *    2  | GX_LIGHT2
 *    3  | GX_LIGHT3
 *    4  | GX_LIGHT4
 *    5  | GX_LIGHT5
 *    6  | GX_LIGHT6
 *    7  | GX_LIGHT7
 * </pre>
 * 
 * @see HSD_Index2LightID
 */
u32
HSD_LightID2Index(GXLightID id)
{
  u32 index;

  //HSD_Report("light id %d =>", id);
  switch (id) {
  case GX_LIGHT0: index = 0; break;
  case GX_LIGHT1: index = 1; break;
  case GX_LIGHT2: index = 2; break;
  case GX_LIGHT3: index = 3; break;
#if DOLPHIN_API_AGE >= 1
  case GX_LIGHT4: index = 4; break;
  case GX_LIGHT5: index = 5; break;
  case GX_LIGHT6: index = 6; break;
  case GX_LIGHT7: index = 7; break;
#endif
  case GX_LIGHT_AMBIENT: index = MAX_GXLIGHT-1; break;
  default: assert(0);
  }
  //HSD_Report(" index %d\n", index);
  return index;
}

/**
 * CfbNXΉ LightID Ԃ
 * 
 * CfbNXLightID̑ΉHSD_LightID2Index()QƁB
 * 
 * @see HSD_LightID2Index
 */
GXLightID
HSD_Index2LightID(u32 index)
{
  GXLightID id;

  //HSD_Report("light index %d =>", index);
  switch (index) {
  case 0: id = GX_LIGHT0; break;
  case 1: id = GX_LIGHT1; break;
  case 2: id = GX_LIGHT2; break;
  case 3: id = GX_LIGHT3; break;
#if DOLPHIN_API_AGE >= 1
  case 4: id = GX_LIGHT4; break;
  case 5: id = GX_LIGHT5; break;
  case 6: id = GX_LIGHT6; break;
  case 7: id = GX_LIGHT7; break;
#endif
  case MAX_GXLIGHT-1: id = GX_LIGHT_AMBIENT; break;
  default: id = GX_LIGHT_NULL;	break;
  }
  //HSD_Report(" id %d\n", id);
  return id;
}

/**
 * AꂽׂĂlobjƂ̉̍\̂폜
 *
 * @see HSD_LObjUnref
 */
void
HSD_LObjRemoveAll(HSD_LObj *lobj)
{
  while (lobj) {
    HSD_LObj *next = lobj->next;
    HSD_LObjDeleteCurrent(lobj);
    HSD_LObjUnref(lobj);
    lobj = next;
  }
}

/**
 * J[ݒ
 * 
 * @see HSD_LObjGetColor
 */
void
HSD_LObjSetColor(HSD_LObj *lobj, GXColor color)
{
  lobj->color = color;
}

/**
 * J[擾
 * 
 * @see HSD_LObjSetColor
 */
void
HSD_LObjGetColor(HSD_LObj *lobj, GXColor *color)
{
  *color = lobj->color;
}

/**
 * Cg̃X|bgp[^ݒ肷B
 * 
 * @warning
 * ̃p[^͌W̒ڎwƂ͋łȂB
 * (HSD_LObjSetAttn{A,K,}())
 * 
 * @param cutoff	JbgItp (Degree)
 * @param spot_func	px֐
 * 
 * @see HSD_LObjSetDistAttn
 * @see HSD_LObjSetAttnA
 * @see HSD_LObjSetAttnK
 * @see HSD_LObjSetAttn
 */
void
HSD_LObjSetSpot(HSD_LObj *lobj, f32 cutoff, GXSpotFn spot_func)
{
  if (!lobj) return;

  lobj->u.spot.cutoff = cutoff;
  lobj->u.spot.spot_func = spot_func;
}

/**
 * p[^ݒ肷B
 * 
 * ref_distanceɂ郉Cg̋ref_brightnessɎw肷B
 * 
 * @warning
 * ̃p[^͌W̒ڎwƂ͋łȂB
 * (HSD_LObjSetAttn{A,K,}())
 * 
 * @see HSD_LObjSetSpot
 * @see HSD_LObjSetAttnA
 * @see HSD_LObjSetAttnK
 * @see HSD_LObjSetAttn
 */
void
HSD_LObjSetDistAttn(HSD_LObj *lobj,
		    f32 ref_distance,
		    f32 ref_brightness,
		    GXDistAttnFn dist_func)
{
  if (!lobj) return;
  
  lobj->u.spot.ref_dist  = ref_distance;
  lobj->u.spot.ref_br    = ref_brightness;
  lobj->u.spot.dist_func = dist_func;
}

/**
 * GAMECUBE̊pxW𒼐ڎw肷B
 * 
 * @warning
 * ̃p[^wHSD_LObjSet{Spot,DistAttn}()Ƃ
 * łȂB
 * 
 * @see HSD_LObjSetSpot
 * @see HSD_LObjSetDistAttn
 * @see HSD_LObjSetAttnK
 * @see HSD_LObjSetAttn
 */
void
HSD_LObjSetAttnA(HSD_LObj *lobj, f32 a0, f32 a1, f32 a2)
{
  if (!lobj) return;
  
  lobj->u.attn.a0 = a0;
  lobj->u.attn.a1 = a1;
  lobj->u.attn.a2 = a2;
}

/**
 * GAMECUBE̋W𒼐ڎw肷B
 * 
 * @warning
 * ̃p[^wHSD_LObjSet{Spot,DistAttn}()Ƃ
 * łȂB
 * 
 * @see HSD_LObjSetSpot
 * @see HSD_LObjSetDistAttn
 * @see HSD_LObjSetAttnA
 * @see HSD_LObjSetAttn
 */
void
HSD_LObjSetAttnK(HSD_LObj *lobj, f32 k0, f32 k1, f32 k2)
{
  if (!lobj) return;

  lobj->u.attn.k0 = k0;
  lobj->u.attn.k1 = k1;
  lobj->u.attn.k2 = k2;
}

/**
 * GAMECUBĚW𒼐ڎw肷B
 * 
 * @warning
 * ̃p[^wHSD_LObjSet{Spot,DistAttn}()Ƃ
 * łȂB
 * 
 * @see HSD_LObjSetSpot
 * @see HSD_LObjSetDistAttn
 * @see HSD_LObjSetAttnA
 * @see HSD_LObjSetAttnK
 */
void
HSD_LObjSetAttn(HSD_LObj *lobj,
	        f32 a0, f32 a1, f32 a2,
	        f32 k0, f32 k1, f32 k2)
{
  HSD_LObjSetAttnA(lobj, a0, a1, a2);
  HSD_LObjSetAttnK(lobj, k0, k1, k2);
}

/**
 * Cgʒuݒ
 * 
 * @see HSD_LObjGetPosition
 */
void
HSD_LObjSetPosition(HSD_LObj *lobj, VecPtr position)
{
  assert(lobj);
  if (!lobj->position) {
    lobj->position = HSD_WObjAlloc();
    assert(lobj->position);
  }
  HSD_WObjSetPosition(lobj->position, position);
}

/**
 * Cgʒu擾
 *
 * @return
 * ʒuݒ肳Ă΍W̃AhXɐݒ肵1ԂB
 * łȂ΁A0ԂB
 * 
 * @see HSD_LObjSetPosition
 */
int
HSD_LObjGetPosition(HSD_LObj *lobj, VecPtr position)
{
  if (lobj && lobj->position) {
    HSD_WObjGetPosition(lobj->position, position);
    return 1;
  } else {
    return 0;
  }
}

/**
 * X|bgCg̒_ݒ
 * 
 * @see HSD_LObjGetInterest
 */
void
HSD_LObjSetInterest(HSD_LObj *lobj, VecPtr interest)
{
  assert(lobj);
  if (!lobj->interest) {
    lobj->interest = HSD_WObjAlloc();
    assert(lobj->interest);
  }
  HSD_WObjSetPosition(lobj->interest, interest);
}

/**
 * X|bgCg̒_擾
 *
 * @return
 * _ݒ肳Ă΍W̃AhXɐݒ肵1ԂB
 * łȂ΁A0ԂB
 * 
 * @see HSD_LObjSetInterest
 */
int
HSD_LObjGetInterest(HSD_LObj *lobj, VecPtr interest)
{
  if (lobj && lobj->interest) {
    HSD_WObjGetPosition(lobj->interest, interest);
    return 1;
  } else {
    return 0;
  }
}

/**
 * ftHgNX̎w
 * 
 * lobjƂăftHgŐIuWFNg̃NXw肷B
 * 
 * włNXhsdLObjA̓oNXłȂ΂ȂȂB
 * 
 * @see HSD_LObjGetDefaultClass
 * @see HSD_LObjAlloc
 * @see HSD_LObjLoadDesc
 */
void
HSD_LObjSetDefaultClass(HSD_LObjInfo *info)
{
  if (info) {
    assert(hsdIsDescendantOf(info, &hsdLObj));
  }
  default_class = info;
}

/**
 * ftHgNX̎擾
 * 
 * lobjƂăftHgŐIuWFNg̃NX擾B
 * 
 * @see HSD_LObjSetDefaultClass
 * @see HSD_LObjAlloc
 * @see HSD_LObjLoadDesc
 */
HSD_LObjInfo *
HSD_LObjGetDefaultClass(void)
{
  return default_class? default_class : &hsdLObj;
}

/**
 * lobj  1쐬
 * 
 * @see HSD_LObjUnref
 * @see HSD_LObjSetDefaultClass
 */
HSD_LObj *
HSD_LObjAlloc(void)
{
  HSD_LObj *new = hsdNew(HSD_LObjGetDefaultClass());
  assert(new);
  return new;
}

/**
 * lobj  j
 *
 * @warning
 * {֐̓o̊J͍lȂBʏ̍폜ɂHSD_LObjUnref(), 
 * ܂HSD_LObjRemoveAllgp邱ƁB
 *
 * @see HSD_LObjUnref
 * @see HSD_LObjRemoveAll
 */
void
HSD_LObjFree(HSD_LObj *lobj)
{
  if (lobj) HSD_CLASS_METHOD(lobj)->destroy(HSD_CLASS(lobj));
}

/**
 * ʒu\wobj擾B
 * 
 * @see HSD_LObjSetPositionWObj
 */
HSD_WObj *
HSD_LObjGetPositionWObj(HSD_LObj *lobj)
{
  return (lobj)? lobj->position : NULL;
}

/**
 * _\wobj擾B
 * 
 * @see HSD_LObjSetInterestWObj
 */
HSD_WObj *
HSD_LObjGetInterestWObj(HSD_LObj *lobj)
{
  return (lobj)? lobj->interest : NULL;
}

/**
 * ʒu\wobjݒ肷B
 *
 * @tip
 * ȑOݒ肳ĂwobjUnrefB
 * 
 * @see HSD_LObjGetPositionWObj
 */
void
HSD_LObjSetPositionWObj(HSD_LObj *lobj, HSD_WObj *wobj)
{
  if (!lobj)
    return;

  HSD_WObjUnref(lobj->position);
  lobj->position = wobj;
}

/**
 * _\wobjݒ肷B
 *
 * @tip
 * ȑOݒ肳ĂwobjUnrefB
 * 
 * @see HSD_LObjGetInterestWObj
 */
void
HSD_LObjSetInterestWObj(HSD_LObj *lobj, HSD_WObj *wobj)
{
  if (!lobj)
    return;

  HSD_WObjUnref(lobj->interest);
  lobj->interest = wobj;
}

static int
LObjLoad(HSD_LObj *lobj, HSD_LightDesc *ldesc)
{
#if 0
  HSD_Report("LObjLightDesc in (%x)\n", ldesc);
#endif

  HSD_LObjSetColor(lobj, ldesc->color);

  HSD_LObjSetFlags(lobj, ldesc->flags);
  switch (ldesc->flags & LOBJ_TYPE_MASK) {
  case LOBJ_AMBIENT:
    break;

  case LOBJ_INFINITE:
    HSD_LObjSetPositionWObj(lobj, HSD_WObjLoadDesc(ldesc->position));
    break;

  case LOBJ_POINT:
    HSD_LObjSetPositionWObj(lobj, HSD_WObjLoadDesc(ldesc->position));
    if (ldesc->attnflags & LOBJ_LIGHT_ATTN) {
      HSD_LObjSetFlags   (lobj, LOBJ_RAW_PARAM);
      HSD_LObjSetAttnK   (lobj, ldesc->u.attn->dist.k0,
				ldesc->u.attn->dist.k1,
				ldesc->u.attn->dist.k2);
    } else {
      HSD_LObjSetDistAttn(lobj, ldesc->u.point->ref_dist,
			        ldesc->u.point->ref_br,
			        ldesc->u.point->dist_func);
    }
    break;

  case LOBJ_SPOT:
    HSD_LObjSetPositionWObj(lobj, HSD_WObjLoadDesc(ldesc->position));
    HSD_LObjSetInterestWObj(lobj, HSD_WObjLoadDesc(ldesc->interest));
    if (ldesc->attnflags & LOBJ_LIGHT_ATTN) {
      HSD_LObjSetFlags   (lobj, LOBJ_RAW_PARAM);
      HSD_LObjSetAttn    (lobj, ldesc->u.attn->angle.a0,
			        ldesc->u.attn->angle.a1,
			        ldesc->u.attn->angle.a2,
			        ldesc->u.attn->dist.k0,
			        ldesc->u.attn->dist.k1,
			        ldesc->u.attn->dist.k2);
    } else {
      HSD_LObjSetDistAttn(lobj, ldesc->u.spot->ref_dist,
			        ldesc->u.spot->ref_br,
			        ldesc->u.spot->dist_func);
      HSD_LObjSetSpot    (lobj, ldesc->u.spot->cutoff,
				ldesc->u.spot->spot_func);
    }
    break;

  default:
    HSD_Report("unexpected lightdesc flags (%x)\n", ldesc->flags);
    HSD_Halt("");
  }
    
#if 0
  HSD_Report("LObjLightDesc out (%x)\n", lobj);
#endif

  return 0;
}

/**
 * Light Desc [hB
 * 
 * @see HSD_LObjRemoveAll
 * @see HSD_LObjSetDefaultClass
 */
HSD_LObj *
HSD_LObjLoadDesc(HSD_LightDesc *ldesc)
{
  HSD_LObj *top, **p = &top;

  
#ifdef XDEBUG
  HSD_Report("loadLightDesc in ldesc (%x)\n", ldesc);
#endif
  for (; ldesc; ldesc=ldesc->next) {
    HSD_ClassInfo *info;

    if (!ldesc->class_name ||
	!(info = hsdSearchClassInfo(ldesc->class_name)))
    {
      *p = HSD_LObjAlloc();
    } else {
      *p = hsdNew(info);
      assert(*p);
    }
    HSD_LOBJ_METHOD(*p)->load(*p, ldesc);
    p = &(*p)->next;
  }
  *p = NULL;

  return top;
}

/**
 * 1 lobj ɑ΂ăAj[Vo^B
 * 
 * @see HSD_LObjAddAnimAll
 */
void
HSD_LObjAddAnim(HSD_LObj *lobj, HSD_LightAnim *lanim)
{
  if (lobj == NULL)
    return;
  
  if (lanim) {
    if (lobj->aobj) {
      HSD_AObjRemove(lobj->aobj);
    }
    lobj->aobj = HSD_AObjLoadDesc(lanim->aobjdesc);
    HSD_WObjAddAnim(HSD_LObjGetPositionWObj(lobj), lanim->position_anim);
    HSD_WObjAddAnim(HSD_LObjGetInterestWObj(lobj), lanim->interest_anim);
  }
}

/**
 * AׂĂ lobj ɑ΂ăAj[Vo^B
 * 
 * @see HSD_LObjAddAnim
 */
void
HSD_LObjAddAnimAll(HSD_LObj *lobj, HSD_LightAnim *lanim)
{
  HSD_LObj *lp;
  HSD_LightAnim *la;

  if (lobj == NULL)
    return;

  for (lp = lobj, la = lanim; lp; lp = next_p(lp), la = next_p(la)) {
    HSD_LObjAddAnim(lp, la);
  }
}

static void
LObjRelease(HSD_Class *o)
{
  HSD_LObj *lobj = HSD_LOBJ(o);

  HSD_AObjRemove(lobj->aobj);
  HSD_WObjUnref(HSD_LObjGetPositionWObj(lobj));
  HSD_WObjUnref(HSD_LObjGetInterestWObj(lobj));

  HSD_PARENT_INFO(&hsdLObj)->release(o);
}

static void
LObjAmnesia(HSD_ClassInfo *info)
{
  if (info == HSD_CLASS_INFO(default_class)) {
    default_class = NULL;
  }
  if (info == HSD_CLASS_INFO(&hsdLObj)) {
    current_lights = NULL;
  }
  HSD_PARENT_INFO(&hsdLObj)->amnesia(info);
}

static void
LObjInfoInit(void)
{
  hsdInitClassInfo(HSD_CLASS_INFO(&hsdLObj),
  		   HSD_CLASS_INFO(&hsdObj),
		   HSD_BASE_CLASS_LIBRARY,
		   "hsd_lobj",
		   sizeof(HSD_LObjInfo),
		   sizeof(HSD_LObj));
  
  HSD_CLASS_INFO(&hsdLObj)->release = LObjRelease;
  HSD_CLASS_INFO(&hsdLObj)->amnesia = LObjAmnesia;

  HSD_LOBJ_INFO(&hsdLObj)->load     = LObjLoad;
  HSD_LOBJ_INFO(&hsdLObj)->update   = LObjUpdateFunc;
}
