/*
 * Copyright 1999-2001 HAL Laboratory, Inc. All rights reserved.
 *
 * $Id: pobj.c,v 1.17 2004/08/06 09:26:03 TypeR Exp $
 */
#ifdef FUJITA
#pragma cplusplus			off
#pragma cpp_extensions		off
#pragma warn_implicitconv	off
#endif

#include <string.h>
#include "sysdolphin/baselib/pobj.h"
#include "sysdolphin/baselib/aobj.h"
#include "sysdolphin/baselib/list.h"
#include "sysdolphin/baselib/memory.h"
#include "sysdolphin/baselib/jobj.h"
#include "sysdolphin/baselib/perf.h"
#include "sysdolphin/baselib/displayfunc.h"
#include "sysdolphin/baselib/state.h"
#include "sysdolphin/baselib/debug.h"
#include "sysdolphin/baselib/initialize.h"

typedef enum {
  SETUP_NORMAL		  = 1 << 0,
  SETUP_REFLECTION	  = 1 << 1,
  SETUP_HIGHLIGHT	  = 1 << 2,
  SETUP_NORMAL_PROJECTION = SETUP_REFLECTION | SETUP_HIGHLIGHT,

  SETUP_JOINT0		  = 1 << 0,
  SETUP_JOINT1		  = 1 << 1,

  SETUP_NONE		  = 0
} PObjSetupFlag;

static void PObjInfoInit(void);

HSD_PObjInfo hsdPObj = { PObjInfoInit };

static HSD_PObjInfo *default_class = NULL;

static f32 (*vertex_buffer)[3] = NULL;
static f32 (*normal_buffer)[3] = NULL;
static u32 vertex_buffer_size = 0;
static u32 normal_buffer_size = 0;

#if WATANABE
static f32 (*draw_vertex_buffer)[3] = NULL;
static f32 (*draw_normal_buffer)[3] = NULL;
static u32 draw_vertex_buffer_size = 0;
static u32 draw_normal_buffer_size = 0;
#endif

static HSD_VtxDescList *prev_vtxdesclist_array = NULL;
static HSD_VtxDescList *prev_vtxdesc = NULL;

static struct {
  void *obj;
  u32 mark;
} mtx_mark[2];

#if DOLPHIN_API_AGE < 1
#define GXSetCurrentMtx(id)	GXSetDefaultMtx(id)
#endif

#if FUJITA
static s32 shadow_mode = 0;
#endif

/**
 * @package Global
 *
 * VFCvAj[VɎgp钸_̍WzꎞIɕێ
 * obt@w肷B
 * ̃obt@w肳ĂȂꍇAŏ̃VFCvAj[V
 * sHSD_Alloc()ɂftHgTCỸobt@B
 *
 * @param buf		obt@p̈ւ̃|C^
 * @param nitems	̈ŕ\钸_(sizeof(buf) / sizeof(float[3]))
 * @return		ȑÕobt@AhXBԂłNULLB
 *
 * @tip
 * ftHgobt@TCYconfig.hŒ`ĂB
 */
void *
HSD_SetVertexBuffer(f32 (*buf)[3], u32 nitems)
{
  void *old = vertex_buffer;

  vertex_buffer = buf;
  vertex_buffer_size = nitems;

  return old;
}
#if WATANABE
void *
HSD_SetDrawVertexBuffer(f32 (*buf)[3], u32 nitems);
void *
HSD_SetDrawVertexBuffer(f32 (*buf)[3], u32 nitems)
{
  void *old = draw_vertex_buffer;

  draw_vertex_buffer = buf;
  draw_vertex_buffer_size = nitems;

  return old;
}
#endif

/**
 * @package Global
 *
 * VFCvAj[VɎgp@xNgzꎞIɕێ
 * obt@w肷B
 * ̃obt@w肳ĂȂꍇAŏ̃VFCvAj[V
 * sHSD_Alloc()ɂftHgTCỸobt@B
 *
 * @param buf		obt@p̈ւ̃|C^
 * @param nitems	̈ŕ\@(sizeof(buf) / sizeof(float[3]))
 * @return		ȑÕobt@AhXBԂłNULLB
 *
 * @tip
 * ftHgobt@TCYconfig.hŒ`ĂB
 * 
 * @warning
 * Normal̂NBTgpꍇAobt@Normal̏ꍇ
 * 3{KvɂȂB
 */
void *
HSD_SetNormalBuffer(f32 (*buf)[3], u32 nitems)
{
  void *old = normal_buffer;

  normal_buffer = buf;
  normal_buffer_size = nitems;

  return old;
}
#if WATANABE
void *
HSD_SetDrawNormalBuffer(f32 (*buf)[3], u32 nitems);
void *
HSD_SetDrawNormalBuffer(f32 (*buf)[3], u32 nitems)
{
  void *old = draw_normal_buffer;

  draw_normal_buffer = buf;
  draw_normal_buffer_size = nitems;

  return old;
}
#endif

/** @package PObj */

/* @@@ SHAPEANIM Ή @@@ */
#ifdef SHAPETEST
float SH_TEST_SHAPE_MAX = 0.0F;
u16   SH_TEST_SHAPE_FLAGS;
float *SH_TEST_SHAPE_BP;
#endif

/**
 * pobj̃tO擾.
 *
 * @see HSD_PObjSetFlags
 * @see HSD_PObjClearFlags
 * @see HSD_PObjModifyFlags
 */
u32
HSD_PObjGetFlags(HSD_PObj *pobj)
{
  return (pobj)? pobj->flags : 0;
}

/**
 * pobj  flags ZbgB
 *
 * @see HSD_PObjGetFlags
 * @see HSD_PObjClearFlags
 * @see HSD_PObjModifyFlags
 */
void
HSD_PObjSetFlags(HSD_PObj *pobj, u32 flags)
{
  if (pobj) pobj->flags |= flags;
}

/**
 * pobj  flags NAB
 *
 * @see HSD_PObjSetFlags
 * @see HSD_PObjGetFlags
 * @see HSD_PObjModifyFlags
 */
void
HSD_PObjClearFlags(HSD_PObj *pobj, u32 flags)
{
  if (pobj) pobj->flags &= ~flags;
}

/**
 *  mask Ŏw肵tO flags ̒lɐݒ肷B
 *
 * @see HSD_PObjSetFlags
 * @see HSD_PObjGetFlags
 * @see HSD_PObjClearFlags
 */
void
HSD_PObjModifyFlags(HSD_PObj *pobj, u32 flags, u32 mask)
{
  if (pobj) pobj->flags = (pobj->flags & ~mask) | (flags & mask);
}

/**
 * 1 pobj Ƃ̉̍\̂ flags Ŏw肳ꂽAj[V폜B
 * 
 * @see HSD_PObjRemoveAnim
 * @see HSD_PObjRemoveAnimAllByFlags
 */
void
HSD_PObjRemoveAnimByFlags(HSD_PObj *pobj, u32 flags)
{
  if (pobj == NULL)
    return;
    
#if FUJITA
  if (pobj_type(pobj) != POBJ_SHAPEANIM) {
	//PObjUpdateFunc()̂ƂAVFCvȊOɂ̓Aj͂ȂB
	return;
  }
#endif
  if (flags & POBJ_ANIM) {
    HSD_AObjRemove(pobj->aobj);
    pobj->aobj = NULL;
  }
}

/**
 * AꂽׂĂ pobj  flags Ŏw肳ꂽAj[V폜B
 * 
 * @see HSD_PObjRemoveAnimByFlags
 * @see HSD_PObjRemoveAnimAll
 */
void
HSD_PObjRemoveAnimAllByFlags(HSD_PObj *pobj, u32 flags)
{
  HSD_PObj *pp;
  
  if (pobj == NULL)
    return;

#if FUJITA
  if (pobj_type(pobj) != POBJ_SHAPEANIM) {
	//PObjUpdateFunc()̂ƂAVFCvȊOɂ̓Aj͂ȂB
	return;
  }
#endif
  for (pp = pobj; pp; pp = pp->next) {
    HSD_PObjRemoveAnimByFlags(pp, flags);
  }
}

/**
 * 1 pobj Ƃ̉̍\̂SẴAj[V폜B
 * 
 * @see HSD_PObjRemoveAnimByFlags
 * @see HSD_PObjRemoveAnimAll
 */
void
HSD_PObjRemoveAnim(HSD_PObj *pobj)
{
#if FUJITA
  if (pobj_type(pobj) != POBJ_SHAPEANIM) {
	//PObjUpdateFunc()̂ƂAVFCvȊOɂ̓Aj͂ȂB
	return;
  }
#endif
  HSD_PObjRemoveAnimByFlags(pobj, POBJ_ANIM);
}

/**
 * AꂽׂĂ pobj Aj[V폜B
 * 
 * @see HSD_PObjRemoveAnim
 * @see HSD_PObjRemoveAnimAllByFlags
 */
void
HSD_PObjRemoveAnimAll(HSD_PObj *pobj)
{
#if FUJITA
  if (pobj_type(pobj) != POBJ_SHAPEANIM) {
	//PObjUpdateFunc()̂ƂAVFCvȊOɂ̓Aj͂ȂB
	return;
  }
#endif
  HSD_PObjRemoveAnimAllByFlags(pobj, POBJ_ANIM);
}


/**
 * 1 pobj Ƃ̉̍\̂ɑ΂Aj[Vo^B
 * 
 * @see HSD_PObjAddAnimAll
 */
void
HSD_PObjAddAnim(HSD_PObj *pobj, HSD_ShapeAnim *shapeanim)
{
  if (pobj == NULL)
    return;

  if (pobj->aobj) {
    HSD_AObjRemove(pobj->aobj);
  }
  pobj->aobj = HSD_AObjLoadDesc(shapeanim->aobjdesc);
}

/**
 * AꂽׂĂ pobj ɑ΂ăAj[Vo^B
 * 
 * @see HSD_PObjAddAnim
 */
void
HSD_PObjAddAnimAll(HSD_PObj *pobj, HSD_ShapeAnim *shapeanim)
{
  HSD_PObj *pp;
  HSD_ShapeAnim *sa;

  if (pobj == NULL || shapeanim == NULL)
    return;

  for (pp = pobj, sa = shapeanim; pp; pp = pp->next, sa = next_p(sa)) {
    HSD_PObjAddAnim(pp, sa);
  }
}

/**
 * 1 pobj Ƃ̉̍\̂ɑ΂flags Ŏw肳ꂽAj[
 * V̊JnwB
 * 
 * @see HSD_PObjReqAnim
 * @see HSD_PObjReqAnimAllByFlags
 */
void
HSD_PObjReqAnimByFlags(HSD_PObj *pobj, float startframe, u32 flags)
{
#if defined(XDEBUG) || defined(SDEBUG)
  HSD_Report("PObjReqAnim in\n");
#endif
  if (pobj == NULL)
    return;
#if FUJITA
  if (pobj_type(pobj) != POBJ_SHAPEANIM) {
	//PObjUpdateFunc()̂ƂAVFCvȊOɂ̓Aj͂ȂB
	return;
  }
#endif
  
  if (flags & POBJ_ANIM) {
    HSD_AObjReqAnim(pobj->aobj, startframe);
  }

#if defined(XDEBUG) || defined(SDEBUG)
  HSD_Report("PObjReqAnim end\n");
#endif
}

/**
 * AꂽׂĂ pobj ɑ΂flags Ŏw肳ꂽAj[V̊JnwB
 * 
 * @see HSD_PObjReqAnimAll
 * @see HSD_PObjReqAnimByFlags
 */
void
HSD_PObjReqAnimAllByFlags(HSD_PObj *pobj, float startframe, u32 flags)
{
  HSD_PObj *pp;

  if (pobj == NULL)
    return;

  for (pp = pobj; pp; pp = pp->next) {
    HSD_PObjReqAnimByFlags(pp, startframe, flags);
  }
}

/**
 * AꂽׂĂ pobj ɑ΂ăAj[V̊JnwB
 * 
 * @see HSD_PObjReqAnim
 * @see HSD_PObjReqAnimAllByFlags
 */
void
HSD_PObjReqAnimAll(HSD_PObj *pobj, float startframe)
{
  HSD_PObjReqAnimAllByFlags(pobj, startframe, POBJ_ANIM);
}

/**
 * pobj ȉ̊Kwɑ΂ăAj[V̊JnwB
 * 
 * @see HSD_PObjReqAnimAll
 * @see HSD_PObjReqAnimByFlags
 */
void
HSD_PObjReqAnim(HSD_PObj *pobj, float startframe)
{
  HSD_PObjReqAnimByFlags(pobj, startframe, POBJ_ANIM);
}

static void
ShapeSetSetAnimResult(HSD_ShapeSet *shape_set, u32 type, FObjData *val)
{
#ifdef SDEBUG
  HSD_Report("ShapeSetSetAnimResult in\n");
#endif
  if (shape_set->flags & SHAPESET_ADDITIVE) {
#ifdef SHXDEBUG
    HSD_Report("HSD_A_S_W%d val->fv (%f)\n", type-HSD_A_S_W0, val->fv);
#endif
    shape_set->blend.bp[type-HSD_A_S_W0] = val->fv;
  } else {
    shape_set->blend.bl = val->fv;
  }
}

/*
 * P pobj Ƃ̉̍\̂ɑ΂ăAj[V̌ʂ\̂ɔf
 */
static void
PObjUpdateFunc(void *obj, u32 type, FObjData *val)
{
  HSD_PObj *pobj = obj;

#ifdef SDEBUG
  HSD_Report("PObjSetAnimResult in pobj (%x)\n", pobj);
#endif
  if (pobj == NULL)
    return;
  
  if (pobj_type(pobj) == POBJ_SHAPEANIM) {
    HSD_ShapeSet *shape_set = pobj->u.shape_set;
    
#ifdef SDEBUG
    HSD_Report("shape_set (%x) shape_set->aobj (%x)\n",
	       shape_set, shape_set->aobj);
#endif
    ShapeSetSetAnimResult(shape_set, type, val);
  }
}

/**
 * 1 pobj Ƃ̉̍\̂̃Aj[Vߎs
 * 
 * @see HSD_PObjAnimAll
 */
void
HSD_PObjAnim(HSD_PObj *pobj)
{
#if defined(XDEBUG) || defined(SDEBUG)
  HSD_Report("PObjAnim in\n");
#endif
  if (pobj == NULL)
    return;

#if FUJITA
  if (pobj_type(pobj) != POBJ_SHAPEANIM) {
	//PObjUpdateFunc()̂ƂAVFCvȊOɂ̓Aj͂ȂB
	return;
  }
#endif
#if HSD_ANIM_METHOD_LEVEL <= 0
  HSD_AObjInterpretAnim(pobj->aobj, pobj, HSD_POBJ_METHOD(pobj)->update);
#else
  HSD_AObjInterpretAnim(pobj->aobj, pobj, PObjUpdateFunc);
#endif

#if defined(XDEBUG) || defined(SDEBUG)
  HSD_Report("PObjAnim out\n");
#endif
}

/**
 * AꂽׂĂ pobj ̃Aj[Vߎs
 * 
 * @see HSD_PObjAnim
 */
void
HSD_PObjAnimAll(HSD_PObj *pobj)
{
  HSD_PObj *pp;

  if (pobj == NULL)
    return;

  for (pp = pobj; pp; pp = pp->next) {
    HSD_PObjAnim(pp);
  }
}

static HSD_Envelope *
HSD_EnvelopeAlloc(void)
{
  HSD_Envelope *envelope = hsdAllocMemPiece(sizeof(HSD_Envelope));
  assert(envelope);
  memset(envelope, 0, sizeof(HSD_Envelope));
  return envelope;
}

static void
HSD_EnvelopeFree(HSD_Envelope *env)
{
  if (env) hsdFreeMemPiece(env, sizeof(HSD_Envelope));
}

static void
HSD_EnvelopeListFree(HSD_SList *list)
{
  while (list) {
    HSD_Envelope *env = list->data;
    while (env) {
      HSD_Envelope *next = env->next;
      HSD_JObjUnrefThis(env->jobj);
      HSD_EnvelopeFree(env);
      env = next;
    }
    list = HSD_SListRemove(list);
  }
}

static HSD_SList *
loadEnvelopeDesc(HSD_EnvelopeDesc **edesc_p)
{
  HSD_SList *list = NULL, **list_p = &list;

  if (!edesc_p) return NULL;

  while (*edesc_p) {
    HSD_Envelope *envelope = NULL, **env_p = &envelope;
    HSD_EnvelopeDesc *edesc = *edesc_p;

    while (edesc->joint) {
      (*env_p) = HSD_EnvelopeAlloc();
      (*env_p)->weight = edesc->weight;
      env_p = &((*env_p)->next);
      edesc++;
    }

    (*list_p) = HSD_SListAlloc();
    (*list_p)->data = envelope;
    list_p = &((*list_p)->next);
    edesc_p++;
  }

  return list;
}

static HSD_ShapeSet *
HSD_ShapeSetAlloc(void)
{
  HSD_ShapeSet *shape_set = hsdAllocMemPiece(sizeof(HSD_ShapeSet));
  assert(shape_set);
  memset(shape_set, 0, sizeof(HSD_ShapeSet));
  return shape_set;
}

static void
HSD_ShapeSetFree(HSD_ShapeSet *shape_set)
{
  if (!shape_set) return;
  hsdFreeMemPiece(shape_set, sizeof(HSD_ShapeSet));
}

static void
HSD_ShapeSetRemove(HSD_ShapeSet *shape_set)
{
  if (!shape_set) return;
  
  if (shape_set->flags & SHAPESET_ADDITIVE)
    HSD_Free(shape_set->blend.bp);

  HSD_ShapeSetFree(shape_set);
}

static HSD_ShapeSet *
loadShapeSetDesc(HSD_ShapeSetDesc *sdesc)
{
  HSD_ShapeSet *shape_set = HSD_ShapeSetAlloc();
  shape_set->flags = sdesc->flags;
  shape_set->nb_shape = sdesc->nb_shape;
#ifdef SHAPETEST
  if (SH_TEST_SHAPE_MAX == 0.0F) {
    SH_TEST_SHAPE_MAX = (float)(shape_set->nb_shape - 1);
    SH_TEST_SHAPE_FLAGS = sdesc->flags;
    if (shape_set->flags & SHAPESET_ADDITIVE)
      SH_TEST_SHAPE_BP = (float *)HSD_Alloc(shape_set->nb_shape * sizeof(float));
  }
#endif
  shape_set->nb_vertex_index = sdesc->nb_vertex_index;
  shape_set->vertex_desc = sdesc->vertex_desc;
  shape_set->vertex_idx_list = sdesc->vertex_idx_list;
  shape_set->nb_normal_index = sdesc->nb_normal_index;
  shape_set->normal_desc = sdesc->normal_desc;
  shape_set->normal_idx_list = sdesc->normal_idx_list;
  if (shape_set->flags & SHAPESET_ADDITIVE) {
    int i;

    shape_set->blend.bp = (float *)HSD_Alloc(shape_set->nb_shape * sizeof(float));
    for (i = 0; i < shape_set->nb_shape; i++)
      shape_set->blend.bp[i] = 0.0F;
  } else
    shape_set->blend.bl = 0.0F;
  return shape_set;
}

static int
PObjLoad(HSD_PObj *pobj, HSD_PObjDesc *desc)
{
  pobj->next = HSD_PObjLoadDesc(desc->next);
  pobj->desc_list = desc->desc_list;
  pobj->flags = desc->flags;
  pobj->dlsize = desc->dlsize;
  pobj->dl =  desc->dl;
#if FUJITA
  pobj->aobj = NULL;
#endif
  switch (pobj_type(pobj)) {
  case POBJ_SHAPEANIM:
    pobj->u.shape_set = loadShapeSetDesc(desc->u.shape_set);
    break;
  case POBJ_ENVELOPE:
    pobj->u.envelope_list = loadEnvelopeDesc(desc->u.envelope_p);
    break;
  case POBJ_SKIN:
    break;
  default:
    HSD_Halt("pobj: unexected type.\n");
  }
  /*
   * _LbV̖KvłƐ錾B
   */
  _HSD_NeedCacheInvalidate(HSD_CACHE_VTX);

  return 0;
}

/**
 * PObjDesc [h
 * 
 * @see HSD_PObjRemove
 * @see HSD_PObjSetDefaultClass
 */
HSD_PObj *
HSD_PObjLoadDesc(HSD_PObjDesc *pobjdesc)
{
  if (pobjdesc != NULL) {
    HSD_PObj *pobj;
    HSD_ClassInfo *info;
    
    if (!pobjdesc->class_name ||
        !(info = hsdSearchClassInfo(pobjdesc->class_name)))
    {
      pobj = HSD_PObjAlloc();
    } else {
      pobj = hsdNew(info);
      assert(pobj);
    }
    HSD_POBJ_METHOD(pobj)->load(pobj, pobjdesc);
    return pobj;
  } else {
    return NULL;
  }
}

/**
 * pobj  next  new A
 * 
 * @warning
 * new  next ͖B
 */
void
HSD_PObjAddNext(HSD_PObj *pobj, HSD_PObj *new)
{
  if (pobj == NULL || new == NULL)
    return;

  new->next = pobj->next;
  pobj->next = new;
}

/**
 * 1 pobj Ƃ̉̍\̂폜
 * 
 * @see HSD_PObjAlloc
 * @see HSD_PObjLoadDesc
 * @see HSD_PObjRemoveAll
 */
void
HSD_PObjRemove(HSD_PObj *pobj)
{
  hsdDelete(pobj);
}

/**
 * AꂽׂĂ pobj 폜
 * 
 * @see HSD_PObjRemove
 */
void
HSD_PObjRemoveAll(HSD_PObj *pobj)
{
  while (pobj) {
    HSD_PObj *next = pobj->next;
    HSD_PObjRemove(pobj);
    pobj = next;
  }
}

/**
 * ftHgNX̎w
 * 
 * pobjƂăftHgŐIuWFNg̃NXw肷B
 * 
 * włNXhsdPObjA̓oNXłȂ΂ȂȂB
 * 
 * @see HSD_PObjGetDefaultClass
 * @see HSD_PObjAlloc
 * @see HSD_PObjLoadDesc
 */
void
HSD_PObjSetDefaultClass(HSD_PObjInfo *info)
{
  if (info) {
    assert(hsdIsDescendantOf(info, &hsdPObj));
  }
  default_class = info;
}

/**
 * ftHgNX̎擾
 * 
 * pobjƂăftHgŐIuWFNg̃NX擾B
 * 
 * @see HSD_PObjSetDefaultClass
 * @see HSD_PObjAlloc
 * @see HSD_PObjLoadDesc
 */
HSD_PObjInfo *
HSD_PObjGetDefaultClass(void)
{
  return default_class? default_class : &hsdPObj;
}

/**
 * pobj P쐬
 *
 * @see HSD_PObjRemove
 * @see HSD_PObjSetDefaultClass
 */
HSD_PObj *
HSD_PObjAlloc(void)
{
  HSD_PObj *pobj = hsdNew(HSD_PObjGetDefaultClass());
  assert(pobj);
  return pobj;
}

/**
 * pobj j
 * 
 * @warning
 * {֐̓oɂčlȂBʏpobj̍폜ɂHSD_PObjRemove()
 * gp邱ƁB
 *
 * @see HSD_PObjRemove
 */
void
HSD_PObjFree(HSD_PObj *pobj)
{
  if (pobj)
    HSD_CLASS_METHOD(pobj)->destroy(HSD_CLASS(pobj));
}

static void
resolveEnvelope(HSD_SList *list, HSD_EnvelopeDesc **edesc_p)
{
  if (!list || !edesc_p)
    return;

  for (; list && *edesc_p; list=list->next, edesc_p++) {
    HSD_Envelope *env = list->data;
    HSD_EnvelopeDesc *edesc = *edesc_p;

    while (env && edesc->joint) {
      HSD_JObjUnrefThis(env->jobj);
      env->jobj = HSD_IDGetData((HSD_ID)edesc->joint, NULL);
      assert(env->jobj);
      HSD_JObjRefThis(env->jobj);
      env = env->next;
      edesc++;
    }
  }
}

/**
 * pobjɊւQƂB
 */
void
HSD_PObjResolveRefs(HSD_PObj *pobj, HSD_PObjDesc *pdesc)
{
  if (!pobj || !pdesc)
    return;

  switch (pobj_type(pobj)) {
  case POBJ_ENVELOPE:
    resolveEnvelope(pobj->u.envelope_list, pdesc->u.envelope_p);
    break;

  case POBJ_SKIN:
    HSD_JObjUnrefThis(pobj->u.jobj);
    pobj->u.jobj = NULL;
    if (pdesc->u.joint) {
      pobj->u.jobj = HSD_IDGetData((HSD_ID)pdesc->u.joint, NULL);
      assert(pobj->u.jobj);
      HSD_JObjRefThis(pobj->u.jobj);
    }
    break;

  default:
    break;
  }
}

/**
 * AꂽׂĂpobjɂāAQƂB
 */
void
HSD_PObjResolveRefsAll(HSD_PObj *pobj, HSD_PObjDesc *pdesc)
{
  while (pobj && pdesc) {
    HSD_PObjResolveRefs(pobj, pdesc);
    pobj = pobj->next;
    pdesc = pdesc->next;
  }
}

/**
 * pobj ȉŕ\钸_EgAO߂B
 * 
 * Rp`ȊÕv~eBu͂Rp`ɊZB
 */
void
HSD_PObjCountVertices(HSD_PObj *pobj, u32 *nbVertices, u32 *nbTriangles)
{
  u32 nbVtx = 0;
  u32 nbTri = 0;
  for (; pobj; pobj = pobj->next) {
    int i, dlsize;
    u8 *dl;
    HSD_VtxDescList *desc_list;
    int desc_size;
    int kind;

    desc_size = 0;
    for (desc_list = pobj->desc_list;
	 desc_list->attr != GX_VA_NULL; desc_list++) {

      switch (desc_list->attr_type) {
      case GX_INDEX8:
	desc_size++;
	break;
      case GX_INDEX16:
	desc_size += 2;
	break;
      case GX_DIRECT:
	{
	  int s;

	  if (desc_list->attr == GX_VA_CLR0 ||
	      desc_list->attr == GX_VA_CLR1) {
	    switch (desc_list->comp_type) {
	    case GX_RGB565:
	    case GX_RGBA4:
	      s = 2;
	      break;
	    case GX_RGB8:
	    case GX_RGBA6:
	      s = 3;
	      break;
	    case GX_RGBA8:
	    case GX_RGBX8:
	      s = 4;
	      break;
	    }
	  } else {
	    switch (desc_list->comp_type) {
	    case GX_U8:
	    case GX_S8:
	      s = 1;
	      break;
	    case GX_U16:
	    case GX_S16:
	      s = 2;
	      break;
	    case GX_F32:
	      s = 4;
	      break;
	    default:
	      goto nextpobj;
	    }
	  }
	  switch (desc_list->attr) {
	  case GX_VA_PNMTXIDX:
	    desc_size++;
	    break;
	  case GX_VA_POS:
	    if (desc_list->comp_cnt == GX_POS_XY)
	      desc_size += s*2;
	    else
	      desc_size += s*3;
	    break;
	  case GX_VA_NRM:
	    if (desc_list->comp_cnt == GX_NRM_XYZ)
	      desc_size += s*3;
	    break;
	  case GX_VA_CLR0:
	  case GX_VA_CLR1:
	    desc_size += s;
	    break;
	  case GX_VA_TEX0:
	  case GX_VA_TEX1:
	  case GX_VA_TEX2:
	  case GX_VA_TEX3:
	  case GX_VA_TEX4:
	  case GX_VA_TEX5:
	  case GX_VA_TEX6:
	  case GX_VA_TEX7:
	    if (desc_list->comp_cnt == GX_TEX_S)
	      desc_size += s;
	    else
	      desc_size += s*2;
	    break;
	  default:
	    goto nextpobj;
	  }
	}
      }
    }
    dlsize = (u32)pobj->dlsize << 5;
    dl = pobj->dl;
    if (dl == NULL)
      continue;
    for (i = 0; i < dlsize;) {
      u32 v;

      kind = dl[i] & GX_OPCODE_MASK;
      i++;
      if (kind == GX_NOP) {
	continue;
      }

      v = ((u16)dl[i] << 8) + (u16)dl[i+1];
      i += 2 + v * desc_size;

      nbVtx += v;
      switch (kind) {
      case GX_DRAW_TRIANGLES:
	nbTri += v / 3;
	break;
      case GX_DRAW_TRIANGLE_STRIP:
      case GX_DRAW_TRIANGLE_FAN:
	nbTri += v - 2;
	break;
      case GX_DRAW_QUADS:
	nbTri += v / 2;
	break;
      }
    }
  nextpobj:
    ;
  }
  if (nbVertices)
    *nbVertices = nbVtx;
  if (nbTriangles)
    *nbTriangles = nbTri;
  return;
}

/**
 * _ݒ̃NA
 * 
 * baselib LbVOĂ钸_ݒNAB
 * 
 * @warning
 * GXSetArray(), GXSetVtxDesc(), GXSetVtxAttrFmt()Ȃǂ
 * gpꍇA{֐păLbV𖳌Ȃ΂ȂȂB
 */
void
HSD_ClearVtxDesc(void)
{
  GXClearVtxDesc();
  prev_vtxdesclist_array = NULL;
  prev_vtxdesc = NULL;
}

static void
setupArrayDesc(HSD_VtxDescList *desc_list)
{
  HSD_VtxDescList *desc;

  if (prev_vtxdesclist_array != desc_list) {
    for (desc = desc_list; desc->attr != GX_VA_NULL; desc++) {
	  //#if FUJITA
#if 0
	  if( shadow_mode && !(desc->attr & (GX_VA_PNMTXIDX|GX_VA_POS))){
	  }else
#endif
		if (desc->attr_type != GX_DIRECT) {
		  GXSetArray(desc->attr, desc->base_ptr, desc->stride);
		}
    }
    prev_vtxdesclist_array = desc_list;
  }
}

static void
setupVtxDesc(HSD_PObj *pobj)
{
  HSD_VtxDescList *desc;

  if (prev_vtxdesc != pobj->desc_list) {
    GXClearVtxDesc();
    for (desc = pobj->desc_list; desc->attr != GX_VA_NULL; desc++) {
	  //#if FUJITA
#if 0
//	  if( shadow_mode && !(desc->attr & (GX_VA_PNMTXIDX|GX_VA_POS)))
//		continue;
	  if( shadow_mode && !(desc->attr & (GX_VA_PNMTXIDX|GX_VA_POS))) {
		GXSetVtxDesc(desc->attr, desc->attr_type);
		continue;
	  }
#endif

      GXSetVtxDesc(desc->attr, desc->attr_type);
      switch (desc->attr) {
      case GX_VA_PNMTXIDX:
      case GX_VA_TEX0MTXIDX:
      case GX_VA_TEX1MTXIDX:
      case GX_VA_TEX2MTXIDX:
      case GX_VA_TEX3MTXIDX:
      case GX_VA_TEX4MTXIDX:
      case GX_VA_TEX5MTXIDX:
      case GX_VA_TEX6MTXIDX:
      case GX_VA_TEX7MTXIDX:
	/*
	 * ȏ̃^Cv͌^u8ɌŒ肳Ă̂ŁA
	 * GXSetVtxAttrFmt()Ă͂ȂB
	 */
        break;
      default:
	GXSetVtxAttrFmt(GX_VTXFMT0, desc->attr,
			desc->comp_cnt, desc->comp_type, desc->frac);
      }
    }
    prev_vtxdesc = pobj->desc_list;
  }
}

/* VFCvAj[V̏ꍇAPOS/NRM/NBT̓^C
 * uĥŁAz͐錾Ȃ
 */
static void
setupShapeAnimArrayDesc(HSD_VtxDescList *desc_list)
{
  HSD_VtxDescList *desc;

  for (desc = desc_list; desc->attr != GX_VA_NULL; desc++) {
    if (desc->attr_type != GX_DIRECT) {
      switch (desc->attr) {
#ifdef WATANABE
      case GX_VA_POS:
	GXSetArray(desc->attr, (void *)draw_vertex_buffer, desc->stride);
//	GXSetArray(desc->attr, (void *)vertex_buffer, desc->stride);
      	break;
      case GX_VA_NRM:
	GXSetArray(desc->attr, (void *)draw_normal_buffer, desc->stride);
//	GXSetArray(desc->attr, (void *)normal_buffer, desc->stride);
      	break;
#else
      case GX_VA_POS:
      case GX_VA_NRM:
#endif
      case GX_VA_NBT:
	break;
      default:
	GXSetArray(desc->attr, desc->base_ptr, desc->stride);
      }
    }
  }
  prev_vtxdesclist_array = NULL;
}

/* VFCvAj[V̏ꍇAPOS/NRM/NBT̓^C
 * uĥŁAattr_type  GX_INDEX{8,16}̑
 * GX_DIRECTɐݒ肷B
 */
static void
setupShapeAnimVtxDesc(HSD_PObj *pobj)
{
  HSD_VtxDescList *desc;

  GXClearVtxDesc();
  for (desc = pobj->desc_list; desc->attr != GX_VA_NULL; desc++) {
    switch (desc->attr) {
#ifdef WATANABE
    case GX_VA_NRM:
    case GX_VA_POS:
      GXSetVtxDesc(desc->attr, desc->attr_type);
      GXSetVtxAttrFmt(GX_VTXFMT0, desc->attr,
		      desc->comp_cnt, GX_F32, 0);
      break;
#else
    case GX_VA_NRM:
    case GX_VA_POS:
#endif
    case GX_VA_NBT:
      GXSetVtxDesc(desc->attr, GX_DIRECT);
      GXSetVtxAttrFmt(GX_VTXFMT0, desc->attr,
		      desc->comp_cnt, GX_F32, 0);
      break;

    case GX_VA_PNMTXIDX:
    case GX_VA_TEX0MTXIDX:
    case GX_VA_TEX1MTXIDX:
    case GX_VA_TEX2MTXIDX:
    case GX_VA_TEX3MTXIDX:
    case GX_VA_TEX4MTXIDX:
    case GX_VA_TEX5MTXIDX:
    case GX_VA_TEX6MTXIDX:
    case GX_VA_TEX7MTXIDX:
      /*
       * ȏ̃^Cv͌^u8ɌŒ肳Ă̂ŁA
       * GXSetVtxAttrFmt()Ă͂ȂB
       */
      GXSetVtxDesc(desc->attr, desc->attr_type);
      break;

    default:
      GXSetVtxDesc(desc->attr, desc->attr_type);
      GXSetVtxAttrFmt(GX_VTXFMT0, desc->attr,
		      desc->comp_cnt, desc->comp_type, desc->frac);
    }
  }
  prev_vtxdesc = NULL;
}

/*
 * _ϊ}gNXir[CO}gNX݁j
 * @ϊ}gNXB
 * (PȂts̓]u)
 */
static void
mkNormalMtx(MtxPtr pmtx, MtxPtr nmtx)
{
#ifdef NRMMTX_NOTINVERSE
  MTXCopy(pmtx, nmtx);
#else
  if (!HSD_MtxInverseTranspose(pmtx, nmtx)) {
    MTXCopy(pmtx, nmtx);
  }
#endif
}

/*
 * @ϊ}gNXn[hEFAɃ[hB
 */
static void
setNormalMtx(MtxPtr nmtx, u32 id)
{
  GXLoadNrmMtxImm(nmtx, id);
  HSD_PerfCountMtxLoad();
}

/* @@@ SHAPEANIM Ή @@@ */
static void
get_shape_vertex_xyz(HSD_ShapeSet *shape_set,
		     int shape_id, int arrayidx, float dst[3])
{
  u8 *index_array = shape_set->vertex_idx_list[shape_id];
  int idx;
  void *src_base;

  if (shape_set->vertex_desc->attr_type == GX_INDEX16) {
    idx = index_array[arrayidx*2];
    idx = (idx<<8) + index_array[arrayidx*2+1];
  } else {
    idx = index_array[arrayidx];
  }

  assert(shape_set->vertex_desc->comp_cnt == GX_POS_XYZ);
  src_base = ((u8 *) shape_set->vertex_desc->base_ptr)
		      + idx * shape_set->vertex_desc->stride;

  /* _z̃oCgI[_̓A[LeN`ˑ */
  if (shape_set->vertex_desc->comp_type == GX_F32) {
    memcpy(dst, src_base, sizeof(f32[3]));
  } else {
    int decimal_point = 1 << shape_set->vertex_desc->frac;
    switch (shape_set->vertex_desc->comp_type) {
    case GX_U8:
      {
	u8 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    case GX_S8:
      {
	s8 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    case GX_U16:
      {
	u16 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    case GX_S16:
      {
	s16 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    default:
      HSD_Halt("unexpected vertex type.\n");
    }
  }
}

static void
get_shape_normal_xyz(HSD_ShapeSet *shape_set,
		     int shape_id, int arrayidx, float dst[3])
{
  u8 *index_array = shape_set->normal_idx_list[shape_id];    
  int idx;
  void *src_base;
  
  if (shape_set->normal_desc->attr_type == GX_INDEX16) {
    idx = index_array[arrayidx*2];
    idx = (idx<<8) + index_array[arrayidx*2+1];
  } else {
    idx = index_array[arrayidx];
  }

  assert(shape_set->normal_desc->comp_cnt == GX_NRM_XYZ);
  src_base = ((u8 *) shape_set->normal_desc->base_ptr)
		      + idx * shape_set->normal_desc->stride;

  /* _z̃oCgI[_̓A[LeN`ˑ */
  if (shape_set->normal_desc->comp_type == GX_F32) {
    memcpy(dst, src_base, sizeof(f32[3]));
  } else {
    int decimal_point = 1 << shape_set->normal_desc->frac;
    switch (shape_set->normal_desc->comp_type) {
    case GX_U8:
      {
	u8 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    case GX_S8:
      {
	s8 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    case GX_U16:
      {
	u16 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    case GX_S16:
      {
	s16 *src = src_base;
	dst[0] = (float)src[0] / decimal_point;
	dst[1] = (float)src[1] / decimal_point;
	dst[2] = (float)src[2] / decimal_point;
      }
      break;
    default:
      HSD_Halt("unexpected normal type.");
    }
  }
}

static void
get_shape_nbt_xyz(HSD_ShapeSet *shape_set,
		  int shape_id, int arrayidx, float *dst)
{
  u8 *index_array = shape_set->normal_idx_list[shape_id];    
  int i, idx;
  void *src_base;
  
  assert(shape_set->normal_desc->attr == GX_VA_NBT);

  if (shape_set->normal_desc->attr_type == GX_INDEX16) {
    idx = index_array[arrayidx*2];
    idx = (idx<<8) + index_array[arrayidx*2+1];
  } else {
    idx = index_array[arrayidx];
  }

  assert(shape_set->normal_desc->comp_cnt == GX_NRM_XYZ);
  src_base = ((u8 *) shape_set->normal_desc->base_ptr)
		      + idx * shape_set->normal_desc->stride;

  /* _z̃oCgI[_̓A[LeN`ˑ */
  if (shape_set->normal_desc->comp_type == GX_F32) {
    memcpy(dst, src_base, sizeof(f32[9]));
  } else {
    int decimal_point = 1 << shape_set->normal_desc->frac;
    switch (shape_set->normal_desc->comp_type) {
    case GX_U8:
      for (i=0; i<9; i++) {
	dst[i] = (float)((u8 *)src_base)[i] / decimal_point;
      }
      break;
    case GX_S8:
      for (i=0; i<9; i++) {
	dst[i] = (float)((s8 *)src_base)[i] / decimal_point;
      }
      break;
    case GX_U16:
      for (i=0; i<9; i++) {
	dst[i] = (float)((u16 *)src_base)[i] / decimal_point;
      }
      break;
    case GX_S16:
      for (i=0; i<9; i++) {
	dst[i] = (float)((s16 *)src_base)[i] / decimal_point;
      }
      break;
    default:
      HSD_Halt("unexpected normal type.");
    }
  }
}

static void
interpretShapeAnimDisplayList(HSD_PObj *pobj,
			      float (*vertex)[3],
			      float (*normal)[3])
{
#ifdef WATANABE
//#pragma	unused(pobj)
//#pragma	unused(vertex)
//#pragma	unused(normal)
	if( vertex!=NULL && vertex_buffer_size ){
		DCStoreRange(vertex,vertex_buffer_size * sizeof(f32) * 3);
	}
	if( normal!=NULL && normal_buffer_size ){
		DCStoreRange(normal,normal_buffer_size * sizeof(f32) * 3);
	}
  GXCallDisplayList(pobj->dl, pobj->dlsize<<5);
#else
  u8 *dl = pobj->dl;
  int length = pobj->dlsize << 5;
  int l;

  for (l=0; l+3<length; ) {
    int n = dl[1] << 8 | dl[2];
    int m = 3;
    int i, j;

    if ((dl[0] & GX_OPCODE_MASK) == GX_NOP) {
      break;
    }
    GXBegin((GXPrimitive)(dl[0]&GX_OPCODE_MASK), (GXVtxFmt)(dl[0]&GX_VAT_MASK), n);
    for (i = 0; i < n; i++) {
      for (j = 0; ;j++) {
	HSD_VtxDescList *desc = &pobj->desc_list[j];
	if (desc->attr == GX_VA_NULL) {
	  break;
	} else {
	  u16 idx = dl[m++];
	  switch (desc->attr) {
	  case GX_VA_PNMTXIDX:
	  case GX_VA_TEX0MTXIDX:
	  case GX_VA_TEX1MTXIDX:
	  case GX_VA_TEX2MTXIDX:
	  case GX_VA_TEX3MTXIDX:
	  case GX_VA_TEX4MTXIDX:
	  case GX_VA_TEX5MTXIDX:
	  case GX_VA_TEX6MTXIDX:
	  case GX_VA_TEX7MTXIDX:
	    GXMatrixIndex1u8(idx);
	    break;

	  case GX_VA_POS:
	    if (desc->attr_type == GX_INDEX16) {
	      idx = (idx<<8) | dl[m++];
	    }
	    GXPosition3f32(vertex[idx][0], vertex[idx][1], vertex[idx][2]);
	    break;

	  case GX_VA_NRM:
	    if (desc->attr_type == GX_INDEX16) {
	      idx = (idx<<8) | dl[m++];
	    }
	    GXNormal3f32(normal[idx][0], normal[idx][1], normal[idx][2]);
	    break;

	  case GX_VA_NBT:
	    if (desc->attr_type == GX_INDEX16) {
	      idx = (idx<<8) | dl[m++];
	    }
	    idx *= 3;
	    GXNormal3f32(normal[idx+0][0], normal[idx+0][1], normal[idx+0][2]);
	    GXNormal3f32(normal[idx+1][0], normal[idx+1][1], normal[idx+1][2]);
	    GXNormal3f32(normal[idx+2][0], normal[idx+2][1], normal[idx+2][2]);
	    break;

	  case GX_VA_TEX0:
	  case GX_VA_TEX1:
	  case GX_VA_TEX2:
	  case GX_VA_TEX3:
	  case GX_VA_TEX4:
	  case GX_VA_TEX5:
	  case GX_VA_TEX6:
	  case GX_VA_TEX7:
	    if (desc->attr_type == GX_INDEX16) {
	      idx = (idx<<8) | dl[m++];
	      GXTexCoord1x16(idx);
	    } else 
	      GXTexCoord1x8(idx);
	    break;

	  case GX_VA_CLR0:
	  case GX_VA_CLR1:
	    if (desc->attr_type == GX_INDEX16) {
	      idx = (idx<<8) | dl[m++];
	      GXColor1x16(idx);
	    } else if (desc->attr_type == GX_INDEX8) {
	      GXColor1x8(idx);
	    } else {
	      switch (desc->comp_type) {
	      case GX_RGB565:
	      case GX_RGBA4:
		GXColor1u16((idx<<8)|dl[m++]);
		break;
	      case GX_RGB8:
	      case GX_RGBA6:
		GXColor3u8(idx, dl[m], dl[m+1]); 
		m += 2;
		break;
	      case GX_RGBA8:
	      case GX_RGBX8:
		GXColor4u8(idx, dl[m], dl[m+1], dl[m+2]);
		m += 3;
		break;
	      }
	    }
	    break;
	  default:
	    if (desc->attr_type == GX_INDEX16) {
	      idx = (idx<<8) | dl[m++];
	    }
	    HSD_Report("attr(%d) is not supported by sysdolphin\n", desc->attr);
	    break;
	  }
	}
      }
    }
    GXEnd();
    l += m;
    dl+= m;
  }
#endif
}

static void
drawShapeAnim(HSD_PObj *pobj)
{
  HSD_ShapeSet *shape_set = pobj->u.shape_set;
  float blend;
  int shape_id, i;
#ifdef SHAPETEST
  extern float SH_TEST_SHAPE_MAX;
  extern float *SH_TEST_SHAPE_BP;
  float SH_TEST_BLEND = 0.5F;
#endif
  int blend_nbt;

#ifdef SHDEBUG
  HSD_Report("setupShapeAnim in nb_vertex_index (%d)\n",
	     shape_set->nb_vertex_index);
#endif

  if (vertex_buffer_size == 0) {
    vertex_buffer_size = HSD_DEFAULT_MAX_SHAPE_VERTICES;
    vertex_buffer = HSD_Alloc(vertex_buffer_size * sizeof(f32[3]));
  }
  assert (vertex_buffer_size >= shape_set->nb_vertex_index);

  if (shape_set->normal_desc && normal_buffer_size == 0) {
    normal_buffer_size = HSD_DEFAULT_MAX_SHAPE_NORMALS;
    normal_buffer = HSD_Alloc(normal_buffer_size * sizeof(f32[3]));
  }

  if (shape_set->normal_desc) {
    if (shape_set->normal_desc->attr == GX_VA_NRM) {
      assert (normal_buffer_size >= shape_set->nb_normal_index);
      blend_nbt = 0;
    } else { // GX_VA_NBT
      assert (normal_buffer_size >= shape_set->nb_normal_index * 3);
      blend_nbt = 1;
    }
  }

  if (shape_set->flags & SHAPESET_AVERAGE) {
    /*
     * Average Animation
     */
#ifdef SHAPETEST
    blend = SH_TEST_BLEND;
#else
    blend = shape_set->blend.bl;
#endif
    shape_id = min(max(0, (int)blend), shape_set->nb_shape-1);
    blend = min(max(0.0, blend-(float)shape_id), 1.0);
    for (i = 0; i < shape_set->nb_vertex_index; i++) {
      float s0[3], s1[3];

      get_shape_vertex_xyz(shape_set, shape_id, i, s0);
      get_shape_vertex_xyz(shape_set,
			   min(shape_id + 1, shape_set->nb_shape-1), i, s1);
      vertex_buffer[i][0] = (s1[0] - s0[0]) * blend + s0[0];
      vertex_buffer[i][1] = (s1[1] - s0[1]) * blend + s0[1];
      vertex_buffer[i][2] = (s1[2] - s0[2]) * blend + s0[2];
    }
    if (shape_set->nb_normal_index) {
      if (blend_nbt) {
	for (i = 0; i < shape_set->nb_normal_index; i++) {
	  float s0[9], s1[9];
	  int j, idx = i * 3;

	  get_shape_nbt_xyz(shape_set, shape_id, i, s0);
	  get_shape_nbt_xyz(shape_set,
			    min(shape_id + 1, shape_set->nb_shape-1), i, s1);
	  for (j=0; j<9; j++) {
	    normal_buffer[idx][j] = (s1[j] - s0[j]) * blend + s0[j];
	  }
	}
      } else {
	for (i = 0; i < shape_set->nb_normal_index; i++) {
	  float s0[3], s1[3];

	  get_shape_normal_xyz(shape_set, shape_id, i, s0);
	  get_shape_normal_xyz(shape_set,
			 min(shape_id + 1, shape_set->nb_shape-1), i, s1);
	  normal_buffer[i][0] = (s1[0] - s0[0]) * blend + s0[0];
	  normal_buffer[i][1] = (s1[1] - s0[1]) * blend + s0[1];
	  normal_buffer[i][2] = (s1[2] - s0[2]) * blend + s0[2];
	}
      }
    }
  } else {
    /*
     * Additive Animation
     * shape_id 0 ̓ftHgVFCvA1 ȍ~̃VFCv͍ɂȂ
     */
    int j;
    float *blend_bp;
    
#ifdef SHAPETEST
    blend_bp = SH_TEST_SHAPE_BP;
#else
    blend_bp = shape_set->blend.bp;
#endif
#ifdef SHXDEBUG
    HSD_Report("=======\n");
    HSD_Report("nb_vertex_index (%d)\n", shape_set->nb_vertex_index);
    HSD_Report("nb_shape (%d)\n", shape_set->nb_shape);
    for (i = 0; i < shape_set->nb_shape; i++)
      HSD_Report("shape (%d) blend (%f)\n", i, blend_bp[i]);
    HSD_Report("=======\n");
#endif
#ifdef WATANABE
	{
		f32		*pbp = (f32 *)HSD_Alloc(sizeof(f32) * shape_set->nb_shape);
		for(j = 0; j < shape_set->nb_shape; j++){
//			pbp[j] = max(0.0f, blend_bp[j]);
			if( blend_bp[j] > 0.0001f || blend_bp[j] < 0.0001f ){
			  pbp[j] = blend_bp[j];
			}else{
			  pbp[j] = 0.0f;
			}
		}
	    for (i = 0; i < shape_set->nb_vertex_index; i++) {
	      get_shape_vertex_xyz(shape_set, 0, i, vertex_buffer[i]);
	      for (j = 0; j < shape_set->nb_shape; j++) {
//		float b = max(0.0, blend_bp[j]);
//		if( b > 0.0f ){
//		if( pbp[j] > 0.0f ){
		if( pbp[j] != 0.0f ){
			float s[3];
			get_shape_vertex_xyz(shape_set, j+1, i, s);
			vertex_buffer[i][0] += s[0] * pbp[j];
			vertex_buffer[i][1] += s[1] * pbp[j];
			vertex_buffer[i][2] += s[2] * pbp[j];
//			vertex_buffer[i][0] += s[0] * b;
//			vertex_buffer[i][1] += s[1] * b;
//			vertex_buffer[i][2] += s[2] * b;
		}
	      }
	    }
		HSD_Free(pbp);
	}
#else
    for (i = 0; i < shape_set->nb_vertex_index; i++) {
      get_shape_vertex_xyz(shape_set, 0, i, vertex_buffer[i]);
      for (j = 0; j < shape_set->nb_shape; j++) {
	float b = max(0.0, blend_bp[j]);
	float s[3];
	get_shape_vertex_xyz(shape_set, j+1, i, s);
	vertex_buffer[i][0] += s[0] * b;
	vertex_buffer[i][1] += s[1] * b;
	vertex_buffer[i][2] += s[2] * b;
      }
    }
#endif
    if (shape_set->nb_normal_index) {
      if (blend_nbt) {
	for (i = 0; i < shape_set->nb_normal_index; i++) {
	  int idx = i * 3;
	  get_shape_nbt_xyz(shape_set, 0, i, normal_buffer[idx]);
	  for (j = 0; j < shape_set->nb_shape; j++) {
	    float b = max(0.0, blend_bp[j]);
	    float s[9];
	    int k;
	    get_shape_nbt_xyz(shape_set, j+1, i, s);
	    for (k=0; k<9; k++) {
	      normal_buffer[idx][k] += s[k] * b;
	    }
	  }
	}
#ifndef WATANABE
      } else {
	for (i = 0; i < shape_set->nb_normal_index; i++) {
	  get_shape_normal_xyz(shape_set, 0, i, normal_buffer[i]);
	  for (j = 0; j < shape_set->nb_shape; j++) {
	    float b = max(0.0, blend_bp[j]);
	    float s[3];
	    get_shape_normal_xyz(shape_set, j+1, i, s);
	    normal_buffer[i][0] += s[0] * b;
	    normal_buffer[i][1] += s[1] * b;
	    normal_buffer[i][2] += s[2] * b;
	  }
	}
#endif
      }
    }
  }
  interpretShapeAnimDisplayList(pobj, vertex_buffer, normal_buffer);

#ifdef SHDEBUG
  HSD_Report("setupShapeAnim out\n");
#endif
}

/**
 * ϊ}gNX̗pr`}[NB
 * 
 * {֐HSD_JObjDisp()ɂāA`ɐ旧1xĂяoB
 * 
 * pr`}[N̏ڍׂHSD_PObjSetMtxMarkQƁB
 * 
 * @param obj	NAɎgpIuWFNg̒l
 * @param mark	NAɎgp}[Nl
 * 
 * @see HSD_PObjSetMtxMark
 * @see HSD_PObjGetMtxMark
 */
void
HSD_PObjClearMtxMark(void *obj, u32 mark)
{
  int i;

  for (i=0; i<2; i++) {
    mtx_mark[i].obj  = obj;
    mtx_mark[i].mark = mark;
  }
}

/**
 * ϊ}gNX̗pr`}[Nݒ肷B
 * 
 * GAMECUBE̍W/@/eNX`ϊ}gNX̃Zbg
 * gp@Ae`B
 * 
 * }[NݒłZbg͂RAeZbg
 * pr}[Nƃ}gNX̐ɎgpIuWFNgi|C^j
 * ݒ肷邱ƂłB
 * 
 * pobj̃ftHg}gNXݒ胁\bh(setup_mtx)̓}[N
 * ׂĕKvȏꍇ̂ݍĐݒȂB
 * pobj̃}gNXݒ胁\bhĒ`ꍇ́AL̐ݒ
 * 悤ɗpr`}[Nݒ肷邱ƁB
 * 
 * <b>Rigid Model</b>
 * 
 * ʏ̃f(Rigid Model)͂Pjobj̕ϊ}gNX̉ê݂
 * 󂯂̂ŁAgpW/@/eNX`ϊ}gNX̃Zbg
 * PłB
 * ̏ꍇAZbgÕ}[NHSD_MTX_RIGIDɐݒ肵AIuWFNgɂ
 * jobjݒ肷B
 * 
 * <b>Shared Vertex Model (Obsolete)</b>
 * 
 * Shared Vertex f͌肳ꂽ`̃Gx[vfŁA
 * v~eBu̕tĂjobjƁA1ʂjobj̉e󂯂邱ƂłB
 * ̃f̓}gNXZbg2gpB
 * 
 * ZbgOɂRigid ModelƓl̐ݒȂB
 * ZbgPɂ̓}[NHSD_MTX_RIGIDAIuWFNg͂1jobjݒ肷B
 * 
 * <b>Envelope Model</b>
 * 
 * Gx[vf͍őPOZbg̃}gNXgpB
 * ŏ̂Q̓ZbgOAPgp邪AĉW}gNX̓ZbgQ
 * \B
 * eZbgɂ̓}[NHSD_MTX_ENVELOPE, IuWFNgɂNULLݒ肷B
 * 
 * <pre>
 *  Model Type    |  0    1    2   (Mtx Set)
 * ---------------+-------------------------
 * Rigid Model    |  o
 * Shared Vertex  |  o    o
 * Envelope Model |  o    o    o
 * </pre>
 *  
 * <pre>
 * j[jbN		l	Ӗ
 * ------------------------------------------------------------------
 * HSD_MTX_NONE		0	gp
 * HSD_MTX_RIGID	1	Pjobj猈܂ϊ}gNX
 * HSD_MTX_ENVELOPE	2	Gx[vf
 * HSD_MTX_CUSTOM	3	[Ugp
 * HSD_MTX_RESERVE1	4	(VXe\)
 * HSD_MTX_RESERVE2	5	(VXe\)
 * HSD_MTX_RESERVE3	6	(VXe\)
 * HSD_MTX_RESERVE4	7	(VXe\)
 * </pre>
 *
 * @see HSD_PObjGetMtxMark
 * @see HSD_PObjClearMtxMark
 */
void
HSD_PObjSetMtxMark(int idx,  void *obj, u32 mark)
{
  if (idx >= 2)
    return;

  if (0 <= idx && idx < 2) {
  } else {
    mtx_mark[idx].obj  = obj;
    mtx_mark[idx].mark = mark;
  }
}

/**
 * ϊ}gNX̗pr`}[N擾B
 * 
 * pr`}[N̏ڍׂHSD_PObjSetMtxMark()QƁB
 * 
 * @param idx		pr`}[ÑCfbNX
 * @param obj		}gNX̐ɎgpꂽIuWFNg𓾂
 * @param mark		pr`}[Nl𓾂
 * 
 * @see HSD_PObjSetMtxMark
 * @see HSD_PObjClearMtxMark
 */
void
HSD_PObjGetMtxMark(int idx,  void **obj, u32 *mark)
{
  assert(obj);
  assert(mark);

  if (idx < 0 || 2 <= idx) {
    *obj  = NULL;
    *mark = HSD_MTX_NONE;
  } else {
    *obj  = mtx_mark[idx].obj;
    *mark = mtx_mark[idx].mark;
  }
}

static PObjSetupFlag
GetSetupFlags(HSD_JObj *jobj)
{
  PObjSetupFlag flags = SETUP_NONE;

  if (jobj->flags & JOBJ_LIGHTING) {
    flags |= SETUP_NORMAL;
  }
  if (_HSD_TObjGetCurrentByType(NULL, TEX_COORD_REFLECTION)) {
    flags |= SETUP_NORMAL | SETUP_REFLECTION;
  }
#ifndef UNUSED_HILIGHTMAP
  if (_HSD_TObjGetCurrentByType(NULL, TEX_COORD_HILIGHT)) {
    flags |= SETUP_NORMAL | SETUP_HIGHLIGHT;
  }
#endif

  return flags;
}

static void
SetupRigidModelMtx(HSD_PObj *pobj, Mtx vmtx, Mtx pmtx, u32 rendermode)
{
#pragma unused(pobj)
#pragma unused(vmtx)
#pragma unused(rendermode)

  HSD_JObj *jobj;
  Mtx n;
  PObjSetupFlag flags;

  jobj = HSD_JObjGetCurrent();

  {
    void *obj;
    u32 mark;

    HSD_PObjGetMtxMark(0, &obj, &mark);
    if (obj == jobj && mark == HSD_MTX_RIGID)
      return;
    HSD_PObjSetMtxMark(0, jobj, HSD_MTX_RIGID);
  }

  GXSetCurrentMtx(GX_PNMTX0);
  GXLoadPosMtxImm(pmtx, GX_PNMTX0);
  HSD_PerfCountMtxLoad();

  flags = GetSetupFlags(jobj);

  //#if FUJITA
#if 0
  if (!shadow_mode && flags & SETUP_NORMAL) {
#else
  if (flags & SETUP_NORMAL) {
#endif
    mkNormalMtx(pmtx, n);
    if (jobj->flags & JOBJ_LIGHTING) {
      setNormalMtx(n, GX_PNMTX0);
    }
    if (flags & SETUP_NORMAL_PROJECTION) {
      GXLoadTexMtxImm(n, GX_TEXMTX0, GX_MTX3x4);
      HSD_PerfCountMtxLoad();
    }
  }
}

static void
SetupSharedVtxModelMtx(HSD_PObj *pobj, Mtx vmtx, Mtx pmtx, u32 rendermode)
{
#pragma unused(rendermode)

  HSD_JObj *jobj;
  Mtx n0, n1, m;
  PObjSetupFlag flags = SETUP_NONE;

  jobj = HSD_JObjGetCurrent();

  {
    void *obj;
    u32 mark;

    HSD_PObjGetMtxMark(0, &obj, &mark);
    if (obj != jobj && mark != HSD_MTX_RIGID)
      flags |= SETUP_JOINT0;
    HSD_PObjSetMtxMark(0, jobj, HSD_MTX_RIGID);

    HSD_PObjGetMtxMark(1, &obj, &mark);
    if (obj != pobj->u.jobj && mark != HSD_MTX_RIGID)
      flags |= SETUP_JOINT1;
    HSD_PObjSetMtxMark(1, pobj->u.jobj, HSD_MTX_RIGID);
  }
  if (flags == SETUP_NONE)
    return;

  flags |= GetSetupFlags(jobj);

  if (flags | SETUP_JOINT0) {
    GXSetCurrentMtx(GX_PNMTX0);
    GXLoadPosMtxImm(pmtx, GX_PNMTX0);
    HSD_PerfCountMtxLoad();

	//#if FUJITA
#if 0
    if (!shadow_mode && flags & SETUP_NORMAL) {
#else
    if (flags & SETUP_NORMAL) {
#endif
      mkNormalMtx(pmtx, n0);
      if (jobj->flags & JOBJ_LIGHTING) {
	setNormalMtx(n0, GX_PNMTX0);
      }
      if (flags & SETUP_NORMAL_PROJECTION) {
	GXLoadTexMtxImm(n0, GX_TEXMTX0, GX_MTX3x4);
	HSD_PerfCountMtxLoad();
      }
    }
  }

  if (flags | SETUP_JOINT1) {
    HSD_JObjSetupMatrix(pobj->u.jobj);
    MTXConcat(vmtx, pobj->u.jobj->mtx, m);
    GXLoadPosMtxImm(m, GX_PNMTX1);
    HSD_PerfCountMtxLoad();

	//#if FUJITA
#if 0
    if (!shadow_mode && flags & SETUP_NORMAL) {
#else
    if (flags & SETUP_NORMAL) {
#endif
      mkNormalMtx(m, n1);
      if (jobj->flags & JOBJ_LIGHTING) {
	setNormalMtx(n1, GX_PNMTX1);
      }
      if (flags & SETUP_NORMAL_PROJECTION) {
	GXLoadTexMtxImm(n1, GX_TEXMTX1, GX_MTX3x4);
	HSD_PerfCountMtxLoad();
      }
    }
  }
}

/*
 * Gx[v}gNX𐶐An[hEFAɃ[hB
 */
static void
SetupEnvelopeModelMtx(HSD_PObj *pobj, Mtx vmtx, Mtx pmtx, u32 rendermode)
{
#pragma unused(rendermode)
#pragma unused(pmtx)

  HSD_JObj *jobj;
  HSD_SList *list;
  int MtxIdx = 0;
  MtxPtr right;
  Mtx mtx;
  PObjSetupFlag flags = SETUP_NONE;
  
  jobj = HSD_JObjGetCurrent();

  HSD_PObjClearMtxMark(NULL, HSD_MTX_ENVELOPE);

  flags = GetSetupFlags(jobj);

  /*
   * right
   * Gx[ṽfWn烋[gWnւ̕ϊ}gNX
   */
#ifdef HSD_EMTX_CACHE
  right = (jobj->flags & JOBJ_SKELETON_ROOT)? NULL : jobj->ermtx;
#else
  right = _HSD_mkEnvelopeModelNodeMtx(jobj, mtx);
#endif /* HSD_EMTX_CACHE */

  for (MtxIdx=0,list=pobj->u.envelope_list;
       MtxIdx < 10 && list;
       MtxIdx++, list=list->next) {
    Mtx mtx, tmp;
    MtxPtr mtxp;
    HSD_Envelope *envelope = list->data;
    GXPosNrmMtx mtx_no = HSD_Index2PosNrmMtx(MtxIdx);
#if HSD_USE_PERF
    int nn = 0;
#endif /* USE_PERF */

    assert(envelope);
    if (envelope->weight >= (1.0F - EPSILON)) {
      HSD_JObjSetupMatrix(envelope->jobj);
      if (right) {
#ifdef HSD_EMTX_CACHE
	mtxp = envelope->jobj->emtx;
#else
	MTXConcat(envelope->jobj->mtx, envelope->jobj->envelopemtx, mtx);
	mtxp = mtx;
#endif
      } else {
	mtxp = envelope->jobj->mtx;
      }
    } else {
      mtx[0][0] = mtx[0][1] = mtx[0][2] = mtx[0][3]
	= mtx[1][0] = mtx[1][1] = mtx[1][2] = mtx[1][3]
	= mtx[2][0] = mtx[2][1] = mtx[2][2] = mtx[2][3] = 0.0F;
      while (envelope) {
	HSD_JObj *jp;

	assert(envelope->jobj);
	jp = envelope->jobj;
	HSD_JObjSetupMatrix(jp);
	assert(jp->mtx);
	assert(jp->envelopemtx);
#ifdef HSD_EMTX_CACHE
	HSD_MtxScaledAdd(envelope->weight, jp->emtx, mtx, mtx);
#else /* HSD_EMTX_CACHE */
	MTXConcat(jp->mtx, jp->envelopemtx, tmp);
	HSD_MtxScaledAdd(envelope->weight, tmp, mtx, mtx);
#endif /* HSD_EMTX_CACHE */
#if HSD_USE_PERF
	nn ++;
#endif /* HSD_USE_PERF */
	envelope = envelope->next;
      }
      mtxp = mtx;
    }
    HSD_PerfCountEnvelopeBlending(nn);
    if (right) {
      MTXConcat(mtxp, right, mtx);
    }
    MTXConcat(vmtx, mtxp, tmp);
    GXLoadPosMtxImm(tmp, mtx_no);
    HSD_PerfCountMtxLoad();
#if FUJITA
	if( pobj->aobj ){
	  //f32 *save_mtx = (f32*)pobj->aobj;
	  //MTXCopy( mtxp, (MtxPtr)&save_mtx[MtxIdx*3*4] );
	  Mtx *save_mtx = (Mtx*)pobj->aobj;
	  MTXCopy( mtxp, (MtxPtr)&save_mtx[MtxIdx] );
	  //OSReport( "%d:(%d %d)\n", MtxIdx, mtx_no, MtxIdx*3*4 );
	  //HSD_DmpMatrix("", (MtxPtr)&save_mtx[MtxIdx*3*4] );
	}
#endif

	//#if FUJITA
#if 0
    if (!shadow_mode && flags & SETUP_NORMAL) {
#else
    if (flags & SETUP_NORMAL) {
#endif
      mkNormalMtx(tmp, mtx);
      if (jobj->flags & JOBJ_LIGHTING) {
	setNormalMtx(mtx, mtx_no);
      }
      if (flags & SETUP_NORMAL_PROJECTION) {
	GXLoadTexMtxImm(mtx, HSD_Index2TexMtx(MtxIdx), GX_MTX3x4);
	HSD_PerfCountMtxLoad();
      }
    }
  }
}

static void
PObjSetupMtx(HSD_PObj *pobj, Mtx vmtx, Mtx pmtx, u32 rendermode)
{
  switch (pobj_type(pobj)) {
  case POBJ_SKIN:
    if (!pobj->u.jobj) {
      SetupRigidModelMtx(pobj, vmtx, pmtx, rendermode);
    } else {
      SetupSharedVtxModelMtx(pobj, vmtx, pmtx, rendermode);
    }
    break;
  case POBJ_SHAPEANIM:
    SetupRigidModelMtx(pobj, vmtx, pmtx, rendermode);
    break;
  case POBJ_ENVELOPE:
    SetupEnvelopeModelMtx(pobj, vmtx, pmtx, rendermode);
    break;
  }
}

static void
PObjDispSimplePrimitive(HSD_PObj *pobj, u32 rendermode)
{
#pragma unused(rendermode)

  setupArrayDesc(pobj->desc_list);
  setupVtxDesc(pobj);

  GXCallDisplayList(pobj->dl, pobj->dlsize<<5);
}

static void
PObjDispShapeAnim(HSD_PObj *pobj, u32 rendermode)
{
#pragma unused(rendermode)

#ifdef	WATANABE
	GXInvalidateVtxCache();
#endif
  setupShapeAnimArrayDesc(pobj->desc_list);
  setupShapeAnimVtxDesc(pobj);

  assert(pobj->u.shape_set);
  drawShapeAnim(pobj);
}

void
HSD_PObjDisp(HSD_PObj *pobj, Mtx vmtx, Mtx pmtx, u32 rendermode)
{
  if (pobj->flags & POBJ_HIDDEN) {
    return;
  }

#if !FUJITA
  switch (pobj->flags&(POBJ_CULLFRONT|POBJ_CULLBACK)) {
  case 0x0:
    HSD_StateSetCullMode(GX_CULL_NONE);
    break;
  case POBJ_CULLFRONT:
    HSD_StateSetCullMode(GX_CULL_FRONT);
    break;
  case POBJ_CULLBACK:
    HSD_StateSetCullMode(GX_CULL_BACK);
    break;
  case POBJ_CULLBACK|POBJ_CULLFRONT:
    return;
  }
#endif

#if FUJITA
  if( rendermode & RENDER_SHADOW ){
	//ep̃_OH
	shadow_mode = 1;
	//Zgق̉eł͂ꂪKvB
	if( rendermode & RENDER_EFFECT ){
	  //JO͎OŁB
	}else{
		switch (pobj->flags&(POBJ_CULLFRONT|POBJ_CULLBACK)) {
		case 0x0:
		  HSD_StateSetCullMode(GX_CULL_NONE);
		  break;
		case POBJ_CULLFRONT:
		  HSD_StateSetCullMode(GX_CULL_FRONT);
		  break;
		case POBJ_CULLBACK:
		  HSD_StateSetCullMode(GX_CULL_BACK);
		  break;
		case POBJ_CULLBACK|POBJ_CULLFRONT:
		  return;
		}
//	  HSD_StateSetCullMode(GX_CULL_FRONT);
	}
#if 0
	{
	  static f32 scl = 0.f;
	  if( HSD_PadGetButton(1) & HSD_PAD_Y ){
		if( HSD_PadGetButton(1) & HSD_PAD_RIGHT ) scl -= 0.00001f;
		if( HSD_PadGetButton(1) & HSD_PAD_LEFT ) scl += 0.00001f;
		OSReport( "%f\n", scl );
	  }
	  //GXSetZScaleOffset( 1.f, scl );
	  GXSetZScaleOffset( 1.f, -0.0021f );
	}
#endif
  }else{
#if 0
	GXSetZScaleOffset( 1.f, 0.f );
#endif

	shadow_mode = 0;
	switch (pobj->flags&(POBJ_CULLFRONT|POBJ_CULLBACK)) {
	case 0x0:
	  HSD_StateSetCullMode(GX_CULL_NONE);
	  break;
	case POBJ_CULLFRONT:
	  HSD_StateSetCullMode(GX_CULL_FRONT);
	  break;
	case POBJ_CULLBACK:
	  HSD_StateSetCullMode(GX_CULL_BACK);
	  break;
	case POBJ_CULLBACK|POBJ_CULLFRONT:
	  return;
	}
  }
#endif

#if HSD_DISP_METHOD_LEVEL > 0
  PObjSetupMtx(pobj, vmtx, pmtx, rendermode);
#else
  HSD_POBJ_METHOD(pobj)->setup_mtx(pobj, vmtx, pmtx, rendermode);
#endif

  if (pobj_type(pobj) == POBJ_SHAPEANIM) {
    PObjDispShapeAnim(pobj, rendermode);
  } else {
    PObjDispSimplePrimitive(pobj, rendermode);
  }
}

/*
 * oJ
 */
static void
PObjRelease(HSD_Class *o)
{
  HSD_PObj *pobj = HSD_POBJ(o);

  if (pobj->aobj) {
    HSD_AObjRemove(pobj->aobj);
  }

  switch (pobj_type(pobj)) {
  case POBJ_SHAPEANIM:
    HSD_ShapeSetRemove(pobj->u.shape_set);
    break;
  case POBJ_ENVELOPE:
    HSD_EnvelopeListFree(pobj->u.envelope_list);
    break;
  case POBJ_SKIN:
    HSD_JObjUnrefThis(pobj->u.jobj);
    break;
  default:
    break;
  }

  HSD_PARENT_INFO(&hsdPObj)->release(o);
}

/*
 * NXď
 */
static void
PObjAmnesia(HSD_ClassInfo *info)
{
  if (info == HSD_CLASS_INFO(default_class)) {
    default_class = NULL;
  }
  if (info == HSD_CLASS_INFO(&hsdPObj)) {
    if (_HSD_MemCheckOwn(vertex_buffer)) {
      vertex_buffer = NULL;
      vertex_buffer_size = 0;
    }
    if (_HSD_MemCheckOwn(normal_buffer)) {
      normal_buffer = NULL;
      normal_buffer_size = 0;
    }
    prev_vtxdesclist_array = NULL;
    prev_vtxdesc = NULL;
  }
  HSD_PARENT_INFO(&hsdPObj)->amnesia(info);
}

/*
 * NX񏉊
 */
static void
PObjInfoInit(void)
{
  hsdInitClassInfo(HSD_CLASS_INFO(&hsdPObj),
  		   HSD_CLASS_INFO(&hsdClass),
		   HSD_BASE_CLASS_LIBRARY,
		   "hsd_pobj",
		   sizeof(HSD_PObjInfo),
		   sizeof(HSD_PObj));
  
  HSD_CLASS_INFO(&hsdPObj)->release   = PObjRelease;
  HSD_CLASS_INFO(&hsdPObj)->amnesia   = PObjAmnesia;
   HSD_POBJ_INFO(&hsdPObj)->disp      = HSD_PObjDisp;
   HSD_POBJ_INFO(&hsdPObj)->setup_mtx = PObjSetupMtx;
   HSD_POBJ_INFO(&hsdPObj)->load      = PObjLoad;
   HSD_POBJ_INFO(&hsdPObj)->update    = PObjUpdateFunc;
}
