////////////////////////////////////////////////////////////////////////////////
//
// Killer7 Model
//
// WK7Model.cpp
//
// Creator       : Watanabe Kazuhisa (GHM)
// Creation Date : 02/08/2003
//
////////////////////////////////////////////////////////////////////////////////

#define	__WK7ARCHIVE_CPP__

//#include <dolphin.h>
//#include <sysdolphin/baselib.h>
//#include "..\Lib\ghmGcFile.h"
//#include "..\Lib\ghmGcMath.h"
#include ".\WK7Model.h"
#include ".\WK7Archive.h"
#include ".\WK7Mirror.h"
#include ".\WK7Shadow.h"
#include ".\WK7HitObj.h"
#include ".\WK7Fog.h"
#include ".\WK7EffectBreakObj.h"
#include "..\Fujita\C_Font.h"
#include "..\Fujita\C_SetAlpha.h"
#include "..\Fujita\C_K7Cloth.h"

//#define	CLASSICAL_SCALING_CHECK
HSD_TrspMask		WK7JObj::mRenderProcessTrspMask   = HSD_TRSP_ALL;
u32					WK7JObj::mRenderProcessRenderMode = 0;
u32					WK7JObj::mCurrentRenderMode = WK7JOBJ_RENDERMODE_NORMAL;
C_EdgeShadow*		WK7JObj::mpEdgeShadow       = NULL;
int					WK7JObj::mEdgeShadowObjNum  = 0;
ghmGcViewVolume*	WK7JObj::mpViewVolume       = NULL;
WK7SdExpInfo*		WK7JObj::mpCurrentSdExpInfo = NULL;
WK7SdExpInfo*		WK7JObj::mpCurrentAttachMotionSdExpInfo = NULL;

extern "C"{
extern	void *HSD_SetDrawVertexBuffer(f32 (*buf)[3], u32 nitems);
extern	void *HSD_SetDrawNormalBuffer(f32 (*buf)[3], u32 nitems);
}

WK7JObjShapeAnim::WK7JObjShapeAnim(HSD_ShapeSet *pShapeSet,u32 vertexBufferSize,u32 normalBufferSize)
{
	mVertexBufferSize = vertexBufferSize;
	mNormalBufferSize = normalBufferSize;
	mBufferIndex      = 0;
	mpVertexBuffer[0] = 
	mpVertexBuffer[1] = 
	mpNormalBuffer[0] = 
	mpNormalBuffer[1] = NULL;
	int			size;
	if( mVertexBufferSize > 0 ){
		size = (int)OSRoundUp32B(mVertexBufferSize * sizeof(f32) * 3) * 2;
		ghmGcMemSetAllocName("JOSHPANV");
		if( (mpVertexBuffer[0] = K7AppSmallMemAlloc(size))!=NULL ){
			mpVertexBuffer[1] = (void *)((u32)mpVertexBuffer[0] + (size>>1));
			int			i,j;
			sysdolVec	*pVertex;
			for(i=0 ; i<2 ; i++){
				pVertex = (sysdolVec *)mpVertexBuffer[i];
				for(j=0 ; j<mVertexBufferSize ; j++){
					get_shape_vertex_xyz(pShapeSet,0,j,pVertex[j]);
				}
			}
		}
	}
	if( mNormalBufferSize > 0 ){
		size = (int)OSRoundUp32B(mNormalBufferSize * sizeof(f32) * 3) * 2;
		ghmGcMemSetAllocName("JOSHPANN");
		if( (mpNormalBuffer[0] = K7AppSmallMemAlloc(size))!=NULL ){
			mpNormalBuffer[1] = (void *)((u32)mpNormalBuffer[0] + (size>>1));
			int			i,j;
			sysdolVec	*pNormal;
			for(i=0 ; i<2 ; i++){
				pNormal = (sysdolVec *)mpNormalBuffer[i];
				for(j=0 ; j<mNormalBufferSize ; j++){
					get_shape_normal_xyz(pShapeSet,0,j,pNormal[j]);
				}
			}
		}
	}
}

WK7JObjShapeAnim::~WK7JObjShapeAnim()
{
	if( mpVertexBuffer[0]!=NULL || mpNormalBuffer[0]!=NULL ){
		GXPixModeSync();
	}
	GHMGCMEM_SAFE_FREE(mpVertexBuffer[0]);
	GHMGCMEM_SAFE_FREE(mpNormalBuffer[0]);
}

void WK7JObjShapeAnim::get_shape_vertex_xyz(HSD_ShapeSet *shape_set,int shape_id, int arrayidx, float dst[3])
{
	u8		*index_array = (u8 *)shape_set->vertex_idx_list[shape_id];
	int		idx;
	void	*src_base;
	
	if( shape_set->vertex_desc->attr_type == GX_INDEX16 ){
		idx = index_array[arrayidx*2];
		idx = (idx<<8) + index_array[arrayidx*2+1];
	}else{
		idx = index_array[arrayidx];
	}
	
	assert(shape_set->vertex_desc->comp_cnt == GX_POS_XYZ);
	
	src_base = (void *)(((u8 *) shape_set->vertex_desc->base_ptr)  + idx * shape_set->vertex_desc->stride);
	
	/* _z̃oCgI[_̓A[LeN`ˑ */
	if( shape_set->vertex_desc->comp_type == GX_F32 ){
		memcpy(dst, src_base, sizeof(f32[3]));
	}else{
		int		decimal_point = 1 << shape_set->vertex_desc->frac;
		switch( shape_set->vertex_desc->comp_type ){
		case GX_U8:{
			u8	*src = (u8 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			} break;
		case GX_S8:{
			s8	*src = (s8 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			} break;
		case GX_U16:{
			u16	*src = (u16 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		case GX_S16:{
			s16	*src = (s16 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		default:
			HSD_Halt("unexpected vertex type.\n");
			break;
		}
	}
}

void WK7JObjShapeAnim::get_shape_normal_xyz(HSD_ShapeSet *shape_set,int shape_id, int arrayidx, float dst[3])
{
	u8		*index_array = (u8 *)shape_set->normal_idx_list[shape_id];
	int		idx;
	void	*src_base;
	
	if( shape_set->normal_desc->attr_type == GX_INDEX16 ){
		idx = index_array[arrayidx*2];
		idx = (idx<<8) + index_array[arrayidx*2+1];
	}else{
		idx = index_array[arrayidx];
	}
	
	assert(shape_set->normal_desc->comp_cnt == GX_NRM_XYZ);
	
	src_base = (void *)(((u8 *)shape_set->normal_desc->base_ptr) + idx * shape_set->normal_desc->stride);
	
	/* _z̃oCgI[_̓A[LeN`ˑ */
	if( shape_set->normal_desc->comp_type == GX_F32 ){
		memcpy(dst,src_base,sizeof(f32[3]));
	}else{
		
		int decimal_point = 1 << shape_set->normal_desc->frac;
		switch (shape_set->normal_desc->comp_type) {
		case GX_U8:{
			u8	*src = (u8 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			} break;
		case GX_S8:{
			s8	*src = (s8 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		case GX_U16:{
			u16	*src = (u16 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		case GX_S16:{
			s16	*src = (s16 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		default:
			HSD_Halt("unexpected normal type.");
			break;
		}
	}
}

bool WK7JObjShapeAnim::FrameProcess(void)
{
	mBufferIndex ^= 1;
	return true;
}

bool WK7JObjShapeAnim::SetShapeBuffer(void)
{
	if( mVertexBufferSize > 0 ){
		HSD_SetVertexBuffer    ((sysdolVec *)mpVertexBuffer[mBufferIndex ^ 1],mVertexBufferSize);
		HSD_SetDrawVertexBuffer((sysdolVec *)mpVertexBuffer[mBufferIndex ^ 1],mVertexBufferSize);
//		HSD_SetDrawVertexBuffer((sysdolVec *)mpVertexBuffer[mBufferIndex    ],mVertexBufferSize);
	}
	if( mNormalBufferSize > 0 ){
		HSD_SetNormalBuffer    ((sysdolVec *)mpNormalBuffer[mBufferIndex ^ 1],mNormalBufferSize);
		HSD_SetDrawNormalBuffer((sysdolVec *)mpNormalBuffer[mBufferIndex ^ 1],mNormalBufferSize);
//		HSD_SetDrawNormalBuffer((sysdolVec *)mpNormalBuffer[mBufferIndex    ],mNormalBufferSize);
	}
	return true;
}
bool WK7JObjShapeAnim::ResetShapeBuffer(void)
{
	HSD_SetVertexBuffer    (NULL,0);
	HSD_SetDrawVertexBuffer(NULL,0);
	HSD_SetNormalBuffer    (NULL,0);
	HSD_SetDrawNormalBuffer(NULL,0);
	return true;
}

void WK7JObj::GetBufferSize(HSD_PObj *pPObj,u32 &vertexBufferSize,u32 &normalBufferSize,HSD_ShapeSet *&pShapeSet)
{
	if( !(pPObj->flags & POBJ_SHAPEANIM) ){
		return;
	}
	if( pPObj->u.shape_set==NULL ){
		return;
	}
	pShapeSet = pPObj->u.shape_set;
	if( vertexBufferSize < pShapeSet->nb_vertex_index ){
		vertexBufferSize = (u32)pShapeSet->nb_vertex_index;
	}
	if( normalBufferSize < pShapeSet->nb_normal_index ){
		normalBufferSize = (u32)pShapeSet->nb_normal_index;
	}
}

void WK7JObj::GetBufferSize(HSD_DObj *pDObj,u32 &vertexBufferSize,u32 &normalBufferSize,HSD_ShapeSet *&pShapeSet)
{
	HSD_PObj		*pPObj;
	for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
		GetBufferSize(pPObj,vertexBufferSize,normalBufferSize,pShapeSet);
	}
}

void WK7JObj::get_shape_normal_xyz(HSD_ShapeSet *shape_set,int shape_id, int arrayidx, float dst[3])
{
	u8		*index_array = (u8 *)shape_set->normal_idx_list[shape_id];
	int		idx;
	void	*src_base;
	
	if( shape_set->normal_desc->attr_type == GX_INDEX16 ){
		idx = index_array[arrayidx*2];
		idx = (idx<<8) + index_array[arrayidx*2+1];
	}else{
		idx = index_array[arrayidx];
	}
	
	assert(shape_set->normal_desc->comp_cnt == GX_NRM_XYZ);
	
	src_base = (void *)(((u8 *)shape_set->normal_desc->base_ptr) + idx * shape_set->normal_desc->stride);
	
	/* _z̃oCgI[_̓A[LeN`ˑ */
	if( shape_set->normal_desc->comp_type == GX_F32 ){
		memcpy(dst,src_base,sizeof(f32[3]));
	}else{
		
		int decimal_point = 1 << shape_set->normal_desc->frac;
		switch (shape_set->normal_desc->comp_type) {
		case GX_U8:{
			u8	*src = (u8 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			} break;
		case GX_S8:{
			s8	*src = (s8 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		case GX_U16:{
			u16	*src = (u16 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		case GX_S16:{
			s16	*src = (s16 *)src_base;
			dst[0] = (float)src[0] / decimal_point;
			dst[1] = (float)src[1] / decimal_point;
			dst[2] = (float)src[2] / decimal_point;
			}break;
		default:
			HSD_Halt("unexpected normal type.");
			break;
		}
	}
}

void WK7JObj::GetVertexF32(Vec *pVertex,void *pBasePtr,int index,f32 base)
{
#pragma unused(base)
	*pVertex = *((Vec*)pBasePtr + index);
}

void WK7JObj::GetVertexU16(Vec *pVertex,void *pBasePtr,int index,f32 base)
{
	WK7VecU16	*pSrcVert = ((WK7VecU16 *)pBasePtr) + index;
	pVertex->x = pSrcVert->x * base;
	pVertex->y = pSrcVert->y * base;
	pVertex->z = pSrcVert->z * base;
}

void WK7JObj::GetVertexS16(Vec *pVertex,void *pBasePtr,int index,f32 base)
{
	WK7VecS16	*pSrcVert = ((WK7VecS16 *)pBasePtr) + index;
	pVertex->x = pSrcVert->x * base;
	pVertex->y = pSrcVert->y * base;
	pVertex->z = pSrcVert->z * base;
}

void WK7JObj::GetVertexU8 (Vec *pVertex,void *pBasePtr,int index,f32 base)
{
	WK7VecU8	*pSrcVert = ((WK7VecU8 *)pBasePtr) + index;
	pVertex->x = pSrcVert->x * base;
	pVertex->y = pSrcVert->y * base;
	pVertex->z = pSrcVert->z * base;
}

void WK7JObj::GetVertexS8 (Vec *pVertex,void *pBasePtr,int index,f32 base)
{
	WK7VecS8	*pSrcVert = ((WK7VecS8 *)pBasePtr) + index;
	pVertex->x = pSrcVert->x * base;
	pVertex->y = pSrcVert->y * base;
	pVertex->z = pSrcVert->z * base;
}

void WK7JObj::GetStF32(WK7St *pSt,void *pBasePtr,int index,f32 base)
{
#pragma unused(base)
	*pSt = *((WK7St*)pBasePtr + index);
}

void WK7JObj::GetStU16(WK7St *pSt,void *pBasePtr,int index,f32 base)
{
	WK7StU16	*pSrcSt = ((WK7StU16 *)pBasePtr) + index;
	pSt->s = pSrcSt->s * base;
	pSt->t = pSrcSt->t * base;
}

void WK7JObj::GetStS16(WK7St *pSt,void *pBasePtr,int index,f32 base)
{
	WK7StS16	*pSrcSt = ((WK7StS16 *)pBasePtr) + index;
	pSt->s = pSrcSt->s * base;
	pSt->t = pSrcSt->t * base;
}

void WK7JObj::GetStU8 (WK7St *pSt,void *pBasePtr,int index,f32 base)
{
	WK7StU8		*pSrcSt = ((WK7StU8 *)pBasePtr) + index;
	pSt->s = pSrcSt->s * base;
	pSt->t = pSrcSt->t * base;
}

void WK7JObj::GetStS8 (WK7St *pSt,void *pBasePtr,int index,f32 base)
{
	WK7StS8		*pSrcSt = ((WK7StS8 *)pBasePtr) + index;
	pSt->s = pSrcSt->s * base;
	pSt->t = pSrcSt->t * base;
}

u32 WK7JObj::GetStride(HSD_VtxDescList *desc,WK7VtxDescInfo &info)
{
	if( desc==NULL ){
		return 0;
	}
	u32		stride = 0;
	info.mpVertexDesc = 
	info.mpNormalDesc = 
	info.mpColorDesc  = 
	info.mpTexDesc    = NULL;
	info.mVertexPos   = 0;
	info.mNormalPos   = 0;
	info.mColorPos    = 0;
	info.mTexPos      = 0;
	for( ; desc->attr!=GX_VA_NULL; desc++){
		switch( desc->attr ){
		case	GX_VA_PNMTXIDX:
		case	GX_VA_TEX0MTXIDX:
		case	GX_VA_TEX1MTXIDX:
		case	GX_VA_TEX2MTXIDX:
		case	GX_VA_TEX3MTXIDX:
		case	GX_VA_TEX4MTXIDX:
		case	GX_VA_TEX5MTXIDX:
		case	GX_VA_TEX6MTXIDX:
		case	GX_VA_TEX7MTXIDX:
			stride += 1;
			break;
		case	GX_VA_POS:
			info.mVertexPos   = stride;
			info.mpVertexDesc = desc;
			if( desc->attr_type==GX_INDEX16 ){
				stride += 2;
			}else{
				stride += 1;
			}
			break;
		case GX_VA_NRM:
		case GX_VA_NBT:
			info.mNormalPos   = stride;
			info.mpNormalDesc = desc;
			if( desc->attr_type==GX_INDEX16 ){
				stride += 2;
			} else {
				stride += 1;
			}
			break;
		case GX_VA_TEX0:
			info.mTexPos   = stride;
			info.mpTexDesc = desc;
		case GX_VA_TEX1:
		case GX_VA_TEX2:
		case GX_VA_TEX3:
		case GX_VA_TEX4:
		case GX_VA_TEX5:
		case GX_VA_TEX6:
		case GX_VA_TEX7:
			if( desc->attr_type==GX_INDEX16 ){
				stride += 2;
			}else{
				stride += 1;
			}
			break;
		case GX_VA_CLR0:
			info.mColorPos   = stride;
			info.mpColorDesc = desc;
		case GX_VA_CLR1:
			if( desc->attr_type==GX_INDEX16 ){
				stride += 2;
			}else if (desc->attr_type==GX_INDEX8 ){
				stride += 1;
			}else{
				//_CNgB
				switch( desc->comp_type ){
				case GX_RGB565:
				case GX_RGBA4:
					stride += 2;
					break;
				case GX_RGB8:
				case GX_RGBA6:
					stride += 3;
					break;
				case GX_RGBA8:
				case GX_RGBX8:
					stride += 4;
					break;
				}
			}
			break;
		default:
			break;
		}
	}
	
	return stride;
}

int WK7JObj::DlGetIndex(u8 *pdl,HSD_VtxDescList *pDesc)
{
	switch( pDesc->attr_type ){
	case	GX_INDEX16:
		return ((int)pdl[0]<<8)|pdl[1];
	case	GX_INDEX8:
		return *pdl;
	default:
		break;
	}
	return -1;
}

WK7JObjBound::WK7JObjBound()
{
	mType   = UNKNOWN;
	mEnable = true;
}

WK7JObjBound::~WK7JObjBound()
{
}

int WK7JObjBound::MakeVertex(vector<C_Vec> &vertlist,HSD_DObj *pDObj)
{
	int		totalvertnum = 0;
	for( ; pDObj!=NULL ; pDObj=pDObj->next){
		totalvertnum += MakeVertex(vertlist,pDObj->pobj);
	}
	return totalvertnum;
}

int WK7JObjBound::MakeVertex(vector<C_Vec> &vertlist,HSD_PObj *pPObj)
{
	int		totalvertnum = 0;
	
	for( ; pPObj!=NULL ; pPObj=pPObj->next){
		WK7VtxDescInfo		info;
		u32					stride;
		stride  = WK7JObj::GetStride(pPObj->desc_list,info);
		if( info.mpVertexDesc==NULL ){
			continue;
		}
		int				dlsize;
		u8				*pdl;
		int				num;
		GXPrimitive		prim;
//		GXVtxFmt		vtxfmt;
		pdl     = pPObj->dl;
		dlsize  = (pPObj->dlsize << 5);
		while( dlsize > 0 ){
			prim    = (GXPrimitive)(pdl[0] & GX_OPCODE_MASK);
//			vtxfmt  = (GXVtxFmt   )(pdl[0] & GX_VAT_MASK   );
			num     = ((int)pdl[1]<<8)|pdl[2];
			dlsize -= 3;
			pdl    += 3;
			switch( prim ){
			case GX_POINTS:
			case GX_LINES:
			case GX_LINESTRIP:
			case GX_TRIANGLES:
			case GX_TRIANGLESTRIP:
			case GX_TRIANGLEFAN:
			case GX_QUADS:{
				u8		*pVertexDL = pdl + info.mVertexPos;
				int		maxid      = -1;
				int		id;
				pdl += stride * num;
				dlsize -= stride * num;
				while( num > 0 ){
					if( maxid < (id=WK7JObj::DlGetIndex(pVertexDL,info.mpVertexDesc)) ){
						maxid = id;
					}
					num --;
					pVertexDL += stride;
				}
				WK7GetVertexFunc		*pGetVertex;
				f32						base;
				switch( info.mpVertexDesc->comp_type ){
				case	GX_U8:  pGetVertex = WK7JObj::GetVertexU8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S8:  pGetVertex = WK7JObj::GetVertexS8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_U16: pGetVertex = WK7JObj::GetVertexU16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S16: pGetVertex = WK7JObj::GetVertexS16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_F32: pGetVertex = WK7JObj::GetVertexF32; base = 1.0f;                                break;
					break;
				default:
					maxid = -1;
					break;
				}
				
				Vec			pos;
				void		*pBasePtr = info.mpVertexDesc->base_ptr;
				maxid++;
				totalvertnum += maxid;
				for(id=0 ; id<maxid ; id++){
					pGetVertex(&pos,pBasePtr,id,base);
					ghmGcMemSetAllocName("C_Vec");
					vertlist.push_back(*(new C_Vec(pos)));
				}
				
				}break;
			case GX_NOP:
			default:
				break;
			}
		}
	}
	
	return totalvertnum;
}

int WK7JObjBound::GetCrossState(Mtx lwmat,ghmGcViewVolume *pViewVolume)
{
#pragma unused(lwmat)
#pragma unused(pViewVolume)
	return ghmGcCollObj::INSIDE;
}

int WK7JObjBound::GetFarClip(f32 dist,Mtx lwmat,ghmGcViewVolume *pViewVolume)
{
#pragma unused(dist)
#pragma unused(lwmat)
#pragma unused(pViewVolume)
	return ghmGcCollObj::INSIDE;
}

bool WK7JObjBound::Initialize(HSD_DObj *pDObj)
{
#pragma	unused(pDObj)
	return true;
}

int WK7JObjBoundSphere::CalcCenter(HSD_DObj *pDObj)
{
	int		totalvertnum = 0;
	for( ; pDObj!=NULL ; pDObj=pDObj->next){
		totalvertnum += CalcCenter(pDObj->pobj);
	}
	if( totalvertnum > 0 ){
		f32		scale = 1.0f / totalvertnum;
		VECScale(&mCenter,&mCenter,scale);
	}
	return totalvertnum;
}

int WK7JObjBoundSphere::CalcCenter(HSD_PObj *pPObj)
{
	int		totalvertnum = 0;
	
	for( ; pPObj!=NULL ; pPObj=pPObj->next){
		WK7VtxDescInfo		info;
		u32					stride;
		stride  = WK7JObj::GetStride(pPObj->desc_list,info);
		if( info.mpVertexDesc==NULL ){
			continue;
		}
		int				dlsize;
		u8				*pdl;
		int				num;
		GXPrimitive		prim;
		pdl     = pPObj->dl;
		dlsize  = (pPObj->dlsize << 5);
		while( dlsize > 0 ){
			prim    = (GXPrimitive)(pdl[0] & GX_OPCODE_MASK);
			num     = ((int)pdl[1]<<8)|pdl[2];
			dlsize -= 3;
			pdl    += 3;
			switch( prim ){
			case GX_POINTS:
			case GX_LINES:
			case GX_LINESTRIP:
			case GX_TRIANGLES:
			case GX_TRIANGLESTRIP:
			case GX_TRIANGLEFAN:
			case GX_QUADS:{
				u8		*pVertexDL = pdl + info.mVertexPos;
				int		maxid      = -1;
				int		id;
				pdl += stride * num;
				dlsize -= stride * num;
				while( num > 0 ){
					if( maxid < (id=WK7JObj::DlGetIndex(pVertexDL,info.mpVertexDesc)) ){
						maxid = id;
					}
					num --;
					pVertexDL += stride;
				}
				WK7GetVertexFunc		*pGetVertex;
				f32					base;
				switch( info.mpVertexDesc->comp_type ){
				case	GX_U8:  pGetVertex = WK7JObj::GetVertexU8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S8:  pGetVertex = WK7JObj::GetVertexS8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_U16: pGetVertex = WK7JObj::GetVertexU16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S16: pGetVertex = WK7JObj::GetVertexS16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_F32: pGetVertex = WK7JObj::GetVertexF32; base = 1.0f;                                break;
					break;
				default:
					maxid = -1;
					break;
				}
				
				Vec			pos;
				void		*pBasePtr = info.mpVertexDesc->base_ptr;
				maxid++;
				totalvertnum += maxid;
				for(id=0 ; id<maxid ; id++){
					pGetVertex(&pos,pBasePtr,id,base);
					VECAdd(&mCenter,&pos,&mCenter);
				}
				
				}break;
			case GX_NOP:
			default:
				break;
			}
		}
	}
	
	return totalvertnum;
}

void WK7JObjBoundSphere::CalcRadius(HSD_DObj *pDObj)
{
	for( ; pDObj!=NULL ; pDObj=pDObj->next){
		CalcRadius(pDObj->pobj);
	}
}

void WK7JObjBoundSphere::CalcRadius(HSD_PObj *pPObj)
{
	for( ; pPObj!=NULL ; pPObj=pPObj->next){
		WK7VtxDescInfo		info;
		u32					stride;
		stride  = WK7JObj::GetStride(pPObj->desc_list,info);
		if( info.mpVertexDesc==NULL ){
			continue;
		}
		int				dlsize;
		u8				*pdl;
		int				num;
		GXPrimitive		prim;
		pdl     = pPObj->dl;
		dlsize  = (pPObj->dlsize << 5);
		while( dlsize > 0 ){
			prim    = (GXPrimitive)(pdl[0] & GX_OPCODE_MASK);
			num     = ((int)pdl[1]<<8)|pdl[2];
			dlsize -= 3;
			pdl    += 3;
			switch( prim ){
			case GX_POINTS:
			case GX_LINES:
			case GX_LINESTRIP:
			case GX_TRIANGLES:
			case GX_TRIANGLESTRIP:
			case GX_TRIANGLEFAN:
			case GX_QUADS:{
				u8		*pVertexDL = pdl + info.mVertexPos;
				int		maxid      = -1;
				int		id;
				pdl += stride * num;
				dlsize -= stride * num;
				while( num > 0 ){
					if( maxid < (id=WK7JObj::DlGetIndex(pVertexDL,info.mpVertexDesc)) ){
						maxid = id;
					}
					num --;
					pVertexDL += stride;
				}
				WK7GetVertexFunc		*pGetVertex;
				f32					base;
				switch( info.mpVertexDesc->comp_type ){
				case	GX_U8:  pGetVertex = WK7JObj::GetVertexU8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S8:  pGetVertex = WK7JObj::GetVertexS8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_U16: pGetVertex = WK7JObj::GetVertexU16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S16: pGetVertex = WK7JObj::GetVertexS16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_F32: pGetVertex = WK7JObj::GetVertexF32; base = 1.0f;                                break;
					break;
				default:
					maxid = -1;
					break;
				}
				
				Vec			pos;
				void		*pBasePtr = info.mpVertexDesc->base_ptr;
				maxid++;
				for(id=0 ; id<maxid ; id++){
					f32		dist;
					pGetVertex(&pos,pBasePtr,id,base);
					if( mBound.mRadius < (dist=GHM_VEC_GET_DIST(&pos,&mCenter)) ){
						mBound.mRadius =  dist;
					}
				}
				
				}break;
			case GX_NOP:
			default:
				break;
			}
		}
	}
}

#define	MINRADIUS		(3.0f)
WK7JObjBoundSphere::WK7JObjBoundSphere()
{
	mCenter.x = 0.0f;
	mCenter.y = 0.0f;
	mCenter.z = 0.0f;
	mBound.mPosition.x = 0.0f;
	mBound.mPosition.y = 0.0f;
	mBound.mPosition.z = 0.0f;
	mBound.mRadius = MINRADIUS;
	mType   = SPHERE;
}

WK7JObjBoundSphere::WK7JObjBoundSphere(HSD_DObj *pDObj)
{
//	int				num;
//	vector<C_Vec>	vertlist;
	mCenter.x = 0.0f;
	mCenter.y = 0.0f;
	mCenter.z = 0.0f;
	mBound.mPosition.x = 0.0f;
	mBound.mPosition.y = 0.0f;
	mBound.mPosition.z = 0.0f;
	mBound.mRadius     = 0.0f;
//	if( (num=MakeVertex(vertlist,pDObj)) > 0 ){
//		u32		i;
//		f32		invnum = 1.0f / num;
//		for(i=0 ; i<(u32)num ; i++){
//			mCenter.x += vertlist[i].x;
//			mCenter.y += vertlist[i].y;
//			mCenter.z += vertlist[i].z;
//		}
//		mCenter.x *= invnum;
//		mCenter.y *= invnum;
//		mCenter.z *= invnum;
//		for(i=0 ; i<(u32)num ; i++){
//			f32		dist = GHM_VEC_GET_DIST(&mCenter,&vertlist[i]);
//			if( mBound.mRadius < dist ){
//				mBound.mRadius = dist;
//			}
//		}
//	}
	CalcCenter(pDObj);
	CalcRadius(pDObj);
	if( mBound.mRadius==0.0f ){
		mBound.mRadius = MINRADIUS;
	}
	mType   = SPHERE;
}

WK7JObjBoundSphere::~WK7JObjBoundSphere()
{
}

bool WK7JObjBoundSphere::Initialize(HSD_DObj *pDObj)
{
	CalcCenter(pDObj);
	CalcRadius(pDObj);
	if( mBound.mRadius==0.0f ){
		mBound.mRadius = MINRADIUS;
	}
	return true;
}

int WK7JObjBoundSphere::GetCrossState(Mtx lwmat,ghmGcViewVolume *pViewVolume)
{
	MTXMultVec(lwmat,&mCenter,&mBound.mPosition);
	return pViewVolume->GetCrossState(mBound);
}

int WK7JObjBoundSphere::GetFarClip(f32 dist,Mtx lwmat,ghmGcViewVolume *pViewVolume)
{
	MTXMultVec(lwmat,&mCenter,&mBound.mPosition);
	return pViewVolume->GetFarClip(mBound,dist);
}

int WK7JObjBoundBox::CalcExtent(HSD_DObj *pDObj,Vec &max,Vec &min)
{
	int		totalvertnum = 0;
	for( ; pDObj!=NULL ; pDObj=pDObj->next){
		totalvertnum += CalcExtent(pDObj->pobj,max,min);
	}
	return totalvertnum;
}

int WK7JObjBoundBox::CalcExtent(HSD_PObj *pPObj,Vec &max,Vec &min)
{
	int		totalvertnum = 0;
	for( ; pPObj!=NULL ; pPObj=pPObj->next){
		WK7VtxDescInfo		info;
		u32					stride;
		stride  = WK7JObj::GetStride(pPObj->desc_list,info);
		if( info.mpVertexDesc==NULL ){
			continue;
		}
		int				dlsize;
		u8				*pdl;
		int				num;
		GXPrimitive		prim;
		pdl     = pPObj->dl;
		dlsize  = (pPObj->dlsize << 5);
		while( dlsize > 0 ){
			prim    = (GXPrimitive)(pdl[0] & GX_OPCODE_MASK);
			num     = ((int)pdl[1]<<8)|pdl[2];
			dlsize -= 3;
			pdl    += 3;
			switch( prim ){
			case GX_POINTS:
			case GX_LINES:
			case GX_LINESTRIP:
			case GX_TRIANGLES:
			case GX_TRIANGLESTRIP:
			case GX_TRIANGLEFAN:
			case GX_QUADS:{
				u8		*pVertexDL = pdl + info.mVertexPos;
				int		maxid      = -1;
				int		id;
				pdl += stride * num;
				dlsize -= stride * num;
				while( num > 0 ){
					if( maxid < (id=WK7JObj::DlGetIndex(pVertexDL,info.mpVertexDesc)) ){
						maxid = id;
					}
					num --;
					pVertexDL += stride;
				}
				WK7GetVertexFunc		*pGetVertex;
				f32					base;
				switch( info.mpVertexDesc->comp_type ){
				case	GX_U8:  pGetVertex = WK7JObj::GetVertexU8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S8:  pGetVertex = WK7JObj::GetVertexS8;  base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_U16: pGetVertex = WK7JObj::GetVertexU16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_S16: pGetVertex = WK7JObj::GetVertexS16; base = 1.0f / (1<<info.mpVertexDesc->frac); break;
				case	GX_F32: pGetVertex = WK7JObj::GetVertexF32; base = 1.0f;                                break;
					break;
				default:
					maxid = -1;
					break;
				}
				
				Vec			pos;
				void		*pBasePtr = info.mpVertexDesc->base_ptr;
				maxid++;
				totalvertnum += maxid;
				for(id=0 ; id<maxid ; id++){
					pGetVertex(&pos,pBasePtr,id,base);
					if( max.x < pos.x ){
						max.x = pos.x;
					}
					if( min.x > pos.x ){
						min.x = pos.x;
					}
					if( max.y < pos.y ){
						max.y = pos.y;
					}
					if( min.y > pos.y ){
						min.y = pos.y;
					}
					if( max.z < pos.z ){
						max.z = pos.z;
					}
					if( min.z > pos.z ){
						min.z = pos.z;
					}
				}
				
				}break;
			case GX_NOP:
			default:
				break;
			}
		}
	}
	return totalvertnum;
}

#define	MINEXTENT		(5.0f)
WK7JObjBoundBox::WK7JObjBoundBox()
{
	mCenter.x = 0.0f;
	mCenter.y = 0.0f;
	mCenter.z = 0.0f;
	mBound.mExtent.x = MINEXTENT;
	mBound.mExtent.y = MINEXTENT;
	mBound.mExtent.z = MINEXTENT;
	mType   = BOX;
}

WK7JObjBoundBox::WK7JObjBoundBox(HSD_DObj *pDObj)
{
	Vec				max;
	Vec				min;
	int				num;
//	vector<C_Vec>	vertlist;
	max.x = 
	max.y = 
	max.z = -MAXFLOAT;
	min.x = 
	min.y = 
	min.z =  MAXFLOAT;
	if( (num=CalcExtent(pDObj,max,min)) > 0  ){
//	if( (num=MakeVertex(vertlist,pDObj)) > 0  ){
//		u32		i;
//		for(i=0 ; i<(u32)num ; i++){
//			if( max.x < vertlist[i].x ){
//				max.x = vertlist[i].x;
//			}
//			if( min.x > vertlist[i].x ){
//				min.x = vertlist[i].x;
//			}
//			if( max.y < vertlist[i].y ){
//				max.y = vertlist[i].y;
//			}
//			if( min.y > vertlist[i].y ){
//				min.y = vertlist[i].y;
//			}
//			if( max.z < vertlist[i].z ){
//				max.z = vertlist[i].z;
//			}
//			if( min.z > vertlist[i].z ){
//				min.z = vertlist[i].z;
//			}
//		}
		mCenter.x = (max.x + min.x) * 0.5f;
		mCenter.y = (max.y + min.y) * 0.5f;
		mCenter.z = (max.z + min.z) * 0.5f;
		mBound.mExtent.x = (max.x - min.x) * 0.5f;
		mBound.mExtent.y = (max.y - min.y) * 0.5f;
		mBound.mExtent.z = (max.z - min.z) * 0.5f;
	}else{
		mCenter.x = 0.0f;
		mCenter.y = 0.0f;
		mCenter.z = 0.0f;
		mBound.mExtent.x = 0.0f;
		mBound.mExtent.y = 0.0f;
		mBound.mExtent.z = 0.0f;
	}
	if( mBound.mExtent.x==0.0f && mBound.mExtent.y==0.0f && mBound.mExtent.z==0.0f ){
		mBound.mExtent.x = MINEXTENT;
		mBound.mExtent.y = MINEXTENT;
		mBound.mExtent.z = MINEXTENT;
	}
	mType   = BOX;
}

WK7JObjBoundBox::~WK7JObjBoundBox()
{
}

int WK7JObjBoundBox::GetCrossState(Mtx lwmat,ghmGcViewVolume *pViewVolume)
{
	Mtx		tm;
	MTXTrans (tm,mCenter.x,mCenter.y,mCenter.z);
	MTXConcat(lwmat,tm,mBound.mCoordinate);
	return pViewVolume->GetCrossState(mBound);
}

int WK7JObjBoundBox::GetFarClip(f32 dist,Mtx lwmat,ghmGcViewVolume *pViewVolume)
{
	Mtx		tm;
	MTXTrans (tm,mCenter.x,mCenter.y,mCenter.z);
	MTXConcat(lwmat,tm,mBound.mCoordinate);
	return pViewVolume->GetFarClip(mBound,dist);
}

bool WK7JObjBoundBox::Initialize(HSD_DObj *pDObj)
{
	Vec				max;
	Vec				min;
	int				num;
	max.x = 
	max.y = 
	max.z = -MAXFLOAT;
	min.x = 
	min.y = 
	min.z =  MAXFLOAT;
	if( (num=CalcExtent(pDObj,max,min)) > 0  ){
		mCenter.x = (max.x + min.x) * 0.5f;
		mCenter.y = (max.y + min.y) * 0.5f;
		mCenter.z = (max.z + min.z) * 0.5f;
		mBound.mExtent.x = (max.x - min.x) * 0.5f;
		mBound.mExtent.y = (max.y - min.y) * 0.5f;
		mBound.mExtent.z = (max.z - min.z) * 0.5f;
	}else{
		mCenter.x = 0.0f;
		mCenter.y = 0.0f;
		mCenter.z = 0.0f;
		mBound.mExtent.x = 0.0f;
		mBound.mExtent.y = 0.0f;
		mBound.mExtent.z = 0.0f;
	}
	if( mBound.mExtent.x==0.0f && mBound.mExtent.y==0.0f && mBound.mExtent.z==0.0f ){
		mBound.mExtent.x = MINEXTENT;
		mBound.mExtent.y = MINEXTENT;
		mBound.mExtent.z = MINEXTENT;
	}
	return true;
}

HSD_PEDesc	shadowPEDesc = {
	//J[AAt@AŒAt@݁AZobt@SăGiuB
	ENABLE_ALPHA_UPDATE|ENABLE_COMPARE|BEFORE_TEX,
	//GXSetAlphaCompare(),ref0,ref1
	//At@RyA͎gȂB
	0,0,
	//GXSetDstAlpha(),dst_alpha
	WK7RENDEROBJ_RENDERID_DEFAULT_ALPHA,
	//GXSetBlendMode(),type,src_factor,dst_factor,logic_op
	//uh͂Ȃ̂src_factorAdst_factor, logic_op͂ǂł悢B
//	GX_BM_NONE,GX_BL_DSTALPHA,GX_BL_INVDSTALPHA,GX_LO_CLEAR,
	GX_BM_BLEND,GX_BL_SRCALPHA,GX_BL_INVSRCALPHA,GX_LO_CLEAR,
	//GXSetZMode(),func
	//Zr͒ʏǂB
	GX_LEQUAL,
	//GXAlphaCompare,comp0,op,comp1
	//At@RyA͎gȂB
	GX_ALWAYS,GX_AOP_AND,GX_ALWAYS
};
bool WK7JObj::SetRenderMode(u32 rendermode,bool next,bool child)
{
	bool	result = true;
	if( mpEdgeShadow==NULL ){
		rendermode &= ~WK7JOBJ_RENDERMODE_SHADOW;
		result = false;
	}else{
//		if( mpJObj!=NULL ){
			if( mRenderMode & WK7JOBJ_RENDERMODE_SHADOW ){
				if( !(rendermode & WK7JOBJ_RENDERMODE_SHADOW) ){
					mpEdgeShadow->Delete(this);
					mEdgeShadowObjNum--;
//					mpEdgeShadow->Delete(mpJObj);
				}
			}else{
				if( rendermode & WK7JOBJ_RENDERMODE_SHADOW ){
					mpEdgeShadow->Add(this);
					mEdgeShadowObjNum++;
//					mpEdgeShadow->Add(mpJObj);
//					SetPE(&shadowPEDesc);
					SetPE(NULL);
				}
			}
//		}
	}
	mRenderMode = rendermode;
	if( child==true && mpChild!=NULL ){
		mpChild->_SetRenderMode(rendermode,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->_SetRenderMode(rendermode,false,child);
		}
	}
	return result;
}

bool WK7JObj::_SetRenderMode(u32 rendermode,bool next,bool child)
{
	mRenderMode = rendermode;
	if( rendermode & WK7JOBJ_RENDERMODE_SHADOW ){
//		SetPE(&shadowPEDesc);
		SetPE(NULL);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->_SetRenderMode(rendermode,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->_SetRenderMode(rendermode,false,child);
		}
	}
	return true;
}

WK7JObj::WK7JObj(HSD_JObj *pJObj,HSD_Joint *pJoint,WK7JObj *pParent,WK7JObj *pPrev)
{
	mpAlphaControl    = NULL;
	mpFog             = NULL;
	mpCurrentFog      = NULL;
	mpShapeAnim       = NULL;
	mpBound           = NULL;
	mMotionRate       = 1.0f;
	mpInterpolateInfo = NULL;
	mpEdgeArray       = NULL;
	mpJObj            = pJObj;
	mpJoint           = pJoint;
	mpParent          = pParent;
	mpPrev            = pPrev;
	mMotionState      = 0;
	mpCurrRotate      = NULL;
	mpAnimJoint       = NULL;
	mpMatAnimJoint    = NULL;
	mpShapeAnimJoint  = NULL;
	mRenderMode       = WK7JOBJ_RENDERMODE_NORMAL;
	mpAttr            = NULL;
	mFlag             = 0;
	mZSortPrio        = WK7JOBJ_ZSORTPRIO_DEFAULT;
	if( pJoint!=NULL ){
		if( mpCurrentSdExpInfo != NULL && mpCurrentSdExpInfo->mVersion >= 127 ){
			mpAttr = (WK7JObjAttr *)(pJoint + 1);
			mZSortPrio = (u32)(WK7JOBJ_ZSORTPRIO_DEFAULT + (mpAttr->mPrio * WK7JOBJ_ZSORTPRIO_SCALE));
			if( mpCurrentSdExpInfo->mVersion >= 171 && mpAttr->mFogDisable ){
				LoadFogDesc(NULL);
			}
		}
		mpName = ::GetDescName(pJoint);
	}else{
		mpName = "";
	}
	if( mpName!=NULL ){
		if( strstr(mpName,kClothJointPrefix)==mpName ){
			mClothJoint = 1;
		}
	}
	if( pJObj!=NULL ){
		
		HSD_JObjSetupMatrix(pJObj);
		
		HSD_DObj		*pDObj;
		HSD_ShapeSet	*pShapeSet = NULL;
		u32				vertexBufferSize,normalBufferSize;
		vertexBufferSize = 0;
		normalBufferSize = 0;
		for(pDObj=HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
			GetBufferSize(pDObj,vertexBufferSize,normalBufferSize,pShapeSet);
		}
		if( pShapeSet!=NULL ){
			ghmGcMemSetAllocName("WK7JObjSA");
			mpShapeAnim = new WK7JObjShapeAnim(pShapeSet,vertexBufferSize,normalBufferSize);
		}
		
#if 0
		if( pShapeSet==NULL && !(pJObj->flags & JOBJ_ENVELOPE_MODEL) ){
			
			WK7JObjBoundSphere	*pSphere;
			WK7JObjBoundBox		*pBox;
			HSD_DObj			*pDObj = HSD_JObjGetDObj(pJObj);
			
			ghmGcMemSetAllocName("WK7JObjBS");
			pSphere = new WK7JObjBoundSphere();
//			pSphere = new WK7JObjBoundSphere(HSD_JObjGetDObj(pJObj));
			ghmGcMemSetAllocName("WK7JObjBB");
			pBox    = new WK7JObjBoundBox   ();
//			pBox    = new WK7JObjBoundBox   (HSD_JObjGetDObj(pJObj));
//			pSphere = NULL;
//			pBox    = NULL;
			
			if( pSphere!=NULL ){
				pSphere->Initialize(pDObj);
			}
			if( pBox!=NULL ){
				pBox->Initialize(pDObj);
			}
			
			if( pSphere==NULL ){
				mpBound = pBox;
				GHM_SAFE_DELETE(pSphere);
			}else if( pBox==NULL ){
				mpBound = pSphere;
				GHM_SAFE_DELETE(pBox);
			}else{
				ghmGcCollSphere		*pSBound = pSphere->GetBound();
				ghmGcCollBox		*pBBound = pBox   ->GetBound();
				f32					svol = pSBound->mRadius * pSBound->mRadius * pSBound->mRadius * (4.0f * PI / 3.0f);
				f32					bvol = 8.0f * pBBound->mExtent.x * pBBound->mExtent.y * pBBound->mExtent.z;
				if( svol <= bvol ){
					mpBound = pSphere;
					GHM_SAFE_DELETE(pBox);
				}else{
					mpBound = pBox;
					GHM_SAFE_DELETE(pSphere);
				}
			}
	//		mpBound = new WK7JObjBoundSphere(HSD_JObjGetDObj(pJObj));
	//		mpBound = new WK7JObjBoundBox(HSD_JObjGetDObj(pJObj));
		}
#endif
		
		SetPE(pJObj,WK7RenderObj::GetNormalPEDesc(),pJoint);
	}
	if( pPrev != NULL ){
		pPrev->mpNext = this;
	}
	if( pJoint->child!=NULL ){
		ghmGcMemSetAllocName("WK7JObj");
		mpChild = new WK7JObj(pJObj->child,pJoint->child,this,NULL);
	}else{
		mpChild = NULL;
	}
	mpNext = NULL;
	if( pPrev==NULL ){
		WK7JObj	*pNext;
		pPrev  = this;
		for(pJObj=pJObj->next,pJoint=pJoint->next ; pJoint!=NULL && pJObj!=NULL ; pJObj=pJObj->next,pJoint=pJoint->next){
			ghmGcMemSetAllocName("WK7JObj");
			if( (pNext = new WK7JObj(pJObj,pJoint,pParent,pPrev))==NULL ){
				break;
			}
			pPrev = pNext;
		}
	}
	mpPreFrameProcessCallBack   = NULL;
	mpPreFrameProcessData       = NULL;
	mpPostFrameProcessCallBack  = NULL;
	mpPostFrameProcessData      = NULL;
	mpPreRenderProcessCallBack  = NULL;
	mpPreRenderProcessData      = NULL;
	mpPostRenderProcessCallBack = NULL;
	mpPostRenderProcessData     = NULL;
	mpReferencedObj             = NULL;
	mpEffectBreakObj            = NULL;
}

WK7JObj::~WK7JObj()
{
	ReleaseAlphaControl();
	
	GHM_SAFE_DELETE(mpFog);
	
	WK7EffectBreakObj::Release(mpEffectBreakObj);
	
	ClearAttachModel();
	
	// qbg`FbNIuWFNgɓĂ烊[X
	WK7HitObj::Release(this);
	
	GHM_SAFE_DELETE(mpCurrRotate);
	GHM_SAFE_DELETE(mpChild);
	GHM_SAFE_DELETE(mpNext );
	GHM_SAFE_DELETE(mpEdgeArray);
	GHM_SAFE_DELETE(mpInterpolateInfo);
	
	// ~[Ƃ̉
	WK7Mirror::ReleaseJObj(this);
	// eeƂ̉
	WK7Shadow::ReleaseJObj(this);
	
	GHM_SAFE_DELETE(mpShapeAnim);
	GHM_SAFE_DELETE(mpBound    );
	
	if( mpParent==NULL ){
		if( mpJObj!=NULL ){
			HSD_JObjUnref(mpJObj); 
			mpJObj = NULL;
		}
	}
}

bool WK7JObj::MakeBound(bool next,bool child)
{
	if( mpJObj!=NULL && mpShapeAnim==NULL && !(mpJObj->flags & JOBJ_ENVELOPE_MODEL) ){
		
		WK7JObjBoundSphere	*pSphere;
		WK7JObjBoundBox		*pBox;
		HSD_DObj			*pDObj = HSD_JObjGetDObj(mpJObj);
		
		GHM_SAFE_DELETE(mpBound);
		
		ghmGcMemSetAllocName("WK7JObjBS");
		pSphere = new WK7JObjBoundSphere();
		ghmGcMemSetAllocName("WK7JObjBB");
		pBox    = new WK7JObjBoundBox   ();
		
		if( pSphere!=NULL ){
			pSphere->Initialize(pDObj);
		}
		if( pBox!=NULL ){
			pBox->Initialize(pDObj);
		}
		
		if( pSphere==NULL ){
			mpBound = pBox;
			GHM_SAFE_DELETE(pSphere);
		}else if( pBox==NULL ){
			mpBound = pSphere;
			GHM_SAFE_DELETE(pBox);
		}else{
			ghmGcCollSphere		*pSBound = pSphere->GetBound();
			ghmGcCollBox		*pBBound = pBox   ->GetBound();
			f32					svol = pSBound->mRadius * pSBound->mRadius * pSBound->mRadius * (4.0f * PI / 3.0f);
			f32					bvol = 8.0f * pBBound->mExtent.x * pBBound->mExtent.y * pBBound->mExtent.z;
			if( svol <= bvol ){
				mpBound = pSphere;
				GHM_SAFE_DELETE(pBox);
			}else{
				mpBound = pBox;
				GHM_SAFE_DELETE(pSphere);
			}
		}
	}
	if( child==true && mpChild!=NULL ){
		mpChild->MakeBound(true,true);
	}
	if( next==true ){
		WK7JObj		*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			pJObj->MakeBound(false,child);
		}
	}
	return true;
}

bool WK7JObj::LoadFogDesc(HSD_FogDesc *pFogDesc)
{
	ReleaseFog();
	ghmGcMemSetAllocName("JOBJFOG");
	mpFog = new WK7Fog(pFogDesc);
	return true;
}

bool WK7JObj::ReleaseFog(void)
{
	GHM_SAFE_DELETE(mpFog);
	return true;
}

bool WK7JObj::SetFogEnable(bool flag)
{
	if( mpFog==NULL ){
		return false;
	}
	mpFog->SetEnable(flag);
	return true;
}

bool WK7JObj::GetFogEnable(void)
{
	if( mpFog==NULL ){
		return false;
	}
	return mpFog->GetEnable();
}

WK7EffectBreakObj* WK7JObj::CreateEffectBreakObj(bool next,bool child,int level)
{
	WK7EffectBreakObj::Release(mpEffectBreakObj);
	return (mpEffectBreakObj = WK7EffectBreakObj::Create(this,next,child,level));
}

bool WK7JObj::ReleaseEffectBreakObj(void)
{
	WK7EffectBreakObj::Release(mpEffectBreakObj);
	return true;
}

const char* WK7JObj::GetNameAttr(void)
{
	if( mpAttr==NULL ){
		return NULL;
	}
	return mpAttr->GetName();
}

bool WK7JObj::SetViewVolume(HSD_CObj *pCObj)
{
	GHM_SAFE_DELETE(mpViewVolume);
	if( pCObj==NULL ){
		return true;
	}
	ghmGcMemSetAllocName("ghmGcVV");
	if( (mpViewVolume = new ghmGcViewVolume)==NULL ){
		return false;
	}
	if( mpViewVolume->SetCurrent(pCObj)==false ){
		GHM_SAFE_DELETE(mpViewVolume);
	}
	return true;
}

bool WK7JObj::SetZSortPrioS(s32 prio,bool next,bool child)
{
	mZSortPrio = (u32)(WK7JOBJ_ZSORTPRIO_DEFAULT + prio * WK7JOBJ_ZSORTPRIO_SCALE);
	if( child==true && mpChild!=NULL ){
		mpChild->SetZSortPrio(mZSortPrio,true,true);
	}
	if( next==true ){
		WK7JObj		*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			pJObj->SetZSortPrio(mZSortPrio,false,child);
		}
	}
	return true;
}

bool WK7JObj::SetZSortPrio(u32 prio,bool next,bool child)
{
	mZSortPrio = prio;
	if( child==true && mpChild!=NULL ){
		mpChild->SetZSortPrio(prio,true,true);
	}
	if( next==true ){
		WK7JObj		*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			pJObj->SetZSortPrio(prio,false,child);
		}
	}
	return true;
}

bool WK7JObj::GetOpa(void)
{
	if( mpJObj==NULL ){
		return false;
	}
	return (HSD_JObjGetFlags(mpJObj) & JOBJ_TRSP_MASK)==JOBJ_OPA;
}

bool WK7JObj::GetXlu(void)
{
	if( mpJObj==NULL ){
		return false;
	}
	return (HSD_JObjGetFlags(mpJObj) & JOBJ_TRSP_MASK)==JOBJ_XLU;
}

bool WK7JObj::GetTexEdge(void)
{
	if( mpJObj==NULL ){
		return false;
	}
	return (HSD_JObjGetFlags(mpJObj) & JOBJ_TRSP_MASK)==JOBJ_TEXEDGE;
}

bool WK7JObj::SetAlphaUpdate(bool flag,bool next,bool child,u32 trsp)
{
	HSD_DObj	*pDObj;
	HSD_MObj	*pMObj;
	
	if( trsp==0xffffffff ||
		((HSD_JObjGetFlags(mpJObj) & JOBJ_TRSP_MASK) & trsp) ){
		for(pDObj = HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
			if( (pMObj = HSD_DObjGetMObj(pDObj)) != NULL ){
				if( pMObj->pe != NULL ){
					if( flag==true ){
						pMObj->pe->flags |=  ENABLE_ALPHA_UPDATE;
					}else{
						pMObj->pe->flags &= ~ENABLE_ALPHA_UPDATE;
					}
				}
			}
		}
	}
	if( child==true && mpChild!=NULL ){
		mpChild->SetAlphaUpdate(flag,true,true ,trsp);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetAlphaUpdate(flag,false,child,trsp);
		}
	}
//	if( next==true && mpNext!=NULL ){
//		mpNext ->SetAlphaUpdate(flag,true,child,trsp);
//	}
	return true;
}

bool WK7JObj::SetNoZUpdate(bool flag,bool next,bool child,u32 trsp)
{
	HSD_DObj	*pDObj;
	HSD_MObj	*pMObj;
	
	if( trsp==0xffffffff ||
		((HSD_JObjGetFlags(mpJObj) & JOBJ_TRSP_MASK) & trsp) ){
		for(pDObj = HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
			if( (pMObj = HSD_DObjGetMObj(pDObj)) == NULL ){
				continue;
			}
			if( flag==true ){
				pMObj->rendermode |=  RENDER_NO_ZUPDATE;
			}else{
				pMObj->rendermode &= ~RENDER_NO_ZUPDATE;
			}
			if( pMObj->pe==NULL ){
				continue;
			}
			if( flag==true ){
				pMObj->pe->flags &= ~ENABLE_ZUPDATE;
			}else{
				pMObj->pe->flags |=  ENABLE_ZUPDATE;
			}
		}
	}
	if( child==true && mpChild!=NULL ){
		mpChild->SetNoZUpdate(flag,true,true ,trsp);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetNoZUpdate(flag,false,child,trsp);
		}
	}
	return true;
}

bool WK7JObj::SetPE(HSD_JObj *pJObj,HSD_PEDesc *pPEDesc,HSD_Joint *pJoint)
{
	HSD_DObj		*pDObj;
	HSD_MObj		*pMObj;
	HSD_DObjDesc	*pDObjDesc;
	HSD_MObjDesc	*pMObjDesc;
	
	if( pJoint!=NULL ){
		pDObjDesc = pJoint->u.dobjdesc;
	}else{
		pDObjDesc = NULL;
	}
	pDObj = HSD_JObjGetDObj(pJObj);
	
	while( pDObj!=NULL && (pJoint==NULL || pDObjDesc!=NULL) ){
		
		if( pDObjDesc!=NULL ){
			pMObjDesc = pDObjDesc->mobjdesc;
		}else{
			pMObjDesc = NULL;
		}
		
		if( (pMObj=HSD_DObjGetMObj(pDObj))!=NULL && (pDObjDesc==NULL || pMObjDesc!=NULL) ){
			if( pMObj->pe!=NULL ){
				hsdFreeMemPiece(pMObj->pe,sizeof(HSD_PEDesc));
				pMObj->pe = NULL;
			}
			if( pPEDesc!=NULL ){
				if( (pMObj->pe = (HSD_PEDesc*)hsdAllocMemPiece(sizeof(HSD_PEDesc)))!=NULL ){
					*pMObj->pe = *pPEDesc;
					if( pMObj->rendermode & RENDER_NO_ZUPDATE ){
						pMObj->pe->flags &= ~ENABLE_ZUPDATE;
					}else{
						pMObj->pe->flags |=  ENABLE_ZUPDATE;
					}
					HSD_PEDesc	*pMObjDescPEDesc;
					if( pMObjDesc!=NULL && (pMObjDescPEDesc=pMObjDesc->pedesc)!=NULL ){
						pMObj->pe->type       = pMObjDescPEDesc->type;
						pMObj->pe->src_factor = pMObjDescPEDesc->src_factor;
						pMObj->pe->dst_factor = pMObjDescPEDesc->dst_factor;
						pMObj->pe->logic_op   = pMObjDescPEDesc->logic_op;
					}
				}
			}
		}
		
	    pDObj=HSD_DObjGetNext(pDObj);
	    if( pDObjDesc != NULL ){
			pDObjDesc = pDObjDesc->next;
		}
		
	}
	
	return true;
}

bool WK7JObj::SetPE(HSD_PEDesc *pPEDesc,bool next,bool child)
{
	if( mpJObj==NULL ){
		return false;
	}
	SetPE(mpJObj,pPEDesc,mpJoint);
	if( child==true && mpChild!=NULL ){
		mpChild->SetPE(pPEDesc,true,true );
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetPE(pPEDesc,false,child);
		}
	}
	return true;
}

bool WK7JObj::SetDstAlpha(HSD_JObj *pJObj,u8 alpha)
{
	HSD_DObj	*pDObj;
	HSD_MObj	*pMObj;
	
	for(pDObj = HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
		if( (pMObj = HSD_DObjGetMObj(pDObj)) != NULL ){
			if( pMObj->pe!=NULL ){
				pMObj->pe->dst_alpha = alpha;
			}
		}
	}
	
	return true;
}

bool WK7JObj::SetDstAlpha(u8 alpha,bool next,bool child)
{
	if( mpJObj==NULL ){
		return false;
	}
	SetDstAlpha(mpJObj,alpha);
	if( child==true && mpChild!=NULL ){
		mpChild->SetDstAlpha(alpha,true,true );
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->SetDstAlpha(alpha,true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetDstAlpha(alpha,false,child);
		}
	}
	return true;
}

WK7AlphaControl::WK7AlphaControl(HSD_JObj *pJObj,f32 alpha,bool zupdate)
{
	if( pJObj==NULL ){
		mpAlpha = NULL;
	}else{
		mpAlpha = new C_SetAlpha(pJObj,alpha,false);
	}
	if( mpAlpha!=NULL ){
		mpAlpha->SetVal(alpha,zupdate);
		mAlpha.SetValue(alpha);
	}
}

WK7AlphaControl::~WK7AlphaControl()
{
	GHM_SAFE_DELETE(mpAlpha);
}

bool WK7AlphaControl::FrameProcess(void)
{
	mAlpha.FrameProcess();
	if( mpAlpha==NULL ){
		return false;
	}
	mpAlpha->SetVal(mAlpha.GetValue(),mZUpDate);
	return true;
}

bool WK7AlphaControl::SetAlpha(f32 alpha,u32 tick)
{
	mAlpha.SetValue(alpha,tick);
	return true;
}

WK7AlphaControl* WK7JObj::GetAlphaControl(bool next,bool child)
{
	WK7AlphaControl	*pAlphaControl;
	if( mpAlphaControl!=NULL ){
		return mpAlphaControl;
	}
	if( child==true && mpChild!=NULL && (pAlphaControl=mpChild->GetAlphaControl(true,true))!=NULL ){
		return pAlphaControl;
	}
	if( next==true ){
		WK7JObj	*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			if( (pAlphaControl=pJObj->GetAlphaControl(false,true))!=NULL ){
				return pAlphaControl;
			}
		}
	}
	return NULL;
}

bool WK7JObj::CreateAlphaControl(f32 alpha,bool zupdate)
{
	if( mpJObj==NULL ){
		return false;
	}
	if( mpAlphaControl!=NULL ){
		return true;
	}
	if( GetAlphaControl(false,true)!=NULL ){
		return false;
	}
	WK7JObj	*pJObj;
	for(pJObj=mpParent ; pJObj!=NULL ; pJObj=pJObj->GetParent()){
		if( pJObj->GetAlphaControl()!=NULL ){
			return false;
		}
	}
	ghmGcMemSetAllocName("JOACtrl");
	if( (mpAlphaControl = new WK7AlphaControl(mpJObj,alpha,zupdate))==NULL ){
		return false;
	}
	return true;
}

bool WK7JObj::ReleaseAlphaControl(void)
{
	GHM_SAFE_DELETE(mpAlphaControl);
	return true;
}

bool WK7JObj::SetAlpha(f32 alpha,u32 tick)
{
	if( mpAlphaControl==NULL ){
		return false;
	}
	mpAlphaControl->SetAlpha(alpha,tick);
	return true;
}

f32 WK7JObj::GetAlpha(void)
{
	if( mpAlphaControl==NULL ){
		return 1.0f;
	}
	return mpAlphaControl->GetAlpha();
}

bool WK7JObj::IsAnimatingAlpha(void)
{
	if( mpAlphaControl==NULL ){
		return false;
	}
	return mpAlphaControl->IsAnimatingAlpha();
}

bool WK7JObj::SetMaterialAmbientColor(HSD_MObj *pMObj,GXColor col)
{
	if( pMObj==NULL ){
		return true;
	}
	HSD_MObjSetAmbientColor (pMObj,col.r,col.g,col.b);
	return true;
}

bool WK7JObj::SetMaterialDiffuseColor(HSD_MObj *pMObj,GXColor col)
{
	if( pMObj==NULL ){
		return true;
	}
	HSD_MObjSetDiffuseColor (pMObj,col.r,col.g,col.b);
	return true;
}

bool WK7JObj::SetMaterialSpecularColor(HSD_MObj *pMObj,GXColor col)
{
	if( pMObj==NULL ){
		return true;
	}
	HSD_MObjSetSpecularColor(pMObj,col.r,col.g,col.b);
	return true;
}

bool WK7JObj::SetMaterialAmbientColor(HSD_DObj *pDObj,GXColor col)
{
	return SetMaterialAmbientColor(HSD_DObjGetMObj(pDObj),col);
}

bool WK7JObj::SetMaterialDiffuseColor(HSD_DObj *pDObj,GXColor col)
{
	return SetMaterialDiffuseColor(HSD_DObjGetMObj(pDObj),col);
}

bool WK7JObj::SetMaterialSpecularColor(HSD_DObj *pDObj,GXColor col)
{
	return SetMaterialSpecularColor(HSD_DObjGetMObj(pDObj),col);
}

bool WK7JObj::SetMaterialAmbientColor(HSD_JObj *pJObj,GXColor col)
{
	HSD_DObj	*pDObj;
	for(pDObj = HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
		SetMaterialAmbientColor(pDObj,col);
	}
	return true;
}

bool WK7JObj::SetMaterialDiffuseColor(HSD_JObj *pJObj,GXColor col)
{
	HSD_DObj	*pDObj;
	for(pDObj = HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
		SetMaterialDiffuseColor(pDObj,col);
	}
	return true;
}

bool WK7JObj::SetMaterialSpecularColor(HSD_JObj *pJObj,GXColor col)
{
	HSD_DObj	*pDObj;
	for(pDObj = HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
		SetMaterialSpecularColor(pDObj,col);
	}
	return true;
}

bool WK7JObj::SetMaterialAmbientColor(GXColor col,bool next,bool child)
{
	SetMaterialAmbientColor(mpJObj,col);
	if( mpChild!=NULL && child==true ){
		mpChild->SetMaterialAmbientColor(col,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMaterialAmbientColor(col,false,child);
		}
	}
	return true;
}

bool WK7JObj::SetMaterialDiffuseColor(GXColor col,bool next,bool child)
{
	SetMaterialDiffuseColor(mpJObj,col);
	if( mpChild!=NULL && child==true ){
		mpChild->SetMaterialDiffuseColor(col,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMaterialDiffuseColor(col,false,child);
		}
	}
	return true;
}

bool WK7JObj::SetMaterialSpecularColor(GXColor col,bool next,bool child)
{
	SetMaterialSpecularColor(mpJObj,col);
	if( mpChild!=NULL && child==true ){
		mpChild->SetMaterialSpecularColor(col,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMaterialSpecularColor(col,false,child);
		}
	}
	return true;
}

bool WK7JObj::SetMaterialAmbientColorLE(HSD_MObj *pMObj,GXColor col)
{
	HSD_Material	*pMaterial;
	if( pMObj==NULL || (pMaterial=pMObj->mat)==NULL ){
		return true;
	}
	if( col.r > pMaterial->ambient.r ){
		col.r = pMaterial->ambient.r;
	}
	if( col.g > pMaterial->ambient.g ){
		col.g = pMaterial->ambient.g;
	}
	if( col.b > pMaterial->ambient.b ){
		col.b = pMaterial->ambient.b;
	}
	HSD_MObjSetAmbientColor (pMObj,col.r,col.g,col.b);
	return true;
}

bool WK7JObj::SetMaterialDiffuseColorLE(HSD_MObj *pMObj,GXColor col)
{
	HSD_Material	*pMaterial;
	if( pMObj==NULL || (pMaterial=pMObj->mat)==NULL ){
		return true;
	}
	if( col.r > pMaterial->diffuse.r ){
		col.r = pMaterial->diffuse.r;
	}
	if( col.g > pMaterial->diffuse.g ){
		col.g = pMaterial->diffuse.g;
	}
	if( col.b > pMaterial->diffuse.b ){
		col.b = pMaterial->diffuse.b;
	}
	HSD_MObjSetDiffuseColor (pMObj,col.r,col.g,col.b);
	return true;
}

bool WK7JObj::SetMaterialSpecularColorLE(HSD_MObj *pMObj,GXColor col)
{
	HSD_Material	*pMaterial;
	if( pMObj==NULL || (pMaterial=pMObj->mat)==NULL ){
		return true;
	}
	if( col.r > pMaterial->specular.r ){
		col.r = pMaterial->specular.r;
	}
	if( col.g > pMaterial->specular.g ){
		col.g = pMaterial->specular.g;
	}
	if( col.b > pMaterial->specular.b ){
		col.b = pMaterial->specular.b;
	}
	HSD_MObjSetSpecularColor(pMObj,col.r,col.g,col.b);
	return true;
}

bool WK7JObj::SetMaterialAmbientColorLE(HSD_DObj *pDObj,GXColor col)
{
	return SetMaterialAmbientColorLE(HSD_DObjGetMObj(pDObj),col);
}

bool WK7JObj::SetMaterialDiffuseColorLE(HSD_DObj *pDObj,GXColor col)
{
	return SetMaterialDiffuseColorLE(HSD_DObjGetMObj(pDObj),col);
}

bool WK7JObj::SetMaterialSpecularColorLE(HSD_DObj *pDObj,GXColor col)
{
	return SetMaterialSpecularColorLE(HSD_DObjGetMObj(pDObj),col);
}

bool WK7JObj::SetMaterialAmbientColorLE(HSD_JObj *pJObj,GXColor col)
{
	HSD_DObj	*pDObj;
	for(pDObj = HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
		SetMaterialAmbientColorLE(pDObj,col);
	}
	return true;
}

bool WK7JObj::SetMaterialDiffuseColorLE(HSD_JObj *pJObj,GXColor col)
{
	HSD_DObj	*pDObj;
	for(pDObj = HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
		SetMaterialDiffuseColorLE(pDObj,col);
	}
	return true;
}

bool WK7JObj::SetMaterialSpecularColorLE(HSD_JObj *pJObj,GXColor col)
{
	HSD_DObj	*pDObj;
	for(pDObj = HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj) ){
		SetMaterialSpecularColorLE(pDObj,col);
	}
	return true;
}

bool WK7JObj::SetMaterialAmbientColorLE(GXColor col,bool next,bool child)
{
	SetMaterialAmbientColorLE(mpJObj,col);
//	if( mpNext!=NULL && next==true ){
//		mpNext->SetMaterialAmbientColorLE(col,true,child);
//	}
	if( mpChild!=NULL && child==true ){
		mpChild->SetMaterialAmbientColorLE(col,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMaterialAmbientColorLE(col,false,child);
		}
	}
	return true;
}

bool WK7JObj::SetMaterialDiffuseColorLE(GXColor col,bool next,bool child)
{
	SetMaterialDiffuseColorLE(mpJObj,col);
//	if( mpNext!=NULL && next==true ){
//		mpNext->SetMaterialDiffuseColorLE(col,true,child);
//	}
	if( mpChild!=NULL && child==true ){
		mpChild->SetMaterialDiffuseColorLE(col,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMaterialDiffuseColorLE(col,false,true);
		}
	}
	return true;
}

bool WK7JObj::SetMaterialSpecularColorLE(GXColor col,bool next,bool child)
{
	SetMaterialSpecularColorLE(mpJObj,col);
//	if( mpNext!=NULL && next==true ){
//		mpNext->SetMaterialSpecularColorLE(col,true,child);
//	}
	if( mpChild!=NULL && child==true ){
		mpChild->SetMaterialSpecularColorLE(col,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMaterialSpecularColorLE(col,false,child);
		}
	}
	return true;
}

bool WK7JObj::CreateEdge(void)
{
	if( ReleaseEdge()==false ){
		return false;
	}
	if( mpJObj==NULL ){
		return true;
	}
	HSD_DObj		*pDObj = HSD_JObjGetDObj(mpJObj);
	if( pDObj==NULL ){
		return true;
	}
	HSD_PObj		*pPObj = HSD_DObjGetPObj(pDObj);
	if( pPObj==NULL ){
		return true;
	}
	ghmGcMemSetAllocName("C_EdgeAr");
	mpEdgeArray = new C_EdgeArray(pPObj,mpJObj);
	return true;
}

bool WK7JObj::ReleaseEdge(void)
{
	GHM_SAFE_DELETE(mpEdgeArray);
	return true;
}

void WK7JObj::Hide(bool next,bool child)
{
	mHide = 1;
	if( mpJObj!=NULL ){
		HSD_JObjSetFlags(mpJObj,JOBJ_HIDDEN);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->Hide(true,true);
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->Hide(true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->Hide(false,child);
		}
	}
//	if( next==true && mpNext!=NULL ){
//		mpNext->Hide(true,child);
//	}
}

void WK7JObj::Disp(bool next,bool child)
{
	mHide = 0;
	if( mpJObj!=NULL ){
		HSD_JObjClearFlags(mpJObj,JOBJ_HIDDEN);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->Disp(true,true);
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->Disp(true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->Disp(false,child);
		}
	}
//	if( next==true && mpNext!=NULL ){
//		mpNext->Disp(true,child);
//	}
}

bool WK7JObj::IsHidden(void)
{
	if( mpJObj==NULL ){
		return true;
	}
	if( mHide ){
		return true;
	}
	if( HSD_JObjGetFlags(mpJObj) & JOBJ_HIDDEN ){
		return true;
	}
	return false;
}

bool WK7JObj::SetCastShadowDisable(bool flag,bool next,bool child)
{
	mCastShadowDisable = (flag==true? 1:0);
	if( child==true && mpChild!=NULL ){
		mpChild->SetCastShadowDisable(flag,true,true);
	}
	if( next==true ){
		WK7JObj	*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			pJObj->SetCastShadowDisable(flag,false,true);
		}
	}
	return true;
}

bool WK7JObj::GetCastShadowDisable(bool next,bool child)
{
	if( mCastShadowDisable ){
		return true;
	}
	if( child==true && mpChild!=NULL ){
		if( mpChild->GetCastShadowDisable(true,true)==true ){
			return true;
		}
	}
	if( next==true ){
		WK7JObj	*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			if( pJObj->GetCastShadowDisable(false,true)==true ){
				return true;
			}
		}
	}
	return false;
}

bool WK7JObj::HasAlpha(bool next,bool child)
{
	
	if( mpJObj!=NULL && (mpJObj->flags & (JOBJ_XLU|JOBJ_TEXEDGE)) ){
//		HSD_DObj		*pDObj;
//		HSD_MObj		*pMObj;
//		for(pDObj=HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj)){
//			if( (pMObj = HSD_DObjGetMObj(pDObj))==NULL || pMObj->pe==NULL ){
//				continue;
//			}
//			if( (pMObj->pe->flags & ENABLE_ZUPDATE) ){
//				
//			}
//		}
		return true;
	}
	
	if( child==true && mpChild!=NULL ){
		if( mpChild->HasAlpha(true,true)==true ){
			return true;
		}
	}
	if( next==true ){
		WK7JObj	*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			if( pJObj->HasAlpha(false,true)==true ){
				return true;
			}
		}
	}
	return true;
}

//HSD_AObj* GetAObj(WK7JObj *pJObj,bool next,bool child)
//{
//	if( pJObj==NULL ){
//		return NULL;
//	}
//	HSD_AObj	*pAObj;
//	HSD_JObj	*pHSDJObj;
//	if( (pHSDJObj=pJObj->GetJObj())!=NULL && pHSDJObj->aobj!=NULL ){
//		return pHSDJObj->aobj;
//	}
//	if( child==true && (pAObj=GetAObj(pJObj->GetChild(),true,true))!=NULL ){
//		return pAObj;
//	}
//	if( next==true ){
//		for(pJObj=pJObj->GetNext() ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
//			if( (pAObj=GetAObj(pJObj,false,child))!=NULL ){
//				return pAObj;
//			}
//		}
//	}
//	return NULL;
//}

bool WK7JObjMotionInterpolateInfo::FrameProcess(WK7JObj *pJObj)
{
	HSD_JObj	*pOrigJObj = pJObj->GetJObj();
	
	if( pOrigJObj==NULL ){
		return false;
	}
	
//	HSD_AObj		*pAObj;
	
	if( pOrigJObj->aobj==NULL ){
		// Aj[VȂԏI
		return false;
//		// ĐwbhXV
//		HSD_JObj	*pHSDJObj;
//		WK7JObj		*pSearchJObj;
//		for(pSearchJObj=pJObj->GetParent() ; pSearchJObj!=NULL ; pSearchJObj=pSearchJObj->GetParent()){
//			if( (pHSDJObj=pSearchJObj->GetJObj())==NULL ){
//				continue;
//			}
//			if( (pAObj=pHSDJObj->aobj)==NULL ){
//				continue;
//			}
//			break;
//		}
//		if( pSearchJObj==NULL ){
//			pAObj = GetAObj(pJObj->GetChild(),true,true);
//		}
////		return false;
//	}else{
//		// ĐwbhXV
//		pAObj = pOrigJObj->aobj;
	}
	
//	if( pAObj!=NULL ){
//		mInterpolateCurrentFrame += pAObj->rate;
//	}else{
//		mInterpolateCurrentFrame += WK7MOTION_DEFAULT_RATE;
//	}
	mInterpolateCurrentFrame += pOrigJObj->aobj->rate;
	
	// ԏI`FbN
	if( mInterpolateCurrentFrame > mInterpolateTotalFrame ){
		return false;
	}
	
	f32				rate,_rate;
	Mtx				m,tm,rm,drm;
	ghmQuaternion	srq,drq;
	ghmQuaternion	*pCurrRotate = pJObj->GetCurrRotate();
	
	rate  = mInterpolateCurrentFrame / mInterpolateTotalFrame;
	_rate = 1.0f - rate;
	
	mInterpolateTranslate.x = pOrigJObj->translate.x * rate + mPrevTranslate.x * _rate;
	mInterpolateTranslate.y = pOrigJObj->translate.y * rate + mPrevTranslate.y * _rate;
	mInterpolateTranslate.z = pOrigJObj->translate.z * rate + mPrevTranslate.z * _rate;
	mInterpolateScale    .x = pOrigJObj->scale    .x * rate + mPrevScale    .x * _rate;
	mInterpolateScale    .y = pOrigJObj->scale    .y * rate + mPrevScale    .y * _rate;
	mInterpolateScale    .z = pOrigJObj->scale    .z * rate + mPrevScale    .z * _rate;
	
	// XP[
	MTXScale (mInterpolateMatrix,mInterpolateScale.x,mInterpolateScale.y,mInterpolateScale.z);
	
	// ]
	if( pCurrRotate==NULL ){
		MTXRotRad(drm,'x',pOrigJObj->rotate.x);
		MTXRotRad(  m,'y',pOrigJObj->rotate.y);
		MTXConcat(  m,drm,drm);
		MTXRotRad(  m,'z',pOrigJObj->rotate.z);
		MTXConcat(  m,drm,drm);
		MatToQuat(drm,&drq);
	}else{
		drq.x = pCurrRotate->x;
		drq.y = pCurrRotate->y;
		drq.z = pCurrRotate->z;
		drq.w = pCurrRotate->w;
	}
	srq.x = mPrevRotate.x;
	srq.y = mPrevRotate.y;
	srq.z = mPrevRotate.z;
	srq.w = mPrevRotate.w;
	mInterpolateRotate.QuatSlerp(&srq,&drq,rate);
	MTXQuat  ( rm,&mInterpolateRotate);
	MTXConcat( rm,mInterpolateMatrix,mInterpolateMatrix);
	
	// sړ
	MTXTrans ( tm,mInterpolateTranslate.x,mInterpolateTranslate.y,mInterpolateTranslate.z);
	MTXConcat( tm,mInterpolateMatrix,mInterpolateMatrix);
	
//	WK7JObj		*pParentJObj;
	WK7JObj		*pReferencedObj;
	
//	if( (pParentJObj=pJObj->GetParent())!=NULL ){
//		MTXConcat(pParentJObj->GetLocalWorldMatrix(),mInterpolateMatrix,m);
	if( HSD_JObjGetParent(pOrigJObj) != NULL ){
		MTXConcat(HSD_JObjGetMtxPtr(HSD_JObjGetParent(pOrigJObj)),mInterpolateMatrix,m);
	}else{
		MTXCopy(mInterpolateMatrix,m);
	}
	HSD_JObjSetUserDefinedMtx(pOrigJObj,m);
	
	if( (pReferencedObj=pJObj->GetReferencedObj())!=NULL ){
		WK7JObj		*pRootObj;
		if( (pRootObj = pReferencedObj->GetParent())!=NULL ){
			for( ; pRootObj->GetParent()!=NULL ; pRootObj=pRootObj->GetParent()){}
			HSD_JObj		*pRootJObj = pRootObj->GetJObj();
			if( pRootJObj!=NULL && (pRootJObj = pRootJObj->parent)!=NULL ){
				HSD_JObjSetupMatrix(pRootJObj);
				MTXConcat(pRootJObj->mtx,m,m);
			}
		}
		HSD_JObj	*pReferJObj = pReferencedObj->GetJObj();
//		if( (pParentJObj=pReferencedObj->GetParent())!=NULL ){
//			MTXConcat(pParentJObj->GetLocalWorldMatrix(),mInterpolateMatrix,m);
//		}else{
//			MTXScale (m,pReferJObj->scale.x,pReferJObj->scale.y,pReferJObj->scale.z);
//			MTXConcat(rm,m,m);
//			MTXConcat(tm,m,m);
//		}
		HSD_JObjSetUserDefinedMtx(pReferJObj,m);
	}
	
//	if( (pParentJObj=pJObj->GetParent())!=NULL ){
//		MTXConcat(pParentJObj->GetLocalWorldMatrix(),mInterpolateMatrix,m);
//	}else{
//		MTXCopy(mInterpolateMatrix,m);
//	}
//	HSD_JObjSetUserDefinedMtx(pOrigJObj,m);
	
	return true;
}

MtxPtr WK7JObj::GetLocalWorldMatrix(void)
{
	if( mpJObj==NULL ){
		return NULL;
	}
	if( mpParent!=NULL ){
		mpParent->GetLocalWorldMatrix();
	}
	HSD_JObjSetupMatrix(mpJObj);
	return mpJObj->mtx;
}

f32 WK7JObj::GetCurrentMotionFrame(bool next,bool child)
{
	f32		frame;
	if( mpJObj!=NULL ){
		if( mpJObj->aobj!=NULL ){
			return mpJObj->aobj->frame;
		}
		HSD_DObj	*pDObj;
		for(pDObj=HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
			if( pDObj->aobj!=NULL ){
				return pDObj->aobj->frame;
			}
			if( pDObj->mobj!=NULL ){
				if( pDObj->mobj->aobj!=NULL ){
					return pDObj->mobj->aobj->frame;
				}
				HSD_TObj	*pTObj;
				for(pTObj=pDObj->mobj->tobj ; pTObj!=NULL ; pTObj=pTObj->next){
					if( pTObj->aobj!=NULL ){
						return pTObj->aobj->frame;
					}
				}
			}
			HSD_PObj	*pPObj;
			for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
				if( pPObj->aobj!=NULL ){
					return pPObj->aobj->frame;
				}
			}
		}
	}
	if( child==true && mpChild!=NULL ){
		if( (frame=mpChild->GetCurrentMotionFrame(true,true)) >= 0.0f ){
			return frame;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (frame=pNext ->GetCurrentMotionFrame(false,child)) >= 0.0f ){
				return frame;
			}
		}
	}
//	if( next ==true && mpNext !=NULL ){
//		if( (frame=mpNext ->GetCurrentMotionFrame(true,child)) >= 0.0f ){
//			return frame;
//		}
//	}
	return -1.0f;
}

f32 WK7JObj::GetTotalMotionFrame(bool next,bool child)
{
	f32		frame;
	if( mpJObj!=NULL ){
		if( mpJObj->aobj!=NULL ){
			return mpJObj->aobj->endframe;
		}
		HSD_DObj	*pDObj;
		for(pDObj=HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
			if( pDObj->aobj!=NULL ){
				return pDObj->aobj->endframe;
			}
			if( pDObj->mobj!=NULL ){
				if( pDObj->mobj->aobj!=NULL ){
					return pDObj->mobj->aobj->endframe;
				}
				HSD_TObj	*pTObj;
				for(pTObj=pDObj->mobj->tobj ; pTObj!=NULL ; pTObj=pTObj->next){
					if( pTObj->aobj!=NULL ){
						return pTObj->aobj->endframe;
					}
				}
			}
			HSD_PObj	*pPObj;
			for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
				if( pPObj->aobj!=NULL ){
					return pPObj->aobj->endframe;
				}
			}
		}
	}
	if( child==true && mpChild!=NULL ){
		if( (frame=mpChild->GetTotalMotionFrame(true,true)) >= 0.0f ){
			return frame;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (frame=pNext ->GetTotalMotionFrame(false,child)) >= 0.0f ){
				return frame;
			}
		}
	}
//	if( next ==true && mpNext !=NULL ){
//		if( (frame=mpNext ->GetTotalMotionFrame(true,child)) >= 0.0f ){
//			return frame;
//		}
//	}
	return -1.0f;
}

void WK7JObj::JObjUpdateFunc(void *obj,u32 type,FObjData *val)
{
	HSD_JObj	*pJObj = (HSD_JObj *)obj;
	
	if( pJObj==NULL ){
		return;
	}
	switch (type) {
//	case HSD_A_J_ROTX:
//		HSD_JObjSetRx(pJObj,val->fv);
//		break;
//	case HSD_A_J_ROTY:
//		HSD_JObjSetRy(pJObj,val->fv);
//		break;
//	case HSD_A_J_ROTZ:
//		HSD_JObjSetRz(pJObj,val->fv);
//		break;
	case HSD_A_J_TRAX:
		HSD_JObjSetTx(pJObj,val->fv);
		break;
	case HSD_A_J_TRAY:
		HSD_JObjSetTy(pJObj,val->fv);
		break;
	case HSD_A_J_TRAZ:
		HSD_JObjSetTz(pJObj,val->fv);
		break;
	case HSD_A_J_SCAX:
		HSD_JObjSetSx(pJObj,val->fv);
		break;
	case HSD_A_J_SCAY:
		HSD_JObjSetSy(pJObj,val->fv);
		break;
	case HSD_A_J_SCAZ:
		HSD_JObjSetSz(pJObj,val->fv);
		break;
	case HSD_A_J_BRANCH:
		if( val->fv > 0.5f ){
			HSD_JObjClearFlagsAll(pJObj,JOBJ_HIDDEN);
		}else{
			HSD_JObjSetFlagsAll  (pJObj,JOBJ_HIDDEN);
		}
		break;
	case HSD_A_J_NODE:
		if( val->fv > 0.5f ){
			HSD_JObjClearFlags(pJObj,JOBJ_HIDDEN);
		}else{
			HSD_JObjSetFlags  (pJObj,JOBJ_HIDDEN);
		}
		break;
	}
}

HSD_ShapeSetDesc* WK7JObj::GetShapeSetDesc(void)
{
	if( mpJoint==NULL || union_type_dobj(mpJoint)==FALSE ){
		return NULL;
	}
	HSD_DObjDesc	*dobjdesc;
	HSD_PObjDesc	*pobjdesc;
	if( (dobjdesc = mpJoint->u.dobjdesc)==NULL || (pobjdesc = dobjdesc->pobjdesc)==NULL ){
		return NULL;
	}
	if (pobj_type(pobjdesc)!=POBJ_SHAPEANIM ){
		return NULL;
	}
	return pobjdesc->u.shape_set;
}

const char** WK7JObj::GetShapeNameList(HSD_ShapeSetDesc *pShapeSet)
{
	if( pShapeSet==NULL ){
		return NULL;
	}
	const char	***pppTable =(const char ***)pShapeSet;
	pppTable--;
	return *pppTable;
}

const char* WK7JObj::GetDescName(const HSD_FObjDesc* desc)
{
	return _GetDescName( (char**)desc );
}

int WK7JObj::SearchName(const char *pShapeAnimName,int shapenum,const char **ppNameList)
{
	if( pShapeAnimName==NULL ){
		return -1;
	}
	int		i;
	for(i=0 ; i<shapenum ; i++){
		if( strcmp(ppNameList[i],pShapeAnimName)==0 ){
			return i;
		}
	}
	return -1;
}

HSD_FObj* WK7JObj::FObjLoadDesc_ShapeAnim(HSD_FObjDesc *fd,HSD_FObj *fobj,int shapenum,const char **ppNameList)
{
	if( fd==NULL ){
		return fobj;
	}
	
	const char	*pShapeAnimName = WK7JObj::GetDescName(fd);
	HSD_FObj	*pnext;
	HSD_FObj	*pnew;
	int			id;
	u8			type;
	
	pnext = FObjLoadDesc_ShapeAnim(fd->next,fobj,shapenum,ppNameList);
	
	if( (id=SearchName(pShapeAnimName,shapenum,ppNameList)) < 0 ){
		return pnext;
	}
	type = (u8)HSD_A_S_WEIGHT(id);
	
	for( ; fobj!=NULL ; fobj=fobj->next){
		if( fobj->type==type ){
			break;
		}
	}
	
	if( fobj!=NULL ){
		fobj->startframe = (s16)fd->startframe;
		fobj->frac_value = fd->frac_value;
		fobj->frac_slope = fd->frac_slope;
		fobj->adhead     = fd->ad;
		fobj->length     = fd->length;
		fobj->flags      = FOBJ_NO_ANIM;
		return pnext;
	}
	
	pnew = HSD_FObjAlloc();
	pnew->next       = pnext;
	pnew->startframe = (short)fd->startframe;
	pnew->type       = type;
	pnew->frac_value = fd->frac_value;
	pnew->frac_slope = fd->frac_slope;
	pnew->adhead     = fd->ad;
	pnew->length     = fd->length;
	pnew->flags      = FOBJ_NO_ANIM;
	return pnew;
}

HSD_AObj* WK7JObj::AObjLoadDesc_ShapeAnim(HSD_AObjDesc *aobjdesc,HSD_AObj *aobj,int shapenum,const char **ppNameList)
{
	if( aobjdesc != NULL ){
		HSD_FObj		*fobj;
		if( aobj==NULL ){
			aobj = HSD_AObjAlloc();
			fobj = NULL;
		}else{
			fobj = HSD_AObjGetFObj(aobj);
		}
		HSD_JObj	*path;
		HSD_AObjSetFlags      (aobj,aobjdesc->flags);
		HSD_AObjSetRewindFrame(aobj,0.0F);
		HSD_AObjSetEndFrame   (aobj,aobjdesc->endframe);
//		HSD_AObjSetFObj(aobj,FObjLoadDesc_ShapeAnim(aobjdesc->fobjdesc,shapenum,ppNameList));
		if( aobj!=NULL ){
			aobj->flags |= AOBJ_NO_ANIM;
			aobj->fobj = FObjLoadDesc_ShapeAnim(aobjdesc->fobjdesc,fobj,shapenum,ppNameList);
		}
		if( aobjdesc->joint != NULL ){
			path = (HSD_JObj *)HSD_IDGetData((HSD_ID)aobjdesc->joint, NULL);
			if (path) {
				HSD_JObjRef(path);
			} else {
				path = HSD_JObjLoadJoint(aobjdesc->joint);
			}
			HSD_AObjSetPath(aobj, path);
		}
		return aobj;
	}
	return NULL;
}

void WK7JObj::PObjAddAnim(HSD_PObj *pobj,HSD_ShapeAnim *shapeanim,bool overwrite)
{
	if( pobj==NULL ){
		return;
	}
	if( overwrite==true && shapeanim->aobjdesc==NULL ){
		return;
	}
	if( pobj->aobj!=NULL && overwrite==false ){
		HSD_AObjRemove(pobj->aobj);
		pobj->aobj = NULL;
	}
	HSD_ShapeSetDesc	*pShapeSetDesc = GetShapeSetDesc();
	if( pShapeSetDesc==NULL ){
		return;
	}
	const char	**ppNameList = WK7JObj::GetShapeNameList(pShapeSetDesc);
	if( ppNameList==NULL ){
		return;
	}
	pobj->aobj = AObjLoadDesc_ShapeAnim(shapeanim->aobjdesc,pobj->aobj,(int)pShapeSetDesc->nb_shape,ppNameList);
}

void WK7JObj::PObjAddAnimAll(HSD_PObj *pobj, HSD_ShapeAnim *shapeanim,bool overwrite)
{
	HSD_PObj		*pp;
//	HSD_ShapeAnim	*sa;
	
	if( pobj==NULL || shapeanim==NULL ){
		return;
	}
	
	for(pp = pobj; pp; pp = pp->next){
		PObjAddAnim(pp,shapeanim,overwrite);
//	for(pp = pobj, sa = shapeanim ; pp; pp = pp->next, sa = next_p(sa) ){
//		PObjAddAnim(pp,sa,overwrite);
	}
}

//void WK7JObj::DObjAddAnim(HSD_DObj *dobj,HSD_ShapeAnim *shapeanim,HSD_ShapeAnimDObj *sd)
//{
//	if( dobj==NULL ){
//		return;
//	}
//	    PObjAddAnimAll(dobj->pobj, sd?sd->shapeanim:NULL);
//	HSD_MObjAddAnim   (dobj->mobj, shapeanim);
//}

void WK7JObj::MObjAddAnim(HSD_DObj *dobj,HSD_MatAnim *matanim)
{
	if( dobj==NULL ){
		return;
	}
	HSD_MObjAddAnim(dobj->mobj,matanim);
}

void WK7JObj::PObjAddAnim(HSD_DObj *dobj,HSD_ShapeAnimDObj *sd,bool overwrite)
{
	if( dobj==NULL ){
		return;
	}
	PObjAddAnimAll(dobj->pobj, sd?sd->shapeanim:NULL,overwrite);
}

//void WK7JObj::DObjAddAnimAll(HSD_DObj *dobj,HSD_MatAnim *matanim,HSD_ShapeAnimDObj *shapeanimdobj)
//{
//	HSD_DObj			*dp;
//	HSD_MatAnim			*ma;
//	HSD_ShapeAnimDObj	*sd;
//	if( dobj==NULL ){
//		return;
//	}
//	for(dp = dobj,ma = matanim,sd = shapeanimdobj ; dp ; dp = dp->next, ma = next_p(ma), sd = next_p(sd) ){
//		DObjAddAnim(dp,ma,sd);
//	}
//}

void WK7JObj::MObjAddAnimAll(HSD_DObj *dobj,HSD_DObjDesc *dobjdesc,HSD_MatAnim *matanim)
{
	HSD_DObj		*dp;
	HSD_DObjDesc	*dd;
	HSD_MatAnim		*ma;
	if( dobj==NULL ){
		return;
	}
	if( mpCurrentAttachMotionSdExpInfo!=NULL && mpCurrentAttachMotionSdExpInfo->mVersion >= 1790 ){
		const char	*pMatAnimName;
		for(ma=matanim ; ma!=NULL ; ma=next_p(ma)){
			if( (pMatAnimName = ::GetDescName(ma))==NULL ){
				continue;
			}
			for(dp=dobj,dd=dobjdesc ; dp!=NULL && dd!=NULL ; dp=dp->next,dd=dd->next){
				if( strcmp(::GetDescName(dd),pMatAnimName)!=0 ){
					continue;
				}
				MObjAddAnim(dp,ma);
			}
		}
	}else{
		for(dp = dobj,ma = matanim ; dp ; dp = dp->next, ma = next_p(ma) ){
			MObjAddAnim(dp,ma);
		}
	}
}

void WK7JObj::PObjAddAnimAll(HSD_DObj *dobj,HSD_ShapeAnimDObj *shapeanimdobj,bool overwrite)
{
	HSD_DObj			*dp;
//	HSD_ShapeAnimDObj	*sd;
	if( dobj==NULL ){
		return;
	}
//	for(dp = dobj,sd = shapeanimdobj ; dp ; dp = dp->next, sd = next_p(sd) ){
//		PObjAddAnim(dp,sd,overwrite);
	for(dp = dobj ; dp ; dp = dp->next){
		PObjAddAnim(dp,shapeanimdobj,overwrite);
	}
}

void WK7JObj::JObjSortAnim(HSD_AObj *aobj)
{
	HSD_FObj **fpp;
	if( aobj==NULL || aobj->fobj==NULL ){
		return;
	}
	for(fpp = &aobj->fobj ; *fpp ; fpp = &(*fpp)->next ){
		if( (*fpp)->type == HSD_A_J_BRANCH ){
			HSD_FObj *t = (*fpp);
			(*fpp) = (*fpp)->next;
			t->next = aobj->fobj;
			aobj->fobj = t;
			return;
		} 
	}
}

HSD_FObj* WK7JObj::FObjLoadDesc(HSD_FObjDesc *fd,HSD_FObj *fobj)
{
	if( fd !=NULL ){
		
		HSD_FObj	*nextfobj;
		
		nextfobj = FObjLoadDesc(fd->next,fobj);
		for( ; fobj!=NULL ; fobj=fobj->next){
			if( fobj->type==fd->type ){
				break;
			}
		}
		
		if( fobj!=NULL ){
			fobj->startframe = (s16)fd->startframe;
			fobj->type       = fd->type;
			fobj->frac_value = fd->frac_value;
			fobj->frac_slope = fd->frac_slope;
			fobj->adhead     = fd->ad;
			fobj->length     = fd->length;
			fobj->flags      = FOBJ_NO_ANIM;
			return nextfobj;
		}
		
		HSD_FObj	*newfobj = HSD_FObjAlloc();
		newfobj->next       = nextfobj;
		newfobj->startframe = (s16)fd->startframe;
		newfobj->type       = fd->type;
		newfobj->frac_value = fd->frac_value;
		newfobj->frac_slope = fd->frac_slope;
		newfobj->adhead     = fd->ad;
		newfobj->length     = fd->length;
		newfobj->flags      = FOBJ_NO_ANIM;
		return newfobj;
		
	}
	
	return fobj;
}

HSD_AObj* WK7JObj::AObjLoadDesc(HSD_AObjDesc *aobjdesc,HSD_AObj *aobj)
{
	if( aobjdesc ){
		HSD_FObj		*fobj;
		if( aobj==NULL ){
			aobj = HSD_AObjAlloc();
			fobj = NULL;
		}else{
			fobj = HSD_AObjGetFObj(aobj);
		}
		HSD_JObj		*path;
		HSD_AObjSetFlags      (aobj,aobjdesc->flags|AOBJ_NO_ANIM);
		HSD_AObjSetRewindFrame(aobj,0.0f);
		HSD_AObjSetEndFrame   (aobj,aobjdesc->endframe);
//		HSD_AObjSetFObj       (aobj,FObjLoadDesc(aobjdesc->fobjdesc,fobj));
		if( aobj!=NULL ){
			aobj->flags |= AOBJ_NO_ANIM;
			aobj->fobj = FObjLoadDesc(aobjdesc->fobjdesc,fobj);
		}
		if( aobjdesc->joint!=NULL ){
			path = (HSD_JObj *)HSD_IDGetData((HSD_ID)aobjdesc->joint,NULL);
			if( path!=NULL ){
				HSD_JObjRef(path);
			}else{
				path = HSD_JObjLoadJoint(aobjdesc->joint);
			}
			HSD_AObjSetPath(aobj,path);
		}
		return aobj;
	}
	return NULL;
}

void WK7JObj::JObjAddAnim(HSD_JObj *jobj,HSD_AnimJoint *ajoint,bool overwrite)
{
	if( jobj==NULL || ajoint==NULL ){
		return;
	}
	if( overwrite==true && ajoint->aobjdesc==NULL ){
		return;
	}
	if( jobj->aobj!=NULL && overwrite==false ){
		HSD_AObjRemove(jobj->aobj);
		jobj->aobj = NULL;
	}
	jobj->aobj = AObjLoadDesc(ajoint->aobjdesc,jobj->aobj);
	JObjSortAnim(jobj->aobj);
	HSD_RObjAddAnimAll(jobj->robj, ajoint->robjanim);
	if( ajoint->flags & JOBJ_ANIM_CLASSICAL_SCALING ){
		HSD_JObjSetFlags  (jobj,JOBJ_CLASSICAL_SCALING);
	}else{
		HSD_JObjClearFlags(jobj,JOBJ_CLASSICAL_SCALING);
	}
}

void WK7JObj::JObjAddAnim(HSD_JObj *jobj,HSD_Joint *joint,HSD_MatAnimJoint *mjoint)
{
	if( jobj==NULL || joint==NULL || union_type_dobj(jobj)==FALSE ){
		return;
	}
	MObjAddAnimAll(jobj->u.dobj,joint->u.dobjdesc,mjoint?mjoint->matanim:NULL);
}

void WK7JObj::JObjAddAnim(HSD_JObj *jobj,HSD_ShapeAnimJoint *sjoint,bool overwrite)
{
	if( jobj==NULL || union_type_dobj(jobj)==FALSE ){
		return;
	}
	PObjAddAnimAll(jobj->u.dobj,sjoint?sjoint->shapeanimdobj:NULL,overwrite);
}

void WK7JObj::AObjInterpretAnimCustom(void)
{
	float		rate;
	HSD_AObj	*aobj = mpJObj->aobj;
	
	if( aobj==NULL || aobj->flags & AOBJ_NO_ANIM ){
		return;
	}
	
	if( aobj->flags & AOBJ_FIRST_PLAY ){
		aobj->flags &= ~AOBJ_FIRST_PLAY;
		rate = 0.0F;
	}else{
		rate = aobj->rate;
		aobj->frame += rate;
	}
	
	if( (aobj->flags & AOBJ_ANIM_LOOP) && aobj->endframe <= aobj->frame ){
		/* [vꍇ */
		if( aobj->startframe < aobj->endframe ){
			HSD_FObjStopAnimAll(aobj->fobj,(void *)mpJObj,JObjUpdateFunc,rate);
			aobj->frame = fmod(aobj->frame - aobj->startframe,aobj->endframe - aobj->startframe) + aobj->startframe;
			HSD_FObjReqAnimAll(aobj->fobj,aobj->frame);
		}else{
			/*
			 * startframe >= endframe ̏ꍇ͍ŏIԂłÂ
			 * Aj[VƉ߂B
			 */
			aobj->frame = aobj->endframe;
		}
		rate = 0.0F;
		aobj->flags |=  AOBJ_ANIM_REWINDED;
	}else{
		aobj->flags &= ~AOBJ_ANIM_REWINDED;
	}
	
	if( aobj->flags & AOBJ_NO_UPDATE ){
		HSD_FObjInterpretAnimAll(aobj->fobj,(void *)mpJObj,NULL,rate);
	}else{
		HSD_FObjInterpretAnimAll(aobj->fobj,(void *)mpJObj,JObjUpdateFunc,rate);
	}
	
	if( !(aobj->flags & AOBJ_ANIM_LOOP) ){
		if( aobj->endframe <= aobj->frame ){
			/* Aj[VI */
			HSD_AObjStopAnim(aobj,(void *)mpJObj,JObjUpdateFunc);
		}
	}
	
//	if( aobj->flags & AOBJ_NO_ANIM ){
//		nb_deadanim++;
//	}else{
//		nb_activeanim++;
//	}
}


bool WK7JObj::SkeltonAnim(void)
{
	if( mpJObj            ==NULL         ||
	    mpJObj->aobj      ==NULL         ||
	    mpJObj->aobj->fobj==NULL         ||
	    !(mpJObj->flags & (JOBJ_SKELETON_ROOT|JOBJ_SKELETON)) ){
		return false;
	}
	
	HSD_FObj		*pFObj = mpJObj->aobj->fobj;
	HSD_FObj		*pFObjRot[3];
	
	pFObjRot[0] = 
	pFObjRot[1] = 
	pFObjRot[2] = NULL;
	
	while( pFObj!=NULL ){
		
		switch( pFObj->type ){
		case	HSD_A_J_ROTX:
			pFObjRot[0] = pFObj;
			break;
		case	HSD_A_J_ROTY:
			pFObjRot[1] = pFObj;
			break;
		case	HSD_A_J_ROTZ:
			pFObjRot[2] = pFObj;
			break;
		default:
			break;
		}
		
		pFObj = pFObj->next;
	}
	if( pFObjRot[0]==NULL && pFObjRot[1]==NULL && pFObjRot[2]==NULL ){
		return false;
	}
	
	HSD_JObjCheckDepend  (mpJObj);
	AObjInterpretAnimCustom();
//	HSD_AObjInterpretAnim(mpJObj->aobj,mpJObj,NULL);
//	HSD_RObjAnimAll      (mpJObj->robj);
	if( union_type_dobj(mpJObj) ){
		HSD_DObjAnimAll(mpJObj->u.dobj);
	}
	
	Vec				crot,nrot;
	Mtx				drm,srm,um,rm,tm;
	ghmQuaternion	srq,drq,crq;
	f32				rate = 0.0f;
	
	if( pFObjRot[0]!=NULL ){
		crot.x = pFObjRot[0]->p0;
		nrot.x = pFObjRot[0]->p1;
		if( pFObjRot[0]->fterm > 0.0f ){
			rate = pFObjRot[0]->time / pFObjRot[0]->fterm;
		}
	}else{
		crot.x = 
		nrot.x = mpJObj->rotate.x;
	}
	if( pFObjRot[1]!=NULL ){
		crot.y = pFObjRot[1]->p0;
		nrot.y = pFObjRot[1]->p1;
		if( pFObjRot[1]->fterm > 0.0f ){
			rate = pFObjRot[1]->time / pFObjRot[1]->fterm;
		}
	}else{
		crot.y = 
		nrot.y = mpJObj->rotate.y;
	}
	if( pFObjRot[2]!=NULL ){
		crot.z = pFObjRot[2]->p0;
		nrot.z = pFObjRot[2]->p1;
		if( pFObjRot[2]->fterm > 0.0f ){
			rate = pFObjRot[2]->time / pFObjRot[2]->fterm;
		}
	}else{
		crot.z = 
		nrot.z = mpJObj->rotate.z;
	}
	
	// XP[
	MTXScale (um,mpJObj->scale.x,mpJObj->scale.y,mpJObj->scale.z);
	
	MTXRotRad(srm,'x',crot.x);
	MTXRotRad( rm,'y',crot.y);
	MTXConcat( rm,srm,srm);
	MTXRotRad( rm,'z',crot.z);
	MTXConcat( rm,srm,srm);
	MTXRotRad(drm,'x',nrot.x);
	MTXRotRad( rm,'y',nrot.y);
	MTXConcat( rm,drm,drm);
	MTXRotRad( rm,'z',nrot.z);
	MTXConcat( rm,drm,drm);
	MatToQuat(drm,&drq);
	MatToQuat(srm,&srq);
	crq.QuatSlerp(&srq,&drq,rate);
	MTXQuat  ( rm,&crq);
	MTXConcat( rm,um,um);
	
	// sړ
	MTXTrans ( tm,mpJObj->translate.x,mpJObj->translate.y,mpJObj->translate.z);
	MTXConcat( tm,um,um);
	
	if( mpParent!=NULL ){
		MTXConcat(mpParent->GetLocalWorldMatrix(),um,um);
	}
	
	HSD_JObjSetUserDefinedMtx(mpJObj,um);
	if( mpReferencedObj!=NULL ){
		WK7JObj		*pRootObj;
		if( (pRootObj = mpReferencedObj->GetParent())!=NULL ){
			for( ; pRootObj->GetParent()!=NULL ; pRootObj=pRootObj->GetParent()){}
			HSD_JObj		*pRootJObj = pRootObj->GetJObj();
			if( pRootJObj!=NULL && (pRootJObj = pRootJObj->parent)!=NULL ){
				HSD_JObjSetupMatrix(pRootJObj);
				MTXConcat(pRootJObj->mtx,um,um);
			}
		}
//		WK7JObj		*pParentJObj;
		HSD_JObj	*pReferJObj = mpReferencedObj->GetJObj();
//		Mtx			m;
//		if( (pParentJObj=mpReferencedObj->GetParent())!=NULL ){
//			MTXConcat(pParentJObj->GetLocalWorldMatrix(),um,m);
//		}else{
//			MTXScale (m,pReferJObj->scale.x,pReferJObj->scale.y,pReferJObj->scale.z);
//			MTXConcat(rm,m,m);
//			MTXConcat(tm,m,m);
//		}
//		HSD_JObjSetUserDefinedMtx(pReferJObj,m);
		HSD_JObjSetUserDefinedMtx(pReferJObj,um);
	}
	
//	if( mpParent!=NULL ){
//		MTXConcat(mpParent->GetLocalWorldMatrix(),um,um);
//	}
//	
//	HSD_JObjSetUserDefinedMtx(mpJObj,um);
	
	if( mpCurrRotate == NULL ){
		ghmGcMemSetAllocName("ghmQuat");
		mpCurrRotate = new ghmQuaternion;
	}
	if( mpCurrRotate != NULL ){
		mpCurrRotate->x = crq.x;
		mpCurrRotate->y = crq.y;
		mpCurrRotate->z = crq.z;
		mpCurrRotate->w = crq.w;
	}
	
	return true;
}

void WK7JObj::SetReferencedObj(WK7JObj *pJObj)
{
	if( mpReferencedObj != NULL ){
		mpReferencedObj->mReferenceObj = 0;
	}
	if( (mpReferencedObj = pJObj) != NULL ){
		pJObj->mReferenceObj = 1;
	}
}

bool WK7JObj::SetCullMode(u32 mode,bool next,bool child)
{
	SetCullMode(mpJObj,mode);
	if( child==true && mpChild!=NULL ){
		mpChild->SetCullMode(mode,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetCullMode(mode,false,true);
		}
	}
	return true;
}

bool WK7JObj::SetCullMode(HSD_JObj *pJObj,u32 mode)
{
	if( pJObj==NULL ){
		return true;
	}
	
	HSD_DObj	*pDObj;
	HSD_PObj	*pPObj;
	if( mode==WK7JOBJ_CULLMODE_REV ){
		for(pDObj=HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
			for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
				switch( pPObj->flags & (POBJ_CULLBACK|POBJ_CULLFRONT) ){
				case	POBJ_CULLFRONT:
					pPObj->flags &= ~POBJ_CULLFRONT;
					pPObj->flags |=  POBJ_CULLBACK;
					break;
				case	POBJ_CULLBACK:
					pPObj->flags &= ~POBJ_CULLBACK;
					pPObj->flags |=  POBJ_CULLFRONT;
					break;
				}
			}
		}
	}else{
		u32		flag;
		switch( mode ){
		case	WK7JOBJ_CULLMODE_BACK:  flag = POBJ_CULLBACK;  break;
		case	WK7JOBJ_CULLMODE_FRONT: flag = POBJ_CULLFRONT; break;
		default:                        flag = 0;              break;
		}
		for(pDObj=HSD_JObjGetDObj(pJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
			for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
				pPObj->flags &= ~(POBJ_CULLBACK|POBJ_CULLFRONT);
				pPObj->flags |=  flag;
			}
		}
	}
	
	return true;
}

int WK7JObj::SetUserDefinedMatrixFlag(bool flag,bool next,bool child)
{
	int		count = 1;
	mUseUserDefinedMatrix = (flag==true? 1 : 0);
	if( mpJObj!=NULL ){
		mpJObj->flags = (JOBJ_USER_DEFINED_MTX|JOBJ_MTX_INDEPEND_PARENT|JOBJ_MTX_INDEPEND_SRT|mpJObj->flags);
	}
	if( child==true && mpChild!=NULL ){
		count += mpChild->SetUserDefinedMatrixFlag(flag,true,true);
	}
	if( next==true ){
		WK7JObj		*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			count += pJObj->SetUserDefinedMatrixFlag(flag,false,true);
		}
	}
	return count;
}

bool WK7JObj::GetUserDefinedMatrixFlag(bool next,bool child)
{
	if( mUseUserDefinedMatrix ){
		return true;
	}
	if( child==true && mpChild!=NULL ){
		if( mpChild->GetUserDefinedMatrixFlag(true,true)==true ){
			return true;
		}
	}
	if( next==true ){
		WK7JObj		*pJObj;
		for(pJObj=mpNext ; pJObj!=NULL ; pJObj=pJObj->GetNext()){
			if( pJObj->GetUserDefinedMatrixFlag(false,true)==true ){
				return true;
			}
		}
	}
	return false;
}

bool WK7JObj::SetUserDefinedMatrix(Mtx mtx)
{
	mUseUserDefinedMatrix = 1;
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetUserDefinedMtx(mpJObj,mtx);
	return true;
}

bool WK7JObj::ResetUserDefinedMatrix(void)
{
	mUseUserDefinedMatrix = 0;
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjResetUserDefinedMtx(mpJObj);
	return true;
}

bool WK7JObj::ResetClothJointUserDefinedMatrix(bool next,bool child)
{
	if( mClothJoint && mpJObj!=NULL ){
		HSD_JObjResetUserDefinedMtx(mpJObj);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->ResetClothJointUserDefinedMatrix(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ResetClothJointUserDefinedMatrix(false,child);
		}
	}
	return true;
}

bool WK7JObj::ResetClothJointToBind(bool next,bool child)
{
	if( mClothJoint && mpJObj!=NULL ){
		ResetMotion(false,false);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->ResetClothJointToBind(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ResetClothJointToBind(false,child);
		}
	}
	return true;
}

void WK7JObj::FrameProcess(bool next,bool child)
{
	if( mpAlphaControl!=NULL ){
		if( mpAlphaControl->FrameProcess()==false ){
			ReleaseAlphaControl();
		}
	}
	if( mpPreFrameProcessCallBack != NULL ){
		mpPreFrameProcessCallBack(this,mpPreFrameProcessData);
	}
	mFrameProcess = 1;
	if( mpEffectBreakObj!=NULL ){
		if( WK7EffectBreakObj::IsValid(mpEffectBreakObj)==false ){
			mpEffectBreakObj = NULL;
		}
	}
	if( mpJObj!=NULL ){
		if( !mReferenceObj && !mUseUserDefinedMatrix ){
			if( SkeltonAnim()==false ){
				GHM_SAFE_DELETE(mpCurrRotate);
				HSD_JObjResetUserDefinedMtx(mpJObj);
				HSD_JObjAnim(mpJObj);
			}
	//#if 0
	//		if( HSD_PadGetTrigger(1) & HSD_PAD_START ){
	//			
	//			HSD_DObj	*pDObj;
	//			for(pDObj=HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
	//				HSD_PObj	*pPObj;
	//				for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
	//					if( pPObj->flags & POBJ_CULLBACK ){
	//						pPObj->flags &= ~POBJ_CULLBACK;
	//						pPObj->flags |=  POBJ_CULLFRONT;
	//					}
	//				}
	//			}
	//			
	//		}
	//#endif
			if( mpInterpolateInfo!=NULL ){
				if( mpInterpolateInfo->FrameProcess(this)==false ){
					GHM_SAFE_DELETE(mpInterpolateInfo);
				}
			}
			if( mpReferencedObj!=NULL ){
				HSD_JObj		*pReferencedJObj = mpReferencedObj->GetJObj();
				Vec				vec;
				HSD_JObjGetTranslate(mpJObj         ,&vec);
				HSD_JObjSetTranslate(pReferencedJObj,&vec);
				vec.x = HSD_JObjGetRx(mpJObj);
				vec.y = HSD_JObjGetRy(mpJObj);
				vec.z = HSD_JObjGetRz(mpJObj);
				HSD_JObjSetRx(pReferencedJObj,vec.x);
				HSD_JObjSetRy(pReferencedJObj,vec.y);
				HSD_JObjSetRz(pReferencedJObj,vec.z);
//				HSD_JObjGetScale(mpJObj         ,&vec);
//				HSD_JObjSetScale(pReferencedJObj,&vec);
				HSD_JObjSetupMatrix(pReferencedJObj);
			}
		}
		if( mpShapeAnim!=NULL ){
			mpShapeAnim->FrameProcess();
#ifdef	_DEBUG
#if 0
			HSD_DObj	*pDObj;
			for(pDObj=HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=HSD_DObjGetNext(pDObj)){
				HSD_PObj	*pPObj;
				for(pPObj=HSD_DObjGetPObj(pDObj) ; pPObj!=NULL ; pPObj=HSD_PObjGetNext(pPObj)){
					if( pobj_type(pPObj)!=POBJ_SHAPEANIM ){
						continue;
					}
					HSD_ShapeSet		*pShapeSet;
					int					i;
					const char			**ppNameList;
					if( (ppNameList = WK7JObj::GetShapeNameList(GetShapeSetDesc()))==NULL ){
						continue;
					}
					if( (pShapeSet=pPObj->u.shape_set)==NULL ){
						continue;
					}
					if( pShapeSet->blend.bp==NULL ){
						continue;
					}
					for(i=0 ; i<pShapeSet->nb_shape ; i++){
						PRINTF("%d : %16s - %.3f\n",i,ppNameList[i],pShapeSet->blend.bp[i]);
					}
				}
			}
#endif
#endif
		}
	}
	if( mpPostFrameProcessCallBack != NULL ){
		mpPostFrameProcessCallBack(this,mpPostFrameProcessData);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->FrameProcess(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->FrameProcess(false,child);
		}
	}
}

#ifdef FUJITA
#pragma cplusplus			off
#pragma cpp_extensions		off
#pragma warn_implicitconv	off

extern	void HSD_ZListSetPrio(u32 prio);
typedef	void HSD_ZListCallbackFunc(void *pData);
extern	void HSD_ZListSetCallbackFunc(HSD_ZListCallbackFunc *pPreCallbackFunc,HSD_ZListCallbackFunc *pPostCallbackFunc,void *pData);

#pragma cplusplus			on
#pragma cpp_extensions		on
#pragma warn_implicitconv	on
#endif

void WK7JObj::_ZListPreCallbackFunc(void *pData)
{
	if( pData==NULL ){
		return;
	}
	WK7JObj		*pJObj = (WK7JObj *)pData;
	pJObj->ZListPreCallbackFunc();
}

void WK7JObj::ZListPreCallbackFunc(void)
{
	if( mpShapeAnim!=NULL ){
		mpShapeAnim->SetShapeBuffer();
	}
	if( mpCurrentFog!=NULL ){
		mpCurrentFog->SetCurrent();
	}
}

void WK7JObj::_ZListPostCallbackFunc(void *pData)
{
	if( pData==NULL ){
		return;
	}
	WK7JObj		*pJObj = (WK7JObj *)pData;
	pJObj->ZListPostCallbackFunc();
}

void WK7JObj::ZListPostCallbackFunc(void)
{
	if( mpCurrentFog!=NULL ){
		mpCurrentFog->EndCurrent();
	}
	if( mpShapeAnim!=NULL ){
		mpShapeAnim->ResetShapeBuffer();
	}
}

int WK7JObj::GetBoundState(void)
{
	if( mpViewVolume!=NULL && mpBound!=NULL && mpBound->GetEnable()==true ){
		if( !mUseUserDefinedMatrix ){
			HSD_JObjSetupMatrix(mpJObj);
		}
		return mpBound->GetCrossState(HSD_JObjGetMtxPtr(mpJObj),mpViewVolume);
	}
	return -1;
}

#include ".\WK7Cam.h"
#include ".\WK7Sight.h"
#define	_SETDS()				//SETDS(JOBJ_TH|(GETDSCNT() & 0xff),this)
void WK7JObj::RenderProcess(bool next,bool child,HSD_TrspMask trsp, u32 rendermode)
{
//	int		a = 0;
//	if( strcmp(mpName,"t10")==0 ){
//		a++;
//	}
#ifdef	_DEBUG
	u16		cnt = GETDSCNT();
#endif
	_SETDS();
	WK7JObj::mRenderProcessTrspMask   = trsp;
	WK7JObj::mRenderProcessRenderMode = rendermode;
	if( !mFrameProcess ){
		FrameProcess(false,false);
	}
	if( mpPreRenderProcessCallBack != NULL ){
		mpPreRenderProcessCallBack(this,mpPreRenderProcessData);
	}
	if( mpFog!=NULL ){
		mpFog->SetCurrent();
	}
	mpCurrentFog = WK7Fog::GetTop();
	if( mpJObj!=NULL && (mRenderMode & mCurrentRenderMode) ){
		bool	visible;
		if( mHide ){
			visible = false;
		}else if( mpViewVolume!=NULL && mpBound!=NULL && mpBound->GetEnable()==true ){
			MtxPtr		mtxptr;
			Mtx			mtx;
			HSD_JObjSetupMatrix(mpJObj);
			HSD_CObj	*pCObj = HSD_CObjGetCurrent();
			if( (mpJObj->flags & JOBJ_BILLBOARD_FIELD) && pCObj!=NULL ){
				Mtx		vmtx,invvmtx;
				HSD_CObjGetViewingMtx (pCObj,vmtx);
				HSD_JObjMakePositionMtx(mpJObj,vmtx,mtx);
				MTXInverse(vmtx,invvmtx);
				MTXConcat (invvmtx,mtx,mtx);
				mtxptr = mtx;
			}else{
				mtxptr = HSD_JObjGetMtxPtr(mpJObj);
			}
			switch( mpBound->GetCrossState(mtxptr,mpViewVolume) ){
			case	ghmGcCollObj::OUTSIDE:
				visible = false;
				break;
			default:
				visible = true;
				break;
			}
		}else{
			visible = true;
		}
		if( visible==true && mpViewVolume!=NULL ){
			WK7Fog		*pFog;
			if( WK7Fog::IsFogging()==true && mpBound!=NULL && (pFog=WK7Fog::GetTop())!=NULL ){
				switch( mpBound->GetFarClip(pFog->GetEndZ(),HSD_JObjGetMtxPtr(mpJObj),mpViewVolume) ){
				case	ghmGcCollObj::OUTSIDE:
					visible = false;
					break;
				default:
					break;
				}
//				WK7JObjBoundSphere	*pBoundSphere = (WK7JObjBoundSphere *)mpBound;
//				ghmGcCollSphere		*pCollShpere  = pBoundSphere->GetBound();
//				MtxPtr				mtxptr = mpViewVolume->GetCoordinate();
//				Vec					pos;
//				MTXMultVec(mtxptr,&pCollShpere->mPosition,&pos);
//				if( pos.z < -(pFog->GetEndZ() + pCollShpere->mRadius) ){
//					visible = false;
//				}
			}
		}
		if( visible==true ){
			HSD_ZListSetCallbackFunc(_ZListPreCallbackFunc,_ZListPostCallbackFunc,this);
			if( mpShapeAnim!=NULL ){
				mpShapeAnim->SetShapeBuffer();
			}
			HSD_ZListSetPrio(mZSortPrio);
			HSD_JObjDisp(mpJObj,0,trsp,rendermode);
			HSD_ZListSetPrio(WK7JOBJ_ZSORTPRIO_DEFAULT);
			if( mpShapeAnim!=NULL ){
				mpShapeAnim->ResetShapeBuffer();
			}
			HSD_ZListSetCallbackFunc(NULL,NULL,NULL);
		}
	}
	SETDSCNT(cnt+1);
	if( mpPostRenderProcessCallBack != NULL ){
		mpPostRenderProcessCallBack(this,mpPostRenderProcessData);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->RenderProcess(true,true,trsp,rendermode);
	}
	if( mpFog!=NULL ){
		mpFog->EndCurrent();
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->RenderProcess(false,child,trsp,rendermode);
		}
	}
}

bool WK7JObj::SetMotionRate(HSD_PObj *pPObj,f32 rate)
{
	if( pPObj==NULL ){
		return true;
	}
	WK7Motion::SetMotionRate(pPObj->aobj,rate);
	return true;
}

bool WK7JObj::SetMotionRate(HSD_TObj *pTObj,f32 rate)
{
	if( pTObj==NULL ){
		return true;
	}
	WK7Motion::SetMotionRate(pTObj->aobj,rate);
	return true;
}

bool WK7JObj::SetMotionRate(HSD_MObj *pMObj,f32 rate)
{
	if( pMObj==NULL ){
		return true;
	}
	f32		crate;
	if( mMotionState & WK7JOBJ_MOTION_MATPLAY ){
		crate = rate;
	}else{
		crate = 0.0f;
	}
	WK7Motion::SetMotionRate(pMObj->aobj,crate);
	SetMotionRate(pMObj->tobj,crate);
	return true;
}

bool WK7JObj::SetMotionRate(HSD_DObj *pDObj,f32 rate)
{
	if( pDObj==NULL ){
		return true;
	}
	f32		crate;
	if( mMotionState & WK7JOBJ_MOTION_SHAPEPLAY ){
		crate = rate;
	}else{
		crate = 0.0f;
	}
	HSD_PObj	*pPObj = pDObj->pobj;
	while( pPObj!=NULL ){
		SetMotionRate(pPObj,crate);
		pPObj = pPObj->next;
	}
	SetMotionRate(pDObj->mobj,rate);
	return true;
}

bool WK7JObj::SetMotionRate(HSD_JObj *pJObj,f32 rate)
{
	if( pJObj==NULL ){
		return true;
	}
	f32		crate;
	if( mMotionState & WK7JOBJ_MOTION_PLAY ){
		crate = rate;
	}else{
		crate = 0.0f;
	}
	WK7Motion::SetMotionRate(pJObj->aobj,crate);
	HSD_DObj	*pDObj;
	pDObj = HSD_JObjGetDObj(pJObj);
	while( pDObj!=NULL ){
		SetMotionRate(pDObj,rate);
		pDObj = pDObj->next;
	}
	return true;
}

bool WK7JObj::SetMotionRate(f32 rate,bool next,bool child)
{
	if( !mMotionOpDisable ){
		mMotionRate = rate;
		RenewMotionRate(false,false);
	}
	if( child==true ){
		if( mpChild!=NULL ){
			mpChild->SetMotionRate(rate,true,true);
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->SetMotionRate(rate,true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMotionRate(rate,false,child);
		}
	}
	return true;
}

bool WK7JObj::RenewMotionRate(bool next,bool child)
{
	if( !mMotionOpDisable ){
		SetMotionRate(mpJObj,mMotionRate);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->RenewMotionRate(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->RenewMotionRate(false,child);
		}
	}
	return true;
}

bool WK7JObj::SetMotionOperation(bool flag,bool next,bool child)
{
	mMotionOpDisable = (flag==true? 0:1);
	if( child==true && mpChild!=NULL ){
		mpChild->SetMotionOperation(flag,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMotionOperation(flag,false,child);
		}
	}
	return true;
}

bool WK7JObj::AttachMotion(bool overwrite)
{
	if( mpJObj==NULL ){
		return false;
	}
	JObjAddAnim(mpJObj,mpAnimJoint,overwrite);
	// ͂
	if( !(GetFlag() & JOBJ_CLASSICAL_SCALING) ){
#ifdef	CLASSICAL_SCALING_CHECK
		OSDebugReport("Warning : Non Classical Scaling Anim - AnimJoint(%s)\n",mpName);
#endif
		SetFlag(JOBJ_CLASSICAL_SCALING,false,false);
	}
	return true;
}

bool WK7JObj::AttachMatMotion(void)
{
	if( mpJObj==NULL || mpJoint==NULL ){
		return false;
	}
	JObjAddAnim(mpJObj,mpJoint,mpMatAnimJoint);
	return true;
}

bool WK7JObj::AttachShapeMotion(bool overwrite)
{
	if( mpJObj==NULL ){
		return false;
	}
	JObjAddAnim(mpJObj,mpShapeAnimJoint,overwrite);
	return true;
}

void WK7JObj::SetCurrentAttachMotionSdExpInfo(WK7Motion *pMotion)
{
	if( pMotion==NULL ){
		mpCurrentAttachMotionSdExpInfo = NULL;
		return;
	}
	mpCurrentAttachMotionSdExpInfo = pMotion->GetSdExpInfo();
}

int WK7JObj::AttachMotion(WK7ModelMotion *pModelMotion,bool next,bool child,bool overwrite)
{
	HSD_AnimJoint	*pAnimJoint;
	int				i,num = pModelMotion->GetModelSetNum();
	for(i=0 ; i<num ; i++){
		if( pModelMotion->GetAnimJoint(i,&pAnimJoint,NULL,NULL)==false || pAnimJoint==NULL ){
			continue;
		}
		SetCurrentAttachMotionSdExpInfo(pModelMotion);
		int		result = AttachMotion(pAnimJoint,next,child,overwrite);
		SetCurrentAttachMotionSdExpInfo();
		return result;
	}
	return 0;
}

int WK7JObj::AttachFindMotion(WK7ModelMotion *pModelMotion,bool next,bool child,bool overwrite)
{
	HSD_AnimJoint	*pAnimJoint;
	if( mpName==NULL || pModelMotion->GetAnimJoint(mpName,&pAnimJoint)==false || pAnimJoint==NULL ){
		return 0;
	}
	SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int	result = AttachMotion(pAnimJoint,next,child,overwrite);
	SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7JObj::AttachMotionAll(WK7ModelMotion *pModelMotion,bool next,bool child,bool overwrite)
{
	int		result;
	result  = AttachMotion     (pModelMotion,next,child,overwrite);
	result += AttachMatMotion  (pModelMotion,next,child);
	result += AttachShapeMotion(pModelMotion,next,child,overwrite);
	return result;
}

int WK7JObj::AttachMotion(HSD_AnimJoint *pAnimJoint,bool next,bool child,bool overwrite)
{
	if( pAnimJoint==NULL ){
		return 0;
	}
	const char		*pAnimJointName;
	HSD_AnimJoint	*pCurrentAnimJoint;
	int				count = 0;
	if( mpJObj!=NULL ){
		for(pCurrentAnimJoint=pAnimJoint ; pCurrentAnimJoint!=NULL ; pCurrentAnimJoint=pCurrentAnimJoint->next){
			if( (pAnimJointName = WK7ModelMotion::GetAnimJointName(pCurrentAnimJoint))==NULL ){
				continue;
			}
			if( mClothJoint ){
				int				i,num = (int)strlen(kClothJointPrefix);
				for(i=0 ; i<num ; i++){
					if( pAnimJointName[i]=='\0' ){
						break;
					}
					if( pAnimJointName[i]!=mpName[i] ){
						break;
					}
				}
				if( i!=num ){
					continue;
				}
				for( ; pAnimJointName[i]==mpName[i] ; i++){
					switch( pAnimJointName[i] ){
					case	'_':
					case	'\0':
						i = -1;
						break;
					default:
						switch( mpName[i] ){
						case	'_':
						case	'\0':
							i = -1;
							break;
						default:
							continue;
						}
						break;
					}
					break;
				}
				if( i >= 0 ){
					continue;
				}
			}else{
				if( strcmp(pAnimJointName,mpName)!=0 ){
					continue;
				}
			}
			if( !mMotionOpDisable ){
				mpAnimJoint = pCurrentAnimJoint;
				mMotionState |=  WK7JOBJ_MOTION_PLAY;
				AttachMotion(overwrite);
				SetMotionRate(1.0f,false,false);
				count = 1;
			}
			break;
		}
	}
	
	if( child==true ){
		if( mpChild!=NULL ){
			if( pCurrentAnimJoint!=NULL ){
				count += mpChild->AttachMotion(pCurrentAnimJoint->child,true,true,overwrite);
			}else{
				count += mpChild->AttachMotion(pAnimJoint,true,true,overwrite);
			}
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			if( pCurrentAnimJoint!=NULL ){
				pJObj->AttachMotion(pCurrentAnimJoint->child,true,true,overwrite);
			}
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			count += pNext->AttachMotion(pAnimJoint,false,child,overwrite);
		}
	}
	return count;
}

int WK7JObj::AttachMotionClassic(WK7ModelMotion *pModelMotion,bool next,bool child)
{
	HSD_AnimJoint	*pAnimJoint;
	if( pModelMotion->GetAnimJoint(0,&pAnimJoint,NULL,NULL)==false || pAnimJoint==NULL ){
		return 0;
	}
	SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int		result = AttachMotionClassic(pAnimJoint,next,child);
	SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7JObj::AttachMotionClassic(HSD_AnimJoint *pAnimJoint,bool next,bool child)
{
	if( pAnimJoint==NULL ){
		return 0;
	}
	int		count = 0;
	if( !mMotionOpDisable ){
		mpAnimJoint = pAnimJoint;
		mMotionState |=  WK7JOBJ_MOTION_PLAY;
		AttachMotion(false);
		SetMotionRate(1.0f,false,false);
		count++;
	}
	if( child==true && mpChild!=NULL ){
		count += mpChild->AttachMotionClassic(pAnimJoint->child,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext,pAnimJoint=pAnimJoint->next ; pNext!=NULL ; pNext=pNext->GetNext(),pAnimJoint=pAnimJoint->next){
			count += pNext->AttachMotionClassic(pAnimJoint,false,child);
		}
	}
	return count;
}

bool WK7JObj::DetachMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState &= ~WK7JOBJ_MOTION_PLAY;
		HSD_JObjRemoveAnimByFlags(mpJObj,JOBJ_ANIM);
		mpAnimJoint = NULL;
	}
	if( child==true ){
		if( mpChild!=NULL ){
			mpChild->DetachMotion(true,true);
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->DetachMotion(true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->DetachMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::ResetMotion(bool next,bool child)
{
	if( mpJObj!=NULL && mpJoint!=NULL && !mMotionOpDisable ){
		HSD_JObjSetTranslate(mpJObj,&mpJoint->translate);
		HSD_JObjSetRx       (mpJObj, mpJoint->rotate.x );
		HSD_JObjSetRy       (mpJObj, mpJoint->rotate.y );
		HSD_JObjSetRz       (mpJObj, mpJoint->rotate.z );
		HSD_JObjSetScale    (mpJObj,&mpJoint->scale    );
		if( mpJoint->flags & JOBJ_HIDDEN ){
			HSD_JObjSetFlags  (mpJObj,JOBJ_HIDDEN);
		}else{
			HSD_JObjClearFlags(mpJObj,JOBJ_HIDDEN);
		}
	}
	if( child==true && mpChild!=NULL ){
		mpChild->ResetMotion(true,true);
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->ResetMotion(true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ResetMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::DetachMotionAll(bool next,bool child)
{
	DetachMotion     (false,false);
	DetachMatMotion  (false,false);
	DetachShapeMotion(false,false);
	if( child==true && mpChild!=NULL ){
		mpChild->DetachMotionAll(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->DetachMotionAll(false,child);
		}
	}
	return true;
}

bool WK7JObj::ResetMotionAll(bool next,bool child)
{
	ResetMotion     (false,false);
	ResetMatMotion  (false,false);
	ResetShapeMotion(false,false);
	if( child==true && mpChild!=NULL ){
		mpChild->ResetMotionAll(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ResetMotionAll(false,child);
		}
	}
	return true;
}

bool WK7JObj::PlayMotionAll(f32 startframe,bool interpolate,f32 interpolateframe,bool next,bool child)
{
	PlayMotion     (startframe,interpolate,interpolateframe,false,false);
	PlayMatMotion  (startframe,false,false);
	PlayShapeMotion(startframe,false,false);
	if( child==true && mpChild!=NULL ){
		mpChild->PlayMotionAll(startframe,interpolate,interpolateframe,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->PlayMotionAll(startframe,interpolate,interpolateframe,false,child);
		}
	}
	return true;
}

bool WK7JObj::ChangeRotInterpolate(const Vec *pRotation)
{
	if( mpJObj==NULL || mpInterpolateInfo==NULL ){
		return false;
	}
	Mtx		rm,tm;
	if( pRotation==NULL ){
		MTXRotRad(tm,'y',PI);
	}else{
		MTXRotRad(tm,'x',pRotation->x);
		MTXRotRad(rm,'y',pRotation->y);
		MTXConcat(rm,tm,tm);
		MTXRotRad(rm,'z',pRotation->z);
		MTXConcat(rm,tm,tm);
	}
	MTXMultVecSR(tm,&mpInterpolateInfo->mPrevTranslate,&mpInterpolateInfo->mPrevTranslate);
	
	MTXQuat  (rm,&mpInterpolateInfo->mPrevRotate);
	MTXConcat(tm,rm,rm);
	MatToQuat(rm,&mpInterpolateInfo->mPrevRotate);
	return true;
}

bool WK7JObj::PlayMotionNA(f32 startframe,bool interpolate,f32 interpolateframe,bool next,bool child)
{
	if( mpJObj!=NULL && (HSD_AObjGetFlags(mpJObj->aobj) & AOBJ_NO_ANIM) ){
		PlayMotion(startframe,interpolate,interpolateframe,false,false);
	}
	if( child==true ){
		if( mpChild!=NULL ){
			mpChild->PlayMotionNA(startframe,interpolate,interpolateframe,true,true);
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->PlayMotionNA(startframe,interpolate,interpolateframe,true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->PlayMotionNA(startframe,interpolate,interpolateframe,false,child);
		}
	}
	return true;
}

bool WK7JObj::PlayMotion(f32 startframe,bool interpolate,f32 interpolateframe,bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		if( interpolate==true && interpolateframe>0.0f ){
			bool	infocreate;
			if( mpInterpolateInfo==NULL ){
				ghmGcMemSetAllocName("WK7JObjMI");
				mpInterpolateInfo = new WK7JObjMotionInterpolateInfo;
				infocreate = true;
			}else{
				infocreate = false;
			}
			if( mpInterpolateInfo!=NULL ){
				if( infocreate==true ){
					if( mpCurrRotate!=NULL ){
						mpInterpolateInfo->mPrevRotate.x = mpCurrRotate->x;
						mpInterpolateInfo->mPrevRotate.y = mpCurrRotate->y;
						mpInterpolateInfo->mPrevRotate.z = mpCurrRotate->z;
						mpInterpolateInfo->mPrevRotate.w = mpCurrRotate->w;
					}else{
						Mtx		m,tm;
						MTXRotRad( m,'x',mpJObj->rotate.x);
						MTXRotRad(tm,'y',mpJObj->rotate.y);
						MTXConcat(tm,m,m);
						MTXRotRad(tm,'z',mpJObj->rotate.z);
						MTXConcat(tm,m,m);
						MatToQuat(m,&mpInterpolateInfo->mPrevRotate);
					}
					mpInterpolateInfo->mPrevScale     = mpJObj->scale;
					mpInterpolateInfo->mPrevTranslate = mpJObj->translate;
					mpInterpolateInfo->mInterpolateScale     = mpInterpolateInfo->mPrevScale;
					mpInterpolateInfo->mInterpolateTranslate = mpInterpolateInfo->mPrevTranslate;
					mpInterpolateInfo->mInterpolateRotate.x  = mpInterpolateInfo->mPrevRotate.x;
					mpInterpolateInfo->mInterpolateRotate.y  = mpInterpolateInfo->mPrevRotate.y;
					mpInterpolateInfo->mInterpolateRotate.z  = mpInterpolateInfo->mPrevRotate.z;
					mpInterpolateInfo->mInterpolateRotate.w  = mpInterpolateInfo->mPrevRotate.w;
				}else{
					mpInterpolateInfo->mPrevScale     = mpInterpolateInfo->mInterpolateScale;
					mpInterpolateInfo->mPrevRotate    = mpInterpolateInfo->mInterpolateRotate;
					mpInterpolateInfo->mPrevTranslate = mpInterpolateInfo->mInterpolateTranslate;
				}
				mpInterpolateInfo->mInterpolateTotalFrame   = interpolateframe;
				mpInterpolateInfo->mInterpolateCurrentFrame = 0.0f;
			}
		}else{
			GHM_SAFE_DELETE(mpInterpolateInfo);
		}
//		HSD_JObjReqAnim(mpJObj,startframe);
		mMotionState |= WK7JOBJ_MOTION_PLAY;
		f32	frame = startframe;
		if( frame < 0.0f ){
			if( mpJObj->aobj!=NULL ){
				frame = mpJObj->aobj->endframe;
			}else{
				frame = 0.0f;
			}
		}
		HSD_JObjReqAnimByFlags(mpJObj,frame,JOBJ_ANIM);
		RenewMotionRate(false,false);
	}
	if( child==true ){
		if( mpChild!=NULL ){
			mpChild->PlayMotion(startframe,interpolate,interpolateframe,true,true);
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->PlayMotion(startframe,interpolate,interpolateframe,true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->PlayMotion(startframe,interpolate,interpolateframe,false,child);
		}
	}
	return true;
}

bool WK7JObj::ReplayMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState |= WK7JOBJ_MOTION_PLAY;
		RenewMotionRate(false,false);
	}
	if( child==true ){
		if( mpChild!=NULL ){
			mpChild->ReplayMotion(true,true);
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->ReplayMotion(true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ReplayMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::ReplayMotionAll(bool next,bool child)
{
	ReplayMotion     (false,false);
	ReplayMatMotion  (false,false);
	ReplayShapeMotion(false,false);
	if( child==true && mpChild!=NULL ){
		mpChild->ReplayMotionAll(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ReplayMotionAll(false,child);
		}
	}
	return true;
}

bool WK7JObj::StopMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState &= ~WK7JOBJ_MOTION_PLAY;
		RenewMotionRate(false,false);
	}
	if( child==true ){
		if( mpChild!=NULL ){
			mpChild->StopMotion(true,true);
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->StopMotion(true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->StopMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::StopMotionAll(bool next,bool child)
{
	StopMotion     (false,false);
	StopMatMotion  (false,false);
	StopShapeMotion(false,false);
	if( child==true && mpChild!=NULL ){
		mpChild->StopMotionAll(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->StopMotionAll(false,child);
		}
	}
	return true;
}

int WK7JObj::AttachMatMotion(WK7ModelMotion *pModelMotion,bool next,bool child)
{
	HSD_MatAnimJoint	*pMatAnimJoint;
	int					i,num = pModelMotion->GetModelSetNum();
	SetCurrentAttachMotionSdExpInfo(pModelMotion);
	for(i=0 ; i<num ; i++){
		if( pModelMotion->GetAnimJoint(i,NULL,&pMatAnimJoint,NULL)==false || pMatAnimJoint==NULL ){
			continue;
		}
		SetCurrentAttachMotionSdExpInfo(pModelMotion);
		int	result = AttachMatMotion(pMatAnimJoint,next,child);
		SetCurrentAttachMotionSdExpInfo();
		return result;
	}
	return 0;
}

int WK7JObj::AttachMatMotion(HSD_MatAnimJoint *pMatAnimJoint,bool next,bool child)
{
	if( pMatAnimJoint==NULL ){
		return 0;
	}
	const char			*pMatAnimJointName;
	HSD_MatAnimJoint	*pCurrentMatAnimJoint;
	int					count = 0;
	
	if( mpJObj!=NULL ){
		pCurrentMatAnimJoint = pMatAnimJoint;
		while( pCurrentMatAnimJoint != NULL ){
			pMatAnimJointName = WK7ModelMotion::GetMatAnimJointName(pCurrentMatAnimJoint);
			if( strcmp(pMatAnimJointName,mpName)==0 ){
				if( !mMotionOpDisable ){
					mpMatAnimJoint = pCurrentMatAnimJoint;
					AttachMatMotion();
					SetMotionRate(1.0f,false,false);
					count = 1;
				}
				break;
			}
			pCurrentMatAnimJoint = pCurrentMatAnimJoint->next;
		}
	}
	
	if( child==true && mpChild!=NULL ){
		if( pCurrentMatAnimJoint!=NULL ){
			count += mpChild->AttachMatMotion(pCurrentMatAnimJoint->child,true,true);
		}else{
			count += mpChild->AttachMatMotion(pMatAnimJoint->child,true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			count += pNext->AttachMatMotion(pMatAnimJoint,false,child);
		}
	}
	return count;
}

int WK7JObj::AttachMatMotionClassic(WK7ModelMotion *pModelMotion,bool next,bool child)
{
	HSD_MatAnimJoint	*pMatAnimJoint;
	if( pModelMotion->GetAnimJoint(0,NULL,&pMatAnimJoint,NULL)==false || pMatAnimJoint==NULL ){
		return 0;
	}
	SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int		result = AttachMatMotionClassic(pMatAnimJoint,next,child);
	SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7JObj::AttachMatMotionClassic(HSD_MatAnimJoint *pMatAnimJoint,bool next,bool child)
{
	if( pMatAnimJoint==NULL ){
		return 0;
	}
	int		count = 0;
	if( !mMotionOpDisable ){
		mpMatAnimJoint = pMatAnimJoint;
		AttachMatMotion();
		SetMotionRate(1.0f,false,false);
		count++;
	}
	if( child==true && mpChild!=NULL ){
		count += mpChild->AttachMatMotionClassic(pMatAnimJoint->child,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext,pMatAnimJoint=pMatAnimJoint->next ; pNext!=NULL ; pNext=pNext->GetNext(),pMatAnimJoint=pMatAnimJoint->next){
			count += pNext->AttachMatMotionClassic(pMatAnimJoint,true,child);
		}
	}
	return count;
}

bool WK7JObj::DetachMatMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState &= ~WK7JOBJ_MOTION_MATPLAY;
		HSD_JObjRemoveAnimByFlags(mpJObj,MOBJ_ANIM|TOBJ_ANIM);
		ResetMotionTObjAll(mpJObj,mpJoint);
		mpMatAnimJoint = NULL;
	}
	if( child==true && mpChild!=NULL ){
		mpChild->DetachMatMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->DetachMatMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::ResetMotionPObj(HSD_PObj *pPObj,HSD_PObjDesc * /*pPObjDesc*/)
{
	switch( pobj_type(pPObj) ){
	case	POBJ_SHAPEANIM:{
		HSD_ShapeSet		*pShapeSet;
		int					i;
		if( (pShapeSet=pPObj->u.shape_set)==NULL ){
			break;
		}
		if( pShapeSet->blend.bp==NULL ){
			break;
		}
		for(i=0 ; i<pShapeSet->nb_shape ; i++){
			 pShapeSet->blend.bp[i] = 0.0F;
		}
		}break;
	case	POBJ_ENVELOPE:
	case	POBJ_SKIN:
	default:
		break;
	}
	return true;
}

bool WK7JObj::ResetMotionTObj(HSD_TObj *pTObj,HSD_TObjDesc *pTObjDesc)
{
	pTObj->imagedesc = pTObjDesc->imagedesc;
	return true;
}

bool WK7JObj::ResetMotionTObj(HSD_MObj *pMObj,HSD_MObjDesc *pMObjDesc)
{
	HSD_TObj		*pTObj;
	HSD_TObjDesc	*pTObjDesc;
	for(pTObj=pMObj->tobj,pTObjDesc=pMObjDesc->texdesc ; pTObj!=NULL && pTObjDesc!=NULL ; pTObj=pTObj->next,pTObjDesc=pTObjDesc->next){
		ResetMotionTObj(pTObj,pTObjDesc);
	}
	return true;
}

bool WK7JObj::ResetMotionTObj(HSD_DObj *pDObj,HSD_DObjDesc *pDObjDesc)
{
	HSD_MObj		*pMObj;
	HSD_MObjDesc	*pMObjDesc;
	if( (pMObj=pDObj->mobj)!=NULL && (pMObjDesc=pDObjDesc->mobjdesc)!=NULL ){
		ResetMotionTObj(pMObj,pMObjDesc);
	}
	return true;
}

bool WK7JObj::ResetMotionTObjAll(HSD_JObj *pJObj,HSD_Joint *pJoint)
{
	HSD_DObj		*pDObj;
	HSD_DObjDesc	*pDObjDesc;
	for(pDObj=HSD_JObjGetDObj(pJObj),pDObjDesc=pJoint->u.dobjdesc ; pDObj!=NULL && pDObjDesc!=NULL ; pDObj=pDObj->next,pDObjDesc=pDObjDesc->next){
		ResetMotionTObj(pDObj,pDObjDesc);
	}
	return true;
}

bool WK7JObj::ResetMotionMObj(HSD_MObj *pMObj,HSD_MObjDesc *pMObjDesc,u32 flag)
{
	if( flag & TOBJ_ANIM ){
		HSD_TObj		*pTObj;
		HSD_TObjDesc	*pTObjDesc;
		for(pTObj=pMObj->tobj,pTObjDesc=pMObjDesc->texdesc ; pTObj!=NULL && pTObjDesc!=NULL ; pTObj=pTObj->next,pTObjDesc=pTObjDesc->next){
			ResetMotionTObj(pTObj,pTObjDesc);
		}
	}
	if( pMObj->mat!=NULL && pMObjDesc->mat!=NULL ){
		memcpy(pMObj->mat,pMObjDesc->mat,sizeof(HSD_Material));
	}
	return true;
}

bool WK7JObj::ResetMotionDObj(HSD_DObj *pDObj,HSD_DObjDesc *pDObjDesc,u32 flag)
{
	if( flag & (MOBJ_ANIM|TOBJ_ANIM) ){
		HSD_MObj		*pMObj;
		HSD_MObjDesc	*pMObjDesc;
		if( (pMObj=pDObj->mobj)!=NULL && (pMObjDesc=pDObjDesc->mobjdesc)!=NULL ){
			ResetMotionMObj(pMObj,pMObjDesc,flag);
		}
	}
	if( flag & POBJ_ANIM ){
		HSD_PObj		*pPObj;
		HSD_PObjDesc	*pPObjDesc;
		for(pPObj=pDObj->pobj,pPObjDesc=pDObjDesc->pobjdesc ; pPObj!=NULL && pPObjDesc!=NULL ; pPObj=pPObj->next,pPObjDesc=pPObjDesc->next){
			ResetMotionPObj(pPObj,pPObjDesc);
		}
	}
	return true;
}

bool WK7JObj::ResetMotionDObjAll(HSD_JObj *pJObj,HSD_Joint *pJoint,u32 flag)
{
	HSD_DObj		*pDObj;
	HSD_DObjDesc	*pDObjDesc;
	for(pDObj=HSD_JObjGetDObj(pJObj),pDObjDesc=pJoint->u.dobjdesc ; pDObj!=NULL && pDObjDesc!=NULL ; pDObj=pDObj->next,pDObjDesc=pDObjDesc->next){
		ResetMotionDObj(pDObj,pDObjDesc,flag);
	}
	return true;
}

bool WK7JObj::ResetMatMotion(bool next,bool child)
{
	if( mpJObj!=NULL && mpJoint!=NULL && !mMotionOpDisable ){
		ResetMotionDObjAll(mpJObj,mpJoint,MOBJ_ANIM|TOBJ_ANIM);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->ResetMatMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ResetMatMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::PlayMatMotion(f32 startframe,bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState |= WK7JOBJ_MOTION_MATPLAY;
		HSD_JObjReqAnimByFlags(mpJObj,startframe,MOBJ_ANIM|TOBJ_ANIM);
		RenewMotionRate(false,false);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->PlayMatMotion(startframe,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->PlayMatMotion(startframe,false,child);
		}
	}
	return true;
}

bool WK7JObj::ReplayMatMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState |= WK7JOBJ_MOTION_MATPLAY;
		RenewMotionRate(false,false);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->ReplayMatMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ReplayMatMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::StopMatMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState &= ~WK7JOBJ_MOTION_MATPLAY;
		RenewMotionRate(false,false);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->StopMatMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->StopMatMotion(false,child);
		}
	}
	return true;
}

int WK7JObj::AttachShapeMotion(WK7ModelMotion *pModelMotion,bool next,bool child,bool overwrite)
{
	HSD_ShapeAnimJoint	*pShapeAnimJoint;
	int					i,num = pModelMotion->GetModelSetNum();
	for(i=0 ; i<num ; i++){
		if( pModelMotion->GetAnimJoint(i,NULL,NULL,&pShapeAnimJoint)==false || pShapeAnimJoint==NULL ){
			continue;
		}
		SetCurrentAttachMotionSdExpInfo(pModelMotion);
		int	result = AttachShapeMotion(pShapeAnimJoint,next,child,overwrite);
		SetCurrentAttachMotionSdExpInfo();
		return result;
	}
	return 0;
}

int WK7JObj::AttachShapeMotion(HSD_ShapeAnimJoint *pShapeAnimJoint,bool next,bool child,bool overwrite)
{
	if( pShapeAnimJoint==NULL ){
		return 0;
	}
	const char			*pShapeAnimJointName;
	HSD_ShapeAnimJoint	*pCurrentShapeAnimJoint;
	int					count = 0;
	
	if( mpJObj!=NULL ){
		pCurrentShapeAnimJoint = pShapeAnimJoint;
		while( pCurrentShapeAnimJoint != NULL ){
			pShapeAnimJointName = WK7ModelMotion::GetShapeAnimJointName(pCurrentShapeAnimJoint);
			if( strcmp(pShapeAnimJointName,mpName)==0 ){
				if( !mMotionOpDisable ){
					mpShapeAnimJoint = pCurrentShapeAnimJoint;
					AttachShapeMotion(overwrite);
					SetMotionRate(1.0f,false,false);
					count = 1;
				}
				break;
			}
			pCurrentShapeAnimJoint = pCurrentShapeAnimJoint->next;
		}
	}
	
	if( child==true && mpChild!=NULL ){
		if( pCurrentShapeAnimJoint!=NULL ){
			count += mpChild->AttachShapeMotion(pCurrentShapeAnimJoint->child,true,true,overwrite);
		}else{
			count += mpChild->AttachShapeMotion(pShapeAnimJoint->child,true,true,overwrite);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			count += pNext->AttachShapeMotion(pShapeAnimJoint,false,child,overwrite);
		}
	}
	return count;
}

int WK7JObj::AttachShapeMotionClassic(WK7ModelMotion *pModelMotion,bool next,bool child)
{
	HSD_ShapeAnimJoint	*pShapeAnimJoint;
	if( pModelMotion->GetAnimJoint(0,NULL,NULL,&pShapeAnimJoint)==false || pShapeAnimJoint==NULL ){
		return 0;
	}
	SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int	result = AttachShapeMotionClassic(pShapeAnimJoint,next,child);
	SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7JObj::AttachShapeMotionClassic(HSD_ShapeAnimJoint *pShapeAnimJoint,bool next,bool child)
{
	if( pShapeAnimJoint==NULL ){
		return 0;
	}
	int		count = 0;
	if( !mMotionOpDisable ){
		mpShapeAnimJoint = pShapeAnimJoint;
		AttachShapeMotion(false);
		SetMotionRate(1.0f,false,false);
		count++;
	}
	if( child==true && mpChild!=NULL ){
		count += mpChild->AttachShapeMotionClassic(pShapeAnimJoint->child,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext,pShapeAnimJoint=pShapeAnimJoint->next ; pNext!=NULL ; pNext=pNext->GetNext(),pShapeAnimJoint=pShapeAnimJoint->next){
			count += pNext->AttachShapeMotionClassic(pShapeAnimJoint,false,child);
		}
	}
	return count;
}

bool WK7JObj::DetachShapeMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState &= ~WK7JOBJ_MOTION_SHAPEPLAY;
		HSD_JObjRemoveAnimByFlags(mpJObj,POBJ_ANIM);
		mpShapeAnimJoint = NULL;
	}
	if( child==true && mpChild!=NULL ){
		mpChild->DetachShapeMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->DetachShapeMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::ResetShapeMotion(bool next,bool child)
{
	if( mpJObj!=NULL && mpJoint!=NULL && !mMotionOpDisable ){
		ResetMotionDObjAll(mpJObj,mpJoint,POBJ_ANIM);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->ResetShapeMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ResetShapeMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::PlayShapeMotion(f32 startframe,bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState |= WK7JOBJ_MOTION_SHAPEPLAY;
		HSD_JObjReqAnimByFlags(mpJObj,startframe,POBJ_ANIM);
		RenewMotionRate(false,false);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->PlayShapeMotion(startframe,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->PlayShapeMotion(startframe,false,child);
		}
	}
	return true;
}

bool WK7JObj::ReplayShapeMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState |= WK7JOBJ_MOTION_SHAPEPLAY;
		RenewMotionRate(false,false);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->ReplayShapeMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ReplayShapeMotion(false,child);
		}
	}
	return true;
}

bool WK7JObj::StopShapeMotion(bool next,bool child)
{
	if( mpJObj!=NULL && !mMotionOpDisable ){
		mMotionState &= ~WK7JOBJ_MOTION_SHAPEPLAY;
		RenewMotionRate(false,false);
	}
	if( child==true && mpChild!=NULL ){
		mpChild->StopShapeMotion(true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->StopShapeMotion(false,child);
		}
	}
	return true;
}

int WK7JObj::GetMotionState(bool next,bool child)
{
	int		state;
	if( mpJObj!=NULL ){
		if( mpJObj->aobj!=NULL ){
			return (int)HSD_AObjGetFlags(mpJObj->aobj);
		}
		HSD_DObj	*pDObj;
		for(pDObj=HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
			if( pDObj->aobj!=NULL ){
				return (int)HSD_AObjGetFlags(pDObj->aobj);
			}
			if( pDObj->mobj!=NULL ){
				if( pDObj->mobj->aobj!=NULL ){
					return (int)HSD_AObjGetFlags(pDObj->mobj->aobj);
				}
				HSD_TObj	*pTObj;
				for(pTObj=pDObj->mobj->tobj ; pTObj!=NULL ; pTObj=pTObj->next){
					if( pTObj->aobj!=NULL ){
						return (int)HSD_AObjGetFlags(pTObj->aobj);
					}
				}
			}
			HSD_PObj	*pPObj;
			for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
				if( pPObj->aobj!=NULL ){
					return (int)HSD_AObjGetFlags(pPObj->aobj);
				}
			}
		}
	}
	if( mpChild!=NULL && child==true && (state=mpChild->GetMotionState(true,true))!=AOBJ_NO_ANIM ){
		return state;
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (state=pNext->GetMotionState(false,child))!=AOBJ_NO_ANIM ){
				return state;
			}
		}
	}
	return AOBJ_NO_ANIM;
}

int WK7JObj::GetShapeMotionState(bool next,bool child)
{
	int		state;
	if( mpJObj!=NULL && union_type_dobj(mpJObj) ){
		HSD_DObj	*pDObj;
		for(pDObj=mpJObj->u.dobj ; pDObj!=NULL ; pDObj=pDObj->next){
			if( pDObj->mobj!=NULL && pDObj->mobj->aobj!=NULL ){
				return (int)HSD_AObjGetFlags(pDObj->mobj->aobj);
			}
		}
	}
	if( mpChild!=NULL && child==true && (state=mpChild->GetShapeMotionState(true,true))!=AOBJ_NO_ANIM ){
		return state;
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (state=pNext->GetShapeMotionState(false,child))!=AOBJ_NO_ANIM ){
				return state;
			}
		}
	}
	return AOBJ_NO_ANIM;
}

bool WK7JObj::SetMotionLoop(bool flag,bool next,bool child)
{
	if( !mMotionOpDisable ){
		if( flag==true ){
			mMotionState |=  WK7JOBJ_MOTION_LOOP;
		}else{
			mMotionState &= ~WK7JOBJ_MOTION_LOOP;
		}
		if( mpJObj!=NULL ){
			if( mpJObj->aobj!=NULL ){
				if( flag==true ){
					HSD_AObjSetFlags  (mpJObj->aobj,AOBJ_ANIM_LOOP);
				}else{
					HSD_AObjClearFlags(mpJObj->aobj,AOBJ_ANIM_LOOP);
				}
			}
			HSD_DObj	*pDObj;
			for(pDObj=HSD_JObjGetDObj(mpJObj) ; pDObj!=NULL ; pDObj=pDObj->next){
				if( pDObj->aobj!=NULL ){
					if( flag==true ){
						HSD_AObjSetFlags  (pDObj->aobj,AOBJ_ANIM_LOOP);
					}else{
						HSD_AObjClearFlags(pDObj->aobj,AOBJ_ANIM_LOOP);
					}
				}
				HSD_MObj	*pMObj;
				if( (pMObj=pDObj->mobj)!=NULL ){
					if( pMObj->aobj!=NULL ){
						if( flag==true ){
							HSD_AObjSetFlags  (pMObj->aobj,AOBJ_ANIM_LOOP);
						}else{
							HSD_AObjClearFlags(pMObj->aobj,AOBJ_ANIM_LOOP);
						}
					}
					HSD_TObj	*pTObj;
					for(pTObj=pMObj->tobj ; pTObj!=NULL ; pTObj=pTObj->next){
						if( pTObj->aobj!=NULL ){
							if( flag==true ){
								HSD_AObjSetFlags  (pTObj->aobj,AOBJ_ANIM_LOOP);
							}else{
								HSD_AObjClearFlags(pTObj->aobj,AOBJ_ANIM_LOOP);
							}
						}
					}
				}
				HSD_PObj	*pPObj;
				for(pPObj=pDObj->pobj ; pPObj!=NULL ; pPObj=pPObj->next){
					if( pPObj->aobj!=NULL ){
						if( flag==true ){
							HSD_AObjSetFlags  (pPObj->aobj,AOBJ_ANIM_LOOP);
						}else{
							HSD_AObjClearFlags(pPObj->aobj,AOBJ_ANIM_LOOP);
						}
					}
				}
			}
		}
	}
	if( child==true ){
		if( mpChild!=NULL ){
			mpChild->SetMotionLoop(flag,true,true);
		}
		WK7JObjAttachModel		*pAttachModel;
		WK7Model				*pModel;
		WK7JObj					*pJObj;
		for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
			if( (pModel = pAttachModel->GetModel())==NULL || (pJObj = pModel->GetTopObj())==NULL ){
				continue;
			}
			pJObj->SetMotionLoop(flag,true,true);
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetMotionLoop(flag,false,child);
		}
	}
	return true;
}

bool WK7JObj::GetMotionLoop(bool next,bool child)
{
	return (GetMotionState(next,child) & AOBJ_ANIM_LOOP)? true : false;
}

u32 WK7JObj::GetFlag(void)
{
	if( mpJObj==NULL ){
		return 0;
	}
	return HSD_JObjGetFlags(mpJObj);
}

bool WK7JObj::SetFlag(u32 flag,bool next,bool child)
{
	if( mpJObj!=NULL ){
		HSD_JObjSetFlags(mpJObj,flag);
	}
	if( mpChild!=NULL && child==true ){
		mpChild->SetFlag(flag,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->SetFlag(flag,false,child);
		}
	}
	return true;
}

bool WK7JObj::ClearFlag(u32 flag,bool next,bool child)
{
	if( mpJObj!=NULL ){
		HSD_JObjClearFlags(mpJObj,flag);
	}
	if( mpChild!=NULL && child==true ){
		mpChild->ClearFlag(flag,true,true);
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			pNext->ClearFlag(flag,false,child);
		}
	}
	return true;
}

HSD_DObj* WK7JObj::GetDObj(const char *pName,bool next,bool child)
{
	HSD_DObj		*d;
	HSD_DObjDesc	*dd;
	
	for(dd=mpJoint->u.dobjdesc,d=mpJObj->u.dobj ; dd ; dd=dd->next,d=d->next){
		if( !strcmp(::GetDescName(dd),pName) ){
			return d;
		}
	}
	if( child==true && mpChild!=NULL ){
		if( (d=mpChild->GetDObj(pName,true,true))!=NULL ){
			return d;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (d=pNext->GetDObj(pName,false,child))!=NULL ){
				return d;
			}
		}
	}
	return NULL;
}

HSD_DObj* WK7JObj::GetDObjByPartialName(const char *pName,bool next,bool child)
{
	HSD_DObj		*d;
	HSD_DObjDesc	*dd;
	
	for(dd=mpJoint->u.dobjdesc,d=mpJObj->u.dobj ; dd ; dd=dd->next,d=d->next){
		if( strstr(::GetDescName(dd),pName)!=NULL ){
			return d;
		}
	}
	if( child==true && mpChild!=NULL ){
		if( (d=mpChild->GetDObjByPartialName(pName,true,true))!=NULL ){
			return d;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (d=pNext->GetDObjByPartialName(pName,false,child))!=NULL ){
				return d;
			}
		}
	}
	return NULL;
}

const char* WK7JObj::GetDObjName(HSD_DObj *pDObj)
{
	if( pDObj==NULL ){
		return NULL;
	}
	HSD_DObj		*d;
	HSD_DObjDesc	*dd;
	
	for(dd=mpJoint->u.dobjdesc,d=mpJObj->u.dobj ; dd ; dd=dd->next,d=d->next){
		if( pDObj==d ){
			return ::GetDescName(dd);
		}
	}
	return NULL;
}

WK7JObj* WK7JObj::GetParent(const char *pName)
{
	if( mpParent==NULL || mpParent->GetName()==NULL ){
		return NULL;
	}
	if( strcmp(mpParent->GetName(),pName)==0 ){
		return mpParent;
	}
	return mpParent->GetParent(pName);
}

bool WK7JObj::HasParent(WK7JObj *pJObj)
{
	if( pJObj==NULL ){
		return false;
	}
	WK7JObj		*pParent;
	for(pParent=GetParent() ; pParent!=NULL ; pParent=pParent->GetParent()){
		if( pParent==pJObj ){
			return true;
		}
	}
	return false;
}

bool WK7JObj::HasChild(WK7JObj *pJObj,bool child)
{
	if( pJObj==NULL ){
		return false;
	}
	WK7JObj		*pChild;
	for(pChild=GetChild() ; pChild!=NULL ; pChild=pChild->GetNext()){
		if( pChild==pJObj ){
			return true;
		}
		if( child==true ){
			if( pChild->HasChild(pJObj,true)==true ){
				return true;
			}
		}
	}
	return false;
}

bool WK7JObj::HasBrother(WK7JObj *pJObj)
{
	if( pJObj==NULL ){
		return false;
	}
	WK7JObj		*pBroter;
	for(pBroter=GetPrev() ; pBroter!=NULL ; pBroter=pBroter->GetPrev()){
		if( pBroter==pJObj ){
			return true;
		}
	}
	for(pBroter=GetNext() ; pBroter!=NULL ; pBroter=pBroter->GetNext()){
		if( pBroter==pJObj ){
			return true;
		}
	}
	return false;
}

WK7JObj* WK7JObj::GetObjPartialName2(const char *pName,bool next,bool child)
{
	if( pName==NULL ){
		return NULL;
	}
	if( strstr(mpName,pName)==mpName ){
		return this;
	}
	WK7JObj		*pObj;
	if( child==true && mpChild!=NULL ){
		if( (pObj=mpChild->GetObjPartialName2(pName,true,true))!=NULL ){
			return pObj;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (pObj=pNext->GetObjPartialName2(pName,false,child))!=NULL ){
				return pObj;
			}
		}
	}
	return NULL;
}

WK7JObj* WK7JObj::GetObj(const char *pName,bool next,bool child)
{
	if( strcmp(mpName,pName)==0 ){
		return this;
	}
	WK7JObj		*pObj;
	if( child==true && mpChild!=NULL ){
		if( (pObj=mpChild->GetObj(pName,true,true))!=NULL ){
			return pObj;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (pObj=pNext->GetObj(pName,false,child))!=NULL ){
				return pObj;
			}
		}
	}
//	if( next==true && mpNext!=NULL ){
//		if( (pObj=mpNext->GetObj(pName,true,child))!=NULL ){
//			return pObj;
//		}
//	}
	return NULL;
}

WK7JObj* WK7JObj::GetObjLC(const char *pName,bool next,bool child)
{
	if( stricmp(mpName,pName)==0 ){
		return this;
	}
	WK7JObj		*pObj;
	if( child==true && mpChild!=NULL ){
		if( (pObj=mpChild->GetObjLC(pName,true,true))!=NULL ){
			return pObj;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (pObj=pNext->GetObjLC(pName,false,child))!=NULL ){
				return pObj;
			}
		}
	}
	return NULL;
}

WK7JObj* WK7JObj::GetObj(HSD_Joint *pJoint,bool next,bool child)
{
	if( mpJoint==pJoint ){
		return this;
	}
	WK7JObj		*pObj;
	if( child==true && mpChild!=NULL ){
		if( (pObj=mpChild->GetObj(pJoint,true,true))!=NULL ){
			return pObj;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (pObj=pNext->GetObj(pJoint,false,child))!=NULL ){
				return pObj;
			}
		}
	}
	return NULL;
}

WK7JObj* WK7JObj::GetObjPartialName(const char *pPartialName,bool next,bool child)
{
	{
		int			i;
		const char	*psrc;
		char		dst[256];
		char		*pdst;
		psrc = mpName;
		pdst = dst;
		for(i=0 ; i<256 ; i++){
			switch( *psrc ){
			case	'\0':
			case	'_':
				*pdst = '\0';
				if( strcmp(dst,pPartialName)==0 ){
					return this;
				}
				pdst = dst;
				if( *psrc=='\0' ){
					i = 256;
				}else{
					i = 0;
					psrc++;
				}
				break;
			default:
				*pdst++ = *psrc++;
				break;
			}
		}
	}
	WK7JObj		*pObj;
	if( child==true && mpChild!=NULL ){
		if( (pObj=mpChild->GetObjPartialName(pPartialName,true,true))!=NULL ){
			return pObj;
		}
	}
	if( next==true ){
		WK7JObj		*pNext;
		for(pNext=mpNext ; pNext!=NULL ; pNext=pNext->GetNext()){
			if( (pObj=pNext->GetObjPartialName(pPartialName,false,child))!=NULL ){
				return pObj;
			}
		}
	}
	return NULL;
}

bool WK7JObj::SetPosition(f32 x,f32 y,f32 z)
{
	if( mpJObj==NULL ){
		return false;
	}
	Vec		pos;
	pos.x = x;
	pos.y = y;
	pos.z = z;
	HSD_JObjSetTranslate(mpJObj,&pos);
	return true;
}

bool WK7JObj::SetPosition(const Vec *pPosition)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetTranslate(mpJObj,(Vec *)pPosition);
	return true;
}

bool WK7JObj::GetPosition(Vec *pPosition)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjGetTranslate(mpJObj,pPosition);
	return true;
}

bool WK7JObj::SetRotation(f32 x,f32 y,f32 z)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetRx(mpJObj,x);
	HSD_JObjSetRy(mpJObj,y);
	HSD_JObjSetRz(mpJObj,z);
	return true;
}

bool WK7JObj::SetRotX(f32 x)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetRx(mpJObj,x);
	return true;
}

bool WK7JObj::SetRotY(f32 y)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetRy(mpJObj,y);
	return true;
}

bool WK7JObj::SetRotZ(f32 z)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetRz(mpJObj,z);
	return true;
}

f32 WK7JObj::GetRotX(void)
{
	if( mpJObj==NULL ){
		return 0.0f;
	}
	return HSD_JObjGetRx(mpJObj);
}

f32 WK7JObj::GetRotY(void)
{
	if( mpJObj==NULL ){
		return 0.0f;
	}
	return HSD_JObjGetRy(mpJObj);
}

f32 WK7JObj::GetRotZ(void)
{
	if( mpJObj==NULL ){
		return 0.0f;
	}
	return HSD_JObjGetRz(mpJObj);
}

bool WK7JObj::SetRotation(const Vec *pRotation)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetRx(mpJObj,pRotation->x);
	HSD_JObjSetRy(mpJObj,pRotation->y);
	HSD_JObjSetRz(mpJObj,pRotation->z);
	return true;
}

bool WK7JObj::GetRotation(f32 *px,f32 *py,f32 *pz)
{
	if( mpJObj==NULL ){
		return false;
	}
	*px = HSD_JObjGetRx(mpJObj);
	*py = HSD_JObjGetRy(mpJObj);
	*pz = HSD_JObjGetRz(mpJObj);
	return true;
}

bool WK7JObj::GetRotation(Vec *pRotation)
{
	if( mpJObj==NULL ){
		return false;
	}
	pRotation->x = HSD_JObjGetRx(mpJObj);
	pRotation->y = HSD_JObjGetRy(mpJObj);
	pRotation->z = HSD_JObjGetRz(mpJObj);
	return true;
}

bool WK7JObj::SetScale(f32 x,f32 y,f32 z)
{
	if( mpJObj==NULL ){
		return false;
	}
	Vec		scale;
	scale.x = x;
	scale.y = y;
	scale.z = z;
	HSD_JObjSetScale(mpJObj,&scale);
	return true;
}

bool WK7JObj::SetScale(const Vec *pScale)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjSetScale(mpJObj,(Vec *)pScale);
	return true;
}

bool WK7JObj::GetScale(f32 *px,f32 *py,f32 *pz)
{
	if( mpJObj==NULL ){
		return false;
	}
	Vec		scale;
	HSD_JObjGetScale(mpJObj,&scale);
	*px = scale.x;
	*py = scale.y;
	*pz = scale.z;
	return true;
}

bool WK7JObj::GetScale(Vec *pScale)
{
	if( mpJObj==NULL ){
		return false;
	}
	HSD_JObjGetScale(mpJObj,pScale);
	return true;
}

bool WK7JObj::CopyUserDefinedMatrix(WK7JObj *pJObj,bool next,bool child)
{
	if( pJObj==NULL || pJObj->GetName()==NULL || mpName==NULL || strcmp(pJObj->GetName(),mpName)!=0 ){
		return false;
	}
	bool		result = true;
	HSD_JObj	*pHSDJObj = pJObj->GetJObj();
	if( pHSDJObj!=NULL && SetUserDefinedMatrix(pHSDJObj->mtx)==false ){	//	if( SetUserDefinedMatrix(pJObj->GetMtxPtr())==false ){
		result = false;
	}
	if( child==true && mpChild!=NULL && mpChild->GetName()!=NULL ){
		WK7JObj	*pChild;
		for(pChild=pJObj->GetChild() ; pChild!=NULL ; pChild=pChild->GetNext()){
			if( pChild->GetName()==NULL ){
				continue;
			}
			if( strcmp(pChild->GetName(),mpChild->GetName())!=0 ){
				continue;
			}
			if( mpChild->CopyUserDefinedMatrix(pChild,true,true)==false ){
				result = false;
			}
			break;
		}
	}
	if( next==true ){
		WK7JObj	*pMyNext;
		for(pMyNext=mpNext ; pMyNext!=NULL ; pMyNext=pMyNext->GetNext()){
			if( pMyNext->GetName()==NULL ){
				continue;
			}
			WK7JObj	*pNext;
			for(pNext=pJObj->GetNext() ; pNext!=NULL ; pNext=pNext->GetNext()){
				if( pNext->GetName()==NULL ){
					continue;
				}
				if( strcmp(pMyNext->GetName(),pNext->GetName())!=0 ){
					continue;
				}
				if( pMyNext->CopyUserDefinedMatrix(pNext,false,true)==false ){
					result = false;
				}
				break;
			}
		}
	}
	return result;
}

bool WK7JObj::CopyTRS(WK7JObj *pJObj,bool next,bool child)
{
	if( pJObj==NULL || pJObj->GetName()==NULL || mpName==NULL || strcmp(pJObj->GetName(),mpName)!=0 ){
		return false;
	}
	HSD_JObj		*pHSDJObj = pJObj->GetJObj();
	bool			result = true;
	if( pHSDJObj!=NULL && mpJObj!=NULL ){
		Vec		pos,rot,scl;
		HSD_JObjGetScale(pHSDJObj,&scl);
		rot.x = HSD_JObjGetRx(pHSDJObj);
		rot.y = HSD_JObjGetRy(pHSDJObj);
		rot.z = HSD_JObjGetRz(pHSDJObj);
		HSD_JObjGetTranslate(pHSDJObj,&pos);
		HSD_JObjSetScale(mpJObj,&scl);
		HSD_JObjSetRx(mpJObj,rot.x);
		HSD_JObjSetRy(mpJObj,rot.y);
		HSD_JObjSetRz(mpJObj,rot.z);
		HSD_JObjSetTranslate(mpJObj,&pos);
	}else{
		result = false;
	}
	if( child==true && mpChild!=NULL && mpChild->GetName()!=NULL ){
		WK7JObj	*pChild;
		for(pChild=pJObj->GetChild() ; pChild!=NULL ; pChild=pChild->GetNext()){
			if( pChild->GetName()==NULL ){
				continue;
			}
			if( strcmp(pChild->GetName(),mpChild->GetName())!=0 ){
				continue;
			}
			if( mpChild->CopyTRS(pChild,true,true)==false ){
				result = false;
			}
			break;
		}
	}
	if( next==true ){
		WK7JObj	*pMyNext;
		for(pMyNext=mpNext ; pMyNext!=NULL ; pMyNext=pMyNext->GetNext()){
			if( pMyNext->GetName()==NULL ){
				continue;
			}
			WK7JObj	*pNext;
			for(pNext=pJObj->GetNext() ; pNext!=NULL ; pNext=pNext->GetNext()){
				if( pNext->GetName()==NULL ){
					continue;
				}
				if( strcmp(pMyNext->GetName(),pNext->GetName())!=0 ){
					continue;
				}
				if( pMyNext->CopyTRS(pNext,false,true)==false ){
					result = false;
				}
				break;
			}
		}
	}
	return result;
}

MtxPtr WK7JObj::GetMtxPtr(void)
{
	if( mpJObj==NULL ){
		return NULL;
	}
	return HSD_JObjGetMtxPtr(mpJObj);
}

bool WK7JObj::GetWorldPosition(Vec *pPosition)
{
	MtxPtr		mtxptr = GetMtxPtr();
	if( mtxptr==NULL ){
		return false;
	}
	pPosition->x = mtxptr[0][3];
	pPosition->y = mtxptr[1][3];
	pPosition->z = mtxptr[2][3];
	return true;
}

bool WK7JObj::SetWorldPosition(f32 x,f32 y,f32 z)
{
	Vec		pos;
	pos.x = x;
	pos.y = y;
	pos.z = z;
	return SetWorldPosition(&pos);
}

bool WK7JObj::SetWorldPosition(const Vec *pPosition)
{
	if( mpParent==NULL ){
		return SetPosition(pPosition);
	}
	MtxPtr		mtxptr = mpParent->GetMtxPtr();
	Mtx			inv;
	if( mtxptr==NULL ){
		return false;
	}
	Vec			pos;
	MTXInverse(mtxptr,inv);
	MTXMultVec(inv,pPosition,&pos);
	HSD_JObjSetTranslate(mpJObj,&pos);
	return true;
}

WK7JObjCallBackFunc* WK7JObj::SetPreFrameProcessCallBack  (WK7JObjCallBackFunc *pCallBackFunc,void *pData)
{
	WK7JObjCallBackFunc	*pOld = mpPreFrameProcessCallBack;
	mpPreFrameProcessCallBack = pCallBackFunc;
	mpPreFrameProcessData     = pData;
	return pOld;
}

WK7JObjCallBackFunc* WK7JObj::SetPostFrameProcessCallBack (WK7JObjCallBackFunc *pCallBackFunc,void *pData)
{
	WK7JObjCallBackFunc	*pOld = mpPostFrameProcessCallBack;
	mpPostFrameProcessCallBack = pCallBackFunc;
	mpPostFrameProcessData     = pData;
	return pOld;
}

WK7JObjCallBackFunc* WK7JObj::SetPreRenderProcessCallBack (WK7JObjCallBackFunc *pCallBackFunc,void *pData)
{
	WK7JObjCallBackFunc	*pOld = mpPreRenderProcessCallBack;
	mpPreRenderProcessCallBack = pCallBackFunc;
	mpPreRenderProcessData     = pData;
	return pOld;
}

WK7JObjCallBackFunc* WK7JObj::SetPostRenderProcessCallBack(WK7JObjCallBackFunc *pCallBackFunc,void *pData)
{
	WK7JObjCallBackFunc	*pOld = mpPostRenderProcessCallBack;
	mpPostRenderProcessCallBack = pCallBackFunc;
	mpPostRenderProcessData     = pData;
	return pOld;
}

#define	WK7JOBJ_LEVELMAX			(64)
#ifdef	_DEBUG
int WK7Model::ReviseJointLevel(HSD_Joint *pJoint,bool next,int count)
{
	int		result,ccount;
	if( pJoint==NULL ){
		return 0;
	}
	count++;
	result = ReviseJointLevel(pJoint->child,true,count);
	if( next==false ){
		return result;
	}
	ccount = count;
	while( pJoint!=NULL && count<WK7JOBJ_LEVELMAX ){
		pJoint = pJoint->next;
		result += ReviseJointLevel(pJoint,false,ccount);
		count++;
	}
	if( pJoint!=NULL ){
		HSD_Joint		*pNext;
		int				lostcount;
		char			str[128];
		for(pNext=pJoint->next,lostcount=0 ; pNext!=NULL ; pNext=pNext->next,lostcount++){}
		pJoint->next = NULL;
		sprintf(str,"Model Joint Level Over !! : Lost Joint Count = %d\n",lostcount);
		OSReport(str);
		K7AppDebugPrint(str);
	}
	return result;
}
#endif

WK7Model::WK7Model(WK7Archive *pArchive,u32 mirrorType)
{
	HSD_JObj		*pJObj;
	mpFog        = NULL;
	mpTopObj     = NULL;
	mTopObjNum   = 0;
	mFlag        = 0;
	mpAttachJObj = NULL;
	mpAlpha      = NULL;
	mAlpha.SetValue(1.0f);
	mMatAmbientColorR.SetValue(0xff);
	mMatAmbientColorG.SetValue(0xff);
	mMatAmbientColorB.SetValue(0xff);
	mMatDiffuseColorR.SetValue(0xff);
	mMatDiffuseColorG.SetValue(0xff);
	mMatDiffuseColorB.SetValue(0xff);
	if( (mpArchive = pArchive)!=NULL ){
		
		int			i,modelSetNum;
		C_ModelSet	*pModelSet;
		WK7JObj		*pObj,*pPrevObj;
		
		WK7JObj::mpCurrentSdExpInfo = pArchive->GetSdExpInfo();
		
		modelSetNum = mpArchive->GetModelSetNum();
		pPrevObj    = NULL;
		for(i=0 ; i<modelSetNum ; i++){
			pModelSet = mpArchive->GetModelSet(i);
#ifdef	_DEBUG
			ReviseJointLevel(pModelSet->joint,false,0);
#endif
			pJObj     = HSD_JObjLoadJoint(pModelSet->joint);
			if( pJObj!=NULL ){
				ghmGcMemSetAllocName("WK7JObj");
				pObj = new WK7JObj(pJObj,pModelSet->joint,NULL,pPrevObj);
			}else{
				pObj = NULL;
			}
			if( pObj!=NULL ){
				if( mpTopObj==NULL ){
					mpTopObj = pObj;
				}
				mTopObjNum++;
				pPrevObj = pObj;
			}
		}
//		if( mpTopObj != NULL ){
//#ifdef	CLASSICAL_SCALING_CHECK
//			// ͂
//			if( !(mpTopObj->GetFlag() & JOBJ_CLASSICAL_SCALING) ){
//				OSDebugReport("Warning : Non Classical Scaling Obj - Archive(type %08x id %08x)\n",pArchive->GetType(),pArchive->GetID());
//			}
//#endif
//			mpTopObj->SetFlag(JOBJ_CLASSICAL_SCALING);
//		}
		mpArchive->ReferenceArchive();
		
		WK7JObj::mpCurrentSdExpInfo = NULL;
		
	}
//	if( mirrorType==0 ){
//		mpMirrorJObj = NULL;
//	}else{
//		if( (mpMirrorJObj = HSD_JObjAlloc())!=NULL ){
//			Vec		scl;
//			HSD_JObjRef(mpMirrorJObj);
//			if( mirrorType & WK7MODEL_MIRROR_X ){
//				scl.x = -1.0f;
//			}else{
//				scl.x =  1.0f;
//			}
//			if( mirrorType & WK7MODEL_MIRROR_Y ){
//				scl.y = -1.0f;
//			}else{
//				scl.y =  1.0f;
//			}
//			if( mirrorType & WK7MODEL_MIRROR_Z ){
//				scl.z = -1.0f;
//			}else{
//				scl.z =  1.0f;
//			}
//			HSD_JObjSetScale(mpMirrorJObj,&scl);
//			HSD_JObjSetFlags(mpMirrorJObj,JOBJ_CLASSICAL_SCALING);
//			HSD_JObjSetupMatrix(mpMirrorJObj);
//			if( mpTopObj!=NULL && mpTopObj->GetJObj()!=NULL ){
//				HSD_JObjAddChild(mpMirrorJObj,mpTopObj->GetJObj());
////				mpTopObj->SetScale   (&scl);
//				mpTopObj->SetCullMode(WK7JOBJ_CULLMODE_REV,true,true);
//			}
//		}
//	}
	mMirrorPos.x    = 0.0f;
	mMirrorPos.y    = 0.0f;
	mMirrorPos.z    = 0.0f;
	mMirrorType     = 0;
	mpMirrorJObj[0] = 
	mpMirrorJObj[1] = NULL;
	SetMirror(mirrorType);
}

WK7Model::~WK7Model()
{
	GHM_SAFE_DELETE(mpFog);
	ClearMirror();
	GHM_SAFE_DELETE(mpAlpha);
	Detach();
	if( mpTopObj!=NULL ){
		mpTopObj->SetRenderMode(WK7JOBJ_RENDERMODE_NORMAL,true,true);
		mpTopObj->SetPE        (NULL,true,true);
	}
	GHM_SAFE_DELETE(mpTopObj);
	if( mpArchive!=NULL ){
		mpArchive->ReleaseReferenceArchive();
	}
}

bool WK7Model::MakeBound(int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->MakeBound(false,true);
	}
	if( mpTopObj==NULL ){
		return false;
	}
	return mpTopObj->MakeBound();
}

bool WK7Model::SetCastShadowDisable(bool flag,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetCastShadowDisable(flag,false,true);
	}
	if( mpTopObj!=NULL ){
		return mpTopObj->SetCastShadowDisable(flag);
	}
	return false;
}

bool WK7Model::GetCastShadowDisable(int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->GetCastShadowDisable(false,true);
	}
	if( mpTopObj!=NULL ){
		return mpTopObj->GetCastShadowDisable();
	}
	return false;
}

bool WK7Model::LoadFogDesc(HSD_FogDesc *pFogDesc)
{
	ReleaseFog();
	ghmGcMemSetAllocName("MDLFOG");
	mpFog = new WK7Fog(pFogDesc);
	return true;
}

bool WK7Model::ReleaseFog(void)
{
	GHM_SAFE_DELETE(mpFog);
	return true;
}

bool WK7Model::SetFogEnable(bool flag)
{
	if( mpFog==NULL ){
		return false;
	}
	mpFog->SetEnable(flag);
	return true;
}

bool WK7Model::GetFogEnable(void)
{
	if( mpFog==NULL ){
		return false;
	}
	return mpFog->GetEnable();
}

bool WK7Model::ClearMirror(void)
{
	int			i;
	for(i=0 ; i<2 ; i++){
		if( mpMirrorJObj[i]!=NULL ){
			HSD_JObjReparent(mpMirrorJObj[i],NULL);
			HSD_JObjUnref   (mpMirrorJObj[i]);
			mpMirrorJObj[i] = NULL;
		}
	}
	mMirrorType = 0;
	return true;
}

bool WK7Model::SetMirror(u32 mirrorType,const Vec *pPos)
{
static	const Vec	zero = { 0.0f,0.0f,0.0f };
	if( pPos==NULL ){
		pPos = &zero;
	}
	mMirrorPos = *pPos;
	ClearMirror();
	Vec		scl = { 1.0f,1.0f,1.0f };
	Vec		trs;
	trs.x = -pPos->x;
	trs.y = -pPos->y;
	trs.z = -pPos->z;
	mMirrorType = mirrorType;
	switch( mirrorType ){
	case	WK7MODEL_MIRROR_X: scl.x = -1.0f; trs.x = -trs.x; break;
	case	WK7MODEL_MIRROR_Y: scl.y = -1.0f; trs.y = -trs.y; break;
	case	WK7MODEL_MIRROR_Z: scl.z = -1.0f; trs.z = -trs.z; break;
	case	0:
		if( mMirrorType && mpTopObj!=NULL ){
			mpTopObj->SetCullMode(WK7JOBJ_CULLMODE_REV,true,true);
		}
		return true;
	default:
		if( mMirrorType && mpTopObj!=NULL ){
			mpTopObj->SetCullMode(WK7JOBJ_CULLMODE_REV,true,true);
		}
		mMirrorType = 0;
		return false;
	}
	int			i;
	for(i=0 ; i<2 ; i++){
		if( (mpMirrorJObj[i] = HSD_JObjAlloc())==NULL ){
			ClearMirror();
			return false;
		}
		HSD_JObjRef(mpMirrorJObj[i]);
	}
	
	HSD_JObjSetTranslate(mpMirrorJObj[0],(Vec *)pPos);
	HSD_JObjSetFlags    (mpMirrorJObj[0],JOBJ_CLASSICAL_SCALING);
	HSD_JObjSetupMatrix (mpMirrorJObj[0]);
	HSD_JObjAddChild    (mpMirrorJObj[0],mpMirrorJObj[1]);
	
	HSD_JObjSetScale    (mpMirrorJObj[1],&scl);
	HSD_JObjSetFlags    (mpMirrorJObj[1],JOBJ_CLASSICAL_SCALING);
	HSD_JObjSetupMatrix (mpMirrorJObj[1]);
	HSD_JObjSetTranslate(mpMirrorJObj[1],&trs);
	
	if( mpTopObj!=NULL && mpTopObj->GetJObj()!=NULL ){
		HSD_JObjAddChild(mpMirrorJObj[1],mpTopObj->GetJObj());
		mpTopObj->SetCullMode(WK7JOBJ_CULLMODE_REV,true,true);
	}
	
	return true;
}

void WK7Model::SetReferencedObj(WK7JObj *pReferJObj,WK7JObj *pBaseJObj)
{
	if( pReferJObj==NULL || pBaseJObj==NULL ){
		return;
	}
	WK7JObj		*pChild;
	WK7JObj		*pReferNext,*pBaseNext;
	pBaseJObj->SetReferencedObj(pReferJObj);
	if( (pChild=pReferJObj->GetChild())!=NULL ){
		SetReferencedObj(pChild,pBaseJObj->GetChild());
	}
	if( pReferJObj->GetPrev()!=NULL ){
		return;
	}
	for(pReferNext=pReferJObj->GetNext(),pBaseNext=pBaseJObj->GetNext() ;
	     pReferNext!=NULL && pBaseNext!=NULL ;
	      pReferNext=pReferNext->GetNext(),pBaseNext=pBaseNext->GetNext()){
		SetReferencedObj(pReferNext,pBaseNext);
	}
}

WK7Model* WK7Model::CreateMirrorModel(u32 mirrorType)
{
	if( mpArchive==NULL || mirrorType==0 ){
		return NULL;
	}
	WK7Model	*pModel = mpArchive->CreateModel(mirrorType);
	if( pModel==NULL ){
		return NULL;
	}
	WK7JObj		*pMirrorTopJObj = pModel->GetTopObj();
	SetReferencedObj(pMirrorTopJObj,mpTopObj);
	return pModel;
}

WK7SdExpInfo* WK7Model::GetSdExpInfo(void)
{
	if( mpArchive==NULL ){
		return NULL;
	}
	return mpArchive->GetSdExpInfo();
}

bool WK7Model::FrameProcess(void)
{
	bool	fade = false;
	if( mMatAmbientColorFade ){
		mMatAmbientColorR.FrameProcess();
		mMatAmbientColorG.FrameProcess();
		mMatAmbientColorB.FrameProcess();
		if( mMatAmbientColorR.IsAnimating()==false &&
		    mMatAmbientColorG.IsAnimating()==false &&
		    mMatAmbientColorB.IsAnimating()==false ){
			mMatAmbientColorFade = 0;
		}
		GXColor		ambient;
		ambient.r = (u8)mMatAmbientColorR.GetValue();
		ambient.g = (u8)mMatAmbientColorG.GetValue();
		ambient.b = (u8)mMatAmbientColorB.GetValue();
		ambient.a = 0xff;
		if( mMatAmbientColorFadeLE ){
			SetMaterialAmbientColorLE(ambient);
		}else{
			SetMaterialAmbientColor(ambient);
		}
	}
	if( mMatDiffuseColorFade ){
		fade = true;
		mMatDiffuseColorR.FrameProcess();
		mMatDiffuseColorG.FrameProcess();
		mMatDiffuseColorB.FrameProcess();
		if( mMatDiffuseColorR.IsAnimating()==false &&
		    mMatDiffuseColorG.IsAnimating()==false &&
		    mMatDiffuseColorB.IsAnimating()==false ){
			mMatDiffuseColorFade = 0;
		}
		GXColor		diffuse;
		diffuse.r = (u8)mMatDiffuseColorR.GetValue();
		diffuse.g = (u8)mMatDiffuseColorG.GetValue();
		diffuse.b = (u8)mMatDiffuseColorB.GetValue();
		diffuse.a = 0xff;
		if( mMatDiffuseColorFadeLE ){
			SetMaterialDiffuseColorLE(diffuse);
		}else{
			SetMaterialDiffuseColor(diffuse);
		}
	}
	f32		alpha = mAlpha.GetValue();
	mAlpha.FrameProcess();
	if( alpha!=mAlpha.GetValue() ){
		_SetAlpha(mAlpha.GetValue());
	}
	if( mpTopObj!=NULL ){
		mpTopObj->FrameProcess(true,true);
//		mpTopObj->FrameProcess(true,false);
	}
	return true;
}

bool WK7Model::RenderProcess(HSD_TrspMask trsp,u32 rendermode)
{
	if( mpTopObj!=NULL ){
		if( mpFog!=NULL ){
			mpFog->SetCurrent();
		}
		if( mCameraAlphaControl ){
			HSD_CObj	*pCObj;
			if( (pCObj=HSD_CObjGetCurrent())!=NULL ){
				Vec			campos,pos;
				MtxPtr		mtxptr = HSD_CObjGetViewingMtxPtr(pCObj);
				Mtx			invview;
				MTXTranspose(mtxptr,invview);
				campos.x = -mtxptr[0][3];
				campos.y = -mtxptr[1][3];
				campos.z = -mtxptr[2][3];
				MTXMultVecSR(invview,&campos,&campos);
				if( GetPosition(&pos)==true ){
					pos.y = 0.0f;
					campos.y = 0.0f;
					f32				dist = GHM_VEC_GET_DIST(&pos,&campos);
					f32				alpha;
					static f32		mindist = 5.0f,maxdist = 15.0f;
					if( dist < mindist ){
						dist = mindist;
					}else if( dist > maxdist ){
						dist = maxdist;
					}
					alpha = mAlpha.GetValue() * (dist - mindist) / (maxdist - mindist);
					GHM_LIMIT(alpha,1.0f,0.0f);
					if( mpAlpha==NULL ){
						if( alpha < 1.0f ){
							ghmGcMemSetAllocName("C_SetAlp");
							mpAlpha = new C_SetAlpha(mpTopObj->GetJObj(),alpha);
						}else{
							GHM_SAFE_DELETE(mpAlpha);
						}
					}else{
						if( alpha >= 1.0f ){
							GHM_SAFE_DELETE(mpAlpha);
						}else{
							mpAlpha->SetVal(alpha);
						}
					}
				}
			}
		}
		mpTopObj->RenderProcess(true,true,trsp,rendermode);
		if( mpFog!=NULL ){
			mpFog->EndCurrent();
		}
	}
	return true;
}

bool WK7JObj::ClearAttachModel(void)
{
	WK7JObjAttachModel		*pAttachModel,*pNext;
	WK7Model				*pModel;
	for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pNext){
		pNext  = pAttachModel->GetNext();
		pModel = pAttachModel->GetModel();
		pModel->Detach();
	}
	return true;
}

bool WK7JObj::AttachModel(WK7Model *pModel)
{
	if( pModel==NULL ){
		return false;
	}
	WK7JObjAttachModel	*pAttachModel = GetAttachModel(pModel);
	if( pAttachModel!=NULL ){
		// do^h~
		return true;
	}
	ghmGcMemSetAllocName("WK7JObjAM");
	pAttachModel = new WK7JObjAttachModel(&mAttachModels,pModel);
	return true;
}

bool WK7JObj::DetachModel(WK7Model *pModel)
{
	if( pModel==NULL ){
		return false;
	}
	WK7JObjAttachModel	*pAtttachModel = GetAttachModel(pModel);
	if( pAtttachModel==NULL ){
		return true;
	}
	delete pAtttachModel;
	return true;
}

WK7JObjAttachModel* WK7JObj::GetAttachModel(WK7Model *pModel)
{
	WK7JObjAttachModel		*pAttachModel;
	for(pAttachModel=(WK7JObjAttachModel *)mAttachModels.GetNext() ; pAttachModel!=NULL ; pAttachModel=pAttachModel->GetNext()){
		if( pAttachModel->GetModel()==pModel ){
			return pAttachModel;
		}
	}
	return NULL;
}

bool WK7Model::Attach(WK7JObj *pAttachJObj)
{
	if( mpAttachJObj!=NULL ){
		if( Detach()==false ){
			return false;
		}
	}
	if( pAttachJObj!=NULL ){
		pAttachJObj->AttachModel(this);
		HSD_JObj		*pAttachObj = pAttachJObj->GetJObj();
		HSD_JObj		*pTopObj    = mpTopObj->GetJObj();
		if( pAttachObj!=NULL && pTopObj!=NULL ){
			HSD_JObjAddChild(pAttachObj,pTopObj);
		}
	}
	mpAttachJObj = pAttachJObj;
	return true;
}

bool WK7Model::Detach(void)
{
	if( mpAttachJObj==NULL ){
		return true;
	}
	HSD_JObj		*pAttachObj = mpAttachJObj->GetJObj();
	HSD_JObj		*pTopObj    = mpTopObj->GetJObj();
	if( pAttachObj!=NULL && pTopObj!=NULL ){
		HSD_JObjReparent(pTopObj,NULL);
	}
	if( mpAttachJObj->DetachModel(this)==false ){
		return false;
	}
	mpAttachJObj = NULL;
	return true;
}

WK7JObj* WK7Model::GetObj(const char *pName,bool child)
{
	if( mpTopObj==NULL ){
		return NULL;
	}
	return mpTopObj->GetObj(pName,true,child);
}

WK7JObj* WK7Model::GetObjLC(const char *pName,bool child)
{
	if( mpTopObj==NULL ){
		return NULL;
	}
	return mpTopObj->GetObjLC(pName,true,child);
}

bool WK7Model::SetAlpha(f32 alpha,u32 tick)
{
	if( tick==0 ){
		_SetAlpha(alpha);
	}else{
		_SetAlpha(mAlpha.GetValue());
	}
	mAlpha.Play(alpha,tick);
	return true;
}

bool WK7Model::_SetAlpha(f32 alpha)
{
	if( mpTopObj==NULL ){
		return false;
	}
	if( alpha > 1.0f ){
		alpha = 1.0f;
	}else if( alpha < 0.0f ){
		alpha = 0.0f;
	}
	if( alpha>=1.0f ){
		GHM_SAFE_DELETE(mpAlpha);
		return true;
	}
	if( mpAlpha==NULL ){
		ghmGcMemSetAllocName("C_SetAl");
		if( (mpAlpha=new C_SetAlpha(mpTopObj->GetJObj(),alpha))==NULL ){
			return false;
		}
	}else{
		mpAlpha->SetVal(alpha);
	}
	return true;
}

f32 WK7Model::GetAlpha(void)
{
	return mAlpha.GetValue();
}

WK7JObj* WK7Model::GetObjPartialName(const char *pPartialName,bool child)
{
	if( mpTopObj==NULL ){
		return NULL;
	}
	return mpTopObj->GetObjPartialName(pPartialName,true,child);
}

WK7JObj* WK7Model::GetObjPartialName2(const char *pPartialName,bool child)
{
	if( mpTopObj==NULL ){
		return NULL;
	}
	return mpTopObj->GetObjPartialName2(pPartialName,true,child);
}

WK7JObj* WK7Model::GetEffectObj(u32 id)
{
	char		efname[16];
	sprintf(efname,"ef%4d",id % 10000);
	return GetObj(efname);
}

WK7JObj* WK7Model::GetObj(int index)
{
	WK7JObj		*pObj;
	int			i;
	for(i=0,pObj=mpTopObj ; i<index && pObj!=NULL ; i++,pObj=pObj->GetNext()){}
	return pObj;
}

WK7JObj* WK7Model::GetObj(HSD_Joint *pJoint,bool child)
{
	if( mpTopObj==NULL ){
		return NULL;
	}
	return mpTopObj->GetObj(pJoint,true,child);
}

bool WK7Model::SetDstAlpha(u8 alpha,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetDstAlpha(alpha,false,true);
	}
	if( mpTopObj!=NULL ){
		mpTopObj->SetDstAlpha(alpha);
	}
	return true;
}

bool WK7Model::SetAlphaUpdate(bool flag,int index,u32 trsp)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetAlphaUpdate(flag,false,true,trsp);
	}
	if( mpTopObj!=NULL ){
		mpTopObj->SetAlphaUpdate(flag,true,true,trsp);
	}
	return true;
}

bool WK7Model::SetNoZUpdate(bool flag,int index,u32 trsp)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetNoZUpdate(flag,false,true,trsp);
	}
	if( mpTopObj!=NULL ){
		mpTopObj->SetNoZUpdate(flag,true,true,trsp);
	}
	return true;
}

bool WK7Model::SetPE(HSD_PEDesc *pPEDesc,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetPE(pPEDesc,false,true);
	}
	if( mpTopObj!=NULL ){
		mpTopObj->SetPE(pPEDesc);
	}
	return true;
}

bool WK7Model::Hide(bool next)
{
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->Hide(next,true);
	return true;
}

bool WK7Model::Disp(bool next)
{
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->Disp(next,true);
	return true;
}

f32 WK7Model::GetCurrentMotionFrame(bool next)
{
	if( mpTopObj==NULL ){
		return -1.0f;
	}
	return mpTopObj->GetCurrentMotionFrame(next);
}

f32 WK7Model::GetTotalMotionFrame(bool next)
{
	if( mpTopObj==NULL ){
		return -1.0f;
	}
	return mpTopObj->GetTotalMotionFrame(next);
}

WK7JObj* WK7Model::AttachMotion(WK7ModelMotion *pModelMotion,const char *pModelName,const char *pMotionName,bool next,bool child,bool overwrite)
{
	if( pModelMotion==NULL ){
		return NULL;
	}
	WK7JObj		*pObj;
	WK7JObj		*pAttachObj;
	if( pModelName==NULL ){
		pObj = GetObj(0);
	}else{
		pObj = GetObj(pModelName);
	}
	if( pObj==NULL ){
		return NULL;
	}
	HSD_AnimJoint	*pAnimJoint;
	if( pMotionName==NULL ){
		pMotionName = pModelName;
	}
	if( pMotionName==NULL ){
		pModelMotion->GetAnimJoint(0,&pAnimJoint);
	}else{
		pModelMotion->GetAnimJoint(pMotionName,&pAnimJoint);
	}
	if( pAnimJoint==NULL ){
		return NULL;
	}
	pAttachObj = pObj;
	if( pMotionName==NULL || pModelName ==NULL || strcmp(pMotionName,pModelName)!=0 ){
		if( pAttachObj->GetChild() != NULL ){
			pAttachObj = pAttachObj->GetChild();
			next = true;
		}
		const char	*pAnimJointName = WK7ModelMotion::GetAnimJointName(pAnimJoint);
		WK7JObj		*pNextObj;
		for(pNextObj=pAttachObj ; pNextObj!=NULL ; pNextObj=pNextObj->GetNext()){
			if( strcmp(pAnimJointName,pNextObj->GetName())==0 ){
				break;
			}
		}
		if( pNextObj==NULL ){
			if( pAnimJoint->child != NULL ){
				pAnimJoint = pAnimJoint->child;
			}
		}
	}
	pAttachObj->AttachMotion(pAnimJoint,next,child,overwrite);
	return pObj;
}

int WK7Model::AttachMotionAll(WK7ModelMotion *pModelMotion,bool overwrite)
{
	int	result;
	result  = AttachMotion     (pModelMotion,overwrite);
	result += AttachMatMotion  (pModelMotion);
	result += AttachShapeMotion(pModelMotion,overwrite);
	return result;
}

int WK7Model::AttachMotion(WK7ModelMotion *pModelMotion,bool overwrite)
{
	if( pModelMotion==NULL ){
		return 0;
	}
	WK7JObj			*pObj;
	WK7JObj			*pAttachObj;
	HSD_AnimJoint	*pAnimJoint;
	const char		*pAnimJointName;
	int				result = 0;
	int				i;
	int				motnum = pModelMotion->GetModelSetNum();
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	for(i=0 ; i<motnum ; i++){
		if( pModelMotion->GetAnimJoint(i,&pAnimJoint,NULL,NULL)==false || pAnimJoint==NULL ){
			continue;
		}
		if( (pAnimJointName = WK7ModelMotion::GetAnimJointName(pAnimJoint))==NULL ){
			continue;
		}
		for(pObj=mpTopObj ; pObj!=NULL ; pObj=pObj->GetNext()){
			if( (pAttachObj = pObj->GetObj(pAnimJointName))==NULL ){
//			if( strcmp(pAnimJointName,pObj->GetName())!=0 ){
				continue;
			}
			result += pAttachObj->AttachMotion(pAnimJoint,false,true,overwrite);
//			result += pObj->AttachMotion(pAnimJoint);
			break;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMotion(WK7ModelMotion *pModelMotion,int motindex,int index,bool autoattr,bool overwrite)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL || pModelMotion==NULL ){
		return 0;
	}
	HSD_AnimJoint	*pAnimJoint;
	if( motindex < 0 ){
		int			motnum = pModelMotion->GetModelSetNum();
		for(motindex=0 ; motindex<motnum ; motindex++){
			if( pModelMotion->GetAnimJoint(motindex,&pAnimJoint,NULL,NULL)==true && pAnimJoint!=NULL ){
				break;
			}
		}
		if( motindex>=motnum ){
			return 0;
		}
	}else{
		if( pModelMotion->GetAnimJoint(motindex,&pAnimJoint,NULL,NULL)==false ){
			return 0;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int		result = pObj->AttachMotion(pAnimJoint,false,true,overwrite);
	if( autoattr==true ){
		SetMotionLoop(pModelMotion->GetLoop());
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMotionClassic(WK7ModelMotion *pModelMotion)
{
	if( pModelMotion==NULL ){
		return 0;
	}
	WK7JObj			*pObj;
	HSD_AnimJoint	*pAnimJoint;
	int				result = 0;
	int				i;
	int				motnum = pModelMotion->GetModelSetNum();
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	for(i=0,pObj=mpTopObj ; i<motnum ; i++,pObj=pObj->GetNext()){
		if( pModelMotion->GetAnimJoint(i,&pAnimJoint,NULL,NULL)==false || pAnimJoint==NULL ){
			continue;
		}
		result += pObj->AttachMotionClassic(pAnimJoint);
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMotionClassic(WK7ModelMotion *pModelMotion,int motindex,int index,bool autoattr)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL || pModelMotion==NULL ){
		return 0;
	}
	HSD_AnimJoint	*pAnimJoint;
	if( motindex < 0 ){
		int			motnum = pModelMotion->GetModelSetNum();
		for(motindex=0 ; motindex<motnum ; motindex++){
			if( pModelMotion->GetAnimJoint(motindex,&pAnimJoint,NULL,NULL)==true && pAnimJoint!=NULL ){
				break;
			}
		}
		if( motindex>=motnum ){
			return 0;
		}
	}else{
		if( pModelMotion->GetAnimJoint(motindex,&pAnimJoint,NULL,NULL)==false ){
			return 0;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int		result = pObj->AttachMotionClassic(pAnimJoint);
	if( autoattr==true ){
		SetMotionLoop(pModelMotion->GetLoop());
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMotion(HSD_AnimJoint *pAnimJoint,int index,bool overwrite)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->AttachMotion(pAnimJoint,false,true,overwrite);
}

int WK7Model::AttachMotionClassic(HSD_AnimJoint *pAnimJoint,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->AttachMotionClassic(pAnimJoint);
}

bool WK7Model::DetachMotionAll(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->DetachMotionAll();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->DetachMotionAll(false,true);
}

bool WK7Model::DetachMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->DetachMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->DetachMotion(false,true);
}

bool WK7Model::ResetMotionAll(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ResetMotionAll();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ResetMotionAll(false,true);
}

bool WK7Model::ResetMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ResetMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ResetMotion(false,true);
}

bool WK7Model::PlayMotionAll(f32 startframe,bool interpolate,f32 interpolateframe,int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->PlayMotionAll(startframe,interpolate,interpolateframe);
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->PlayMotionAll(startframe,interpolate,interpolateframe,false,true);
}

bool WK7Model::PlayMotion(f32 startframe,bool interpolate,f32 interpolateframe,int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->PlayMotion(startframe,interpolate,interpolateframe);
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->PlayMotion(startframe,interpolate,interpolateframe,false,true);
}

bool WK7Model::PlayMotionNA(f32 startframe,bool interpolate,f32 interpolateframe,int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->PlayMotionNA(startframe,interpolate,interpolateframe);
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->PlayMotionNA(startframe,interpolate,interpolateframe,false,true);
}

bool WK7Model::ReplayMotionAll(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ReplayMotionAll();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ReplayMotionAll(false,true);
}

bool WK7Model::ReplayMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ReplayMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ReplayMotion(false,true);
}

bool WK7Model::StopMotionAll(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->StopMotionAll();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->StopMotionAll(false,true);
}

bool WK7Model::StopMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->StopMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->StopMotion(false,true);
}

WK7JObj* WK7Model::AttachMatMotion(WK7ModelMotion *pModelMotion,const char *pModelName,const char *pMotionName,bool childOnly,bool next,bool child)
{
	WK7JObj		*pObj;
	if( pModelMotion==NULL ){
		return NULL;
	}
	if( pModelName==NULL ){
		pObj = GetObj(0);
	}else{
		pObj = GetObj(pModelName);
	}
	if( pObj==NULL ){
		return NULL;
	}
	HSD_MatAnimJoint	*pMatAnimJoint;
	if( pMotionName==NULL ){
		pModelMotion->GetAnimJoint(0,NULL,&pMatAnimJoint);
	}else{
		pModelMotion->GetAnimJoint(pMotionName,NULL,&pMatAnimJoint);
	}
	if( pMatAnimJoint==NULL ){
		return NULL;
	}
	if( childOnly==true ){
		pMatAnimJoint = pMatAnimJoint->child;
		pObj          = pObj->GetChild();
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	pObj->AttachMatMotion(pMatAnimJoint,next,child);
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return pObj;
}

int WK7Model::AttachMatMotion(WK7ModelMotion *pModelMotion)
{
	if( pModelMotion==NULL ){
		return 0;
	}
	WK7JObj				*pObj;
	HSD_MatAnimJoint	*pMatAnimJoint;
	const char			*pMatAnimJointName;
	int					i;
	int					result = 0;
	int					motnum = pModelMotion->GetModelSetNum();
	
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	for(i=0 ; i<motnum ; i++){
		if( pModelMotion->GetAnimJoint(i,NULL,&pMatAnimJoint,NULL)==false || pMatAnimJoint==NULL ){
			continue;
		}
		if( (pMatAnimJointName = WK7ModelMotion::GetMatAnimJointName(pMatAnimJoint))==NULL ){
			continue;
		}
		for(pObj=mpTopObj ; pObj!=NULL ; pObj=pObj->GetNext()){
			if( strcmp(pMatAnimJointName,pObj->GetName())!=0 ){
				continue;
			}
			result += pObj->AttachMatMotion(pMatAnimJoint,false,true);
			break;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMatMotion(WK7ModelMotion *pModelMotion,int motindex,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL || pModelMotion==NULL ){
		return 0;
	}
	HSD_MatAnimJoint	*pMatAnimJoint;
	if( motindex < 0 ){
		int			motnum = pModelMotion->GetModelSetNum();
		for(motindex=0 ; motindex<motnum ; motindex++){
			if( pModelMotion->GetAnimJoint(motindex,NULL,&pMatAnimJoint,NULL)==true && pMatAnimJoint!=NULL ){
				break;
			}
		}
		if( motindex>=motnum ){
			return 0;
		}
	}else{
		if( pModelMotion->GetAnimJoint(motindex,NULL,&pMatAnimJoint,NULL)==false ){
			return 0;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int	result = pObj->AttachMatMotion(pMatAnimJoint);
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMatMotionClassic(WK7ModelMotion *pModelMotion)
{
	if( pModelMotion==NULL ){
		return 0;
	}
	WK7JObj				*pObj;
	HSD_MatAnimJoint	*pMatAnimJoint;
	int					i;
	int					result = 0;
	int					motnum = pModelMotion->GetModelSetNum();
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	for(i=0,pObj=mpTopObj ; i<motnum ; i++,pObj=pObj->GetNext()){
		if( pModelMotion->GetAnimJoint(i,NULL,&pMatAnimJoint,NULL)==false || pMatAnimJoint==NULL ){
			continue;
		}
		result += pObj->AttachMatMotionClassic(pMatAnimJoint);
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMatMotionClassic(WK7ModelMotion *pModelMotion,int motindex,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL || pModelMotion==NULL ){
		return 0;
	}
	HSD_MatAnimJoint	*pMatAnimJoint;
	if( motindex < 0 ){
		int			motnum = pModelMotion->GetModelSetNum();
		for(motindex=0 ; motindex<motnum ; motindex++){
			if( pModelMotion->GetAnimJoint(motindex,NULL,&pMatAnimJoint,NULL)==true && pMatAnimJoint!=NULL ){
				break;
			}
		}
		if( motindex>=motnum ){
			return 0;
		}
	}else{
		if( pModelMotion->GetAnimJoint(motindex,NULL,&pMatAnimJoint,NULL)==false ){
			return 0;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int	result = pObj->AttachMatMotionClassic(pMatAnimJoint);
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachMatMotion(HSD_MatAnimJoint *pMatAnimJoint,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->AttachMatMotion(pMatAnimJoint);
}

int WK7Model::AttachMatMotionClassic(HSD_MatAnimJoint *pMatAnimJoint,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->AttachMatMotionClassic(pMatAnimJoint);
}

bool WK7Model::DetachMatMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->DetachMatMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->DetachMatMotion(false,true);
}

bool WK7Model::ResetMatMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ResetMatMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ResetMatMotion(false,true);
}

bool WK7Model::PlayMatMotion(f32 startframe,int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->PlayMatMotion(startframe);
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->PlayMatMotion(startframe,false,true);
}

bool WK7Model::ReplayMatMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ReplayMatMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ReplayMatMotion(false,true);
}

bool WK7Model::StopMatMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->StopMatMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->StopMatMotion(false,true);
}

WK7JObj* WK7Model::AttachShapeMotion(WK7ModelMotion *pModelMotion,const char *pModelName,const char *pMotionName,bool childOnly,bool next,bool child,bool overwrite)
{
	if( pModelMotion==NULL ){
		return NULL;
	}
	WK7JObj		*pObj;
	if( pModelName==NULL ){
		pObj = GetObj(0);
	}else{
		pObj = GetObj(pModelName);
	}
	if( pObj==NULL ){
		return NULL;
	}
	HSD_ShapeAnimJoint	*pShapeAnimJoint;
	if( pMotionName==NULL ){
		pModelMotion->GetAnimJoint(0,NULL,NULL,&pShapeAnimJoint);
	}else{
		pModelMotion->GetShapeAnimJoint(pMotionName,NULL,NULL,&pShapeAnimJoint);
	}
	if( pShapeAnimJoint==NULL ){
		return NULL;
	}
	if( childOnly==true ){
		pShapeAnimJoint = pShapeAnimJoint->child;
		pObj            = pObj->GetChild();
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	pObj->AttachShapeMotion(pShapeAnimJoint,next,child,overwrite);
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return pObj;
}

int WK7Model::AttachShapeMotion(WK7ModelMotion *pModelMotion,bool overwrite)
{
	if( pModelMotion==NULL ){
		return 0;
	}
	WK7JObj				*pObj;
	HSD_ShapeAnimJoint	*pShapeAnimJoint;
	const char			*pShapeAnimJointName;
	int					i;
	int					result = 0;
	int					motnum = pModelMotion->GetModelSetNum();
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	for(i=0 ; i<motnum ; i++){
		if( pModelMotion->GetAnimJoint(i,NULL,NULL,&pShapeAnimJoint)==false || pShapeAnimJoint==NULL ){
			continue;
		}
		if( (pShapeAnimJointName = WK7ModelMotion::GetShapeAnimJointName(pShapeAnimJoint))==NULL ){
			continue;
		}
		for(pObj=mpTopObj ; pObj!=NULL ; pObj=pObj->GetNext()){
			if( strcmp(pShapeAnimJointName,pObj->GetName())!=0 ){
				continue;
			}
			result += pObj->AttachShapeMotion(pShapeAnimJoint,false,true,overwrite);
			break;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachShapeMotion(WK7ModelMotion *pModelMotion,int motindex,int index,bool overwrite)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL || pModelMotion==NULL ){
		return 0;
	}
	HSD_ShapeAnimJoint	*pShapeAnimJoint;
	if( motindex < 0 ){
		int			motnum = pModelMotion->GetModelSetNum();
		for(motindex=0 ; motindex<motnum ; motindex++){
			if( pModelMotion->GetAnimJoint(motindex,NULL,NULL,&pShapeAnimJoint)==true && pShapeAnimJoint!=NULL ){
				break;
			}
		}
		if( motindex>=motnum ){
			return 0;
		}
	}else{
		if( pModelMotion->GetAnimJoint(motindex,NULL,NULL,&pShapeAnimJoint)==false ){
			return 0;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int	result = pObj->AttachShapeMotion(pShapeAnimJoint,overwrite);
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachShapeMotionClassic(WK7ModelMotion *pModelMotion)
{
	if( pModelMotion==NULL ){
		return 0;
	}
	WK7JObj				*pObj;
	HSD_ShapeAnimJoint	*pShapeAnimJoint;
	int					i;
	int					result = 0;
	int					motnum = pModelMotion->GetModelSetNum();
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	for(i=0,pObj=mpTopObj ; i<motnum ; i++,pObj=pObj->GetNext()){
		if( pModelMotion->GetAnimJoint(i,NULL,NULL,&pShapeAnimJoint)==false || pShapeAnimJoint==NULL ){
			continue;
		}
		result += pObj->AttachShapeMotionClassic(pShapeAnimJoint);
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachShapeMotionClassic(WK7ModelMotion *pModelMotion,int motindex,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	HSD_ShapeAnimJoint	*pShapeAnimJoint;
	if( motindex < 0 ){
		int			motnum = pModelMotion->GetModelSetNum();
		for(motindex=0 ; motindex<motnum ; motindex++){
			if( pModelMotion->GetAnimJoint(motindex,NULL,NULL,&pShapeAnimJoint)==true && pShapeAnimJoint!=NULL ){
				break;
			}
		}
		if( motindex>=motnum ){
			return 0;
		}
	}else{
		if( pModelMotion->GetAnimJoint(motindex,NULL,NULL,&pShapeAnimJoint)==false ){
			return 0;
		}
	}
	WK7JObj::SetCurrentAttachMotionSdExpInfo(pModelMotion);
	int	result = pObj->AttachShapeMotionClassic(pShapeAnimJoint);
	WK7JObj::SetCurrentAttachMotionSdExpInfo();
	return result;
}

int WK7Model::AttachShapeMotion(HSD_ShapeAnimJoint *pShapeAnimJoint,int index,bool overwrite)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->AttachShapeMotion(pShapeAnimJoint,overwrite);
}

int WK7Model::AttachShapeMotionClassic(HSD_ShapeAnimJoint *pShapeAnimJoint,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->AttachShapeMotionClassic(pShapeAnimJoint);
}

bool WK7Model::DetachShapeMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->DetachShapeMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->DetachShapeMotion(false,true);
}

bool WK7Model::ResetShapeMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ResetShapeMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ResetShapeMotion(false,true);
}

bool WK7Model::PlayShapeMotion(f32 startframe,int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->PlayShapeMotion(startframe);
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->PlayShapeMotion(startframe,false,true);
}

bool WK7Model::ReplayShapeMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ReplayShapeMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->ReplayShapeMotion(false,true);
}

bool WK7Model::StopShapeMotion(int index)
{
	WK7JObj		*pObj;
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->StopShapeMotion();
		}
		return true;
	}
	pObj = GetObj(index);
	if( pObj==NULL ){
		return 0;
	}
	return pObj->StopShapeMotion(false,true);
}

int WK7Model::GetMotionState(void)
{
	if( mpTopObj==NULL ){
		return AOBJ_NO_ANIM;
	}
	return mpTopObj->GetMotionState();
}

int WK7Model::GetShapeMotionState(void)
{
	if( mpTopObj==NULL ){
		return AOBJ_NO_ANIM;
	}
	return mpTopObj->GetShapeMotionState();
}

bool WK7Model::SetMotionLoop(bool flag,int index)
{
	WK7JObj		*pObj;
	bool		next;
	if( index >= 0 ){
		pObj = GetObj(index);
		next = false;
	}else{
		pObj = mpTopObj;
		next = true;
	}
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetMotionLoop(flag,next,true);
}

bool WK7Model::GetMotionLoop(int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->GetMotionLoop(false,true);
}

bool WK7Model::SetMotionRate(f32 rate,bool next)
{
	if( mpTopObj==NULL ){
		return false;
	}
	return mpTopObj->SetMotionRate(rate,next,true);
}

bool WK7Model::RenewMotionRate(bool next)
{
	if( mpTopObj==NULL ){
		return false;
	}
	return mpTopObj->RenewMotionRate(next,true);
}

bool WK7Model::SetPosition(f32 x,f32 y,f32 z,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetPosition(x,y,z);
}

bool WK7Model::SetPosition(const Vec *pPosition,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetPosition(pPosition);
}

bool WK7Model::GetPosition(Vec *pPosition,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->GetPosition(pPosition);
}

bool WK7Model::SetRotation(f32 x,f32 y,f32 z,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetRotation(x,y,z);
}

bool WK7Model::SetRotX(f32 x,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetRotX(x);
}

bool WK7Model::SetRotY(f32 y,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetRotY(y);
}

bool WK7Model::SetRotZ(f32 z,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetRotZ(z);
}

bool WK7Model::SetRotation(const Vec *pRotation,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetRotation(pRotation);
}

bool WK7Model::GetRotation(f32 *px,f32 *py,f32 *pz,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->GetRotation(px,py,pz);
}

bool WK7Model::GetRotation(Vec *pRotation,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->GetRotation(pRotation);
}

bool WK7Model::SetScale(f32 x,f32 y,f32 z,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetScale(x,y,z);
}

bool WK7Model::SetScale(const Vec *pScale,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->SetScale(pScale);
}

bool WK7Model::GetScale(f32 *px,f32 *py,f32 *pz,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->GetScale(px,py,pz);
}

bool WK7Model::GetScale(Vec *pScale,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->GetScale(pScale);
}

bool WK7Model::CopyUserDefinedMatrix(WK7Model *pModel)
{
	if( pModel==NULL || mpTopObj==NULL ){
		return false;
	}
	return mpTopObj->CopyUserDefinedMatrix(pModel->GetTopObj());
}

bool WK7Model::CopyTRS(WK7Model *pModel)
{
	if( pModel==NULL || mpTopObj==NULL ){
		return false;
	}
	return mpTopObj->CopyTRS(pModel->GetTopObj());
}

MtxPtr WK7Model::GetMtxPtr(int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return NULL;
	}
	return pObj->GetMtxPtr();
}

bool WK7Model::GetWorldPosition(Vec *pPosition,int index)
{
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	return pObj->GetWorldPosition(pPosition);
}

bool WK7Model::SetFlag(u32 flag,int index)
{
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->SetFlag(flag);
		}
		return true;
	}
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	pObj->SetFlag(flag,false);
	return true;
}

bool WK7Model::ClearFlag(u32 flag,int index)
{
	if( index < 0 ){
		if( mpTopObj!=NULL ){
			mpTopObj->ClearFlag(flag);
		}
		return true;
	}
	WK7JObj		*pObj = GetObj(index);
	if( pObj==NULL ){
		return false;
	}
	pObj->ClearFlag(flag,false);
	return true;
}

bool WK7Model::InitializeAll(void)
{
	ghmGcMemSetAllocName("C_EdgeSh");
	WK7JObj::mpEdgeShadow = new C_EdgeShadow;
	return true;
}

bool WK7Model::TerminateAll(void)
{
	GHM_SAFE_DELETE(WK7JObj::mpEdgeShadow);
	return true;
}

bool WK7Model::FrameProcessEdgeShadow(void)
{
	if( WK7JObj::mEdgeShadowObjNum <= 0 ){
		return false;
	}
	if( WK7JObj::mpEdgeShadow==NULL ){
		return false;
	}
	WK7JObj::mpEdgeShadow->Proc();
	return true;
}

bool WK7Model::RenderProcessEdgeShadow(void)
{
	if( WK7JObj::mEdgeShadowObjNum <= 0 ){
		return false;
	}
	if( WK7JObj::mpEdgeShadow==NULL ){
		return false;
	}
//	GXSetDither(GX_FALSE);
	WK7JObj::SetCurrentRenderMode(WK7JOBJ_RENDERMODE_SHADOW);
	WK7JObj::mpEdgeShadow->DispAll(0,0x00000008);
//	WK7JObj::mpEdgeShadow->DispStart();
//	WK7JObj::mpEdgeShadow->DispEnd();
	WK7JObj::SetCurrentRenderMode(WK7JOBJ_RENDERMODE_NORMAL);
	return true;
}

bool WK7Model::RenderProcessEdgeShadowPaint(void)
{
	if( WK7JObj::mEdgeShadowObjNum <= 0 ){
		return false;
	}
	if( WK7JObj::mpEdgeShadow==NULL ){
		return false;
	}
	static u32	color = 0x000000c0;
	WK7JObj::mpEdgeShadow->DispShadow(color);
	return true;
}

bool WK7Model::ClearAlpha(u32 alpha)
{
	
	Mtx				view;
	Mtx44			proj;
	Vec				vPos;
	Vec				vUp;
	Vec				vAt;
	u32				color = 0x00000000|alpha;
	f32				z     = 0.0f;
	
	vUp .x =    0.0f;
	vUp .y = -  1.0f;
	vUp .z =    0.0f;
	vPos.x =    0.0f;
	vPos.y =    0.0f;
	vPos.z =    0.0f;
	vAt .x =  vPos.x;
	vAt .y =  vPos.y;
	vAt .z =  300.0f;
	
	// [Jr[ϊ}gNXݒ
	MTXLookAt       (view,&vPos,&vUp,&vAt);
	MTXOrtho        (proj,0,-WK7_SCREEN_HEIGHT,0,WK7_SCREEN_WIDTH,0.0f,1024.0f);
	
	GXLoadPosMtxImm (view,GX_PNMTX0);
	GXLoadNrmMtxImm (view,GX_PNMTX0);
	GXSetCurrentMtx (GX_PNMTX0);
	GXSetProjection (proj,GX_ORTHOGRAPHIC);
	
	GXSetScissor    (0,0,WK7_SCREEN_WIDTH,WK7_SCREEN_HEIGHT);
	GXSetViewport   (0.0f,0.0f,WK7_SCREEN_WIDTH,WK7_SCREEN_HEIGHT,0.0f,1.0f);
	
	GXSetDither(GX_FALSE);
	GXSetDstAlpha(0,0);
	GXSetAlphaCompare(GX_ALWAYS,0,GX_AOP_AND,GX_ALWAYS,0);
	
	GXSetNumChans     (1);
	GXSetChanCtrl     (GX_COLOR0A0,GX_FALSE,GX_SRC_VTX,GX_SRC_VTX,GX_LIGHT_NULL,GX_DF_NONE,GX_AF_NONE);
	
	// uhȂ
	GXSetBlendMode    (GX_BM_NONE,GX_BL_SRCALPHA,GX_BL_INVSRCALPHA,GX_LO_CLEAR);
//	GXSetBlendMode    (GX_BM_BLEND,GX_BL_SRCALPHA,GX_BL_INVSRCALPHA,GX_LO_CLEAR);
	
	GXSetCullMode     (GX_CULL_NONE);
	
	GXSetZMode (GX_FALSE,GX_ALWAYS,GX_FALSE);
	GXSetZCompLoc   (GX_TRUE);
	GXSetColorUpdate(GX_FALSE);
	GXSetAlphaUpdate(GX_TRUE);
	
	GXClearVtxDesc    ();
	GXSetVtxDesc      (GX_VA_POS ,GX_DIRECT);
	GXSetVtxDesc      (GX_VA_CLR0,GX_DIRECT);
	
	GXSetVtxAttrFmt   (GX_VTXFMT0,GX_VA_POS ,GX_POS_XYZ ,GX_F32  ,0);
	GXSetVtxAttrFmt   (GX_VTXFMT0,GX_VA_CLR0,GX_CLR_RGBA,GX_RGBA8,0);
	
	GXSetNumTexGens   (0);
	
	GXSetNumTevStages (1);
	GXSetTevOp        (GX_TEVSTAGE0,GX_PASSCLR);
	GXSetTevOrder     (GX_TEVSTAGE0,GX_TEXCOORD_NULL,GX_TEXMAP_NULL,GX_COLOR0A0);
	
	GXBegin(GX_QUADS,GX_VTXFMT0,4);
	
		GXPosition3f32(0.0f            ,0.0f             ,z);
		GXColor1u32   (color);
		
		GXPosition3f32(WK7_SCREEN_WIDTH,0.0f             ,z);
		GXColor1u32   (color);
		
		GXPosition3f32(WK7_SCREEN_WIDTH,WK7_SCREEN_HEIGHT,z);
		GXColor1u32   (color);
		
		GXPosition3f32(0.0f            ,WK7_SCREEN_HEIGHT,z);
		GXColor1u32   (color);
		
	GXEnd();
	
	return true;
}

bool WK7Model::SetMaterialAmbientColor(GXColor col,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetMaterialAmbientColor(col,false,true);
	}
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->SetMaterialAmbientColor(col);
	return true;
}

bool WK7Model::SetMaterialDiffuseColor(GXColor col,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetMaterialDiffuseColor(col,false,true);
	}
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->SetMaterialDiffuseColor(col);
	return true;
}

bool WK7Model::SetMaterialSpecularColor(GXColor col,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetMaterialSpecularColor(col,false,true);
	}
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->SetMaterialSpecularColor(col);
	return true;
}

bool WK7Model::SetMaterialAmbientColorLE(GXColor col,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetMaterialAmbientColorLE(col,false,true);
	}
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->SetMaterialAmbientColorLE(col);
	return true;
}

bool WK7Model::SetMaterialDiffuseColorLE(GXColor col,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetMaterialDiffuseColorLE(col,false,true);
	}
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->SetMaterialDiffuseColorLE(col);
	return true;
}

bool WK7Model::SetMaterialSpecularColorLE(GXColor col,int index)
{
	if( index >= 0 ){
		WK7JObj		*pObj;
		pObj = GetObj(index);
		if( pObj==NULL ){
			return false;
		}
		return pObj->SetMaterialSpecularColorLE(col,false,true);
	}
	if( mpTopObj==NULL ){
		return false;
	}
	mpTopObj->SetMaterialSpecularColorLE(col);
	return true;
}

bool WK7Model::SetMaterialAmbientColor(GXColor col,u32 tick)
{
	mMatAmbientColorFade   = 1;
	mMatAmbientColorFadeLE = 0;
	mMatAmbientColorR.Play(col.r,tick);
	mMatAmbientColorG.Play(col.r,tick);
	mMatAmbientColorB.Play(col.r,tick);
	return true;
}

bool WK7Model::SetMaterialAmbientColorLE(GXColor col,u32 tick)
{
	mMatAmbientColorFade   = 1;
	mMatAmbientColorFadeLE = 1;
	mMatAmbientColorR.Play(col.r,tick);
	mMatAmbientColorG.Play(col.r,tick);
	mMatAmbientColorB.Play(col.r,tick);
	return true;
}

bool WK7Model::SetMaterialDiffuseColor(GXColor col,u32 tick)
{
	mMatDiffuseColorFade   = 1;
	mMatDiffuseColorFadeLE = 0;
	mMatDiffuseColorR.Play(col.r,tick);
	mMatDiffuseColorG.Play(col.r,tick);
	mMatDiffuseColorB.Play(col.r,tick);
	return true;
}

bool WK7Model::SetMaterialDiffuseColorLE(GXColor col,u32 tick)
{
	mMatDiffuseColorFade   = 1;
	mMatDiffuseColorFadeLE = 1;
	mMatDiffuseColorR.Play(col.r,tick);
	mMatDiffuseColorG.Play(col.r,tick);
	mMatDiffuseColorB.Play(col.r,tick);
	return true;
}

